#include "include.h"
#include "global.h"
#include "zoneman.h"
#include "modeman.h"

/* Zone comparison status values for collapse_zone()
 *
 * 	CONTAINED     = -2
 * 	NO_INTERSECT  = -1
 * 	EQUAL	      =  0
 * 	INTERSECT     =  1
 * 	CONTAINS      =  2
 *
 */

int collapse_zones(struct thread_type *);
int general_zone_comp(struct zone_type *, struct zone_type *);
struct zone_type *zone_intersect(struct zone_type *, struct zone_type *);

int collapse_zones(struct thread_type *t) {
  struct mode_type *m, *n;
  struct zone_type *z;
  int zc;

  printf("    Collapsing zones for modes in thread %d.\n", t->index);
  sprintf(tmpstring, "Collapsing zones for modes in thread %d.\n", t->index);
  output(tmpstring);
  for(m=t->starting_mode; m; m=m->cnext) {
    for(n=t->starting_mode; n; n=n->cnext) {
      if(m->index == n->index) continue;
      if(strcmp(m->name, n->name)) continue;
      if(dvl_comp(m, n) != EQUAL) continue;
      zc = general_zone_comp(m->zone, n->zone);
      switch(zc) {
        case CONTAINS:
          free_zone(m->zone);
          m->zone = zone_copy(n->zone);
          break;
        case CONTAINED:
          free_zone(n->zone);
          n->zone = zone_copy(m->zone);
          break;
        case OVERLAP:
          z = zone_intersect(m->zone, n->zone);
          free_zone(m->zone);
          free_zone(n->zone);
          m->zone = z;
          n->zone = zone_copy(z);
          break;
        case NO_OVERLAP:
        case EQUAL:
          continue;
      }
    }
  }
  sprintf(tmpstring, "END of collapsing zones for modes in thread %d.\n", 
                     t->index);
  output(tmpstring);
  merge_iden_red_multi(t);
}

int general_zone_comp(struct zone_type *z1, struct zone_type *z2) {
  int i, j;
  int ans = EQUAL;

/*
  if(!z1 && !z2) return EQUAL;
  if(z1 && !z2) return CONTAINS;
  if(!z1 && z2) return CONTAINED;
*/

  if(z1->relop[0][0] == TRUE && z2->relop[0][0] == TRUE) return EQUAL;

/*
  if(z1->relop[0][0] == FALSE && z2->relop[0][0] == FALSE) return EQUAL;
  if(z1->relop[0][0] == TRUE && z2->relop[0][0] == FALSE) return CONTAINS;
  if(z1->relop[0][0] == FALSE && z2->relop[0][0] == TRUE) return CONTAINED;
*/

  if(both_out_of_bounds(z1, z2)) {
/*
    boob = TRUE;
    sprintf(tmpstring,
            "(max_const = %d) Both zones out of bounds: ", max_const); 
    output(tmpstring);
    print_zone(z1);
    output(" AND "); print_zone(z2); output("\n");
*/
    return EQUAL;
  }
  for(i=0; i<=composed_ta; i++) {
    for(j=0; j<=composed_ta; j++) {
      if(i==0 && j==0) continue;
      if(i == 0 && z1->relop[i][j] == LEQ && z1->diff[i][j] == 0 
                && z2->relop[i][j] == -1) continue;
      if(i == 0 && z2->relop[i][j] == LEQ && z2->diff[i][j] == 0 
                && z1->relop[i][j] == -1) continue;
      if(z1->relop[i][j] == -1 && z2->relop[i][j] == -1) continue;
      if(z1->relop[i][j] != -1 && z2->relop[i][j] == -1) {
        if(ans == CONTAINS || ans == OVERLAP) ans = OVERLAP;
        else ans = CONTAINED;
      }
      if(z1->relop[i][j] == -1 && z2->relop[i][j] != -1) {
        if(ans == CONTAINED || ans == OVERLAP) ans = OVERLAP;
        else ans = CONTAINS;
      }
      if(z1->relop[i][j] != -1 && z2->relop[i][j] != -1) {
/*
        sprintf(tmpstring, "ans = %d, comparing: i=%d and j=%d, ", ans, i, j);
        output(tmpstring);
*/
        switch(z1->relop[i][j]) {
          case LESS:
            switch(z2->relop[i][j]) {
              case LESS:
                if(z1->diff[i][j]<z2->diff[i][j]) {
                  if(ans == CONTAINS || ans == OVERLAP) ans = OVERLAP;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i][j]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = OVERLAP;
                }
                break;
              case LEQ:
                if(z1->diff[i][j]<=z2->diff[i][j]) {
                  if(ans == CONTAINS || ans == OVERLAP) ans = OVERLAP;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i][j]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = OVERLAP;
                }
                break;
              case EQ:
                if(z1->diff[i][j]<=z2->diff[i][j]) {
                  return NO_OVERLAP;
                }
                else if(z1->diff[i][j]>z2->diff[i][j]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = OVERLAP;
                }
                break;
            }
            break;
          case LEQ:
            switch(z2->relop[i][j]) {
              case LESS:
                if(z1->diff[i][j]<z2->diff[i][j]) {
                  if(ans == CONTAINS || ans == OVERLAP) ans = OVERLAP;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>=z2->diff[i][j]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = OVERLAP;
                }
                break;
              case LEQ:
                if(z1->diff[i][j]<z2->diff[i][j]) {
                  if(ans == CONTAINS || ans == OVERLAP) ans = OVERLAP;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i][j]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = OVERLAP;
                }
                break;
              case EQ:
                if(z1->diff[i][j]<z2->diff[i][j]) {
                  return NO_OVERLAP;
                }
                else if(z1->diff[i][j]>=z2->diff[i][j]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = OVERLAP;
                }
                break;
            }
            break;
          case EQ:
            switch(z2->relop[i][j]) {
              case LESS:
                if(z1->diff[i][j]<z2->diff[i][j]) {
                  if(ans == CONTAINS || ans == OVERLAP) ans = OVERLAP;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>=z2->diff[i][j]) {
                  return NO_OVERLAP;
                }
                break;
              case LEQ:
                if(z1->diff[i][j]<=z2->diff[i][j]) {
                  if(ans == CONTAINS || ans == OVERLAP) ans = OVERLAP;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i][j]) {
                  return NO_OVERLAP;
                }
                break;
              case EQ:
                if(z1->diff[i][j] != z2->diff[i][j]) {
                  return NO_OVERLAP;
                }
                break;
            }
            break;
        } /* end switch */
/*
        sprintf(tmpstring, "ans = %d\n", ans);
        output(tmpstring);
*/
      } 
    }
  }

/*
  output("ZONE comparison: ");
  print_zone(z1); 
  output(" AND "); 
  print_zone(z2);
  sprintf(tmpstring, " IS %d.\n", ans); 
  output(tmpstring);
*/
  return ans;
}

struct zone_type *zone_intersect(struct zone_type *z1, struct zone_type *z2) {
  struct zone_type *z = alloc_zone(NULL);
  int i, j;

  if(verbose >= 0) {
    output("Z1:"); print_zone(z1); output("\n");
    output("Z2:"); print_zone(z2); output("\n");
  }
  
  for(i=0; i<=composed_ta; i++) {
    for(j=0; j<=composed_ta; j++) {
      if(z1->relop[i][j] == -1 && z2->relop[i][j] == -1) continue;
      if(z1->relop[i][j] != -1 && z2->relop[i][j] == -1) {
        z->relop[i][j] = z1->relop[i][j];
        z->diff[i][j] = z1->diff[i][j];
      }
      if(z1->relop[i][j] == -1 && z2->relop[i][j] != -1) {
        z->relop[i][j] = z2->relop[i][j];
        z->diff[i][j] = z2->diff[i][j];
      }
      if(z1->relop[i][j] != -1 && z2->relop[i][j] != -1) {
        if(z1->diff[i][j] < z2->diff[i][j]) {
          z->diff[i][j] = z1->diff[i][j];
          z->relop[i][j] = z1->relop[i][j];
        }
        else if(z1->diff[i][j] > z2->diff[i][j]) {
          z->diff[i][j] = z2->diff[i][j];
          z->relop[i][j] = z2->relop[i][j];
        }
        else {
          z->diff[i][j] = z1->diff[i][j];
          if(   (z1->relop[i][j] == LESS && z2->relop[i][j] == LEQ)
             || (z1->relop[i][j] == LEQ && z2->relop[i][j] == LESS)
            )
            z->relop[i][j] = LESS;
          else if(   (z1->relop[i][j] == EQ && z2->relop[i][j] == LEQ)
                  || (z1->relop[i][j] == LEQ && z2->relop[i][j] == EQ)
                 )
                 z->relop[i][j] = EQ;
          else z->relop[i][j] = z1->relop[i][j];
        }
      }
    }
  }
  if(verbose >= 0) {
    output("Intersection:");
    print_zone(z);
    output("\n");
  }
  return z;
}

