#include "include.h"
#include "global.h"
#include "zoneman.h"
#include "modeman.h"
#include "collapse.h"

int enclose_zones(struct thread_type *);
struct zone_type *zone_union(struct zone_type *, struct zone_type *);

int enclose_zones(struct thread_type *t) {
  struct mode_type *m, *n;
  struct zone_type *z;
  int zc;

  printf("    Enclosing zones for modes in thread %d.\n", t->index);
  sprintf(tmpstring, "Enclosing zones for modes in thread %d.\n", t->index);
  output(tmpstring);
  for(m=t->starting_mode; m; m=m->cnext) {
    for(n=t->starting_mode; n; n=n->cnext) {
      if(m->index == n->index) continue;
      if(strcmp(m->name, n->name)) continue;
      if(dvl_comp(m, n) != EQUAL) continue;
      zc = general_zone_comp(m->zone, n->zone);
      switch(zc) {
        case CONTAINS:
          free_zone(n->zone);
          n->zone = zone_copy(m->zone);
          break;
        case CONTAINED:
          free_zone(m->zone);
          m->zone = zone_copy(n->zone);
          break;
        case OVERLAP:
          z = zone_union(m->zone, n->zone);
          free_zone(m->zone);
          free_zone(n->zone);
          m->zone = z;
          n->zone = zone_copy(z);
          break;
        case NO_OVERLAP:
        case EQUAL:
          continue;
      }
    }
  }
  sprintf(tmpstring, "END of enclosing zones for modes in thread %d.\n", 
                     t->index);
  output(tmpstring);
  merge_iden_red_multi(t);
}

struct zone_type *zone_union(struct zone_type *z1, struct zone_type *z2) {
  struct zone_type *z = alloc_zone(NULL);
  int i, j;

  if(verbose >= 0) {
    output("Z1:"); print_zone(z1); output("\n");
    output("Z2:"); print_zone(z2); output("\n");
  }
 
  for(i=0; i<=composed_ta; i++) {
    for(j=0; j<=composed_ta; j++) {
      if( (z1->relop[i][j] == -1 && z2->relop[i][j] == -1) ||
          (z1->relop[i][j] != -1 && z2->relop[i][j] == -1) ||
          (z1->relop[i][j] == -1 && z2->relop[i][j] != -1)
        )
        continue;
      if(z1->relop[i][j] != -1 && z2->relop[i][j] != -1) {
        if(z1->diff[i][j] > z2->diff[i][j]) {
          z->diff[i][j] = z1->diff[i][j];
          z->relop[i][j] = z1->relop[i][j];
        }
        else if(z1->diff[i][j] < z2->diff[i][j]) {
          z->diff[i][j] = z2->diff[i][j];
          z->relop[i][j] = z2->relop[i][j];
        }
        else {
          z->diff[i][j] = z1->diff[i][j];
          if(   (z1->relop[i][j] == LESS && z2->relop[i][j] == LEQ)
             || (z1->relop[i][j] == LEQ && z2->relop[i][j] == LESS)
            )
            z->relop[i][j] = LEQ;
          else if(   (z1->relop[i][j] == EQ && z2->relop[i][j] == LEQ)
                  || (z1->relop[i][j] == LEQ && z2->relop[i][j] == EQ)
                 )
                 z->relop[i][j] = LEQ;
          else z->relop[i][j] = z1->relop[i][j];
        }
      }
    }
  }
  output("Union:");
  print_zone(z);
  output("\n");
  return z;
}

