automata Sender : i = 1..3,
         Medium : j = 1;
clock x[1..total(Sender)], y[1..total(Medium)];
register channel, users, cdetect;

--------------------------------------------------------------------------
automaton Sender[i] {
initially Idle and 
          channel = 0 and      --  channel=0 => free, channel=1 =>in_use
          users = 0 and        --  users => total number of users
          cdetect = 0 and      --  cdetect=0 => no collision detected
          x[i] = 0;            --  x[i] => local clock
  mode Idle {
    invariant True;
    when channel = 0 may goto Ready; x[i] := 0;
  }
  mode Ready{ 
    invariant x[i] = 0;
    when x[i] = 0 and channel = 1 may goto Collision; x[i] := 0;
    when x[i] = 0 and channel = 0 may goto Transmit;
                                  x[i] := 0; users := users + 1;
  }
  mode Collision{
    invariant True;
    when x[i] <= 2 may goto Ready; x[i] := 0;
  }
  mode Transmit{
    invariant True;
    when x[i] <= 10 and cdetect = 1 may goto Collision; 
                                    x[i] := 0; users := users - 1;
    when x[i] = 10 may goto Idle; users := users - 1;
  }
}


automaton Medium[j] {
initially Idle and cdetect = 0 and channel = 0 and users = 0 and y[j] = 0;  
  mode Idle {
    invariant True;
    when users >= 1 may goto Busy; y[j] := 0; channel := 1;
  }
  mode Busy {
    invariant True;
    when y[j] = 1 and users > 1 may goto Collision; 
                                cdetect := 1; y[j] := 0;
    when y[j] >= 1 and users = 0 may goto Idle;
  }
  mode Collision {
    invariant True;
    when users = 0 may goto Idle; cdetect := 0; channel := 0;
  }
}

--verify all_paths ( not {(~(lock = i) and x[i] <= 1) or (lock = 3)}  
--                   and (mode(A[1]) != M1)
--                  until >= 1
--                   {(mode(A[2]) != M3) or (mode(A[1]) = M1)}
--                   and (mode(A[2]) = M3)
--                 )
--       and exists_path (eventually {
--                         (henceforth not {(mode(A[1]) = M4) and 
--                                          (mode(A[2]) = M4)})
--                                   })
--;

--verify all_paths (henceforth not{(mode(A[1]) = M4) and (mode(A[2]) = M4)}
--                         and not{(mode(A[1]) = M4) and (mode(A[3]) = M4)}
--                         and not{(mode(A[2]) = M4) and (mode(A[3]) = M4)}
--                 );

--verify ~(mode(A[1]) = M4 and 
--         ~(mode(A[2]) = M4 or (mode(A[1]) = M3 or mode(A[1])=M4) and
--           mode(A[2]) = M2
--          )
--         or mode(A[2]) = M1
--        )
--        and 
--        (mode(A[2]) != M1) or 
--       ~(mode(A[1]) = M3 or mode(A[2]) = M3);


manipulation

----------------------------------------
--   A user-defined sequential merge  --
----------------------------------------
--    graph g[1..total_automata-1], g[1];
--
--    for(i:=1; i<total_automata; i++) {
--      if(i = 1) {
--        g[1] := merge_graph(A[1], A[2]);
--      }
--      else {
--        g[i] := merge_graph(g[i-1], A[i+1]);
--      }
--      look_ahead(g[i-1], g[i]);
--      shield_clock(g[i]);
--      normalize_region(g[i]);
--    }
--    model_check(g[total_automata-1]);
--    print_graph(g[total_automata-1]);
----------------------------------------

----------------------------------------
--   System-defined sequential merge  --
----------------------------------------
    graph g[0];
    g[0] := merge_graph(Sender[1], Medium[1]);
--    sequential_merge();
--      pairwise_merge();
--    look_ahead();
    read_write();
    shield_clock();
    normalize_region();
--    print_graph();
--    model_check();
----------------------------------------

----------------------------------------
--    if(num_modes(g[1]) < num_modes(g[2])) {
--      model_check(g[1]);
--    }
--    if(num_trans(g[2]) < num_trans(A[1])) {
--      model_check(A[1]);
--    }

--  graph gg, y2, x3, g[1..total_automata-1];
--    for(i:=0; i<total_automata; i++) {
--      if(i=1) {
--       g[1] := merge_graph(A[1], A[2]);
--        look_ahead(g[2], x3);
--        if(i <= 3) {
--          look_ahead(g[2], x3);
--        }
--      }
--      else {
--        x3 = merge_graph(A[1], A[3]);
--        look_ahead();
--        look_ahead(g[1], A[3]);
--      }
--    }
--    for(i:=0; i<total_automata; i++) {
--      g[2] := merge_graph(g[1], y2);
--      for(i=0; i<total_automata; i++) {
--        g[2] := merge_graph(g[1], y2);
--        look_ahead(A[3], y2);
--      }
--      look_ahead(A[3], y2);
--    }
--    look_ahead(A[1], A[3]);
--    model_check(g[2]);
