define(W,1)
define(SEQ,2)
define(DATA,0)
define(ACK,1)
define(NAK,2)
define(SND,3)
define(RCV,4)

automata S, R, SR, RS, U;
register msg, sendseq, rack, recseq, lack, 
         sc_pty, sc_pno, cs_pty, cs_pno, cs_ack,
         rc_pty, rc_pno, rc_ack, cr_pty, cr_pno, cr_ack,
         resnd_pno,
         cs_hck, cs_dck, cr_hck, cr_dck,
         sc_ctl, sc_ack, cs_ctl, rc_ctl, cr_ctl;

automaton S {
initially S_Idle and sendseq = 1 & rack = 0 & lack = 0 & recseq = 0 &
          sc_ctl = SND & cr_ctl = SND & rc_ctl = SND & cs_ctl = SND &
          sc_pno = 0 & cs_pno = 0 & rc_pno = 0 & cr_pno = 0 &
          resnd_pno = 0 & cs_ack = 0 & cr_ack = 0 & rc_ack = 0 & 
          sc_pty = DATA & sc_ack = 0 & cs_pty = DATA & rc_pty = DATA & 
          cr_pty = DATA;
  mode S_Idle {
    invariant True;
    when msg = 1 may goto S_Ready;
    when cs_ctl = RCV may goto S_GetPkt;
  }
  mode S_Ready {
    invariant True;
    when sendseq - rack <= W (mod SEQ) & sendseq != rack (mod SEQ) 
              may goto S_Giving; 
              sc_pty:=DATA; sc_pno:=sendseq; sendseq:=sendseq+1 (mod SEQ);
    when sendseq - rack > W (mod SEQ) may goto S_OutOfWindow;
    when sendseq = rack (mod SEQ) may goto S_OutOfWindow;
  }
  mode S_Giving {
    invariant True;
    when sc_ctl = SND may goto S_Idle;
                      lack:=recseq; sc_ctl:=RCV; msg:=0;
  }
  mode S_OutOfWindow {
    invariant True;
    when True may goto S_TimeOut;
    when cs_ctl = RCV may goto S_GetPkt;
  }
  mode S_TimeOut {
    invariant True;
    when cs_ctl = RCV may goto S_GetPkt;
  }
  mode S_GetPkt {
    invariant True;
    when cs_ctl = RCV & cs_hck = 1 & cs_pty = ACK & cs_ack != sendseq &
         sendseq - cs_ack <= W (mod SEQ) & cs_ack - rack <= W (mod SEQ)
         may goto S_GetPkt;
         rack := cs_ack; cs_ctl := SND;
    when cs_pty != ACK & cs_pty != NAK may goto S_Idle;
    when cs_hck = 1 & cs_pty = NAK & cs_pno != sendseq & 
         sendseq - cs_pno <= W (mod SEQ) & cs_pno - rack <= W (mod SEQ)
         may goto S_ResndNakdPkt;
         resnd_pno := cs_pno;
  }
  mode S_ResndNakdPkt {
    invariant True;
    when sc_ctl = SND may goto S_Idle;
                      sc_pty := DATA; sc_pno := resnd_pno; sc_ack := recseq;
                      sc_ctl := RCV;
  }
}

automaton R {
initially R_Idle and cr_hck = 1 & cr_dck = 1 & 
                     sendseq = 1 & recseq = 0 & lack = 0 &
                     sc_ctl = SND & cr_ctl = SND & rc_ctl = SND & cs_ctl = SND &
                     sc_pno = 0 & cs_pno = 0 & rc_pno = 0 & cr_pno = 0 &
                     cs_ack = 0 & cr_ack = 0 & rc_ack = 0 &
                  sc_pty = DATA & cs_pty = DATA & rc_pty = DATA & cr_pty = DATA;
  mode R_Idle {
    invariant True;
    when cr_ctl = RCV & cr_hck = 1 may goto R_GetPkt; cr_ctl := SND;
  }
  mode R_GetPkt {
    invariant True;
    when cr_pty = NAK may goto R_BugFix;
    when cr_pty = DATA & cr_pno != lack & cr_pno - lack <= W (mod SEQ)
                      may goto R_GetData;
    when cr_pty = ACK may goto R_Idle;
    when cr_pno = lack may goto R_Idle;
    when cr_pno - lack > W (mod SEQ) may goto R_Idle;
  }
  mode R_BugFix {
    invariant True;
    when rc_ctl = SND may goto R_Idle; 
                lack := recseq; rc_ack := recseq;
		rc_pty := ACK; rc_pno := recseq; rc_ctl := RCV;
  }
  mode R_GetData {
    invariant True;
    when True may goto R_Idle;
  }
}

automaton SR {
initially SR_Getting and sc_ctl = SND & cr_ctl = SND;
  mode SR_Getting {
    invariant True;
    when sc_ctl = RCV & cr_ctl = SND may goto SR_Giving;
         cr_ctl := RCV; sc_ctl := SND;
         cr_pty := sc_pty; cr_pno := sc_pno;
  }
  mode SR_Giving {
    invariant True;
    when True may goto SR_Getting;
  }
}

automaton RS {
initially RS_Getting and rc_ctl = SND & cs_ctl = SND;
  mode RS_Getting {
    invariant True;
    when rc_ctl = RCV & cs_ctl = SND may goto RS_Giving;
         rc_ctl := SND; cs_ctl := RCV; cs_pty := rc_pty;
         cs_pno := rc_pno; cs_ack := rc_ack;
  }
  mode RS_Giving {
    invariant True;
    when True may goto RS_Getting;
  }
}

automaton U {
initially U_Snd and msg = 0;
  mode U_Snd {
    invariant True;
    when msg = 0 may goto U_Rcv; msg := 1;
    when msg = 0 may goto U_Rcv;
  }
  mode U_Rcv {
    invariant True;
    when True may goto U_Snd;
  }
}

manipulation
    sequential_merge();
    read_write();
--    shield_variables();
