#define R 2
#define A 3
#define W 10
#define MAX_RCR 5
#define AEC_MAX 5

#define STATUS_RUNNING	0
#define STATUS_COMPLETE	1
#define	STATUS_ABORT	2

process count = 3;

local clock x;
local discrete status; 

global discrete RCR, AEC;
global discrete TIDOK, tidok_sent; 

global synchronizer i_send_invoke, i_send_ack, i_send_abort, i_rcv_ack, i_rcv_abort;
global synchronizer r_rcv_invoke, r_rcv_ack, r_rcv_abort, r_send_ack,
r_send_abort;

/* automaton: network */

mode network_initial true {
  when ?i_send_invoke true may goto network_i_invoke; 	/* 1 */ 
  when ?i_send_ack true may goto network_i_ack;		/* 2 */ 
  when ?i_send_abort true may goto network_i_abort;	/* 3 */ 
  when ?r_send_ack true may goto network_r_ack;		/* 4 */ 
  when ?r_send_abort true may goto network_r_abort;	/* 5 */ 
}

/* we will never lost any packet, though it might be delayed */

mode network_i_invoke true {
  when !r_rcv_invoke true may goto network_initial;	/* 6 */ 
}


mode network_i_ack true {
  when !r_rcv_ack true may goto network_initial;	/* 7 */ 
}

mode network_i_abort true { 
  when !r_rcv_abort true may goto network_initial;	/* 8 */ 
}

mode network_r_ack true { 
  when !i_rcv_ack true may goto network_initial;		/* 9 */ 
}

mode network_r_abort true { 
  when !i_rcv_abort true may goto network_initial;	/* 10 */ 
}



/* automaton: initializer */

mode sender_init_null true {
  /* ignore event TR-Invoke.req */
  when !i_send_invoke true may RCR:= 0; x:= 0; TIDOK:=0; tidok_sent:=0;
				status:= STATUS_RUNNING; 
				goto sender_init_result_wait;		/* 11 */ 
}

mode sender_init_result_wait true {
  /* ignore event TR-Abort.req */
  when !i_send_abort true may status:= STATUS_ABORT; goto sender_init_null;	/* 12 */ 

  /* goto sub-mode while event RcvAck occurs */
  when ?i_rcv_ack true may goto sender_init_result_wait_ack;			/* 13 */ 

  /* ignore action TR-Abort.ind */
  when ?i_rcv_abort true may status:= STATUS_ABORT; goto sender_init_null; 	/* 14 */ 

  /* TimerTo_R */
  when !i_send_invoke x >= R and RCR < MAX_RCR and tidok_sent = 0 
  may RCR++1; x:= 0;								/* 15 */ 

  /* ignore TIDok sent with Ack */
  when !i_send_ack  x >= R and RCR < MAX_RCR and tidok_sent = 1 
  may RCR++1; x:= 0;								/* 16 */ 

  /* ignore indication */
  when x >= R and RCR = MAX_RCR may status:= STATUS_ABORT; goto sender_init_null; /* 17 */ 
}

mode sender_init_result_wait_ack true {
  /* ignore action Stop Timer and TR-Abort.cnf */
  when TIDOK = 0 may status:= STATUS_COMPLETE; goto sender_init_null;		/* 18 */ 

  /* doing TIDve */
  when !i_send_ack TIDOK = 1 may tidok_sent:= 1; RCR++1; x:= 0; 	/* 19 */ 
  goto sender_init_result_wait;

} 




mode rcvr_resp_listen true {
  /* simplification of Invalid-TID-test */
  when ?r_rcv_invoke true may status:= STATUS_RUNNING; 
				goto rcvr_resp_listen_invoke; /* 20 */ 

  /* ignore TR-Invoke.ind */
  when ?r_rcv_invoke true may AEC:=0; x:=0; status:= STATUS_RUNNING; 
				goto rcvr_resp_invoke_resp_wait; /* 21 */ 

}

mode rcvr_resp_listen_invoke true {
  when !r_send_ack true may TIDOK:=1; goto rcvr_resp_tidok_wait;	/* 22 */ 
}

mode rcvr_resp_tidok_wait true {
  /* ignore TR-Invoke.ind */
  when ?r_rcv_ack true may TIDOK:=0; AEC:=0; x:=0; goto rcvr_resp_invoke_resp_wait; /* 23 */ 

  when ?r_rcv_abort true may status:= STATUS_ABORT; goto rcvr_resp_listen;	/* 24 */ 

  /* goto sub-mode while event RcvInvoke occurs */
  when ?r_rcv_invoke true may goto rcvr_resp_tidok_wait_invoke;	/* 25 */ 
}

mode rcvr_resp_tidok_wait_invoke true { 
  /* ignore RID problem */
  when !r_send_ack true may goto rcvr_resp_tidok_wait; 		/* 26 */ 

} 

mode rcvr_resp_invoke_resp_wait true {

/*
EVENT:		TR-Invoke.res
CONDITION:	(ExitInfo)
ACTION:		Queue(A) Ack PDU (with InfoTPI)
		Start timer, W
NEXT STATE:	INVOKE RESP WAIT
*/
when !r_send_ack true may x:= 0; goto rcvr_resp_wait_timeout; /* 27 */ 

/* ignore event TR-Abort.req */
when !r_send_abort true may status:= STATUS_ABORT; goto rcvr_resp_listen; /* 28 */ 

/* ignore event RcvInvoke */
when ?r_rcv_invoke true may ; /* 29 */ 

/* TimerTO_A */
when x >= A and AEC < AEC_MAX may AEC++1; x := 0;  /* 30 */ 

/*
when !r_send_abort x >= A and AEC = AEC_MAX may status:= STATUS_ABORT; 
goto rcvr_resp_listen;
*/

when !r_send_ack x >= A and AEC = AEC_MAX may
goto rcvr_resp_wait_timeout;				/* 31 */ 

}

mode rcvr_resp_wait_timeout true {

/* goto sub-mode while event RcvInvoke occurs */
when ?r_rcv_invoke true may goto rcvr_resp_wait_timeout_invoke;	/* 32 */ 

/* ignore T-TRAbort.ind */
when ?r_rcv_abort true may status:= STATUS_ABORT; goto rcvr_resp_listen; /* 33 */ 

/* Transaction succeed! Clear Transaction */
when x >= W and true may status:= STATUS_COMPLETE; goto rcvr_resp_listen; /* 34 */ 

/* ignore event TR-Abort.req */
when !r_send_abort true may status:= STATUS_ABORT; goto rcvr_resp_listen;	/* 35 */ 
}

mode rcvr_resp_wait_timeout_invoke true {

when !r_send_ack true may goto rcvr_resp_wait_timeout;/* 36  */ 

}

initially AEC = 0 and RCR = 0
and network_initial[1] and x[1] = 0 
and sender_init_null[2] and x[2] = 0 and status[2] = STATUS_RUNNING 
and rcvr_resp_listen[3] and x[3] = 0 and status[3] = STATUS_RUNNING ;


risk
status[2] = STATUS_COMPLETE and status[3] = STATUS_ABORT; 





