#include "global.h"
#include "print.h"
#include "modeman.h"
#include "bddman.h"
#include "zoneman.h"
#include "back.h"
#include "reduce.h"

int total_mode_mu = 0;
int mode_mu_count = 0;
int mm_trans_count = 0;
struct thread_type *BT_THREAD = 0;
struct a23tree_type *mode_mu_tree;
struct mode_mu_link_type *destination_cycle = NULL, *source_cycle = NULL;
struct mode_mu_type *mmlist;

int gen_btgraph(struct thread_type *t);
/*
int back_reduce(struct thread_type *t);
*/
struct mode_mu_type *alloc_mode_mu();
int free_mmlist(struct mode_mu_type *);
int is_initial_mode(struct mode_type *);
int back_path(struct mode_type *, struct transition_type *);
int consistent_inv_mu(struct mode_type *, struct clkconj_type *, bdd);
struct mode_mu_link_type *
       append_mm(struct mode_mu_type *, struct mode_mu_link_type *);
int back_dfs(struct mode_mu_type *);
int calculate_weight(struct mode_type *, struct mode_type *);
struct mode_link_type *copy_mode_link_cycle(struct mode_link_type *);
struct mode_link_type *insert_mode(struct mode_type *, struct mode_link_type *);
int exist_mode(struct mode_type *, struct mode_link_type *);
int print_mode_link_cycle(struct mode_link_type *);
int mode_mu_comp(struct mode_mu_type *, struct mode_mu_type *);
struct mode_link_type *delete_mode(struct mode_type *, struct mode_link_type *);
int free_mode_mu(struct mode_mu_type *);
int mode_mu_count_inc(struct mode_mu_type *);
int mode_mu_parent_set(struct mode_mu_type *, struct a23tree_type *);
int print_mode_mu(struct mode_mu_type *);
struct clkconj_type *back_delta(struct clkconj_type *, int);
int initial_consistent(struct mode_mu_type *);
struct clkconj_type *reset_clk_assign(struct clkconj_type *, 
                                      struct assign_type *);
bdd reset_prp_assign(bdd, struct assign_type *);
struct clkconj_type *weakest_clk_precondition(struct transition_type *,
                                              struct mode_mu_type *,
                                              struct clkconj_type *);
bdd weakest_prp_precondition(struct transition_type *, struct mode_mu_type *);
int remove_dormant_cycles(struct thread_type *);

struct clkconj_type * back_clkconj_and(struct clkconj_type *, 
                                       struct clkconj_type *);
int back_compatible(int, int, int, int, int *, int *);

/**** FUNCTIONS ****/

int gen_btgraph(struct thread_type *t) {
  struct mode_type *M;
  struct transition_type *e;
  struct trans_link_type *elim_cycle = NULL, *tl;

  BT_THREAD = t;
  printf("    Generating BT-graph for thread %d using backward search.\n", 
         t->index);
  if(verbose >= 0) {
    sprintf(tmpstring,
            "\nGenerating BT-graph for thread %d using backward search:\n", 
            t->index);
    output(tmpstring);
  }

  for(M=t->starting_mode; M; M=M->cnext) {
    if(M->out_trans_cycle) {
      for(e=M->out_trans_cycle->next; ; e=e->next) {
        init_23tree(&(mode_mu_tree));
        free_mmlist(mmlist); 
        mmlist = (struct mode_mu_type *)NULL;
        total_mode_mu = 0;
        mode_mu_count = 0;
        mm_trans_count = 0;
        full_back_path(M, e);         /* will never become valid */
        free_23tree(mode_mu_tree);
/*
        print_mmlist(mmlist);
*/
        if(btgraph_kclosure(mmlist) == TRUE) {
          tl=(struct trans_link_type *)malloc(sizeof(struct trans_link_type));
          tl->transition = e;
          if(elim_cycle) {
            tl->next = elim_cycle->next;
            elim_cycle->next = tl;
            elim_cycle = tl;
          }
          else
            elim_cycle = tl->next = tl;
        }
        if(e==M->out_trans_cycle) break;
      }
    }
    if(elim_cycle) {
      for(tl = elim_cycle->next; ; tl=tl->next) {
        e = tl->transition;
        eliminate_transition(e, M);

        if(verbose >= 0) {
          sprintf(tmpstring, 
                  "ELIMINATED transition %d from mode %s (%d) to %s (%d).\n",
                  e->index, M->name, M->index, e->mode->name, e->mode->index);
          output(tmpstring);
        }

        t->trans_count--;
        if(BDD_AND(e->trigger_bdd, M->dval_bdd) == BDD_FALSE)
          t->invalid_trans_count--;
        if(tl == elim_cycle) break;
      }
    }
    free_trans_link_cycle(&elim_cycle);
  }
  printf("\tRemoving irreachable modes in thread %d.\n\t", t->index);
  remove_dormant_cycles(t);
  remove_irreachable_mode(t);
  print_counts(t->index);
}

/*
int back_reduce(struct thread_type *t) {
  struct mode_type *M;
  struct transition_type *e;
  struct trans_link_type *elim_cycle = NULL, *tl;

  printf("Reducing thread %d using backward search.\n", t->index);
  if(verbose >= 0) {
    sprintf(tmpstring,
            "\nReducing thread %d using backward search:\n", t->index);
    output(tmpstring);
  }

  for(M=t->starting_mode; M; M=M->cnext) {
    if(M->out_trans_cycle) {
      for(e=M->out_trans_cycle->next; ; e=e->next) {
        init_23tree(&(mode_mu_tree));
        if(back_path(M, e) == FALSE){         
          tl=(struct trans_link_type *)malloc(sizeof(struct trans_link_type)); 
          tl->transition = e;
          if(elim_cycle) {
            tl->next = elim_cycle->next;
            elim_cycle->next = tl;
            elim_cycle = tl;
          }
          else 
            elim_cycle = tl->next = tl;
        }
        free_23tree(mode_mu_tree);
        if(e==M->out_trans_cycle) break;
      }
    }
    if(elim_cycle) {
      for(tl = elim_cycle->next; ; tl=tl->next) {
        e = tl->transition;
        eliminate_transition(e, M);

        if(verbose >= 0) {
          sprintf(tmpstring, 
                  "ELIMINATED transition %d from mode %s (%d) to %s (%d).\n",
                  e->index, M->name, M->index, e->mode->name, e->mode->index);
          output(tmpstring);
        }

        t->trans_count--;
        t->invalid_trans_count--;
        if(tl == elim_cycle) break;
      }
    }
    free_trans_link_cycle(&elim_cycle);
  }
  printf("\tRemoving irreachable modes in thread %d.\n", t->index);
  remove_dormant_cycles(t);
  remove_irreachable_mode(t);
  print_counts(t->index);
}
*/

int free_trans_link_cycle(struct trans_link_type **trans_link_cycle) {
  struct trans_link_type *x, *tl;

  if(*trans_link_cycle) {
    x = (*trans_link_cycle)->next; /* break the cycle into a list */
    (*trans_link_cycle)->next = NULL;
    tl = x;
    while(tl) {
      x = tl->next;
      free(tl);
      tl = x;
    }
    *trans_link_cycle = NULL;
  }
}

struct mode_mu_type *alloc_mode_mu() {
  struct mode_mu_type *mm; 
  mm = (struct mode_mu_type *)malloc(sizeof(struct mode_mu_type));
  mm->index = 0;
  mm->visited = 0;
  mm->count = 0;
  mm->parent = (struct a23tree_type *)NULL;
  mm->Mode = (struct mode_type *)NULL;
  mm->mode_cycle = (struct mode_link_type *)NULL;
  mm->mu_clk = (struct clkconj_type *)NULL;
  mm->mu_prp = (bdd)NULL;
  mm->out_mm_trans_cycle = (struct mm_transition_type *)NULL;
  mm->in_mm_trans_cycle = (struct mm_transition_type *)NULL;
  mm->next = (struct mode_mu_type *)NULL;
  return mm;
}

struct mode_mu_type *mm_prepend(struct mode_mu_type *mm, 
                                struct mode_mu_type *list) {
  total_mode_mu++;
  if(list) {
    mm->next = list;
    list = mm;
  }
  else {
    mm->next = NULL;
    list = mm;
  }
  return list;
}

int free_mmlist(struct mode_mu_type *list) {
  struct mode_mu_type *mm, *mms;
  if(list) {
    for(mm = list, mms=list->next; mm; mm=mms, mms=mms->next) {
      free_mode_mu(mm);
      if(mms == NULL) break;
    }
  }
  return;
}

int is_initial_mode(struct mode_type *M) {
  char *tmp_name = str_preserve(M->name);
  char *tc;
  tc = strtok(tmp_name, ".");
  while(tc) {
    if(strcmp(tc, thread[total_automata-1].starting_mode->name)) {
      free(tmp_name);
      return FALSE;
    }
    tc = strtok((char *)NULL, ".");
  }
  free(tmp_name);
  return TRUE;
}

int full_back_path(struct mode_type *M, struct transition_type *e) {
  bdd mu_prp, tmp_bdd;
  struct clkconj_type *mu_clk;
  struct mode_type *m, *Mm;
  struct mode_mu_type *x, *nx;
  struct mode_link_type *S = NULL;  /* a set of modes in the form of a cycle */
  struct mode_mu_link_type *mml = NULL;

  if(verbose >= 0) {
    sprintf(tmpstring, 
          "\nConsidering mode %s (%d) and transition %d to mode %s (%d).\n", 
          M->name, M->index, e->index, e->mode->name, e->mode->index);
    output(tmpstring);
  }

  mu_prp = BDD_AND(M->boolexp->boolexp_bdd, e->trigger_bdd);
  tmp_bdd = BDD_AND(mu_prp, M->dval_bdd);
 
  if(tmp_bdd != BDD_FALSE) {
    if(verbose >= 0)
      output("\tA valid transition.\n");
    return TRUE;
  }

  mu_clk = clkconj_and(M->xconj_cycle, e->clk_trigger_cycle);
  if(is_initial_mode(M)) return TRUE;

  for(m = thread[total_automata-1].starting_mode; m; m = m->cnext) {

    if(verbose >= 0) {
      sprintf(tmpstring, 
              "    Considering A^1 mode %s (%d).\n", m->name, m->index);
      output(tmpstring);
    }

    if(consistent_inv_mu(m, mu_clk, mu_prp)) {

      if(verbose >= 0) {
        output("\t... consistent mode.invariance and mu.\n");
      }

      S = (struct mode_link_type *)malloc(sizeof(struct mode_link_type));
      S->mode = m;
      S->next = S;      /* represent S = {m} as a cycle */

      x = alloc_mode_mu();
      x->index = ++mode_mu_count;
      x->Mode = M;
      x->mode_cycle = S;
      x->mu_clk = mu_clk;
      x->mu_prp = mu_prp;
      x->visited = FALSE;

      nx = (struct mode_mu_type *)
            add_23tree(x, mode_mu_tree, mode_mu_comp, free_mode_mu,
                       mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

      if(nx == x) {
        if(verbose >= 0) {
          sprintf(tmpstring,
                  "\tCREATED new mode_mu(%d) <%s (%d), ", 
                  x->index, x->Mode->name, x->Mode->index);
          output(tmpstring);
          print_mode_link_cycle(x->mode_cycle);
          output(">.\n");
        }
      }
      else if(nx->visited == TRUE) {
        if(verbose >= 0) {
          sprintf(tmpstring, "\tFOUND visited mode_mu(%d) <%s (%d), ",
                  nx->index, nx->Mode->name, nx->Mode->index);
          output(tmpstring);
          print_mode_link_cycle(nx->mode_cycle);
          output(">.\n");
        }
        mode_mu_count--;
        continue;
      }
      else {
        if(verbose >= 0) {
          sprintf(tmpstring, "\tFOUND unvisited mode_mu(%d) <%s (%d), ",
                  nx->index, nx->Mode->name, nx->Mode->index);
          output(tmpstring);
          print_mode_link_cycle(nx->mode_cycle);
          output(">.\n");
        }
        mode_mu_count--;
      }
      mmlist = mm_prepend(x, mmlist);
      destination_cycle = append_mm(x, destination_cycle);
      back_dfs(x);
    }
    else { 
      if(verbose >= 0) {
        output("\t... inconsistent mode.invariance and mu.\n");
      }
    }
  }
  print_mmlist(mmlist);
  return FALSE;
}

/*
int back_path(struct mode_type *M, struct transition_type *e) {
  bdd mu_prp, tmp_bdd;
  struct clkconj_type *mu_clk;
  struct mode_type *m, *Mm;
  struct mode_mu_type *x, *nx;
  struct mode_link_type *S = NULL;  
  struct mode_mu_link_type *mml = NULL;

  if(verbose >= 0) {
    sprintf(tmpstring, 
          "\nConsidering mode %s (%d) and transition %d to mode %s (%d).\n", 
          M->name, M->index, e->index, e->mode->name, e->mode->index);
    output(tmpstring);
  }

  mu_prp = BDD_AND(M->boolexp->boolexp_bdd, e->trigger_bdd);
  tmp_bdd = BDD_AND(mu_prp, M->dval_bdd);
 
  if(tmp_bdd != BDD_FALSE) {
    if(verbose >= 0)
      output("\tA valid transition.\n");
    return TRUE;
  }

  mu_clk = clkconj_and(M->xconj_cycle, e->clk_trigger_cycle);
  if(is_initial_mode(M)) return TRUE;

  for(m = thread[total_automata-1].starting_mode; m; m = m->cnext) {

    if(verbose >= 0) {
      sprintf(tmpstring, 
              "    Considering A^1 mode %s (%d).\n", m->name, m->index);
      output(tmpstring);
    }

    if(consistent_inv_mu(m, mu_clk, mu_prp)) {

      if(verbose >= 0) {
        output("\t... consistent mode.invariance and mu.\n");
      }

      S = (struct mode_link_type *)malloc(sizeof(struct mode_link_type));
      S->mode = m;
      S->next = S;      

      x = alloc_mode_mu();
      x->index = ++mode_mu_count;
      x->Mode = M;
      x->mode_cycle = S;
      x->mu_clk = mu_clk;
      x->mu_prp = mu_prp;
      x->visited = FALSE;

      nx = (struct mode_mu_type *)
            add_23tree(x, mode_mu_tree, mode_mu_comp, free_mode_mu,
                       mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

      if(nx == x) {
        if(verbose >= 0) {
          sprintf(tmpstring,
                  "\tCREATED new mode_mu(%d) <%s (%d), ", 
                  x->index, x->Mode->name, x->Mode->index);
          output(tmpstring);
          print_mode_link_cycle(x->mode_cycle);
          output(">.\n");
        }
      }
      else if(nx->visited == TRUE) {
        if(verbose >= 0) {
          sprintf(tmpstring, "\tFOUND visited mode_mu(%d) <%s (%d), ",
                  nx->index, nx->Mode->name, nx->Mode->index);
          output(tmpstring);
          print_mode_link_cycle(nx->mode_cycle);
          output(">.\n");
        }
        mode_mu_count--;
        continue;
      }
      else {
        if(verbose >= 0) {
          sprintf(tmpstring, "\tFOUND unvisited mode_mu(%d) <%s (%d), ",
                  nx->index, nx->Mode->name, nx->Mode->index);
          output(tmpstring);
          print_mode_link_cycle(nx->mode_cycle);
          output(">.\n");
        }
        mode_mu_count--;
      }

      mmlist = mm_prepend(x, mmlist);
      destination_cycle = append_mm(x, destination_cycle);
      if(back_dfs(x) == TRUE) {
        return TRUE;
      }
    }
    else { 
      if(verbose >= 0) {
        output("\t... inconsistent mode.invariance and mu.\n");
      }
    }
  }
  return FALSE;
}
*/

int consistent_inv_mu(struct mode_type *m, struct clkconj_type *mu_clk,
                      bdd mu_prp) {
  struct clkconj_type *c;
  bdd b;

  c = clkconj_and(m->xconj_cycle, mu_clk);
  if(c->op == FALSE) {
    free_clkconj(c);
/*
    output("inconsistent clock constraints\n");
*/
    return FALSE;
  }
  free_clkconj(c);

  b = BDD_AND(m->boolexp->boolexp_bdd, mu_prp);
  if(b == BDD_FALSE) {
    bdd_free(bddm, b);
/*
    output("inconsistent proposition BDDs\n");
*/
    return FALSE;
  }
  bdd_free(bddm, b);

  return TRUE;
}

struct mode_mu_link_type *
append_mm(struct mode_mu_type *mm, struct mode_mu_link_type *mm_cycle) {
  struct mode_mu_link_type *mml = 
         (struct mode_mu_link_type *)malloc(sizeof(struct mode_mu_link_type));

  mml->mode_mu = mm;
  if(mm_cycle) {
    mml->next = mm_cycle->next;
    mm_cycle->next = mml;
    mm_cycle = mml;
  }
  else
    mm_cycle = mml->next = mml;
  return mm_cycle;
}

int back_dfs(struct mode_mu_type *x) {
  int wt = 0;
  struct clkconj_type *mu1, *mu_clk_before, *mu_clk_after, *tmp_clkconj;
  bdd mu_prp_before, mu_prp_after;
  struct transition_type *f;
  struct mode_type *cm, *M, *m, *n;
  struct mode_mu_type *y, *ny, *z, *nz;
  struct mode_link_type *ml, *S, *S1;

  if(x->visited == TRUE) {
    if(verbose >= 0) {
      output("\tVISITED!\n");
    }
    return FALSE;
  }
  x->visited = TRUE;
  if(verbose >= 1) {
    sprintf(tmpstring, "\tVISITED mode_mu <%s (%d), ", 
            x->Mode->name, x->Mode->index);
    output(tmpstring);
    print_mode_link_cycle(x->mode_cycle);
    output(">.\n");
  }
  
  for(ml = x->mode_cycle->next; ; ml=ml->next) {
    m = ml->mode;
    M = x->Mode;

    if(verbose >= 0) {
      sprintf(tmpstring, 
              "\tBackward searching from <%s (%d), %s (%d)> with: \n", 
              M->name, M->index, m->name, m->index);
      output(tmpstring);
      output("\tmode set: ");
      print_mode_link_cycle(x->mode_cycle);
      output("\n\tclock constraint: ");
      print_clkconj(x->mu_clk);
      output("\n");
      output("\tbdd: ");
      print_bdd(x->mu_prp);
    }

    mu1 = back_delta(x->mu_clk, FALSE);
    mu1 = conjoin(mu1, DELTA);

    if(m->in_trans_cycle) {
      for(f=m->in_trans_cycle->next; ; f=f->next) {

        if(verbose >= 0) {
          sprintf(tmpstring, 
       "\tConsidering transition (%d) from A^1 mode %s (%d) to mode %s (%d).\n",
                 f->index, f->mode->name, f->mode->index, m->name, m->index);
          output(tmpstring);
        }

        mu_clk_before = weakest_clk_precondition(f, x, mu1);
        mu_prp_before = weakest_prp_precondition(f, x);

        if(verbose >= 1) {
          output("\t    mu_clk_before: "); 
          print_clkconj(mu_clk_before);
          output("\n\t    mu_prp_before: "); 
          print_bdd(mu_prp_before);
        }

        tmp_clkconj = clkconj_copy(mu_clk_before);
        mu_clk_after = assign(tmp_clkconj, f->assign_cycle);
        if(f->permutation) {
          tmp_clkconj= permute_clkconj(mu_clk_after, f->permutation->perm);
          free_clkconj(mu_clk_after);
          mu_clk_after = tmp_clkconj;
        }
        tmp_clkconj = clkconj_and(mu_clk_after, mu1);
        free_clkconj(mu_clk_after);
        mu_clk_after = tmp_clkconj;

        if(verbose >= 1) {
          output("\t    mu_clk_after: "); 
          print_clkconj(mu_clk_after);
          output("\n");
        }

        tmp_clkconj = clkconj_and(mu_clk_after, x->mu_clk);
        if(tmp_clkconj->op == FALSE) {
          free_clkconj(tmp_clkconj);
          free_clkconj(mu_clk_before);
          free_clkconj(mu_clk_after);
          if(verbose >= 0)
            output("\tmu_clk_after is INCONSISTENT with mu_clk.\n");
          if(f==m->in_trans_cycle) break;
          else continue;
        }
        free_clkconj(tmp_clkconj);

        mu_prp_after = bdd_assign(mu_prp_before, f->assign_cycle);

        if(verbose >= 1) {
          output("\t    mu_prp_after: "); 
          print_bdd(mu_prp_after);
        }

        if(BDD_AND(mu_prp_after, x->mu_prp) == BDD_FALSE) {
          if(verbose >= 0)
            output("\t    mu_prp_after is INCONSISTENT with mu_prp.\n");
          free_clkconj(mu_clk_before);
          free_clkconj(mu_clk_after);
          if(f==m->in_trans_cycle) break;
          else continue;
        }

        S = copy_mode_link_cycle(x->mode_cycle);
        S = delete_mode(m, S);
        S = insert_mode(f->mode, S);

        y = alloc_mode_mu();
        y->index = ++mode_mu_count;
        y->Mode = x->Mode;
        y->mode_cycle = S;
        y->mu_clk = clkconj_copy(mu_clk_before);
        y->mu_prp = mu_prp_before;
        y->visited = TRUE;

        if(initial_consistent(y)) {
          source_cycle = append_mm(y, source_cycle);
          if(verbose >= 0) {
            output("\tINITIAL modes consistent reached (by A^1 transition).\n");
          }
          ny = (struct mode_mu_type *)
               add_23tree(y, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                          mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);
  
          wt = calculate_weight(f->mode, m);
          add_mm_transition(ny, x, f->mode, m, wt);

          if(ny == y) {
            mmlist = mm_prepend(y, mmlist);
            if(verbose >= 0) {
              sprintf(tmpstring, 
                      "\tCREATED new mode_mu(%d) <%s (%d), ", 
                      y->index, y->Mode->name, y->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(y->mode_cycle);
              output(">.\n");
            }
          }
          else if(ny->visited == TRUE) {
            if(verbose >= 0) {
              sprintf(tmpstring, "\tFOUND visited mode_mu(%d) <%s (%d), ", 
                      ny->index, ny->Mode->name, ny->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(ny->mode_cycle);
              output(">.\n");
            }
            mode_mu_count--;
            if(f==m->in_trans_cycle) break;
            else continue;
          }
          else {
            if(verbose >= 0) {
              sprintf(tmpstring, "\tFOUND unvisited mode_mu(%d) <%s (%d), ", 
                      ny->index, ny->Mode->name, ny->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(ny->mode_cycle);
              output(">.\n");
            }
            mode_mu_count--;
          }
          return TRUE;
        }

        for(n = thread[total_automata-1].starting_mode; n; n = n->cnext) {
          if(n->index == f->mode->index) continue;
          if(consistent_inv_mu(n, mu_clk_before, mu_prp_before)) {
          
            if(verbose >= 0) {
              sprintf(tmpstring, 
          "\tJumping to next consistent A^1 mode %s (%d) from mode %s (%d).\n", 
                      n->name, n->index, f->mode->name, f->mode->index);
              output(tmpstring);
            }

            S1 = copy_mode_link_cycle(S);
            S1 = insert_mode(n, S1);
 
            z = alloc_mode_mu();
            z->index = ++mode_mu_count;
            z->Mode = M;
            z->mode_cycle = S1;
            z->mu_clk = mu_clk_before;
            z->mu_prp = mu_prp_before;
            z->visited = FALSE;

            nz = (struct mode_mu_type *)
                 add_23tree(z, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                         mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

            wt = calculate_weight(f->mode, m);
            add_mm_transition(nz, x, f->mode, m, wt);
/*
            wt = calculate_weight(n, m);
            add_mm_transition(nz, x, n, m, wt);
*/
            if(nz == z) {
              mmlist = mm_prepend(z, mmlist);
              if(verbose >= 0) {
                sprintf(tmpstring,
                        "\tCREATED new mode_mu(%d) <%s (%d), ", 
                        z->index, z->Mode->name, z->Mode->index);
                output(tmpstring);
                print_mode_link_cycle(z->mode_cycle);
                output(">.\n");
              }
            }
            else if(nz->visited == TRUE) {
              if(verbose >= 0) {
                sprintf(tmpstring, "\tFOUND visited mode_mu(%d) <%s (%d), ",
                        nz->index, nz->Mode->name, nz->Mode->index);
                output(tmpstring);
                print_mode_link_cycle(nz->mode_cycle);
                output(">.\n");
              }
              mode_mu_count--;
              continue;
            }
            else {
              if(verbose >= 0) {
                sprintf(tmpstring, "\tFOUND unvisited mode_mu(%d) <%s (%d), ",
                       nz->index, nz->Mode->name, nz->Mode->index);
                output(tmpstring);
                print_mode_link_cycle(nz->mode_cycle);
                output(">.\n");
              }
              mode_mu_count--;
            }

            if(back_dfs(nz) == TRUE) {
              return TRUE;
            }
          }
        } /* end for(n = ...) */
        if(f==m->in_trans_cycle) break;
      }
    }


    if(M->in_trans_cycle) {
      for(f=M->in_trans_cycle->next; ; f=f->next) {

        if(verbose >= 0) {
          sprintf(tmpstring, 
      "\tConsidering transition (%d) from A^k mode %s (%d) to mode %s (%d) .\n",
                  f->index, f->mode->name, f->mode->index, M->name, M->index);
          output(tmpstring);
        }

        mu_clk_before = weakest_clk_precondition(f, x, mu1);
        mu_prp_before = weakest_prp_precondition(f, x);

        if(verbose >= 1) {
          output("\t    mu_clk_before: "); 
          print_clkconj(mu_clk_before);
          output("\n");
          output("\t    mu_prp_before: "); 
          print_bdd(mu_prp_before);
        }

        tmp_clkconj = clkconj_copy(mu_clk_before);
        mu_clk_after = assign(tmp_clkconj, f->assign_cycle);

        if(f->permutation) {
          tmp_clkconj= permute_clkconj(mu_clk_after, f->permutation->perm);
          free_clkconj(mu_clk_after);
          mu_clk_after = tmp_clkconj;
        }

        tmp_clkconj = clkconj_and(mu_clk_after, mu1);
        free_clkconj(mu_clk_after);
        mu_clk_after = tmp_clkconj;

        if(verbose >= 0) {
          output("\t    mu_clk_after: "); 
          print_clkconj(mu_clk_after);
          output("\n");
        }

        tmp_clkconj = clkconj_and(mu_clk_after, x->mu_clk);
        if(tmp_clkconj->op == FALSE) {
          free_clkconj(tmp_clkconj);
          if(verbose >= 0)
            output("\t    mu_clk_after is INCONSISTENT with mu_clk.\n");
          if(f==M->in_trans_cycle) break;
          else continue;
        }
        free_clkconj(tmp_clkconj);

        mu_prp_after = bdd_assign(mu_prp_before, f->assign_cycle);

        if(verbose >= 1) {
          output("\t    mu_prp_after: "); 
          print_bdd(mu_prp_after);
        }

        if(BDD_AND(mu_prp_after, x->mu_prp) == BDD_FALSE) {
          if(verbose >= 0)
            output("\t    mu_prp_after is INCONSISTENT with mu_prp.\n");
          if(f==M->in_trans_cycle) break;
          else continue;
        }

        S = copy_mode_link_cycle(x->mode_cycle);
        y = alloc_mode_mu();
        y->index = ++mode_mu_count;
        y->Mode = f->mode;
        y->mode_cycle = S;
        y->mu_clk = mu_clk_before;
        y->mu_prp = mu_prp_before;
        y->visited = TRUE;

        if(initial_consistent(y)) {
          source_cycle = append_mm(y, source_cycle);
          if(verbose >= 0) {
            output("\tINITIAL consistent modes reached (by A^k transition).\n");
          }
          ny = (struct mode_mu_type *)
                add_23tree(y, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                          mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

          add_mm_transition(ny, x, f->mode, M, 0);

          if(ny == y) {
            mmlist = mm_prepend(y, mmlist);
            if(verbose >= 0) {
              sprintf(tmpstring, "\tCREATED new A^k mode_mu(%d) <%s (%d), ", 
                      ny->index, ny->Mode->name, ny->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(ny->mode_cycle);
              output(">.\n");
            }
          }
          else if(ny->visited == TRUE) {
            if(verbose >= 0) {
              sprintf(tmpstring, "\tFOUND visited A^k mode_mu(%d) <%s (%d), ", 
                      ny->index, ny->Mode->name, ny->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(ny->mode_cycle);
              output(">.\n");
            }
            mode_mu_count--;
            if(f==M->in_trans_cycle) break;
            else continue;
          }
          else {
            if(verbose >= 0) {
              sprintf(tmpstring, "\tFOUND unvisited A^k mode_mu(%d) <%s (%d), ", 
                      ny->index, ny->Mode->name, ny->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(ny->mode_cycle);
              output(">.\n");
            }
            mode_mu_count--;
          }
          return TRUE;
        }


        if(consistent_inv_mu(f->mode, mu_clk_before, mu_prp_before)) {

          if(verbose >= 1) {
            output("\tIn-trans-mode is consistent with mu_before.\n");
          }

          S = copy_mode_link_cycle(x->mode_cycle);

          z = alloc_mode_mu();
          z->index = ++mode_mu_count;
          z->Mode = f->mode;
          z->mode_cycle = S;
          z->mu_clk = mu_clk_before;
          z->mu_prp = mu_prp_before;
          z->visited = FALSE;

          nz = (struct mode_mu_type *)
               add_23tree(z, mode_mu_tree, mode_mu_comp, free_mode_mu, 
                          mode_mu_count_inc, mode_mu_parent_set, print_mode_mu);

          add_mm_transition(nz, x, f->mode, M, 0);

          if(nz == z) {
            mmlist = mm_prepend(z, mmlist);
            if(verbose >= 0) {
              sprintf(tmpstring,
                      "\tCREATED new A^k mode_mu(%d) <%s (%d), ", 
                      z->index, z->Mode->name, z->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(z->mode_cycle);
              output(">.\n");
            }
          }
          else if(nz->visited == TRUE) {
            if(verbose >= 0) {
              sprintf(tmpstring, "\tFOUND visited A^k mode_mu(%d) <%s (%d), ",
                      nz->index, nz->Mode->name, nz->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(nz->mode_cycle);
              output(">.\n");
            }
            mode_mu_count--;
            if(f==M->in_trans_cycle) break;
            else continue;
          }
          else {
            if(verbose >= 0) {
              sprintf(tmpstring, "\tFOUND unvisited A^k mode_mu(%d) <%s (%d), ",
                      nz->index, nz->Mode->name, nz->Mode->index);
              output(tmpstring);
              print_mode_link_cycle(nz->mode_cycle);
              output(">.\n");
            }
            mode_mu_count--;
          }

          if(back_dfs(z) == TRUE) {
            return TRUE;
          }
        }
        else
          if(verbose >= 1)
            output("\tIn-trans-mode is inconsistent with mu_before.\n");

        if(f==M->in_trans_cycle) break;
      } /* end for(n = ...) */
    }

    if(ml == x->mode_cycle) {
      if(verbose >= 0) {
        sprintf(tmpstring, 
                "NO PATH to initial modes from <%s (%d), ", 
                x->Mode->name, x->Mode->index);
        output(tmpstring);
        print_mode_link_cycle(x->mode_cycle);
        output(">.\n");
      }
      return FALSE;
    }
  } /* END for(ml=..) */
}

int calculate_weight(struct mode_type *m1, struct mode_type *m2) {
  int wt = 0;
  if(strcmp(m1->name, thread[total_automata-1].starting_mode->name)) {
    wt++;
  }
  if(strcmp(m2->name, thread[total_automata-1].starting_mode->name)) {
    wt--;
  }
  return wt;
}

struct mode_link_type *copy_mode_link_cycle(struct mode_link_type *mlc) {
  struct mode_link_type *ans = NULL, *ptr, *copy;
  if(mlc) {
    for(ptr = mlc->next; ; ptr = ptr->next) {
      copy = (struct mode_link_type *)malloc(sizeof(struct mode_link_type));
      copy->mode = ptr->mode;
      if(ans) {
        copy->next = ans->next;
        ans->next = copy;
        ans = copy;
      }
      else
        ans = copy->next = copy;
      if(ptr == mlc) break;
    }
  }
  return ans;
}

struct mode_link_type *insert_mode(struct mode_type *m, 
                                   struct mode_link_type *mlc) {
  struct mode_link_type *copy;

  if(verbose >= 1) {
    sprintf(tmpstring, 
            "\tInserting mode %s (%d) into cycle: ", m->name, m->index);
    output(tmpstring);
    print_mode_link_cycle(mlc);
    output("\n");
  }

  if(exist_mode(m, mlc)) return mlc;
  copy = (struct mode_link_type *)malloc(sizeof(struct mode_link_type));
  copy->mode = m;
  if(mlc) {
    copy->next = mlc->next;
    mlc->next = copy;
    mlc = copy;
  }
  else
    mlc = copy->next = copy;

  if(verbose >= 1) {
    output("\tAfter insertion, ");
    print_mode_link_cycle(mlc);
    output("\n");
  }

  return mlc;
}

int exist_mode(struct mode_type *m, struct mode_link_type *mlc) {
  struct mode_link_type *ptr;

  if(mlc) {
    for(ptr = mlc->next; ; ptr=ptr->next) {
      if(ptr->mode->index == m->index) return TRUE;
      if(ptr == mlc) break;
    }
  }
  return FALSE;
}

struct mode_link_type *delete_mode(struct mode_type *m, 
                                   struct mode_link_type *mlc) {
  struct mode_link_type *pptr, *ptr;

  if(verbose >= 1) {
    sprintf(tmpstring, 
            "\tDeleting mode %s (%d) from cycle: ", m->name, m->index);
    output(tmpstring);
    print_mode_link_cycle(mlc);
    output("\n");
  }

  if(mlc) {
    for(pptr= mlc, ptr = mlc->next; ; pptr=ptr, ptr=ptr->next) {
      if(ptr->mode->index == m->index) break;
      if(ptr == mlc) break;
    }
    if(pptr == ptr) {
      mlc = NULL;
      free(ptr);
    }
    else {
      pptr->next = ptr->next;
      if(mlc == ptr) mlc = pptr;
      free(ptr);
    }
  }

  if(verbose >= 1) {
    output("\tAfter deletion, ");
    print_mode_link_cycle(mlc);
    output("\n");
  }
  return mlc;
}

int print_mode_link_cycle(struct mode_link_type *mlc) {
  struct mode_link_type *ptr;
  output("{");
  if(mlc) {
    for(ptr = mlc->next; ; ptr = ptr->next) {
      sprintf(tmpstring, "%s (%d)", ptr->mode->name, ptr->mode->index);
      output(tmpstring);
      if(ptr == mlc) break;
      else output(", ");
    }
  }
  output("}");
}

int mode_mu_comp(struct mode_mu_type *mm1, struct mode_mu_type *mm2) {
  int comp;
  struct mode_link_type *ml, *nl;
  struct mode_type *m, *n;

  if(verbose >=3 ) {
    output("Comparing mode_mu's:\n");
    sprintf(tmpstring, "(1) <%s(%d), ", mm1->Mode->name, mm1->Mode->index);
    output(tmpstring);
    print_mode_link_cycle(mm1->mode_cycle);
    output("\nmu_clk: ");
    print_clkconj(mm1->mu_clk);
    output("\nmu_prp: ");
    print_bdd(mm1->mu_prp);
    output("\n");

    sprintf(tmpstring, "(2) <%s(%d), ", mm2->Mode->name, mm2->Mode->index);
    output(tmpstring);
    print_mode_link_cycle(mm2->mode_cycle);
    output("\nmu_clk: ");
    print_clkconj(mm2->mu_clk);
    output("\nmu_prp: ");
    print_bdd(mm2->mu_prp);
    output("\n");
  }

  comp = mode_comp(mm1->Mode, mm2->Mode);
  if(comp != EQUAL) return comp;

  comp = bdd_comp(mm1->mu_prp, mm2->mu_prp);
  if(comp != EQUAL) return comp;

  comp = zone_comp(zone(mm1->mu_clk), zone(mm2->mu_clk));
  if(comp != EQUAL) return comp;

  for(ml=mm1->mode_cycle->next, nl=mm2->mode_cycle->next; ;
      ml=ml->next, nl=nl->next) {
    if(!exist_mode(ml->mode, mm2->mode_cycle)) return -1;
    if(!exist_mode(nl->mode, mm1->mode_cycle)) return 1;
/*
    comp = mode_comp(ml->mode, nl->mode);
    if(comp != EQUAL) return comp;
    if( (ml == mm1->mode_cycle) || (nl == mm2->mode_cycle) ) break;
*/
    if( (ml == mm1->mode_cycle) && (nl != mm2->mode_cycle) ) return -1;
    if( (ml != mm1->mode_cycle) && (nl == mm2->mode_cycle) ) return 1;
    if( (ml == mm1->mode_cycle) && (nl == mm2->mode_cycle) ) break;
  }

  if(verbose >= 3) 
    output("EQUAL mode_mu's.\n");
  return(comp); 
}

int free_mode_mu(struct mode_mu_type *mm) {
  struct mode_link_type *ml;
  struct mm_transition_type *mmt;

  if(mm) {
/*
    if(mm->mu_clk) free_clkconj(mm->mu_clk);
*/
    if(mm->mu_prp) bdd_free(bddm, mm->mu_prp);
    if(mm->mode_cycle) {
      ml = mm->mode_cycle->next;
      mm->mode_cycle->next = NULL;
      mm->mode_cycle = ml;
      while(ml) {
        mm->mode_cycle = ml->next;
        free(ml);
        ml = mm->mode_cycle;
      }
    }
    if(mm->out_mm_trans_cycle) {
      mmt = mm->out_mm_trans_cycle->next;
      mm->out_mm_trans_cycle->next = NULL;
      mm->out_mm_trans_cycle = mmt;
      while(mmt) {
        mm->out_mm_trans_cycle = mmt->next;
        free_mm_transition(mmt);
        mmt = mm->out_mm_trans_cycle;  
      }
    }
    free(mm);
  }
}

int mode_mu_count_inc(struct mode_mu_type *m) {
  (m->count)++;
}

int mode_mu_parent_set(struct mode_mu_type *m, struct a23tree_type *parent) {
  m->parent = parent;
}

int print_mode_mu(struct mode_mu_type *mm) {
  output("Printing mode_mu:\n");
  output("\tmode: ");
  print_mode(mm->Mode);
  output("\tmu_clk: ");
  print_clkconj(mm->mu_clk);
  output("\tmu_prp: ");
  print_bdd(mm->mu_prp);
}

struct clkconj_type *back_delta(struct clkconj_type *mu_clk, int ret_type) {
  struct clkconj_type *c, *x, *tmp;
  struct variable_type *v;

  c = clkconj_copy(mu_clk);

  if(c) {
    x = c->next;
    for(; ; x=x->next) {
      if(x->clk2 == 0)
        x->clk2 = DELTA;
      if(x->clk1 == 0)
        x->clk1 = DELTA;
      if(x == c) break;
    }
    x = alloc_clkconj();
    x->clk1 = DELTA;       /* DELTA LEQ 0 */
    if(strict_semantics || ret_type == FALSE)
      x->op = LEQ;
    else
      x->op = LESS;
    tmp = clkconj_and(c, x);
    free_clkconj(c);
    c = tmp;
    tmp = NULL;

    for(v=var_tail->next; ; v=v->next) {
      if(v->type == V_CLOCK) {
        x = alloc_clkconj();
        x->clk1 = v->index;
        x->op = GEQ;
        tmp = clkconj_and(c, x);
        free_clkconj(c);
        c = tmp;
        tmp = NULL;
      }
      if(v == var_tail) break;
    }
/*
    output("After back_delta: ");
    print_clkconj(c); output("\n");
*/
  }
  return c;
}

int initial_consistent(struct mode_mu_type *x) {
  int i = 0;
  char *tmp_name = NULL, *tc = NULL; 
  struct mode_type *m;

/*
  tmp_name = str_preserve(x->Mode->name);
  tc = strtok(tmp_name, ".");
  if(strcmp(tc, thread[0].starting_mode->name)) {
    free(tmp_name);
    return FALSE;
  }

  while(tc = strtok((char *)NULL, ".")) {
    if(strcmp(tc, thread[++i].starting_mode->name)) {
      free(tmp_name);
      return FALSE;
    }
  }
*/
  if(x->Mode->index != BT_THREAD->starting_mode->index) return FALSE;

  if(verbose >= 2) {
    output("Inside initial_consistent().\n");
    print_mode_link_cycle(x->mode_cycle);
    output("\n");
  }

  if(x->mode_cycle->next != x->mode_cycle) {
    return FALSE;
  }
  tmp_name = str_preserve(x->mode_cycle->mode->name);
  
  tc = strtok(tmp_name, ".");
  if(strcmp(tc, thread[0].starting_mode->name)) {
    free(tmp_name);
    return FALSE;
  }

  while(tc = strtok((char *)NULL, ".")) {
    if(strcmp(tc, thread[++i].starting_mode->name)) {
      free(tmp_name); 
      return FALSE;
    }
  }

  m = compose_modes(x->Mode, x->mode_cycle->mode);

  if(!consistent_inv_mu(m, x->mu_clk, x->mu_prp)) {
    free(tmp_name); 
    free_mode(m);
    return FALSE;
  }
  free(tmp_name); 
  free_mode(m);
  return TRUE;
}

struct clkconj_type *reset_clk_assign(struct clkconj_type *c, 
                                      struct assign_type *a) {
  struct assign_type *aa;

/*
  output("Resetting assignment: "); print_assignment1(a); output("\n");
  output("for: "); print_clkconj(c); output("\n");
*/
  if(a) {
    for(aa = a->next; ; aa = aa->next) {
      if(get_var_type(aa->var) == V_CLOCK) {
        c = conjoin(c, aa->var);
      }
      if(aa == a) break;
    }
  }
/*
  output("After resetting:\n");
  print_clkconj(c);
  output("\n");
*/
  return c;
}

bdd reset_prp_assign(bdd f, struct assign_type *a) {
  struct assign_type *x;
  struct proposition_type *p;
  int i=0;
  bdd assoc[bdd_vars(bddm)+1];
  bdd t = f;

  if(verbose >= 2) {
    output("Resetting assignment (");
    print_assignment1(a);
    output(")\n in BDD: ");
    print_bdd(f);
  }

  p = prop_tail->next;
  if(a) {
    x = a->next;
    for(; ; x=x->next) {
      if(get_var_type(x->var) == V_REGISTER) {
        for(; ; p = p->next) {
          if(x->var == p->var)
            assoc[i++] = p->prop_bdd;
          if(p == prop_tail) break;
        }
      }
      if(x == a) break;
    }
/*
    sprintf(tmpstring, "No. of vars in temp assoc = %d\n", i);
*/
    assoc[i] = (bdd)NULL;
/*
    output(tmpstring);
*/
    bdd_temp_assoc(bddm, assoc, 0);
    t = bdd_exists(bddm, t);
  }
  return t;
}

struct clkconj_type *weakest_clk_precondition(struct transition_type *f,
                                              struct mode_mu_type *x,
                                              struct clkconj_type *mu1) {
  /* Here, mu1 = conjoin(back_delta(x->mu_clk, FALSE), DELTA); */

  struct clkconj_type *tmp_clkconj = NULL, *c = NULL;
  int *tmp_perm = inverse_permutation(f->permutation->perm);

  c = permute_clkconj(mu1, tmp_perm);
  c = reset_clk_assign(c, f->assign_cycle);

  tmp_clkconj = back_clkconj_and(c, f->clk_trigger_cycle);
  free_clkconj(c);
  c = tmp_clkconj;

  tmp_clkconj = back_clkconj_and(c, f->mode->xconj_cycle);
  free_clkconj(c);
  c = tmp_clkconj;

  tmp_clkconj = NULL; 
  return c;
}
 
bdd weakest_prp_precondition(struct transition_type *f, 
                             struct mode_mu_type *x) {
  bdd b = x->mu_prp;
  int *tmp_perm = inverse_permutation(f->permutation->perm);

/*
  output("Calculating weakest precondition of BDD:\n");
  print_bdd(b);
  sprintf(tmpstring, 
          "by back-traversing transition %d with trigger BDD:\n", f->index);
  output(tmpstring);
  if(f->trigger_bdd)
    print_bdd(f->trigger_bdd);
  else output("NULL BDD\n");
*/

  b = permute_bdd(b, tmp_perm);

  b = reset_prp_assign(b, f->assign_cycle);

/*
  output("After resetting assignment, ");
  print_bdd(b);
*/
  b = BDD_AND(b, f->trigger_bdd);
/*
  output("After conjunction with trigger, ");
  print_bdd(b);
*/
  b = BDD_AND(b, f->mode->boolexp->boolexp_bdd);
/*
  output("Weakest precondition is:\n");
  print_bdd(b);
*/
  return b;
}

int remove_dormant_cycles(struct thread_type *th) {
  struct mode_type *m;
  struct transition_type *ptr, *tr, *str, *ttr;

  for(m=th->starting_mode; m; m=m->cnext) {
    m->status = 0;
  }

  dfs_visit_mode(th->starting_mode);

  for(m=th->starting_mode; m; m=m->cnext) {
    if(m->status == VISITED) continue;
    if(m->in_trans_cycle) {
      for(ptr=m->in_trans_cycle, tr=ptr->next, str=tr->next; ; 
          tr=str, str=str->next) {
        for(ttr=tr->mode->out_trans_cycle->next; ; ttr=ttr->next) {
          if(ttr->index == tr->index) {
            break;
          }
          if(ttr==tr->mode->out_trans_cycle) break;
        }
        eliminate_transition(ttr, tr->mode);
        th->trans_count--;
        if(ttr->valid == FALSE) th->invalid_trans_count--;
        if(m->in_trans_cycle == NULL) break;
      }
    }
  }

}

int dfs_visit_mode(struct mode_type *m) {
  struct transition_type *tr;

  if(m->status == VISITED) return;
  
  m->status = VISITED;

  if(m->out_trans_cycle) {
    for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
      dfs_visit_mode(tr->mode);
      if(tr == m->out_trans_cycle) break;
    }
  }
  return;
}

int add_mm_transition(struct mode_mu_type *mm1, 
                      struct mode_mu_type *mm2, 
                      struct mode_type *m1, struct mode_type *m2,
                      int weight) {

  struct mm_transition_type *mmt;

/*
  if(mm1->out_mm_trans_cycle) {
    for(mmt = mm1->out_mm_trans_cycle->next; ;mmt = mmt->next) {
      if( (mmt->mode_mu->index == mm2->index) && 
          (mmt->from_mode->index == m1->index) && 
          (mmt->to_mode->index == m2->index) ) 
        return;   
      if(mmt == mm1->out_mm_trans_cycle) break;
    }
  }
*/

  /* for out-going trans */
  mmt = (struct mm_transition_type *) malloc(sizeof(struct mm_transition_type));
  mmt->index = ++mm_trans_count;
  mmt->weight = weight;
  mmt->mode_mu = mm2;
  mmt->from_mode = m1;
  mmt->to_mode = m2;
  if(mm1->out_mm_trans_cycle) {
    mmt->next = mm1->out_mm_trans_cycle->next;
    mm1->out_mm_trans_cycle->next = mmt;
    mm1->out_mm_trans_cycle = mmt;
  }
  else 
    mm1->out_mm_trans_cycle = mmt->next = mmt;

  /* for in-coming trans */
  mmt = (struct mm_transition_type *) malloc(sizeof(struct mm_transition_type));
  mmt->index = mm_trans_count;
  mmt->weight = weight;
  mmt->mode_mu = mm1;
  mmt->from_mode = m1;
  mmt->to_mode = m2;
  if(mm2->in_mm_trans_cycle) {
    mmt->next = mm2->in_mm_trans_cycle->next;
    mm2->in_mm_trans_cycle->next = mmt;
    mm2->in_mm_trans_cycle = mmt;
  }
  else 
    mm2->in_mm_trans_cycle = mmt->next = mmt;

  return;
}

int free_mm_transition(struct mode_mu_type *m, struct mm_transition_type *mmt) {
  struct mm_transition_type *ptr;
  for(ptr=mmt->mode_mu->in_mm_trans_cycle->next; ;ptr = ptr->next) { 
    if(ptr->index == mmt->index) break;
    if(ptr==mmt->mode_mu->in_mm_trans_cycle);
  }
  free(ptr);  /* in-coming trans */
  free(mmt);  /* out-going trans */
}


struct clkconj_type *
back_clkconj_and(struct clkconj_type *x1, struct clkconj_type *x2) {
  int tmp_clk, i, j;
  int nx1_count = 0;
  int nx2_count = 0;
  int comp = DIFFERENT;
  int comp_new = DIFFERENT;
  int *tmp_op, *tmp_diff;
  struct clkconj_type *nx1, *nx2, *tmp, 
                      *y1, *y2, *py1, *py2, 
                      *p_true_ptr1 = NULL, *true_ptr1 = NULL;

/*
  output("Back conjunction of clkconj: ");
  print_clkconj(x1);
  output(" AND ");
  print_clkconj(x2);
  output(" is:\n");
*/

  if(!x1 && !x2) {
    tmp = alloc_clkconj();
    tmp->op = TRUE;
    return tmp;
  }

  nx1 = clkconj_copy(x1);
  nx2 = clkconj_copy(x2);

  if(x1 == NULL || x1->op == TRUE) {
    free_clkconj(nx1);
    return nx2;
  }
  else if(x1->op == FALSE) {
    free_clkconj(nx1);
    free_clkconj(nx2);
    tmp = alloc_clkconj();
    tmp->op = FALSE;
    return tmp;
  }

  if(x2 == NULL || x2->op == TRUE) {
    free_clkconj(nx2);
    return nx1;
  }
  else if(x2->op == FALSE) {
    free_clkconj(nx1);
    free_clkconj(nx2);
    tmp = alloc_clkconj();
    tmp->op = FALSE;
    return tmp;
  }

  for(y1=nx1->next; ; y1=y1->next) {
    nx1_count++;
    if(y1 == nx1) break;
  }

  for(y2=nx2->next; ; y2=y2->next) {
    nx2_count++;
    if(y2 == nx2) break;
  }

  for(j=1, py2=nx2, y2=nx2->next; j<=nx2_count; j++, py2 = y2, y2=y2->next) {
    for(i=1, py1=nx1, y1=nx1->next; i<=nx1_count; i++, py1 = y1, y1=y1->next) {
      if(y1->clk1 == y2->clk2 && y1->clk2 == y2->clk1) {
        if(y1->clk1) {
          tmp_clk = y2->clk1;  
          y2->clk1 = y2->clk2;
          y2->clk2 = tmp_clk;
          y2->op = reverse_op(y2->op);
          y2->diff = -(y2->diff);
        }
        else {
          tmp_clk = y1->clk1; 
          y1->clk1 = y1->clk2;
          y1->clk2 = tmp_clk;
          y1->op = reverse_op(y1->op);
          y1->diff = -(y1->diff);
        }
      }
      if(y1->clk1 == y2->clk1 && y1->clk2 == y2->clk2) {
/*        
        printf("y1: "); print_one_clkconj(y1); printf("\n");
        printf("y2: "); print_one_clkconj(y2); printf("\n");
*/
        tmp_op = (int *)malloc(sizeof(int));
        tmp_diff = (int *)malloc(sizeof(int));

        comp_new =  back_compatible( y1->op, y1->diff, y2->op, y2->diff, 
                                tmp_op, tmp_diff);
        switch(comp_new) { 
          case OVERLAPPING : /*printf("Got OVERLAPPING\n"); */
                             switch(comp) {
                               case DIFFERENT: comp = OVERLAPPING;
                                               break;
                               case TRUE: break;
                             }
                             break;

          case FALSE: /* printf("Got FALSE\n"); */
                      comp = FALSE;
                      break;

          case TRUE: /* printf("Got TRUE\n"); */
                     comp = TRUE; 
/*
                     printf("Got TRUE for %s, %d, %s, %d\n", 
                            relop_string(y1->op), y1->diff, 
                            relop_string(y2->op), y2->diff);
*/
                     p_true_ptr1 = py1;
                     true_ptr1 = y1;
                     tmp = alloc_clkconj();
                     tmp->clk1 = y1->clk1;
                     tmp->clk2 = y1->clk2;
                     tmp->op = *tmp_op;
                     tmp->diff = *tmp_diff;
                     break;

          default: 
            printf("Error: invalid return value from back_compatible() = %d\n", 
                      comp_new);
              break;

        } /* END switch(comp_new) */

      } /* END if(y1->clk1 == y2->clk1 && y1->clk2 == y2->clk2) */

      if(comp == FALSE) break;
/*
      if(y1 == nx1) break;
*/

    } /* END for(; ; py1 = y1, y1=y1->next) */

    switch(comp) { 
      case TRUE:                            /* replace true_ptr1 with tmp */
/*
                 printf("Before replacing ");
                 print_one_clkconj(true_ptr1);
                 printf("nx1: ");
                 print_clkconj(nx1);
                 printf("\n");
*/

                 if(p_true_ptr1 == true_ptr1) {
                   tmp->next = tmp;
                   nx1 = tmp;
                 }
                 else {
                   p_true_ptr1->next = tmp;
                   tmp->next = true_ptr1->next; /* remember to free true_ptr1 */
                   if(nx1 == true_ptr1) 
                     nx1 = tmp;
                 }

/*
                 printf("After replacement, nx1: ");
                 print_clkconj(nx1);
                 printf("\n");
*/
                 break;
      case FALSE: free_clkconj(nx1);
                  free_clkconj(nx2);
                  tmp = alloc_clkconj();
                  tmp->op = FALSE;
                  return tmp;
                  break;
      case DIFFERENT:
      case OVERLAPPING:  
                        if(nx2->next == nx2)
                          nx2 = (struct clkconj_type *)NULL;
                        else {
                          if(y2 == nx2) 
                            nx2 = py2;
                          py2->next = y2->next; 
                        }
/*
                        nx2_count--;
*/

                        y2->next = nx1->next;
                        nx1->next = y2;        /* append y2 into nx1 */
                        nx1 = y2;
                        nx1_count++;
                        y2 = py2;
                        break;
    }
    if(!nx2) break;
/*
    if(y2 == nx2) break;
*/
    comp = comp_new = DIFFERENT; /* reset the comparison results */
  }    /* END for(; ; py2 = y2, y2=y2->next) */

  if(true_ptr1) 
    true_ptr1->next = true_ptr1;
  free_clkconj(true_ptr1);
  free_clkconj(nx2);
/*
  print_clkconj(nx1);
  output("\n");
*/
  return nx1;
}


/* 
   check if (op1, num1) and (op2, num2) are compatible intervals:
      (1) if they are overlapping intervals, return OVERLAPPING
      (2) if they can be reduced into 1 interval, return TRUE 
             ( the resulting interval is in (*op3, *num3) )
      (3) if they are not compatible, return FALSE
*/

int back_compatible(int op1, int num1, int op2, int num2, int *op3, int *num3) {
  if(op1 == LESS || op1 == LEQ) {

    if(op2 == LESS || op2 == LEQ) {
      if(num1 == num2) {
        switch(op1) {
          case LEQ: *op3 =  LEQ;
                    break;
          case LESS: switch(op2) {
                      case LESS: *op3 = LESS;
                                 break;
                      case LEQ: *op3 = LEQ;
                                break;
                    }
        }
      }
      else if(op1 == LEQ && op2 == LEQ && num1 == -num2) {
        *op3 = EQ;
        *num3 = (num1 >= 0) ? num1 : num2;
        return TRUE;
      }
      else {
        *op3 = (num1 > num2) ? op1 : op2;
      }

      *num3 = (num1 >= num2) ? num1 : num2;
      return TRUE;
    }

    else
    if(op2 == EQ) {
      if(num2 < num1) {
        *op3 = op1;
        *num3 = num1;
        return TRUE;
      }
      else if(num2 > num1) {
             free(op3);
             free(num3);
             return FALSE;
      }
      else {                    /* num1 == num2 */
        if(op1 == LESS) return FALSE;
        else {
          *op3 = op1;
          *num3 = num1;
          return TRUE;
        }
      }
    }

    else
    if(op2 == GEQ || op2 == GREATER) {
      if(num2 < num1) {
        free(op3);
        free(num3);
        return OVERLAPPING;
      }
      else if(num2 > num1) {
        free(op3);
        free(num3);
        return FALSE;
      }
      else {                 /* num1 == num2 */
        switch(op1) {
          case LESS: free(op3);
                     free(num3);
                     return FALSE;
                     break;
          case LEQ: switch(op2) {
                      case GREATER: free(op3);
                                    free(num3);
                                    return FALSE;
                                    break;
                      case GEQ: *op3 = EQ;
                                *num3 = num1;
                                return TRUE;
                                break;
                    }
        }
      }
    }

  } /* END if(op1 == LESS || op1 == LEQ) */

  if(op1 == GREATER || op1 == GEQ) {

    if(op2 == GREATER || op2 == GEQ) {
      if(num2 == num1) {
        switch(op1) {
          case GEQ: *op3 =  GEQ;
                    break;
          case GREATER: switch(op2) {
                        case GREATER: *op3 = GREATER;
                                   break;
                        case GEQ: *op3 = GEQ;
                                  break;
                      }
        }
      }
      else
        *op3 = (num1 < num2) ? op1 : op2;

      *num3 = (num1 <= num2) ? num1 : num2;
      return TRUE;
    }

    else
    if(op2 == EQ) {
      if(num2 > num1) {
        *op3 = op1;
        *num3 = num1;
        return TRUE;
      }
      else if(num2 < num1) {
             free(op3);
             free(num3);
             return FALSE;
      }
      else {                    /* num1 == num2 */
        if(op1 == GREATER) return FALSE;
        else {      /* op1 == GEQ */
          *op3 = op1;
          *num3 = num1;
          return TRUE;
        }
      }
    }

    else
    if(op2 == LEQ || op2 == LESS) {
      if(num2 > num1) {
        free(op3);
        free(num3);
        return OVERLAPPING;
      }
      else if(num2 < num1) {
        free(op3);
        free(num3);
        return FALSE;
      }
      else {                 /* num1 == num2 */
        switch(op1) {
          case GREATER: free(op3);
                        free(num3);
                        return FALSE;
                        break;
          case GEQ: switch(op2) {
                      case LESS: free(op3);
                                 free(num3);
                                 return FALSE;
                                 break;
                      case LEQ: *op3 = EQ;
                                *num3 = num1;
                                return TRUE;
                                break;
                    }
        }
      }
    }

  } /* END if(op1 == GREATER || op1 == GEQ) */

  if(op1 == EQ) {
    if(op2 == GREATER || op2 == GEQ) {
      if(num2 > num1) {
        free(op3);
        free(num3);
        return FALSE;
      }
      else if(num2 < num1) {
        *op3 = op2;
        *num3 = num2;
        return TRUE;
      }
      else {         /* num2 == num1 */
        switch(op2) {
          case GREATER: free(op3);
                        free(num3);
                        return FALSE; break;
          case GEQ: *op3 = GEQ;
                    *num3 = num2;
                    return TRUE;
        }
      }
    }
    if(op2 == LESS || op2 == LEQ) {
      if(num2 < num1) {
        free(op3);
        free(num3);
        return FALSE;
      }
      else if(num2 > num1) {
        *op3 = op2;
        *num3 = num2;
        return TRUE;
      }
      else {         /* num2 == num1 */
        switch(op2) {
          case LESS: free(op3);
                     free(num3); 
                     return FALSE; break;
          case LEQ: *op3 = EQ;
                    *num3 = num1;
                    return TRUE;
        }
      }
    }
    if(op2 == EQ) {
      if(num1 == num2) {
        *op3 = EQ;
        *num3 = num1;
        return TRUE;
      }
      else {
        free(op3);
        free(num3);
        return FALSE;
      }
    }
  } /* END if(op1 == EQ) */
}

int print_mmlist(struct mode_mu_type *list) {
  struct mode_mu_type *mm;
  struct mm_transition_type *mmt;
  if(verbose >= 0) {
    sprintf(tmpstring, 
         "Listing Mode-Mu List for Back-Trace Graph of thread %d (Totally %d nodes).\n",
         BT_THREAD->index, total_mode_mu);
    output(tmpstring);
  }
  if(list) {
    for(mm=list; mm; mm=mm->next) {
      if(verbose >= 0) {
        sprintf(tmpstring, "MODE_MU (%d){%s(%d), ", 
                mm->index, mm->Mode->name, mm->Mode->index);
        output(tmpstring);
        print_mode_link_cycle(mm->mode_cycle);
        output("}\n");
      }
      if(mm->out_mm_trans_cycle) {
        for(mmt = mm->out_mm_trans_cycle->next; ; mmt=mmt->next) {
          if(verbose >= 0) {
            sprintf(tmpstring, 
                    "  MMT(%d): weight=%d, to mode_mu (%d){%s(%d), ",
                    mmt->index, mmt->weight, mmt->mode_mu->index, 
                    mmt->mode_mu->Mode->name, mmt->mode_mu->Mode->index);
            output(tmpstring);
            print_mode_link_cycle(mmt->mode_mu->mode_cycle);
            sprintf(tmpstring, "}, [from: %s(%d), to: %s(%d)]\n", 
                   mmt->from_mode->name, mmt->from_mode->index, 
                   mmt->to_mode->name, mmt->to_mode->index);
            output(tmpstring);
          }
          if(mmt == mm->out_mm_trans_cycle) break;
        }
      }
    }
  }
  if(verbose >= 0) {
    sprintf(tmpstring, 
            "End of Mode-Mu List for Back-Trace Graph of thread %d.\n", 
            BT_THREAD->index);
    output(tmpstring);
  }
}

int mm_new_index(struct mode_mu_type *mm, struct mode_mu_type *list) {
  int i;
  struct mode_mu_type *ptr;

  if(list) {
    for(i=1, ptr=list; ptr; i++, ptr=ptr->next) {
      if(mm->index == ptr->index) return i;
    }
  }
}

int btgraph_kclosure(struct mode_mu_type *list) {
  int bound = -3, i, j, k, lend[total_mode_mu+1][total_mode_mu+1], ans = FALSE;
  struct mode_mu_type *mm, *mmd;
  struct mm_transition_type *mmt;

  if(list) {
    for(i = 1, mm = list; mm; i++, mm = mm->next) {
      lend[0][i] = lend[i][0] = mm->index;
    }

    for(i = 0; i<=total_mode_mu; i++) {
      for(j = 0; j<=total_mode_mu; j++) {
        lend[i][j] = bound;   /* valid values are -1, 0, 1 only */
      }
    }

    for(mm = list; mm; mm = mm->next) {
      if(mm->out_mm_trans_cycle) {
        for(mmt = mm->out_mm_trans_cycle->next; ; mmt=mmt->next) {
          lend[mm_new_index(mm, mmlist)]
              [mm_new_index(mmt->mode_mu, mmlist)] = mmt->weight;
          if(mmt == mm->out_mm_trans_cycle) break;
        }
      }
    }
    
    for(i=1; i<=total_mode_mu; i++) {
      for(j=1; j<=total_mode_mu; j++) {
        if(lend[j][i] > bound) {
          for(k=1; k<=total_mode_mu; k++) {
            if(lend[i][k] > bound) {
              if(lend[j][k] < lend[j][i] + lend[i][k]) {
                lend[j][k] = lend[j][i] + lend[i][k];
              }
            }
          }
        }
      }
    }
    
    for(i=1; i<=total_mode_mu; i++) {
      if(lend[i][i] != bound && lend[i][i] < 0) return FALSE;
    }

    for(mm = list; mm; mm = mm->next) {
      if(source_mm(mm)) {
        for(mmd = list; mmd; mmd = mmd->next) {
          if(destination_mm(mmd)) {
            if( (lend[mm_new_index(mm, mmlist)]
                     [mm_new_index(mmd, mmlist)] < 0) &&
                (lend[mm_new_index(mm, mmlist)]
                     [mm_new_index(mmd, mmlist)] != bound)
              ) {
              return FALSE;
            }
          }
        }
      }
    }
    return TRUE;
  }
}

/*
int source_mm(struct mode_mu_type *mm) {
  if(strcmp(mm->Mode->name, BT_THREAD->starting_mode->name)) 
    return FALSE;
  if(mm->mode_cycle->next != mm->mode_cycle) 
    return FALSE;
  if(strcmp(mm->mode_cycle->mode->name, thread[0].starting_mode->name)) 
    return FALSE;

  return TRUE;
}
*/

int source_mm(struct mode_mu_type *mm) {
  struct mode_mu_link_type *mml;
  for(mml = source_cycle->next; ; mml = mml->next) {
    if(mml->mode_mu->index == mm->index) return TRUE;
    if(mml == source_cycle) break;
  }
  return FALSE;
}

int destination_mm(struct mode_mu_type *mm) {
  struct mode_mu_link_type *mml;
  for(mml = destination_cycle->next; ; mml = mml->next) {
    if(mml->mode_mu->index == mm->index) return TRUE;
    if(mml == destination_cycle) break;
  }
  return FALSE;
}

