#include "include.h"
#include "global.h"
#include "print.h"
#include "bddman.h"
#include "modeman.h"
#include "zoneman.h"
#include "reduce.h"
#include "back.h"
#include "shield.h"
#include "lookahead-gen.h"


int gor_count = 0;
int gio_count = 0;
int gis_trans_count = 0;
int CTA = 0;
int HS_THREAD_T, HS_THREAD_t;
struct clkconj_type *literal_cycle = NULL;
struct a23_tree_type *GIO_TREE = (struct a23_tree_type *)NULL, 
                     *GOR_TREE = (struct a23_tree_type *)NULL;
struct generic_thread_type *GIS_THREAD = (struct generic_thread_type *)NULL;
struct gor_type *last_gor = (struct gor_type *)NULL,
                *TOP = (struct gor_type *)NULL;


int generic_merge(int Ti, int ti);
struct gor_type *copy_gor(struct gor_type *g);
struct gor_type *alloc_gor();
struct gio_type *copy_gio(struct gio_type *g);
struct mode_type *copy_mode(struct mode_type *);
struct gio_link_type *copy_gio_cycle(struct gor_type *g);
struct gio_type *alloc_gio();

struct gor_type *initial_gor(int T, int t);
int print_initial_gor_status(struct gor_type *);
int gor_comp(struct gor_type *, struct gor_type *);
int free_gor(struct gor_type *);
int gor_count_inc(struct gor_type *);
int gor_parent_set(struct gor_type *, struct a23tree_type *);
int print_gor(struct gor_type *);

int gio_comp(struct gio_type *g1, struct gio_type *g2);
int free_gio(struct gio_type *g);
int free_gio_cycle(struct gor_type *);
int gio_count_inc(struct gio_type *g);
int gio_parent_set(struct gio_type *g, struct a23tree_type *parent);
int print_gio(struct gio_type *g);

int gis_transition_n1_first(struct gor_type *x);
int gis_transition_n2_first(struct gor_type *, struct gor_type *);
int gis_transition_n2_first_split(struct gor_type *, struct gor_type *);

int print_new_gor_status(struct gor_type *ng);
int print_trans_status(struct transition_type *tr);
int print_new_gio_status(struct gio_type *m, struct transition_type *tr,
                         struct gio_type *nm);
int print_zone_status(struct mode_type *nm);
struct mode_type *post_prop(struct mode_type *m, struct transition_type *tr, 
                            struct mode_type **nm);
int append_gio(struct gio_link_type **, struct gor_type **, int);
struct gis_transition_type *alloc_gis_trans();
struct gis_transition_type *add_gis_transition(struct transition_type *tr, 
                                               struct gor_type *gp, 
                                               struct gor_type *gs);

struct gor_type *extend_gor(struct gor_type *gor, struct gio_type *gio,
                            struct transition_type *tr);
int extend_zones(struct gor_type *gor, struct gio_type *gio);
struct gor_type *conjoin_gor(struct gor_type *gor, struct gio_type *gio,
                             struct mode_type *, struct mode_type *);
struct gio_type *conjoin_gio(struct gio_type *gio, struct mode_type *om1, 
                             struct mode_type *om2);
struct gor_type *set_gor(struct gor_type *);
int *gen_perm_gis(int, int);

struct clkconj_type *get_literals(struct gio_type *g) {
  struct clkconj_type *c, *tmpc, *tmp, *ans = NULL;
/*
  sprintf(tmpstring, "Finding literal set of gio %d with clkconj:\n", g->index);
  output(tmpstring);
  print_clkconj(g->mode->xconj_cycle);
  output("\n");
*/
  if(literal_cycle) {
    for(c=literal_cycle->next; ; c=c->next) {
      tmpc = clkconj_copy1(c);
      tmp = clkconj_and(tmpc, g->mode->xconj_cycle);
      if(tmp->op != FALSE)
        append_clkconj(tmpc, &ans);
      else
        free_clkconj(tmpc);
      free_clkconj(tmp);
      if(c == literal_cycle) break;
    }
  }
/*
  output("Literal: "); print_literals(ans); output("\n");
*/
  return ans;
}

int init_gis_thread() {
  GIS_THREAD = (struct generic_thread_type *)
               malloc(sizeof(struct generic_thread_type));
  GIS_THREAD->gor_count = 0;
  GIS_THREAD->gis_trans_count = 0;
  GIS_THREAD->starting_gor = NULL;
}

int generic_merge(int T, int t) {
  int i;
  struct gor_type *x, *y, *z;
  struct gio_link_type *gl;
  struct mode_type *m;

  HS_THREAD_T = T;
  HS_THREAD_t = t;
  CTA = t + 1;

  printf("Generic merging automata threads (%d) and (%d):\n", T, t);
  if(verbose >= 0) {
    output("\n************************************************************\n");
    sprintf(tmpstring, "Generic merging threads %d and %d: \n", T, t);
    output(tmpstring);
  }

  init_gis_thread();
  GIS_THREAD->starting_gor = initial_gor(T, t);
  last_gor = GIS_THREAD->starting_gor;
  GIS_THREAD->gor_count = 1;
  GIS_THREAD->gis_trans_count = 0;

  if(GIS_THREAD->starting_gor) {
    TOP = GIS_THREAD->starting_gor;
    TOP->next = NULL;
    for( ; TOP; ) {
      if(verbose >= 0) {
        output("\nCurrent stack status (gor indices): ");
        for(x = TOP; x; x=x->next) {
          sprintf(tmpstring, "%d", x->index);
          output(tmpstring);
          if(x->next != NULL) 
            output(", ");
        }
        output("\n");
      }
      x = TOP;
      TOP = TOP->next;
      x->next = (struct gor_type *)NULL;
      if(verbose >= 0) {
        sprintf(tmpstring, "Predecessor gor (%d) {", x->index);
        output(tmpstring);
        for(gl=x->gio_cycle->next; ; gl=gl->next) {
          sprintf(tmpstring, "(%d)[%s(%d)]", 
                  gl->gio->index, gl->gio->mode->name, gl->gio->mode->index);
          output(tmpstring);
          if(gl == x->gio_cycle) break;
          else output(", ");
        }
        output("}\n");
      }

        if(verbose >= 0)
          output("---- Starting submode1 transitions ...\n");
        gis_transition_n1_first(x);
        if(verbose >= 0)
          output("---- Finished submode1 transitions ...\n");

        y = copy_gor(x);
        z = copy_gor(x);
        if(verbose >= 0)
          output("---- Starting submode2 transitions (no splitting) ...\n");
        sprintf(tmpstring, "GOR (%d) is a copy of GOR (%d).\n", 
                z->index, x->index);
        output(tmpstring);
        gis_transition_n2_first(z, x);
        if(verbose >= 0)
          output("---- Finished submode2 transitions (no splitting) ...\n");

        if(verbose >= 0)
          output("---- Starting submode2 transitions (splitting) ...\n");
        sprintf(tmpstring, "GOR (%d) is a copy of GOR (%d).\n", 
                y->index, x->index);
        output(tmpstring);
        gis_transition_n2_first_split(y, x);
        if(verbose >= 0)
          output("---- Finished submode2 transitions (splitting) ...\n");
    }
  }
  else {
    fprintf(stderr, "A conflict in initial gor merging.\n");
    return (-1);
  }
  last_gor->cnext = (struct gor_type *)NULL;
  print_gis(GIS_THREAD->starting_gor);
  check_trans(GIS_THREAD->starting_gor);
  printf("\tRemoving irreachable modes in thread %d.\n\t", T);
  remove_dormant_cycles(&thread[T]);
  remove_irreachable_mode(&thread[T]);
  print_counts(T);
  print_thread(thread[T]);
  return (1);
}

struct gor_type *copy_gor(struct gor_type *g) {
  struct gor_type *ng = alloc_gor();
  ng->count = g->count;
  ng->parent = g->parent;
  ng->gio_cycle = copy_gio_cycle(g);
/*
  ng->out_gis_trans_cycle = g->out_gis_trans_cycle; 
  ng->in_gis_trans_cycle = g->in_gis_trans_cycle; 
*/
  return ng;
}

struct gor_type *alloc_gor() {
  struct gor_type *g;
  g = (struct gor_type *)malloc(sizeof(struct gor_type));
  g->index = gor_count++;
  g->count = 0;
  g->parent = (struct a23tree_type *)NULL;
  g->gio_cycle = (struct gio_link_type *)NULL;
  g->out_gis_trans_cycle = (struct gis_transition_type *)NULL;
  g->in_gis_trans_cycle = (struct gis_transition_type *)NULL;
  g->next = (struct gor_type *)NULL;
  g->cnext = (struct gor_type *)NULL;
  return g;
}

struct gio_type *copy_gio(struct gio_type *g) {
  struct gio_type *ng = alloc_gio();
  ng->index = g->index;
  ng->count = g->count;
  ng->parent = g->parent;
  ng->mode = copy_mode(g->mode);
  ng->clk_literal = clkconj_copy(g->clk_literal);
  ng->bdd_literal = g->bdd_literal;
  return ng;
}

struct mode_type *copy_mode(struct mode_type *m) {
  struct mode_type *nm = alloc_mode();
  nm->name = str_preserve(m->name);
  nm->submode1 = m->submode1;
  nm->submode2 = m->submode2;
  nm->boolexp = alloc_boolexp();
  nm->boolexp->boolexp_bdd = m->boolexp->boolexp_bdd;
  nm->dval_bdd = m->dval_bdd;
  nm->zone = zone_copy(m->zone);
  nm->xconj_cycle = clkconj_copy(m->xconj_cycle);
  return nm;
}

struct gio_link_type *copy_gio_cycle(struct gor_type *g) {
  struct gio_link_type *x, *y;
  struct gor_type *ans_gor = alloc_gor();

  for(x = g->gio_cycle->next; ; x = x->next) {
    y = (struct gio_link_type *)malloc(sizeof(struct gio_link_type));
    y->gio = copy_gio(x->gio);
    append_gio(&y, &ans_gor, FALSE);
    if(x == g->gio_cycle) break;
  }
  return ans_gor->gio_cycle;
}

struct gio_type *alloc_gio() {
  struct gio_type *g;
  g = (struct gio_type *)malloc(sizeof(struct gio_type));
  g->index = gio_count++;
  g->count = 0;
  g->parent = (struct a23tree_type *)NULL;
  g->mode = (struct mode_type *)NULL;
  g->clk_literal = (struct clkconj_type *)NULL;
  g->bdd_literal = (bdd)NULL;
  return g;
}

struct gor_type *initial_gor(int T, int t) {
  struct gor_type 		*gn;
  struct mode_type              *ns, *s1, *s2, *nns;
  int                           i, var_id;
  struct transition_type        *tr;
  struct clkconj_type 		*tmp_clkconj;

  gn = alloc_gor();
  gn->gio_cycle = (struct gio_link_type *)malloc(sizeof(struct gio_link_type));
  gn->gio_cycle->gio = alloc_gio();
  ns = alloc_mode();
  gn->gio_cycle->gio->mode = ns;
  gn->gio_cycle->next = gn->gio_cycle;

  ns->submode1 = s1 = thread[T].starting_mode;
  ns->submode2 = s2 = thread[t].starting_mode;
  ns->name = assign_mode_name(ns, (int *)NULL, 0);

  ns->boolexp = alloc_boolexp();
  ns->dval_bdd = (bdd)NULL;
  
  ns->boolexp->boolexp_bdd = BDD_AND(s1->boolexp->boolexp_bdd,
                                     s2->boolexp->boolexp_bdd);

  if(thread[T].init_boolexp && thread[t].init_boolexp) {
    ns->dval_bdd = BDD_AND( BDD_AND(thread[T].init_boolexp->boolexp_bdd, 
                                    thread[t].init_boolexp->boolexp_bdd), 
                            BDD_AND(s1->boolexp->boolexp_bdd, 
                                    s2->boolexp->boolexp_bdd)
                   );
  }
  else if(thread[T].init_boolexp) {
    ns->dval_bdd = BDD_AND(thread[T].init_boolexp->boolexp_bdd, 
                           BDD_AND(s1->boolexp->boolexp_bdd, 
                                   s2->boolexp->boolexp_bdd)
                   );
  }
  else if(thread[t].init_boolexp) {
    ns->dval_bdd = BDD_AND(thread[t].init_boolexp->boolexp_bdd, 
                           BDD_AND(s1->boolexp->boolexp_bdd, 
                                   s2->boolexp->boolexp_bdd)
                   );
  }
  else {
    ns->dval_bdd = BDD_AND(s1->boolexp->boolexp_bdd, s2->boolexp->boolexp_bdd);
  }

  ns->xconj_cycle = clkconj_and(thread[T].init_clkconj, 
                                thread[t].init_clkconj);

  tmp_clkconj = clkconj_and(ns->xconj_cycle, 
                            clkconj_and(s1->xconj_cycle, s2->xconj_cycle));
  free_clkconj(ns->xconj_cycle);
  ns->xconj_cycle = tmp_clkconj;
  tmp_clkconj = NULL;

  ns->xconj_cycle = advance(ns->xconj_cycle, FALSE);
  tmp_clkconj = conjoin(ns->xconj_cycle, DELTA);	
  ns->xconj_cycle = tmp_clkconj;
  tmp_clkconj = NULL;

  ns->zone = zone(ns->xconj_cycle);
  ns->zone = process_zone_max_const(ns->zone);
  ns->xconj_cycle = zone2clkconj(ns->zone);
  gn->gio_cycle->gio->clk_literal = get_literals(gn->gio_cycle->gio);

  print_initial_gor_status(gn);
        
  init_23tree(&GIO_TREE);
  init_23tree(&GOR_TREE);

  add_23tree(gn->gio_cycle->gio, GIO_TREE, gio_comp, free_gio, gio_count_inc, 
             gio_parent_set, print_gio);

  add_23tree(gn, GOR_TREE, gor_comp, free_gor, gor_count_inc, 
             gor_parent_set, print_gor);

  return(gn);
}
  /* initial_gor() */

int print_initial_gor_status(struct gor_type *gn) {
  struct mode_type *ns = gn->gio_cycle->gio->mode;
  if(verbose == 0) {
    sprintf(tmpstring, "gor(%d): {%s[%d(%d, %d)]}: \n", 
            gn->index, ns->name, ns->index, 
            ns->submode1->index, ns->submode2->index);
    output(tmpstring); 
/*
    output("inv bdd: ");
    print_bdd(ns->boolexp->boolexp_bdd);
*/
    output("\tZone: ");
    print_zone(ns->zone);
    output("\n");
    output("\tBDD: ");
    print_bdd(ns->dval_bdd);
    output("\tLiterals: ");
    print_literals(gn->gio_cycle->gio->clk_literal);
  }

  if(verbose > 0) {
    sprintf(tmpstring, 
            "\n###  Initial GOR created (%d) {%s[%d(%d, %d)]}  ###\n", 
            gn->index,  ns->name, ns->index, 
            ns->submode1->index, ns->submode2->index);
    output(tmpstring);
    sprintf(tmpstring, "BDD of mode (%s)[%d(%d, %d)] is: \n\t", 
                 ns->name, ns->index, ns->submode1->index, ns->submode2->index);
    output(tmpstring);
    print_bdd(ns->boolexp->boolexp_bdd);
    sprintf(tmpstring, "ZONE of mode (%s)[%d(%d, %d)] is:\n\t", 
                 ns->name, ns->index, ns->submode1->index, ns->submode2->index);
    output(tmpstring);
    print_zone(ns->zone);
    output("\n");
  }
}

int get_gio_count(struct gor_type *g) {
  struct gio_link_type *gl;
  int count = 0;
  if(g) {
    for(gl = g->gio_cycle->next; ; gl=gl->next) {
      count++;
      if(gl == g->gio_cycle) break;
    }
  }
  return count;
}

int gor_comp(struct gor_type *g1, struct gor_type *g2) {
  int comp = EQUAL, g1_gio_count, g2_gio_count;
  struct gio_link_type *gl1, *gl2;
  struct mode_type *m;
 
/*
  sprintf(tmpstring, "Comparing GORs %d and %d.\n", g1->index, g2->index);
  output(tmpstring);
*/
  if(g1 && g2) {
    g1_gio_count = get_gio_count(g1);
    g2_gio_count = get_gio_count(g2);
    if(g1_gio_count < g2_gio_count) return -1;
    if(g1_gio_count > g2_gio_count) return 1;
    for(gl1 = g1->gio_cycle->next, gl2=g2->gio_cycle->next; ; 
        gl1 = gl1->next, gl2 = gl2->next) {
      comp = gio_comp(gl1->gio, gl2->gio);
      if(comp != EQUAL) {
/*
        sprintf(tmpstring, "comp result of gio_comp() = %d.\n", comp);
        output(tmpstring);
*/
        return comp;
      }
      if(gl1 == g1->gio_cycle && gl2 == g2->gio_cycle) break;
    }
    return EQUAL;
  }
  else fprintf(stderr, "Fatal error in gor_comp().\n");
/*
  if(g1) {
    for(gl = g1->gio_cycle->next; ; gl=gl->next) {
      if(gio_in_gor(gl->gio, g2) == FALSE) {
        comp = 1; 
        break;
      }
      if(gl == g1->gio_cycle) break;
    }
  }
  if(g2) {
    for(gl = g2->gio_cycle->next; ; gl=gl->next) {
      if(gio_in_gor(gl->gio, g1) == FALSE) {
        comp = -1; 
        break;
      }
      if(gl == g2->gio_cycle) break;
    }
  }
  return comp;
*/
}

int gio_in_gor(struct gio_type *gio, struct gor_type *gor) {
  struct gio_link_type *gp;
  if(gor) {
    for(gp = gor->gio_cycle->next; ; gp = gp->next) {
      if(gio_comp(gp->gio, gio) == EQUAL) return TRUE;
      if(gp == gor->gio_cycle) break;
    }
  }
  return FALSE;
}

int gio_comp(struct gio_type *g1, struct gio_type *g2) {
  return mode_comp(g1->mode, g2->mode);
}

int free_gor(struct gor_type *g) {
}

int free_gio(struct gio_type *g) {
/*
  free_mode(g->mode);
  free_clkconj(g->clk_literal);
  free(g);
*/
}

int free_gio_cycle(struct gor_type *g) {
/*
  struct gio_link_type *gl;
  if(g->gio_cycle) {
    gl = g->gio_cycle->next;
    g->gio_cycle->next = NULL;
    while(gl) {
      g->gio_cycle = gl->next;
      free_gio(gl->gio);
      free(gl);
      gl = g->gio_cycle; 
    }
  }
*/
  g->gio_cycle = NULL;
}

gor_count_inc(struct gor_type *g) {
  (g->count)++;
}

gio_count_inc(struct gio_type *g) {
  (g->count)++;
}

gor_parent_set(struct gor_type *g, struct a23tree_type *parent) {
  g->parent = parent;
}

gio_parent_set(struct gio_type *g, struct a23tree_type *parent) {
  g->parent = parent;
}

int print_gor(struct gor_type *g) {
  struct gio_link_type *gl;
  struct gis_transition_type *gtr;

  sprintf(tmpstring, "GOR (%d): {", g->index);
  output(tmpstring);
  for(gl = g->gio_cycle->next; ; gl = gl->next) {
    sprintf(tmpstring, "%d", gl->gio->index);
    output(tmpstring);
    if(gl == g->gio_cycle) break;
    else output(", ");
  }
  output("}\n");
  for(gl = g->gio_cycle->next; ; gl = gl->next) {
    print_gio(gl->gio);
    if(gl == g->gio_cycle) break;
  }
  if(g->out_gis_trans_cycle) {
    output("\tOUT: ");
    for(gtr = g->out_gis_trans_cycle->next; ; gtr = gtr->next) {
      sprintf(tmpstring, "%d(%d)", gtr->gor->index, gtr->tr->index);
      output(tmpstring);
      if(gtr == g->out_gis_trans_cycle) break;
      else output(", ");
    }
    output("\n");
  }
  if(g->in_gis_trans_cycle) {
    output("\tIN: ");
    for(gtr = g->in_gis_trans_cycle->next; ; gtr = gtr->next) {
      sprintf(tmpstring, "%d(%d)", gtr->gor->index, gtr->tr->index);
      output(tmpstring);
      if(gtr == g->in_gis_trans_cycle) break;
      else output(", ");
    }
    output("\n");
  }
}

print_gio(struct gio_type *g) {
  sprintf(tmpstring, "    GIO (%d): %s(%d)\n", 
          g->index, g->mode->name, g->mode->index);
  output(tmpstring);
  output("\tZone: "); 
  print_zone(g->mode->zone); 
  output("\n");
  output("\tClkconj: ");
  print_clkconj(g->mode->xconj_cycle);
  output("\n");
  output("\tBDD: "); 
  print_bdd(g->mode->dval_bdd);
  output("\tLiterals: ");
  print_literals(g->clk_literal);
  output("\n");
}

/* A^k transition */

int gis_transition_n1_first(struct gor_type *g) {
  int i;
  bdd tmp_bdd;
  struct gor_type *ng, *nng;
  struct gio_type *ngio, *nngio;
  struct gio_link_type *giol, *nngiol;
  struct mode_type *m, *m1, *m2, *nm, *nm1;
  struct transition_type *tr1; 
  struct gis_transition_type *ntr1;

  m1 = g->gio_cycle->gio->mode->submode1;
  if(m1->out_trans_cycle) {
   for (i = 0, tr1 = m1->out_trans_cycle->next; i < m1->out_trans_count;
       i++, tr1 = tr1->next) {
    print_trans_status(tr1);
    ng = alloc_gor();
    print_new_gor_status(ng);
    for(giol = g->gio_cycle->next; ; giol = giol->next) {
      m = giol->gio->mode;
      m2 = m->submode2;
      nm = alloc_mode();
      nm->submode1 = nm1 = tr1->mode;
      nm->submode2 = m2;
      nm->name = assign_mode_name(nm, (int *)NULL, 0);
      ngio = alloc_gio();
      ngio->mode = nm;
      print_new_gio_status(giol->gio, tr1, ngio);
      nm = post_prop(m, tr1, &nm);
      if(nm == NULL) {
        free_gio(ngio);
        free_gio_cycle(ng);
        break;
      }
      nm->zone = post_xconj(m, nm1->xconj_cycle, m2->xconj_cycle, 
		            tr1->clk_trigger_cycle, tr1->assign_cycle, tr1, 
			    &(nm->xconj_cycle));
      nm->xconj_cycle = zone2clkconj(nm->zone);
      if(nm->zone->relop[0][0] == FALSE) {
        if(verbose >= 0) output("\tFalse ZONE.\n");
        free_gio(ngio);
        free_gio_cycle(ng);
        break;
/*
        if(giol == g->gio_cycle) break;
        else continue;
*/
      }
      else {
        print_zone_status(nm);
        ngio->clk_literal = get_literals(ngio);
        nngio = (struct gio_type *)
	        add_23tree(ngio, GIO_TREE, gio_comp, free_gio, gio_count_inc, 
                           gio_parent_set, print_gio);
        if(nngio != ngio) {
          sprintf(tmpstring, "New gio VISITED (%d).\n", nngio->index);
          output(tmpstring);
        }
        nngiol = (struct gio_link_type *)malloc(sizeof(struct gio_link_type));
        nngiol->gio = nngio;
        append_gio(&nngiol, &ng, TRUE);
      } 
      if(giol == g->gio_cycle) break;
    }
    if(ng->gio_cycle == NULL) {
      free_gor(ng);
      continue;
    }
    if(consistent_gor(ng)) {
      nng = (struct gor_type *)
            add_23tree(ng, GOR_TREE, gor_comp, free_gor, gor_count_inc,
                       gor_parent_set, print_gor);
      ntr1 = add_gis_transition(tr1, g, nng);
      GIS_THREAD->gis_trans_count++;
      if(nng == ng) {
        GIS_THREAD->gor_count++;
        last_gor->cnext = nng;
        last_gor = nng;
        nng->next = TOP;
        TOP = nng;
        if(verbose >= 0) {
          sprintf(tmpstring, "Pushed gor (%d) into stack.\n", ng->index);
          output(tmpstring);
          print_gor(ng); 
        }
      }
      else {
        sprintf(tmpstring, "New GOR already VISITED (%d).\n", nng->index); 
        output(tmpstring);
      }
    }
    else {
      sprintf(tmpstring, "New GOR (%d) inconsistent, freeing it.\n", ng->index);
      output(tmpstring);
      free_gor(ng);
    }
   }
  }
}
      /* gis_transition_n1_first() */

int print_new_gor_status(struct gor_type *ng) {
  if(verbose >= 0) {
    sprintf(tmpstring, "### New GOR created (%d).\n", ng->index);
    output(tmpstring);
  }
}

int print_trans_status(struct transition_type *tr) {
  if(verbose >= 0) {
    sprintf(tmpstring, "    considering transition %d (-> %s) with ", 
                       tr->index, tr->mode->name);
    output(tmpstring);
    print_perm(tr->permutation->perm);
  }
}

int print_new_gio_status(struct gio_type *gio, struct transition_type *tr,
                         struct gio_type *ngio) {
  struct mode_type *m = gio->mode, 
                   *nm = ngio->mode;
  if(verbose == 0) {
    sprintf(tmpstring, "gio(%d): (%s)[%d(%d, %d)]: \n\t", 
            ngio->index, nm->name, nm->index, 
            nm->submode1->index, nm->submode2->index);
    output(tmpstring);
  }
  if(verbose > 0) {
    sprintf(tmpstring, 
            "\nNew GIO created (%d) (%s)[%d(%d, %d)] from parent GIO (%d)(%s)[%d(%d, %d)]\n",
            ngio->index, nm->name, nm->index, 
            nm->submode1->index, nm->submode2->index,
	    gio->index, m->name, m->index, 
            m->submode1->index, m->submode2->index);
    output(tmpstring);
  }
  if(verbose > 1) {
    sprintf(tmpstring, 
            "Parent gio(%d): (%s)[%d(%d, %d)] has invariant BDD: \n\t",
	    gio->index, m->name, m->index, 
            m->submode1->index, m->submode2->index);
    output(tmpstring);
    print_bdd(m->boolexp->boolexp_bdd);
    sprintf(tmpstring, 
    	    "Parent gio(%d): (%s)[%d(%d, %d)] has discrete valuation BDD: \n\t",
	    gio->index, m->name, m->index, 
            m->submode1->index, m->submode2->index);
    output(tmpstring);
    print_bdd(m->dval_bdd);
    sprintf(tmpstring, 
            "Transition (%d) has trigger BDD: \n\t", tr->index);
    output(tmpstring);
    if(tr->trigger_bdd) 
      print_bdd(tr->trigger_bdd);
    else output("NULL BDD\n");
  }
}
    
int print_zone_status(struct mode_type *nm) {
  if(verbose == 0) {
    output("\tZone: ");
    print_zone(nm->zone);
    output("\n");
  }
  if(verbose > 0) {
    sprintf(tmpstring, "ZONE of mode(%s) [%d(%d, %d)] is:\n\t", 
             nm->name, nm->index, nm->submode1->index, nm->submode2->index);
    output(tmpstring);
    print_zone(nm->zone);
    output("\n");
  }
}

/*
 * Assign the boolexp and dval_bdd of nm which is a new mode obtained
 * by traversing the transition tr from mode m 
 */

struct mode_type *post_prop(struct mode_type *m, struct transition_type *tr, 
                            struct mode_type **nm) {
  bdd tmp_bdd;

/*
  sprintf(tmpstring, "Taking transition (%d) from mode %s(%d) to %s(%d):\n", 
          tr->index, m->name, m->index, (*nm)->name, (*nm)->index);
  output(tmpstring);
  print_trans(tr);
*/
  
  if((BDD_AND(m->boolexp->boolexp_bdd, tr->trigger_bdd)) == BDD_FALSE) {
    if(verbose >= 0)
      output("Invalid transition (BDD: invariant & trigger = false).\n");
    free_mode(*nm);
    return NULL;
  }

  (*nm)->boolexp = alloc_boolexp();
  (*nm)->dval_bdd = (bdd)NULL;
  (*nm)->boolexp->boolexp_bdd = BDD_AND((*nm)->submode1->boolexp->boolexp_bdd,
   	 			        (*nm)->submode2->boolexp->boolexp_bdd); 
  if(compact_reduction) 
    (*nm)->boolexp->boolexp_bdd = permute_bdd((*nm)->boolexp->boolexp_bdd, 
                                              tr->permutation->perm);
  tmp_bdd = BDD_AND(m->dval_bdd, tr->trigger_bdd);
  if(tmp_bdd == BDD_FALSE) {
    if(verbose >= 0) 
      output("Invalid mode (BDD: discrete variable valuation = false).\n");
    free_mode(*nm);
    return NULL;
  }
  if( BDD_AND(((*nm)->dval_bdd = bdd_assign(m->dval_bdd, tr->assign_cycle)),
               (*nm)->boolexp->boolexp_bdd
              ) == BDD_FALSE ) {
    if(verbose >= 0) 
      output("Invalid mode (BDD: discrete var valuation = false).\n");
    free_mode(*nm);
    return NULL;
  }
/*
  sprintf(tmpstring, "After assignment, BDD: ");
  output(tmpstring);
  print_bdd((*nm)->dval_bdd);
*/
    
  if(compact_reduction) 
    (*nm)->dval_bdd = permute_bdd((*nm)->dval_bdd, tr->permutation->perm);

/*
  sprintf(tmpstring, "After permutation, BDD: ");
  output(tmpstring);
  print_bdd((*nm)->dval_bdd);
*/
    
  if(verbose == 0) {
    output("inv bdd: "); 
    print_bdd((*nm)->boolexp->boolexp_bdd);
    output("\tdval bdd: "); 
    print_bdd((*nm)->dval_bdd);
  }
  if(verbose > 0) {
    sprintf(tmpstring, "Invariant BDD of mode (%s) [ %d(%d, %d) ] is:\n\t", 
                       (*nm)->name, (*nm)->index, 
                       (*nm)->submode1->index, (*nm)->submode2->index);
    output(tmpstring);
    print_bdd((*nm)->boolexp->boolexp_bdd);
    sprintf(tmpstring, 
	    "Discrete variable valuation BDD of mode(%s) [%d(%d, %d)] is:\n\t",
     	     (*nm)->name, (*nm)->index, 
             (*nm)->submode1->index, (*nm)->submode2->index);
    output(tmpstring);
    print_bdd((*nm)->dval_bdd);
  }
  return (*nm);
}

int print_append_status(struct gio_link_type **g, struct gor_type **ng) {
  sprintf(tmpstring, "Appending gio (%d) to gor (%d).\n", 
          (*g)->gio->index, (*ng)->index);
  output(tmpstring);
}

/* the GIOs in a gio_cycle are ordered according to the gio_comp() ascending
 * values; if two values are same, then they are ordered according to their
 * index values.
 */
int append_gio(struct gio_link_type **g, struct gor_type **ng, int set) {
  struct gio_link_type  *glp, *gl;
  int comp = EQUAL;

  if((*ng)->gio_cycle) {
    for(glp=(*ng)->gio_cycle, gl=glp->next; ; glp=gl, gl=gl->next) {
      comp = gio_comp(gl->gio, (*g)->gio);
      if(comp > 0) {
        (*g)->next = gl;
        glp->next = *g;
        return;
      } 
      else if(comp == EQUAL) {
        if(set == TRUE) return;
        else if( set == FALSE) {
          if((*g)->gio->index > gl->gio->index) { 
            if(gl == (*ng)->gio_cycle) break;
            else continue;
          }
          else {
            (*g)->next = gl;
            glp->next = *g;
            return;
          }
        }
      }
      if(gl == (*ng)->gio_cycle) break;
    }
    (*g)->next = (*ng)->gio_cycle->next;
    (*ng)->gio_cycle->next = *g;
    (*ng)->gio_cycle = *g;
  }
  else 
    (*ng)->gio_cycle = (*g)->next = *g;
}

struct gis_transition_type *add_gis_transition(struct transition_type *tr, 
                                               struct gor_type *gp, 
                                               struct gor_type *gs) {
  struct gis_transition_type *ntr;

  /* for out-going trans */
  ntr = alloc_gis_trans();
  ntr->index = gis_trans_count++;
  ntr->gor = gs;
  ntr->tr = tr;
  if(gp->out_gis_trans_cycle) {               
    ntr->next = gp->out_gis_trans_cycle->next;
    gp->out_gis_trans_cycle->next = ntr;
    gp->out_gis_trans_cycle = ntr;
  }
  else
    gp->out_gis_trans_cycle = ntr->next = ntr;

  /* for in-coming trans */
  ntr = alloc_gis_trans();
  ntr->gor = gp;
  ntr->tr = tr;
  ntr->index = gis_trans_count;
  if(gs->in_gis_trans_cycle) {
    ntr->next = gs->in_gis_trans_cycle->next;
    gs->in_gis_trans_cycle->next = ntr;
    gs->in_gis_trans_cycle = ntr;
  }
  else
    gs->in_gis_trans_cycle = ntr->next = ntr;

  return gp->out_gis_trans_cycle;
}

struct gis_transition_type *alloc_gis_trans() {
  struct gis_transition_type *ntr = (struct gis_transition_type *)
                                    malloc(sizeof(struct gis_transition_type));
  ntr->tr = (struct transition_type *)NULL;
  ntr->gor = (struct gor_type *)NULL;
  ntr->next = (struct gis_transition_type *)NULL;
  return ntr;
}

int consistent_gor(struct gor_type *g) {
  struct gio_link_type *giol;
  struct clkconj_type *c = NULL;
  bdd f = BDD_TRUE;

  for(giol = g->gio_cycle->next; ; giol=giol->next) {
    f = BDD_AND(f, giol->gio->mode->dval_bdd);
/*
    c = clkconj_and(c, giol->gio->clk_literal);
*/
    c = clkconj_and(c, giol->gio->mode->xconj_cycle);
    if(f == BDD_FALSE) return FALSE;
    if(c->op == FALSE) return FALSE;
    if(giol == g->gio_cycle) break;
  }
  return TRUE;
}

int gis_transition_n2_first(struct gor_type *g, struct gor_type *og) {
  int i, *perm = NULL;
  bdd tmp_bdd;
  struct gor_type *ng, *nng;
  struct gio_type *ngio, *nngio, *ngm;
  struct gio_link_type *gml, *giol, *nngiol;
  struct mode_type *m, *m1, *m2, *om2, *nm, *nm2;
  struct transition_type *tr2, *ptr2; 
  struct gis_transition_type *ntr2;
  struct clkconj_type *submode1_clkconj = NULL, *submode2_clkconj = NULL;

  for(gml = g->gio_cycle->next; ; gml = gml->next) {
    m2 = gml->gio->mode->submode2;
    if(m2->out_trans_cycle) {
     for(i = 0, tr2 = m2->out_trans_cycle->next; i < m2->out_trans_count;
        i++, tr2 = tr2->next) {
      print_trans_status(tr2);
      ng = alloc_gor();
      print_new_gor_status(ng);

      extend_zones(g, gml->gio);

      for(giol = g->gio_cycle->next; ; giol = giol->next) {

        if(giol->gio->index != gml->gio->index) {
          perm = gen_perm_gis(CTA, CTA+1);
          ptr2 = alloc_transition();
          ptr2 = copy_transition(ptr2, tr2);
          sprintf(tmpstring, "Original transition (%d):\n", ptr2->index);
          output(tmpstring);
          print_trans(ptr2); 
          ptr2 = permute_transition(ptr2, perm);
          sprintf(tmpstring, "Permuted transition (%d):\n", ptr2->index);
          output(tmpstring);
          print_trans(ptr2); 
        }
        else ptr2 = tr2;

        m = giol->gio->mode;
        m1 = m->submode1;
        om2 = m->submode2;
        nm = alloc_mode();
/*
        nm->submode1 = m1;
*/
        if(giol->gio->index != gml->gio->index) 
          nm->submode1 = m;
        else nm->submode1 = m1;

        nm->submode2 = nm2 = ptr2->mode;
        nm->name = assign_mode_name(nm, (int *)NULL, 0);
        ngio = alloc_gio();
        ngio->mode = nm;
        print_new_gio_status(giol->gio, tr2, ngio);
        nm = post_prop(m, ptr2, &nm);
        if(nm == NULL) {
          free_gio(ngio);
          free_gio_cycle(ng);
          break;
/*
          if(giol == g->gio_cycle) break;
          else continue;
*/
        }
        if(giol->gio->index != gml->gio->index) {
          submode1_clkconj = m->xconj_cycle;
          submode2_clkconj = permute_clkconj(nm2->xconj_cycle, perm);
        }
        else {
          submode1_clkconj = m1->xconj_cycle;
          submode2_clkconj = nm2->xconj_cycle;
        }
        
        nm->zone = post_xconj(m, submode1_clkconj, submode2_clkconj, 
		             ptr2->clk_trigger_cycle, ptr2->assign_cycle, ptr2, 
			      &(nm->xconj_cycle));
        nm->xconj_cycle = zone2clkconj(nm->zone);
        if(nm->zone->relop[0][0] == FALSE) {
          if(verbose >= 0) output("\tFalse ZONE.\n");
          free_gio(ngio);
          free_gio_cycle(ng);
          break;
/*
          if(giol == g->gio_cycle) break;
          else continue;
*/
        }
        else {
          print_zone_status(nm);
          ngio->clk_literal = get_literals(ngio);
          nngio = (struct gio_type *)
	          add_23tree(ngio, GIO_TREE, gio_comp, free_gio, gio_count_inc, 
                             gio_parent_set, print_gio);
          if(nngio != ngio) {
            sprintf(tmpstring, "New gio VISITED (%d).\n", nngio->index);
            output(tmpstring);
          }
          nngiol = (struct gio_link_type *)malloc(sizeof(struct gio_link_type));
          nngiol->gio = nngio;
          print_append_status(&nngiol, &ng);
          append_gio(&nngiol, &ng, TRUE);
        } 
/*
        if(giol->gio->index == gml->gio->index) ngm = nngio;
*/
        if(giol->gio->index != gml->gio->index) 
          nngio = conjoin_gio(nngio, m1, om2);
        if(giol == g->gio_cycle) break;
      }
      if(ng->gio_cycle == NULL) {
        free_gor(ng);
        continue;
      }
/*
      ng = conjoin_gor(ng, ngm, m1, om2);
*/
      ng = set_gor(ng);
      output("After conjoining and transforming into a set, new GOR:\n"); 
      print_gor(ng);

      if(consistent_gor(ng)) {
        nng = (struct gor_type *)
              add_23tree(ng, GOR_TREE, gor_comp, free_gor, gor_count_inc,
                         gor_parent_set, print_gor);
        ntr2 = add_gis_transition(tr2, og, nng);
        GIS_THREAD->gis_trans_count++;
        if(nng == ng) {
          GIS_THREAD->gor_count++;
          last_gor->cnext = nng;
          last_gor = nng;
          nng->next = TOP;
          TOP = nng;
          if(verbose >= 0) {
            sprintf(tmpstring, "Pushed gor (%d) into stack\n", ng->index);
            output(tmpstring);
            print_gor(ng); 
          }
        }
        else {
          sprintf(tmpstring, "New GOR already VISITED (%d).\n", nng->index); 
          output(tmpstring);
        }
      }
      else {
        sprintf(tmpstring, 
                "New GOR (%d) inconsistent, freeing it.\n", ng->index);
        output(tmpstring); 
        free_gor(ng);
      }
     }
    }
    if(gml == g->gio_cycle) break;
  }
}
  /* gis_transition_n2_first() */

int gis_transition_n2_first_split(struct gor_type *g, struct gor_type *og) {
  int i, *perm = NULL;
  bdd tmp_bdd;
  struct gor_type *eg, *ng, *nng;
  struct gio_type *gm, *ngm, *gio, *ngio, *nngio;
  struct gio_link_type *gml, *giol, *nngiol;
  struct mode_type *m, *m1, *m2, *om2, *nm, *nm2;
  struct transition_type *tr2, *ptr2; 
  struct gis_transition_type *ntr2;
  struct clkconj_type *submode1_clkconj = NULL, *submode2_clkconj = NULL;

  for(gml = g->gio_cycle->next; ; gml = gml->next) {
    m2 = gml->gio->mode->submode2;
    if(m2->out_trans_cycle) {
     for(i = 0, tr2 = m2->out_trans_cycle->next; i < m2->out_trans_count;
        i++, tr2 = tr2->next) {
      print_trans_status(tr2);
      ng = alloc_gor();
      print_new_gor_status(ng);
      eg = extend_gor(g, gml->gio, tr2);
      for(giol = eg->gio_cycle->next; ; giol = giol->next) {
        if(giol->gio->index != gml->gio->index) {
          perm = gen_perm_gis(CTA, CTA+1);
          ptr2 = alloc_transition();
          ptr2 = copy_transition(ptr2, tr2);
          sprintf(tmpstring, "Original transition (%d):\n", ptr2->index);
          output(tmpstring);
          print_trans(ptr2); 
          ptr2 = permute_transition(ptr2, perm);
          sprintf(tmpstring, "Permuted transition (%d):\n", ptr2->index);
          output(tmpstring);
          print_trans(ptr2); 
        }
        else ptr2 = tr2;
        m = giol->gio->mode;
        m1 = m->submode1;
        om2 = m->submode2;
        nm = alloc_mode();
        if(giol->gio->index != gml->gio->index) 
          nm->submode1 = m;
        else nm->submode1 = m1;
        nm->submode2 = nm2 = ptr2->mode;
        nm->name = assign_mode_name(nm, (int *)NULL, 0);
        ngio = alloc_gio();
        ngio->mode = nm;
        print_new_gio_status(giol->gio, ptr2, ngio);
        nm = post_prop(m, ptr2, &nm);
        if(nm == NULL) {
          free_gio(ngio);
          free_gio_cycle(ng);
          break;
/*
          if(giol == eg->gio_cycle) break;
          else continue;
*/
        }
        if(giol->gio->index != gml->gio->index) {
          submode1_clkconj = m->xconj_cycle;
          submode2_clkconj = permute_clkconj(nm2->xconj_cycle, perm);
        }
        else {
          submode1_clkconj = m1->xconj_cycle;
          submode2_clkconj = nm2->xconj_cycle;
        }
        
        nm->zone = post_xconj(m, submode1_clkconj, submode2_clkconj,
		              ptr2->clk_trigger_cycle, ptr2->assign_cycle, 
                              ptr2, &(nm->xconj_cycle));
        nm->xconj_cycle = zone2clkconj(nm->zone);
        if(nm->zone->relop[0][0] == FALSE) {
          if(verbose >= 0) output("\tFalse ZONE.\n");
          free_gio(ngio);
          free_gio_cycle(ng);
          break;
/*
          if(giol == eg->gio_cycle) break;
          else continue;
*/
        }
        else {
          print_zone_status(nm);
          ngio->clk_literal = get_literals(ngio);
          nngio = (struct gio_type *)
	          add_23tree(ngio, GIO_TREE, gio_comp, free_gio, gio_count_inc, 
                             gio_parent_set, print_gio);
          if(nngio != ngio) {
            sprintf(tmpstring, "New gio VISITED (%d).\n", nngio->index);
            output(tmpstring);
          }
          nngiol = (struct gio_link_type *)malloc(sizeof(struct gio_link_type));
          nngiol->gio = nngio;
          print_append_status(&nngiol, &ng);
          append_gio(&nngiol, &ng, TRUE);
        } 
        if(giol->gio->index != gml->gio->index)
          nngio = conjoin_gio(nngio, m1, om2);
        if(giol == eg->gio_cycle) break;
      }
      if(ng->gio_cycle == NULL) {
        free_gor(ng);
        continue;
      }
/*
      ng = conjoin_gor(ng, ngm, m1, om2);
*/
      ng = set_gor(ng);
      output("After conjoining and transforming into a set, new GOR:\n"); 
      print_gor(ng);
      if(consistent_gor(ng)) {
        nng = (struct gor_type *)
              add_23tree(ng, GOR_TREE, gor_comp, free_gor, gor_count_inc,
                         gor_parent_set, print_gor);
        ntr2 = add_gis_transition(tr2, og, nng);
        GIS_THREAD->gis_trans_count++;
        if(nng == ng) {
          GIS_THREAD->gor_count++;
          last_gor->cnext = nng;
          last_gor = nng;
          nng->next = TOP;
          TOP = nng;
          if(verbose >= 0) {
            sprintf(tmpstring, "Pushed gor (%d) into stack\n", ng->index);
            output(tmpstring);
            print_gor(ng); 
          }
        }
        else {
          sprintf(tmpstring, "New GOR already VISITED (%d).\n", nng->index); 
          output(tmpstring);
        }
      }
      else {
        sprintf(tmpstring, 
                "New GOR (%d) inconsistent, freeing it.\n", ng->index);
        output(tmpstring); 
        free_gor(ng);
      }
     }
    }
    if(gml == g->gio_cycle) break;
  }
}
  /* gis_transition_n2_first_split() */

struct gor_type *extend_gor(struct gor_type *gor, struct gio_type *gio,
                            struct transition_type *tr) {
  struct gio_type *gp = copy_gio(gio);
  struct gio_link_type *gpl = (struct gio_link_type *)
                              malloc(sizeof(struct gio_link_type));
  struct gor_type *ngor = copy_gor(gor);

  sprintf(tmpstring, "Extending GOR (%d) into GOR (%d) using GIO (%d):\n", 
          gor->index, ngor->index, gio->index);
  output(tmpstring);
  gp->index = gio_count++;
  gpl->gio = gp;
  append_gio(&gpl, &ngor, FALSE);
  extend_zones(ngor, gio);
  print_gor(ngor);
  return ngor;
}

int extend_zones(struct gor_type *gor, struct gio_type *gio) {
  int i;
  struct gio_link_type *g;
  sprintf(tmpstring, "  extending zones using zone of gio (%d).\n", gio->index);
  output(tmpstring);
  for(g = gor->gio_cycle->next; ; g=g->next) {
    if(g->gio->index != gio->index) {
      for(i=0; i<=total_clocks; i++) {
        g->gio->mode->zone->relop[CTA+1][i] = gio->mode->zone->relop[CTA][i];
        g->gio->mode->zone->diff[CTA+1][i] = gio->mode->zone->diff[CTA][i];
        g->gio->mode->zone->relop[i][CTA+1] = gio->mode->zone->relop[i][CTA];
        g->gio->mode->zone->diff[i][CTA+1] = gio->mode->zone->diff[i][CTA];
      }
    }
    if(g == gor->gio_cycle) break;
  }
}

struct gor_type *conjoin_gor(struct gor_type *gor, struct gio_type *gio,
                             struct mode_type *om1, struct mode_type *om2) {
  struct gio_link_type *g;

  for(g = gor->gio_cycle->next; ; g=g->next) {
    if(g->gio->index != gio->index) {
      g->gio = conjoin_gio(g->gio, om1, om2);
    }
    if(g == gor->gio_cycle) break;
  }
  return gor;
}

struct gio_type *conjoin_gio(struct gio_type *gio, struct mode_type *om1, 
                             struct mode_type *om2) {
  struct clkconj_type *c;
  if(gio) {
    free(gio->mode->name);
    gio->mode->name = str_preserve(gio->mode->submode1->name);
    gio->mode->submode1 = om1;
    gio->mode->submode2 = om2;
    c = zone2clkconj(gio->mode->zone);
    c = conjoin(c, CTA+1);
    free_zone(gio->mode->zone);
    gio->mode->zone = zone(c);
    gio->mode->zone = process_zone_max_const(gio->mode->zone);
    free_clkconj(gio->mode->xconj_cycle);
    gio->mode->xconj_cycle = zone2clkconj(gio->mode->zone);
  }
  return gio;
}

int *gen_perm_gis(int p1, int p2) {
  int i;
  int *perm = (int *)malloc(sizeof(int) * (total_clocks + 1));
  for(i=0; i<=total_clocks; i++) {
    perm[i] = i;
  }
  perm[p1] = p2;
  perm[p2] = p1;
  return perm;
}


int print_gis(struct gor_type *gis) {
  struct gor_type *gor;
  sprintf(tmpstring, "Printing GIS: %d nodes, %d transitions.\n", 
         GIS_THREAD->gor_count, GIS_THREAD->gis_trans_count);
  output(tmpstring);
  for(gor = gis; gor; gor=gor->cnext) {
    print_gor(gor);
  }
  output("END of printing GIS.\n");
}

struct gor_type *set_gor(struct gor_type *gor) {
  struct gio_link_type *gl, *gln;
  int gio_count = get_gio_count(gor);
  int elim[gio_count], i, j;

  for(i=0; i<gio_count; i++)
    elim[i] = FALSE;
  if(gor) {
    for(gl=gor->gio_cycle->next, i=0; i<gio_count; gl=gl->next, i++) {
      if(elim[i] == TRUE) continue;
      for(gln=gor->gio_cycle->next, j=0; j<gio_count; gln=gln->next, j++) {
        if(gl == gln) continue;
        if(elim[j] == TRUE) continue;
        if(gio_comp(gl->gio, gln->gio) == EQUAL) {
          if(gl->gio->index < gln->gio->index) elim[j] = TRUE;
          else if(gl->gio->index > gln->gio->index) {
            elim[i] = TRUE;
            break;
          }
        }
      }
    }
    for(gln=gor->gio_cycle, gl=gln->next, i=0; i<gio_count; i++) {
      if(elim[i] == TRUE) {
        gln->next = gl->next;
        if(gl == gor->gio_cycle) gor->gio_cycle = gln;
        free(gl);
        gl = gln->next;
      }
      else {
        gln = gl;
        gl = gl->next;
      }
    }
  }
  return gor;
}


/* Check if the transitions in thread[HS_THREAD_T] appear in GIS_THREAD */

int check_trans(struct gor_type *gis) {
  int tt = thread[HS_THREAD_T].trans_count;
  int *ts = (int *)malloc(sizeof(int)*tt), 
      i = 0, sum = 0;
  struct transition_type **trlist;
  struct mode_type **mlist;
  struct mode_type *m, *M;
  struct transition_type *tr, *e;
  struct gor_type *gor;
  struct gis_transition_type *gtr;

  printf("Checking and eliminating transitions in thread %d.\n", HS_THREAD_T);
  sprintf(tmpstring, 
          "Checking and eliminating transitions in thread %d.\n", HS_THREAD_T);
  output(tmpstring);
  sprintf(tmpstring, "Thread %d has %d transitions initially.\n", 
          HS_THREAD_T, tt);
  output(tmpstring);

  trlist = (struct transition_type **)
           malloc(sizeof(struct transition_type *)*tt);;
  mlist = (struct mode_type **) malloc(sizeof(struct mode_type *)*tt);;

  for(i=0; i<tt; i++) 
    ts[i] = FALSE;

  for(i = 0, m = thread[HS_THREAD_T].starting_mode; m; m = m->cnext) {
    if(m->out_trans_cycle) {
      for(tr = m->out_trans_cycle->next; ; tr = tr->next) {
        mlist[i] = m;
        trlist[i++] = tr;
        if(tr == m->out_trans_cycle) break;
      }
    }
  }
  for(gor = gis; gor; gor=gor->cnext) {
    sprintf(tmpstring, "Considering GOR %d.\n", gor->index);
    output(tmpstring);
    if(gor->out_gis_trans_cycle) {
      for(gtr = gor->out_gis_trans_cycle->next; ; gtr = gtr->next) {
        sprintf(tmpstring, "  Found transition %d in GIS.\n", gtr->tr->index);
        output(tmpstring);
        for(i=0; i<tt; i++) {
          if((trlist[i])->index == gtr->tr->index)
            ts[i] = TRUE;
        }
        if(gtr == gor->out_gis_trans_cycle) break;
      }
    }
  }
  for(i=0; i<tt; i++) {
    if(ts[i] == FALSE) {
      e = trlist[i];
      M = mlist[i];
      eliminate_transition(e, M);
      if(verbose >= 0) {
        sprintf(tmpstring,
                "ELIMINATED transition %d from mode %s (%d) to %s (%d).\n",
                e->index, M->name, M->index, e->mode->name, e->mode->index);
        output(tmpstring);
      }
      thread[HS_THREAD_T].trans_count--;
      sum++;
    }
  }
  sprintf(tmpstring, "Totally %d transitions eliminated from thread %d.\n", 
          sum, HS_THREAD_T);
  output(tmpstring);
}
