#include "include.h"
#include "global.h"
#include "print.h"
#include "bddman.h"
#include "modeman.h"
#include "zoneman.h"
#include "reduce.h"
#include "back.h"
#include "shield.h"
#include "lookahead.h"

int QS_THREAD = 0;
int qs_count = 0;
int qst_count = 0;
int total_ew_var = 0;
struct var_val_type *EW;
struct qs_node_type *qs_cycle = NULL;

int compute_val(struct variable_type *, struct thread_type *, int **);
int print_vv(struct var_val_type *);
int ext_write(struct variable_type *, struct thread_type *);
int init_EW(struct thread_type *);
struct qs_node_type *alloc_qs_node();
int free_qs_node(struct qs_node_type *);
int free_in_qs_trans_cycle(struct qs_node_type *);
int free_in_qs_trans(struct qs_transition_type *);
int free_out_qs_trans_cycle(struct qs_node_type *);
int free_out_qs_trans(struct qs_transition_type *);
int gen_quotient_structure(struct thread_type *);
int print_qs(struct qs_node_type *);
struct qs_transition_type *add_qs_transition(struct qs_node_type *,
                                             struct qs_node_type *,
                                             struct transition_type *);
struct qs_node_type *get_consistent_qs(struct mode_type *, 
                                       struct qs_node_type *);
struct qs_node_type *exec_trans(struct qs_node_type *, 
                                struct transition_type *, 
                                struct qs_node_type *);
int qs_consistent(struct qs_node_type *, struct mode_type *);
int new_mode_index(struct mode_type *);
int gen_qs_trans(struct qs_node_type *, struct thread_type *);
int *max_wt_vect(struct qs_node_type *, struct qs_node_type *);
int print_vector(int *);
int *add_vectors(int *, int *);
int *max_vectors(int *, int *);
int ***qs_kclosure(struct qs_node_type *);
int print_wt_vect_matrix(int ***w);
struct qs_node_link_type *gen_source_nodes(struct qs_node_type *);
int initial_qs(struct qs_node_type *);
int qs_elim_trans(struct thread_type *, int ***, struct qs_node_type *,
                  struct qs_node_link_type *);
int path_exists(struct transition_type *, struct thread_type *, 
                int ***, struct qs_node_type *, struct qs_node_link_type *);
int non_neg_vector(int *);


int compute_val(struct variable_type *v, struct thread_type *th, int **vals) {
  int total_int = 0;
  struct mode_type *m;
  struct transition_type *tr;
  struct proposition_type *pr;
  bdd f = BDD_FALSE;
  bdd t = BDD_TRUE;

  if(verbose >= 3) { 
    sprintf(tmpstring, 
        "Computing values of variable %s in thread %d.\n", v->name, th->index); 
    output(tmpstring);
  }
  for(m=th->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {

        if( (tr->trigger_bdd != NULL) && (tr->trigger_bdd != BDD_TRUE) ) {
          f = BDD_OR(f, tr->trigger_bdd);
/*
          output("After OR with trigger_bdd: "); 
          print_bdd(tr->trigger_bdd);
          output("f: "); 
          print_bdd(f);
*/
        }

        if(tr->assign_cycle != NULL) {
          t = bdd_assign(t, tr->assign_cycle);
/*
          output("After assign with assign_cycle: "); 
          print_assignment1(tr->assign_cycle);
          output("t: "); 
          print_bdd(t);
*/
        }

        if(t != BDD_TRUE) {
          f = BDD_OR(f, t);
/*
          output("After OR of f and t ... "); 
          output("f: "); 
          print_bdd(f);
*/
          t = BDD_TRUE;
        }

        if(tr == m->out_trans_cycle) break;
      }
    }
  }
/*
  if(verbose >= 3) { 
    output("\tCalculated bdd is:\n");
    print_bdd(f);
  }
*/
  for(pr=prop_tail->next; ; pr=pr->next) {
    if(v->index == pr->var) {
      if(bdd_and(bddm, pr->prop_bdd, f) != BDD_FALSE) {
        total_int++;
        (*vals) = (int *)realloc((*vals), sizeof(int)*total_int);
        (*vals)[total_int-1] = pr->value;
      }
    }
    if(pr == prop_tail) break;
  }
  return total_int;
}

int print_vv(struct var_val_type *vv) {
  int i;

  sprintf(tmpstring, "EW variable: %s, {", vv->var->name);
  output(tmpstring);
  for(i=0; i < vv->total_val; i++) {
    if(vv->val_list[i] == EXT_VAL) 
      output("EXT_VAL");
    else {
      sprintf(tmpstring, "%d", vv->val_list[i]);
      output(tmpstring);
    }
    if((i+1) < vv->total_val) 
      output(", ");
  }
  output("}\n");
  return;
}

int ext_write(struct variable_type *v, struct thread_type *th) {
  return TRUE;
}

int init_EW(struct thread_type *th) {
  struct variable_type *v;
  struct var_val_type *vv;

  for(v = var_tail->next; ; v=v->next) {
    if(v->type == V_REGISTER) {
      total_ew_var++;
      vv = (struct var_val_type *)
           malloc(sizeof(struct var_val_type));
      vv->var = v;
      vv->total_val = compute_val(v, th, &(vv->val_list));
      if(ext_write(v, th))
        vv->total_val++;
        vv->val_list = (int *)realloc(vv->val_list, 
                                      (vv->total_val) * sizeof(int));
        vv->val_list[(vv->total_val)-1] = EXT_VAL;
      if(verbose >= 0) 
        print_vv(vv);
      if(EW) {
        vv->next = EW->next;
        EW->next = vv;
      }
      else vv->next = vv;
      EW = vv;
    } 
    if(v == var_tail) break;
  }
  return;
}

struct qs_node_type *alloc_qs_node() {
  struct qs_node_type *qs;
  qs = (struct qs_node_type *)malloc(sizeof(struct qs_node_type));
  qs->index = qs_count++;
  qs->mode = NULL;
  qs->val_vect = NULL;
  qs->val_bdd = NULL;
  qs->out_qs_trans_cycle = NULL;
  qs->in_qs_trans_cycle = NULL;
  qs->next = NULL;
  return qs;
}

int free_qs_node(struct qs_node_type *qs) {
/*
  bdd_free(bddm, qs->val_bdd);
*/
  free(qs->val_vect);
  free_in_qs_trans_cycle(qs);
  free_out_qs_trans_cycle(qs);
  free(qs);
}

int free_in_qs_trans_cycle(struct qs_node_type *qsn) {
  struct qs_transition_type *x, *qst;

  if(qsn->in_qs_trans_cycle) {
    x = qsn->in_qs_trans_cycle->next; /* break the cycle into a list */
    qsn->in_qs_trans_cycle->next = NULL;
    qst = x;
    while(qst) {
      x = qst->next;
      free_in_qs_trans(qst);
      qst = x;
    }
    qsn->in_qs_trans_cycle = NULL;
  }
}

int free_out_qs_trans_cycle(struct qs_node_type *qsn) {
  struct qs_transition_type *x, *qst;

  if(qsn->out_qs_trans_cycle) {
    x = qsn->out_qs_trans_cycle->next; /* break the cycle into a list */
    qsn->out_qs_trans_cycle->next = NULL;
    qst = x;
    while(qst) {
      x = qst->next;
      free_out_qs_trans(qst);
      qst = x;
    }
    qsn->out_qs_trans_cycle = NULL;
  }
}

int free_in_qs_trans(struct qs_transition_type *qst) {
  struct qs_transition_type *x, *y;

  if(qst) {
    if(qst->node->out_qs_trans_cycle) {
      if(qst->node->out_qs_trans_cycle == qst->node->out_qs_trans_cycle->next) {
        free(qst->node->out_qs_trans_cycle->wt_vect);
        free(qst->node->out_qs_trans_cycle);
        qst->node->out_qs_trans_cycle = NULL;
      }
      else {
        for(x=qst->node->out_qs_trans_cycle, y=x->next; ; x=y, y=y->next) {
          if(y->index == qst->index) {
            x->next = y->next;
            free(y->wt_vect);
            free(y);
            if(qst->node->out_qs_trans_cycle->index == qst->index)
              qst->node->out_qs_trans_cycle = x;
            break;
          }
        }
      }
    }
    free(qst->wt_vect);
    free(qst);
  }
}

int free_out_qs_trans(struct qs_transition_type *qst) {
  struct qs_transition_type *x, *y;

  if(qst) {
    if(qst->node->in_qs_trans_cycle) {
      if(qst->node->in_qs_trans_cycle == qst->node->in_qs_trans_cycle->next) {
        free(qst->node->in_qs_trans_cycle->wt_vect);
        free(qst->node->in_qs_trans_cycle);
        qst->node->in_qs_trans_cycle = NULL;
      }
      else {
        for(x=qst->node->in_qs_trans_cycle, y=x->next; ; x=y, y=y->next) {
          if(y->index == qst->index) {
            x->next = y->next;
            free(y->wt_vect);
            free(y);
            if(qst->node->in_qs_trans_cycle->index == qst->index)
              qst->node->in_qs_trans_cycle = x;
            break;
          }
        }
      }
    }
    free(qst->wt_vect);
    free(qst);
  }
}

int gen_quotient_structure(struct thread_type *t) {
  struct mode_type *m;
  struct qs_node_type *qs;
  struct var_val_type *vv;
  struct proposition_type *pr;
  struct qs_node_link_type *source_cycle;
  int ***lend;
  int i = 0, j = 0;

  QS_THREAD = t->index;
  init_EW(t);
  qs_count = qst_count = 0;
  for(m=t->starting_mode; m; m=m->cnext) {
    i = 0;
    j = 0;
    if(EW->next == EW) {   /* only one EW variable */
      for(vv=EW->next; ; vv=vv->next) {
        for(j=0; j<vv->total_val; j++) {
          qs = alloc_qs_node();
          qs->mode = m;
          qs->val_vect = (int *)malloc(sizeof(int)*total_ew_var);
          qs->val_vect[i] = vv->val_list[j];
          for(pr = prop_tail->next; ; pr=pr->next) {
            if(pr->value == qs->val_vect[i]) {
              qs->val_bdd = BDD_AND(qs->val_bdd, pr->prop_bdd);
            }
            else
              qs->val_bdd = BDD_AND(qs->val_bdd, bdd_not(bddm, pr->prop_bdd));
            if(pr == prop_tail) break;
          }
          if(BDD_AND(m->boolexp->boolexp_bdd, qs->val_bdd) == BDD_FALSE) {
            free_qs_node(qs); 
            continue;
          }
          if(qs_cycle) {
            qs->next = qs_cycle->next;
            qs_cycle->next = qs;
          }
          else qs->next = qs;
          qs_cycle = qs;
        }
        if(vv == EW) break;
      }
    } /* END if(EW->next == EW) */
    else {
      fprintf(stderr, 
              "ERROR: Upgrade program to more than one global EW variable.\n"); 
      fprintf(stderr, 
              "REF: lookahead.c: gen_quotient_structure().");
    }
  }
  /* REMEMBER to generalize the above loop to more than one EW variable */

  gen_qs_trans(qs_cycle, t);

  if(verbose >= 0) {
    sprintf(tmpstring, 
        "Quotient Structure for thread %d (%d nodes, %d transitions).\n",
        t->index, qs_count, qst_count);
    output(tmpstring);
/*
    print_qs(qs_cycle);
*/
    sprintf(tmpstring,
        "END of Quotient Structure for thread %d (%d nodes, %d transitions).\n",
        t->index, qs_count, qst_count);
    output(tmpstring);
  }

  lend = qs_kclosure(qs_cycle);
  source_cycle = gen_source_nodes(qs_cycle);
  qs_elim_trans(t, lend, qs_cycle, source_cycle);
}

int print_qs(struct qs_node_type *qsc) {
  int i;
  struct qs_node_type *x;
  struct qs_transition_type *qst;

  if(qsc) {
    for(x=qsc->next; ; x=x->next) {
      sprintf(tmpstring, "QS_node (%d) <%s(%d), {", 
                         x->index, x->mode->name, x->mode->index);
      output(tmpstring);
      for(i=0; i<total_ew_var; i++) {
        if(x->val_vect[i] == EXT_VAL)
          output("EXT_VAL");
        else {
          sprintf(tmpstring, "%d", x->val_vect[i]);
          output(tmpstring);
        }
        if(i+1 < total_ew_var) output(", ");
      }
      output("}>\n");
/*
      output("\tBDD: "); 
      print_bdd(x->val_bdd);
*/
      if(x->out_qs_trans_cycle) {
        for(qst=x->out_qs_trans_cycle->next; ; qst=qst->next) {
          sprintf(tmpstring, "\tOut trans (%d) TO node %d <%s(%d), {",
                  qst->index, qst->node->index, qst->node->mode->name,
                                                qst->node->mode->index); 
          output(tmpstring);
          for(i=0; i<total_ew_var; i++) {
            if(qst->node->val_vect[i] == EXT_VAL)
              output("EXT_VAL");
            else {
              sprintf(tmpstring, "%d", qst->node->val_vect[i]);
              output(tmpstring);
            }
            if(i+1 < total_ew_var) output(", ");
          }
          output("}, [");
          if(qst->wt_vect) {
            for(i=0; i<automaton_mode_count; i++) {
              sprintf(tmpstring, "%d", qst->wt_vect[i]);
              output(tmpstring);
              if(i+1 < automaton_mode_count) 
                output(" ");
            }
          }
          output("]>\n\t  ");
          print_trans(qst->e);
          if(qst == x->out_qs_trans_cycle) break;
        }
      }
      if(x == qsc) break;
    }
  }
}

/* ADD a qs transition from qs node n1 to n2 using transition e */

struct qs_transition_type *add_qs_transition(struct qs_node_type *n1,
                                             struct qs_node_type *n2,
                                             struct transition_type *e) {
  struct qs_transition_type *qst;

  /* out-going transition of n1 */
  qst=(struct qs_transition_type *)malloc(sizeof(struct qs_transition_type));
  qst->index = qst_count;
  qst->wt_vect = NULL;
  qst->node = n2;
  qst->e = e;
  if(n1->out_qs_trans_cycle) {
    qst->next = n1->out_qs_trans_cycle->next;
    n1->out_qs_trans_cycle->next = qst;
  }
  else
    qst->next = qst;
  n1->out_qs_trans_cycle = qst;
  
  /* in-coming transition of n2 */
  qst=(struct qs_transition_type *)malloc(sizeof(struct qs_transition_type));
  qst->index = qst_count++;
  qst->wt_vect = NULL;
  qst->node = n1;
  qst->e = e;
  if(n2->in_qs_trans_cycle) {
    qst->next = n2->in_qs_trans_cycle->next;
    n2->in_qs_trans_cycle->next = qst;
  }
  else
    qst->next = qst;
  n2->in_qs_trans_cycle = qst;
  
  return n1->out_qs_trans_cycle;
}


struct qs_node_type *
get_consistent_qs(struct mode_type *m, struct qs_node_type *qsc) {
  struct qs_node_type *x = NULL;
  int val;

  if(qsc) {
    for(x = qsc->next; ; x = x->next) {
      if(m->index == x->mode->index) {
        if(BDD_AND(m->dval_bdd, x->val_bdd) != BDD_FALSE) 
        return x;
      }
      if(x == qsc) break;
    }
  }
  return x;
}

struct qs_node_type *
exec_trans(struct qs_node_type *qsp, struct transition_type *tr, 
           struct qs_node_type *qsc) {
  struct qs_node_type *qss = NULL;
  bdd b;

  if(qsc) {
    b = bdd_assign(qsp->val_bdd, tr->assign_cycle);
    for(qss=qsc->next; ; qss=qss->next) {
      if(qss->mode->index == qsp->mode->index) {
        if(BDD_AND(b, qss->val_bdd) != BDD_FALSE) 
          return qss;
      }
      if(qss == qsc) break;
    }
  }
  return qss;
}

int qs_consistent(struct qs_node_type *qsp, struct mode_type *m) {
  struct clkconj_type *c;

  c = clkconj_and(qsp->mode->xconj_cycle, m->xconj_cycle);
  if(c->op == FALSE) 
    return FALSE;
  if(BDD_AND(qsp->val_bdd, m->boolexp->boolexp_bdd) == BDD_FALSE)
    return FALSE;
  return TRUE;
}

int new_mode_index(struct mode_type *sm) {
  int i;
  struct mode_type *m;

  for(m=thread[total_automata-1].starting_mode, i=0; m; m=m->cnext, i++) {
    if(m->index == sm->index) {
      return i;
    }
  }
}

int gen_qs_trans(struct qs_node_type *qsc, struct thread_type *th) {
  struct mode_type *m;
  struct transition_type *tr;
  struct qs_transition_type *qst;
  struct qs_node_type *qsp, *qss;
  int i, *wt_v = NULL;

  for(m=th->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
        qsp = get_consistent_qs(m, qsc);
        qss = get_consistent_qs(tr->mode, qsc);
        qst = add_qs_transition(qsp, qss, tr);

        wt_v = (int *)malloc(sizeof(int) * automaton_mode_count);
        for(i=0; i<automaton_mode_count; i++) {
          wt_v[i] = 0;
        }
        qst->wt_vect = wt_v;
        wt_v = NULL;

        if(tr == m->out_trans_cycle) break;
      }
    }
  }

  for(m=thread[total_automata-1].starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
        for(qsp=qsc->next; ; qsp=qsp->next) {
          if(qs_consistent(qsp, m)) {
            qss = exec_trans(qsp, tr, qsc);
            qst = add_qs_transition(qsp, qss, tr);

            wt_v = (int *)malloc(sizeof(int) * automaton_mode_count);
            for(i=0; i<automaton_mode_count; i++) {
              if(m->index == tr->mode->index) wt_v[i] = 0;
              else if(i == new_mode_index(m)) wt_v[i] = -1;
              else if(i == new_mode_index(tr->mode)) wt_v[i] = 1;
              else wt_v[i] = 0;
            }
            qst->wt_vect = wt_v;
            wt_v = NULL;
          }
          if(qsp == qsc) break;
        }
        if(tr == m->out_trans_cycle) break;
      }
    }
  }

}

int *max_wt_vect(struct qs_node_type *qs1, struct qs_node_type *qs2) {
  int bound = -3, i, 
      *max = (int *)malloc(sizeof(int) * automaton_mode_count);
  struct qs_transition_type *qst;
  
  for(i=0; i<automaton_mode_count; i++) {
    max[i] = bound;
  }
  if(qs1->out_qs_trans_cycle) {
    for(qst=qs1->out_qs_trans_cycle->next; ; qst=qst->next) {
      if(qst->node->index == qs2->index) {
        for(i=0; i<automaton_mode_count; i++) {
          if(qst->wt_vect) {
            if(max[i] < qst->wt_vect[i]) 
              max[i] = qst->wt_vect[i];
          }
          else {
            if(max[i] < 0) 
              max[i] = 0;
          }
        }
      }
      if(qst == qs1->out_qs_trans_cycle) break; 
    }
  }
  return max;
}

int print_vector(int *v) {
  int i;
  output("[");
  if(v) {
    for(i=0; i<automaton_mode_count; i++) {
      sprintf(tmpstring, "%d", v[i]);
      output(tmpstring);
      if(i+1 < automaton_mode_count) output(" ");
    }
  }
  output("]");
  return;
}

int *add_vectors(int *v1, int *v2) {
  int i, *ans = (int *)malloc(sizeof(int) * automaton_mode_count);

/*
  if(v1 == NULL) {
    if(v2 != NULL) {
      for(i=0; i < automaton_mode_count; i++) {
        ans[i] = v2[i];
      }
      return ans;
    }
    else {              
      for(i=0; i < automaton_mode_count; i++) {
        ans[i] = 0;
      }
      return ans;
    }
  }
  else {
    if(v2 == NULL) {
      for(i=0; i < automaton_mode_count; i++) {
        ans[i] = v1[i];
      }
      return ans;
    }
    else {             
*/
      for(i=0; i < automaton_mode_count; i++) {
        ans[i] = v1[i] + v2[i];
      }
      return ans;
/*
    }
  }
*/
}

int *max_vectors(int *v1, int *v2) {
  int i, *ans = (int *)malloc(sizeof(int) * automaton_mode_count);
  
/*
  if(v1 == NULL) {
    if(v2 != NULL) {
      for(i=0; i < automaton_mode_count; i++) {
        ans[i] = (v2[i] > 0) ? v2[i] : 0;
      }
      return ans;
    }
    else {           
      for(i=0; i < automaton_mode_count; i++) {
        ans[i] = 0;
      }
      return ans;
    }
  }
  else {
    if(v2 == NULL) {
      for(i=0; i < automaton_mode_count; i++) {
        ans[i] = (v1[i] > 0) ? v1[i] : 0;
      }
      return ans;
    }
    else {            
*/
      for(i=0; i < automaton_mode_count; i++) {
        if(v1[i] > v2[i]) ans[i] = v1[i];
        else ans[i] = v2[i];
/*
        ans[i] = (v1[i] > v2[i]) ? v1[i] : v2[i];
*/
      }
      return ans;
/*
    }
  }
*/
}

int ***qs_kclosure(struct qs_node_type *qsc) {
  int bound = -3, 
      i, j, k, 
      ***lend,          /* lend: a [qs_count][qs_count] array of int pointers */
      *d = NULL, *tmp = NULL,
      ans = FALSE;
  struct qs_node_type *qs, *qsd;
  struct qs_transition_type *qst;

  lend = (int ***)malloc(qs_count * sizeof(int **));
  for(i=0; i<qs_count; i++) {
    lend[i] = (int **)malloc(qs_count * sizeof(int *));
  }

  if(qsc) {
                                   /* initialize all vectors to NULL */
    for(i = 0; i<qs_count; i++) {
      for(j = 0; j<qs_count; j++) {
        lend[i][j] = (int *)NULL;
      }
    }
                                   /* set the weights to max of trans wt_vect */
    for(qs=qsc->next; ; qs=qs->next) {
      if(qs->out_qs_trans_cycle) {
        for(qst=qs->out_qs_trans_cycle->next; ; qst=qst->next) {
          lend[qs->index][qst->node->index] = max_wt_vect(qs, qst->node);
          if(qst == qs->out_qs_trans_cycle) break;
        }
      }
      if(qs == qsc) break; 
    }

/*
    print_wt_vect_matrix(lend);
*/
                                   /* KLEENE's CLOSURE */
    for(i=0; i<qs_count; i++) {
      for(j=0; j<qs_count; j++) {
        for(k=0; k<qs_count; k++) {
          if(lend[j][i] != NULL && lend[i][k] != NULL) {
            d = add_vectors(lend[j][i], lend[i][k]);

/*
              sprintf(tmpstring, "SUM(lend[%d][%d]", j, i);
              output(tmpstring);
              print_vector(lend[j][i]); 
              sprintf(tmpstring, ", lend[%d][%d]", i, k);
              output(tmpstring);
              print_vector(lend[i][k]);
              output(") = ");
              print_vector(d);
              output("\n");
*/
/*
            sprintf(tmpstring, "lend[%d][%d]: ", j, k);
            output(tmpstring);
            print_vector(lend[j][k]);
*/

            if(lend[j][k]) {
/*
              output("MAX(d");
              print_vector(d); 
              sprintf(tmpstring, ", lend[%d][%d]", j, k);
              output(tmpstring);
              print_vector(lend[j][k]);
              output(") = ");
*/
              tmp = max_vectors(d, lend[j][k]);
              free(lend[j][k]);
              lend[j][k] = tmp;

              free(d);
/*
              print_vector(lend[j][k]);
              output("\n");
*/
            }
            else {
              lend[j][k] = d;
            }
/*
            output("--> ");
            print_vector(lend[j][k]);
            sprintf(tmpstring, " (i = %d).\n", i);
            output(tmpstring);
*/
          }
        }
      }
    }

    print_wt_vect_matrix(lend);
  }
  return lend;
}

int print_wt_vect_matrix(int ***w) {
  int i, j, k, *v;

  if(verbose >= 0) {
    output("Weight Vectors of QS transitions.\n");
  }
  for(i=0; i<qs_count; i++) {
    for(j=0; j<qs_count; j++) {
      sprintf(tmpstring, "wv[%d][%d] = [", i, j);
      output(tmpstring);
      for(k=0; k<automaton_mode_count; k++) {
        v = w[i][j];
        if(v) {
          sprintf(tmpstring, "%d", v[k]);
          output(tmpstring);
        }
        if(k+1 < automaton_mode_count) 
          output(" ");
      }
      output("]");
      if(j%4 == 0 || j+1 >= qs_count) output("\n");
      else output(", ");
    }
  }
  if(verbose >= 0) {
    output("END of Weight Vectors of QS transitions.\n");
  }
}

struct qs_node_link_type *gen_source_nodes(struct qs_node_type *qsc) {
  struct qs_node_link_type *qsl = NULL, *x;
  struct qs_node_type *qs;
  if(qsc) {
    for(qs = qsc->next; ; qs = qs->next) {
      if(initial_qs(qs)) {
        x=(struct qs_node_link_type *)malloc(sizeof(struct qs_node_link_type));
        x->qs_node = qs;
        if(qsl) {
          x->next = qsl->next;
          qsl->next = x;
        }
        else
          x->next = x;
        qsl = x;
      }
      if(qs == qsc) break;
    }
  }
  print_source_cycle(qsl);
  return qsl;
}

int print_source_cycle(struct qs_node_link_type *qsl) {
  struct qs_node_link_type *x;
  if(qsl) {
    output("Source Cycle consists of the following QS nodes.\n");
    for(x = qsl->next; ; x=x->next) {
      sprintf(tmpstring, "QS_NODE %d", x->qs_node->index);
      output(tmpstring);
      if(x == qsl) break;
      else output(", ");
    }
    output("\n");
  }
  else 
    output("EMPTY SOURCE CYCLES\n");
}

int initial_qs(struct qs_node_type *qs) {
  if(is_initial_mode(qs->mode) == FALSE) return FALSE;
  if(BDD_AND(qs->val_bdd, 
             thread[QS_THREAD].starting_mode->dval_bdd) 
     == BDD_FALSE) 
    return FALSE;
  return TRUE;
}

int qs_elim_trans(struct thread_type *t, int ***lend, 
                  struct qs_node_type *qs_cycle,
                  struct qs_node_link_type *source_cycle) {
  struct mode_type *m;
  struct transition_type *e;
  struct trans_link_type *tl, *elim_cycle = NULL;
  
  for(m=t->starting_mode; m; m=m->cnext) {
    if(m->out_trans_cycle) {
      for(e=m->out_trans_cycle->next; ; e=e->next) {
        if(path_exists(e, t, lend, qs_cycle, source_cycle) == FALSE) {
          tl=(struct trans_link_type *)malloc(sizeof(struct trans_link_type));
          tl->transition = e;
          if(elim_cycle) {
            tl->next = elim_cycle->next;
            elim_cycle->next = tl;
            elim_cycle = tl;
          }
          else
            elim_cycle = tl->next = tl;
        }
        if(e==m->out_trans_cycle) break;
      }
    }

    if(elim_cycle) {
      for(tl = elim_cycle->next; ; tl=tl->next) {
        e = tl->transition;
        eliminate_transition(e, m);

        if(verbose >= 0) {
          sprintf(tmpstring,
                  "ELIMINATED transition %d from mode %s(%d) to %s(%d).\n",
                  e->index, m->name, m->index, e->mode->name, e->mode->index);
          output(tmpstring);
        }

        t->trans_count--;
        if(tl == elim_cycle) break;
      }
    }
    free_trans_link_cycle(&elim_cycle);
  }
  printf("\tRemoving irreachable modes in thread %d.\n\t", t->index);
  remove_dormant_cycles(t);
  remove_irreachable_mode(t);
  print_counts(t->index);
}

int path_exists(struct transition_type *e, struct thread_type *t, 
                int ***lend, struct qs_node_type *qs_cycle, 
                struct qs_node_link_type *source_cycle) {
  struct qs_node_link_type *ql;
  struct qs_node_type *qs;

  if(source_cycle) {
    for(ql = source_cycle->next; ; ql=ql->next) {
      for(qs = qs_cycle->next; ; qs=qs->next) {
        if(qs->mode->index == e->mode->index) { 
          if(lend[ql->qs_node->index][qs->index])
            if(non_neg_vector(lend[ql->qs_node->index][qs->index])) {

              sprintf(tmpstring,
       "Found path from initial mode %s(%d) to mode %s(%d), lend[%d][%d] = ", 
                ql->qs_node->mode->name, ql->qs_node->mode->index,
                e->mode->name, e->mode->index,
                ql->qs_node->index, qs->index);
              output(tmpstring);
              print_vector(lend[ql->qs_node->index][qs->index]);
              output("\n");

              return TRUE;
            }
        }
        if(qs == qs_cycle) break;
      }
      if(ql == source_cycle) break;
    }
  }
  return FALSE;
}

int non_neg_vector(int *v) {
  int i;
  for(i=0; i<automaton_mode_count; i++) {
    if(v[i] < 0) return FALSE;
  }
  return TRUE;
}

int test_kcl() {
  int ***W;
  struct qs_node_type *n1, *n2, *n3;
  struct qs_transition_type *t1, *t2, *t3, *t4;

  n1 = (struct qs_node_type *)malloc(sizeof(struct qs_node_type));
  n2 = (struct qs_node_type *)malloc(sizeof(struct qs_node_type));
  n3 = (struct qs_node_type *)malloc(sizeof(struct qs_node_type));
  
n1->index = 0;
n2->index = 1;
n3->index = 2;

  t1 = (struct qs_transition_type *)malloc(sizeof(struct qs_transition_type));
  t2 = (struct qs_transition_type *)malloc(sizeof(struct qs_transition_type));
  t3 = (struct qs_transition_type *)malloc(sizeof(struct qs_transition_type));
  t4 = (struct qs_transition_type *)malloc(sizeof(struct qs_transition_type));

  t1->node = n2;
  t2->node = n1;
  t3->node = n3;
  t4->node = n1;

  t1->wt_vect = (int *)malloc(sizeof(int) * automaton_mode_count);
  t2->wt_vect = (int *)malloc(sizeof(int) * automaton_mode_count);
  t3->wt_vect = (int *)malloc(sizeof(int) * automaton_mode_count);
  t4->wt_vect = (int *)malloc(sizeof(int) * automaton_mode_count);

  t1->wt_vect[0] = 1;
  t1->wt_vect[1] = -1;
  t2->wt_vect[0] = 0;
  t2->wt_vect[1] = 0;
  t3->wt_vect[0] = -1;
  t3->wt_vect[1] = 1;
  t4->wt_vect[0] = 0;
  t4->wt_vect[1] = 0;

  n1->out_qs_trans_cycle = t1;
  t1->next = t1;

  t2->next = t3;
  t3->next = t2;

  n2->out_qs_trans_cycle = t2;

  t4->next = t4;
  n3->out_qs_trans_cycle = t4;

  n1->next = n2;
  n2->next = n3;
  n3->next = n1;
  
  qs_count = 3;
  W = qs_kclosure(n1);
  print_wt_vect_matrix(W);
}
