#include "include.h"
#include "global.h"
#include "print.h"
#include "bddman.h"
#include "modeman.h"
#include "zoneman.h"
#include "reduce.h"
#include "back.h"
#include "red_pred.h"
#include "lookahead.h"

struct thread_type *AUX_THREAD_T = (struct thread_type *)NULL;
struct thread_type *AUX_THREAD_t = (struct thread_type *)NULL;
struct thread_type *AUX_THREAD_Tt = (struct thread_type *)NULL;


int aux_merge(struct thread_type *T,  		/* A^k automaton */
              struct thread_type *t,  		/* A^1 automaton */
              struct thread_type Tt);  		/* A^{k+1} automaton */



int aux_merge(struct thread_type *T,  		/* A^k automaton */
              struct thread_type *t,  		/* A^1 automaton */
              struct thread_type Tt)  		/* A^{k+1} automaton */
{
  int i, nc1, nc2;
  struct mode_type *mode_queue = (struct mode_type *)NULL;
  struct mode_type *x;

  AUX_THREAD_T = T;
  AUX_THREAD_t = t;
  AUX_THREAD_Tt = Tt;

  printf("Auxiliary merging automata threads (%d) and (%d) into thread (%d):\n",
         T->index, t->index, Tt->index);
  if(verbose >= 0) {
    output("\n************************************************************\n");
    sprintf(tmpstring, "Auxiliary composing threads %d and %d into %d: \n", 
            T->index, t->index, Tt->index);
    output(tmpstring);
  }

  Tt->starting_mode = aux_initial_mode(T, t);
  last_mode = Tt->starting_mode;
  Tt->mode_count = 1;
  Tt->trans_count = 0;

  if(Tt->starting_mode) {
    TOP = Tt->starting_mode;
    TOP->next = NULL;
    for( ; TOP; ) {
      if(verbose > 0) {
        x = TOP;
        output("\nCurrent stack status: ");
        for(; x; x=x->next) {
          sprintf(tmpstring, "mode %d(%d, %d) ", x->index, x->submode1->index,
                                                           x->submode2->index);
          output(tmpstring);
        }
        output("\n");
      }
      x = TOP;
      TOP = TOP->next;
      x->next = (struct mode_type *)NULL;
      if(verbose >= 2) {
        sprintf(tmpstring, "Processing mode (%s) [ %d(%d, %d) ]\n",
                     x->name, x->index, x->submode1->index, x->submode2->index);
        output(tmpstring);
      }
      if(verbose == 0) {
        sprintf(tmpstring, "--- Predecessor mode (%s) [ %d(%d, %d) ]\n",
                     x->name, x->index, x->submode1->index, x->submode2->index);
        output(tmpstring);
      }
      nc1 = x->submode1->out_trans_count;
      nc2 = x->submode2->out_trans_count;

      if(partial_order_reduction || partial_order_with_symmetry_reduction)
        reset_partial_order();

      if(nc1) {
        if(verbose >= 2) {
          sprintf(tmpstring,"Submode1: No. of outgoing transitions = %d\n",nc1);
          output(tmpstring);
        }
        transition_m1_first(x);
        if(verbose >= 2)
          output("---- Finished submode1 transitions ...\n");
      }
      if(nc2) {
        if(verbose >= 2) {
          sprintf(tmpstring,
                  "\nSubmode2: No. of outgoing transitions = %d\n",nc2);
          output(tmpstring);
        }
        transition_m2_first(x);
        if(verbose >= 2)
          output("---- Finished submode2 transitions ...\n");
      }
      if(partial_order_with_symmetry_reduction) {
        if(x->accessed) free(x->accessed);
        if(x->history) free(x->history);
      }
    }
  }
  else {
    fprintf(stderr, "A conflict in initial mode merging.\n");
    return (-1);
  }
  last_mode->cnext = (struct mode_type *)NULL;
  return (1);
}
