#include "global.h"
#include "print.h"
#include "modeman.h"
#include "bddman.h"
#include "zoneman.h"
#include "back.h"
#include "shield.h"

int *write_0;
int *write_i;
int *read_0;
int *read_i;

/* for partial order reduction */
int trans_type(struct transition_type *);
int init_partial_order();
int reset_partial_order();
int record_trans(struct transition_type *, struct mode_type *);
int po_reduce(struct transition_type *, struct mode_type *);

int history_complete(int *);
int print_read_write();
int print_history_accessed(struct mode_type *);
int pos_reduce(struct transition_type *, struct mode_type *);

int trans_type(struct transition_type *t) {
  struct assign_type *a;
  struct boolconj_type *b;

  if(t->assign_length) {
    a = t->assign_cycle->next;
    for(; ; a=a->next) {
      if(get_var_type(a->var) == V_REGISTER) {
        if(a->value == 0)
          return WRITE_0_TRANS;
        else return WRITE_I_TRANS;
      }
      if(a==t->assign_cycle) break;
    }
  }
  if(t->prp_trigger_cycle) {
    b = t->prp_trigger_cycle->next;
    for(; ; b=b->next) {
      if(get_var_type(b->var) == V_REGISTER) {
        if(b->constant == 0)
          return READ_0_TRANS;
        else return READ_I_TRANS;
      }
      if(b==t->prp_trigger_cycle) break;
    }
  }
  printf("Neither read/write transition.\n");
  return -1;
}

int init_partial_order() {
  int i;
  write_0 = (int *)malloc(sizeof(int)*(total_automata+1));
  write_i = (int *)malloc(sizeof(int)*(total_automata+1));
  read_0 = (int *)malloc(sizeof(int)*(total_automata+1));
  read_i = (int *)malloc(sizeof(int)*(total_automata+1));

  for(i=0; i<=total_automata; i++) {
    write_0[i] = 0;
    write_i[i] = 0;
    read_0[i] = 0;
    read_i[i] = 0;
  }
  return;
}

int reset_partial_order() {
  int i;
  for(i=0; i<=total_automata; i++) {
    write_0[i] = 0;
    write_i[i] = 0;
    read_0[i] = 0;
    read_i[i] = 0;
  }
  return;
}

int record_trans(struct transition_type *t, struct mode_type *m) {
  int i, sum = 0;
  switch(trans_type(t)) {
    case WRITE_0_TRANS:
      write_0[t->automaton_index] = 1;
      break;
    case WRITE_I_TRANS:
      write_i[t->automaton_index] = 1;
      break;
    case READ_0_TRANS:
      read_0[t->automaton_index] = 1;
      break;
    case READ_I_TRANS:
      read_i[t->automaton_index] = 1;
      break;
    default:
      fprintf(stderr, "Error: a transition which neither reads nor writes!\n");
  }
  if(partial_order_with_symmetry_reduction) {
    t->mode->accessed = (int *)malloc(sizeof(int)*(total_automata+1));
    t->mode->history = (int *)malloc(sizeof(int)*(total_automata+1));
    if(m->accessed) {
      for(i=0; i<=total_automata; i++) {
        t->mode->accessed[i] = m->accessed[i];
        t->mode->history[i] = m->history[i];
      }
    }
    else {
      for(i=0; i<=total_automata; i++) {
        t->mode->accessed[i] = 0;
      }
    }
    if(trans_type(t) == WRITE_I_TRANS) {
      for(i=0; i<=total_automata; i++) {
        if(write_i[i] == 1)  {
          t->mode->accessed[i] = 1;
          t->mode->history[i]  = 1;
        }
      }
      for(i=1; i<=total_automata; i++) {
        if(write_i[i] == 1) sum++;
      }
      if(sum == 2) {
        t->mode->status = SECOND_WRITE;
      }
      if(m->status == SECOND_WRITE)
        t->mode->status = SECOND_WRITE;
    }
    if(trans_type(t) == READ_0_TRANS) {
      t->mode->history[t->automaton_index] = -1;
      for(i=0; i<=total_automata; i++) {
        if(read_0[i] == 1)  t->mode->accessed[i] = -1;
      }
    }
    output("\tRecorded transition ... ");
    output("\tt->mode->accessed =( ");
    for(i=1; i<=total_automata; i++) {
      sprintf(tmpstring, "%d ", t->mode->accessed[i]);
      output(tmpstring);
    }
    output(")\n");
  }
}

int po_reduce(struct transition_type *t, struct mode_type *m) {
  int i;

  if(trans_type(t) == READ_I_TRANS) {
    for(i=1; i<=total_automata; i++) {
      if(read_i[i] == 1)
        return TRUE;
    }
    return FALSE;
  }
  else return FALSE;
}

int history_complete(int *h) {
  int i;
  if(h == NULL) return FALSE;
  for(i=1; i<=total_automata; i++) {
    if(h[i] < 0) return FALSE;
  }
  return TRUE;
}

int print_read_write() {
  int i;
  output("\t read_0 = ( ");
  for(i=1; i<=total_automata; i++) {
    sprintf(tmpstring, "%d ", read_0[i]);
    output(tmpstring);
  }
  output("), write_i = ( ");
  for(i=1; i<=total_automata; i++) {
    sprintf(tmpstring, "%d ", write_i[i]);
    output(tmpstring);
  }
  output(")");
}

int print_history_accessed(struct mode_type *m) {
  int i;
  if(m->accessed) {
    output(", m->accessed = ( ");
    for(i=1; i<=total_automata; i++) {
      sprintf(tmpstring, "%d ", m->accessed[i]);
      output(tmpstring);
    }
    output(")");
  }
  if(m->history) {
    output(", m->history = ( ");
    for(i=1; i<=total_automata; i++) {
      sprintf(tmpstring, "%d ", m->history[i]);
      output(tmpstring);
    }
    output(")");
  }
  output("\n");
}

int pos_reduce(struct transition_type *t, struct mode_type *m) {
  int i, sum = 0;

  if(verbose >= 0) {
    sprintf(tmpstring, "POS reducing transition %d from mode %d(%s).\n",
                        t->index, m->index, m->name);
    output(tmpstring);
  }
  print_read_write();
  print_history_accessed(m);

  if(trans_type(t) == READ_0_TRANS) {
    if(m->accessed == NULL) return FALSE;
    if(m->accessed[t->automaton_index] == -1) {
      output("\t Reduction Reason: read accessed.\n");
      return TRUE;
    }
    else return FALSE;
  }
  else if(trans_type(t) == WRITE_I_TRANS) {
    for(i=1; i<=total_automata; i++) {
      if(write_i[i] == 1) sum++;
    }
    if(sum >= 2) {
      output("\t Reduction Reason: more than 2 writes.\n");
      return TRUE;
    }
    if(m->status == SECOND_WRITE && sum >= 1) {
      output("\t Reduction Reason: second-write and more than 1 write.\n");
      return TRUE;
    }
    if(m->status == SECOND_WRITE && sum < 1) {
      if(history_complete(m->history)) {
        output("\t NOT reduced: history is complete.\n");
        return FALSE;
      }
      else if(m->accessed[t->automaton_index] == 1) {
        output("\t Reduction Reason: second-write and accessed.\n");
        return TRUE;
      }
    }
    return FALSE;
  }
  else return FALSE;
}


