#include <unistd.h>
#ifdef __SUNOS
#include <sys/time.h>
#endif
#include <sys/resource.h>
#include "include.h"
#include "global.h"
#include "bddman.h"
#include "modeman.h"
#include "zoneman.h"
#include "symmetry.h"
#include "shieldclk.h"
#include "readwrite.h"
#include "transman.h"
#include "mcheck.h"
#include "action.h"
#include "proc.h"
#include "print.h"
#include "gui.h"
#include "shieldvar.h"
#include "getresources.h"

extern FILE *yyin;
extern void lexinit();
extern void yyparse();
/*extern int getrusage(int, struct rusage *);*/

char *tool_name = "State Graph Manipulator";
char *tool_abbreviation = "SGM";
float tool_version = 2.0;
char *authors = "Farn Wang  and  Pao-Ann Hsiung";
char *prompt = "sgm> ";

MODE SGM_MODE = INTERACTIVE;

int tmp_int;
int composed_ta = 0;			/* Already composed num. of automata */
char *tmpstring;  			/* TMP string for sprintf() */
int verbose = -2;			/* verbosity level for logfile */
int stdoutput = FALSE;			/* whether to output to std console */
int lastcomp = FALSE;         		/* last composition */

					/* Defaults: */
int clock_shielding = FALSE;		/* Do not shield clocks */
int region_normalization = FALSE;	/* Do not normalize regions */
int internal_trans_bypass = FALSE;	/* Do not bypass internal transitions*/
int variables_shielding = FALSE;	/* Do not shield variables */
int model_checking = FALSE;		/* Do not model check */
int zone_enclosing = FALSE;		/* Do not enclose zones */
int zone_collapsing = FALSE;		/* Do not collapse zones */
int on_the_fly_normalization = FALSE;	/* Do not normalize during merge */ 
int keep_graphs = FALSE;		/* Do not keep graphs in memory */
int strict_semantics = FALSE;		/* Loose semantics: transition is
                                                            NOT instantaneous */
int changed;			        /* for converging LA */

int sibling_trans_multi_reduction = TRUE; /* STMR technique */

int lazy_bdd_merging = TRUE;
int look_forward_reduction = FALSE;

int stats = FALSE;

/*
int old_trans_num;
int new_trans_num;
*/

int EIV = FALSE;        /* Elimination of Invalid runs that were Valid before */
int EOI = FALSE;        /* Elimination of runs that find their 
                          Own index trigger Invalid */
int EDV = FALSE;        /* Eliminate delusively valid runs: runs that seem to
                          be valid but are not due to insufficient info in
                          the lazy discrete variable valuation (bdd) */
int EOT = FALSE;       /* Eliminate orthogonal (corresponding) transitions
                          leaving only one representative, so that the
                          state-graph can be reconstructed for model-checking */

/*
int look_ahead_reduction = FALSE;
int zone_reduction = FALSE;
int partial_order_reduction = FALSE;
int partial_order_with_symmetry_reduction = FALSE;
*/

FILE *logfile;
FILE *out;
char *infilename = (char *)NULL;
char *logfilename = (char *)NULL;


struct a23tree_type *bdd_tree;
struct a23tree_type *zone_tree;
struct a23tree_type *perm_tree;

struct thread_type *thread;
struct manipulator_type *manipulator_list;

int BASIC_THREAD_COUNT, TOTAL_THREAD_COUNT;
struct thread_type *CURRENT_GRAPH, *COMPOSED_GRAPH, *REDUCED_GRAPH;


/* Function Declarations */

void process_files(int, char *[]);
void program_init(void);
void printline(void);
void intro_output(void);
void process_flags(int, char *[]);
void process_one_flag(char *, int);
int ds_init(void);
int sequential_merge(void);
int manipulate_graph(struct thread_type *t, struct thread_type *t_1);
int pairwise_merge(void);
int new_manipulator(int);
struct manipulator_type *alloc_manipulator(void);
int action_to_manipulator();
int exec_all_manipulators();
ERROR exec_manipulator(struct manipulator_type *);

struct timeval        system_start_time;
struct timeval        system_end_time;
struct timezone       time_zone_p;
struct rusage         usage;

/*********************************************************************
 * MAIN PROGRAM 
 *********************************************************************/

#ifndef _GUI
int main(int argc, char *argv[]) {
/*
  float                 f = 0.0;
*/

  program_init();
  if(SGM_MODE == BATCH || SGM_MODE == INTERACTIVE) intro_output();   
  process_flags(argc, argv);

                                     /* get system starting time */
/*
  gettimeofday(&system_start_time, &time_zone_p); 
*/
  process_files(argc, argv);

  lexinit();

  printf("Parsing automata specifications ... ");
  fflush(stdout);
  yyparse(); 
  if(SGM_ERROR == E_NONE) printf("done\n");

  ds_init();

  BASIC_THREAD_COUNT = system_ta_num;
  TOTAL_THREAD_COUNT = (2 * BASIC_THREAD_COUNT) - 1;

  if(system_ta_num > 1) {
    if(system_ta_num == 2) 
      lastcomp = TRUE;
    switch(merge_type) {
      case SM: sequential_merge();
               break;
      case PM: pairwise_merge();
               break;
      default: exec_action_list(action_list, NON_INDEXED);
               break;
    }
    if(model_checking) { 
      model_check(COMPOSED_GRAPH, vspec);
    }
  }
  else {
    sprintf(tmpstring, 
          "Number of automata (%d) too few for composition.\n", system_ta_num);
    output(tmpstring);
  }

/*
  if(model_check(thread[TOTAL_THREAD_COUNT-1].starting_mode, vspec))
    printf("Specification satisfied!\n");
  else
    printf("Specification NOT satisfied!\n");
*/


  /* get system ending time */
/*
  gettimeofday(&system_end_time, &time_zone_p);
  printline();
  if (system_end_time.tv_usec - system_start_time.tv_usec >= 0) {
    printf("System elapse time : %d.%06d seconds.\n",
            (int)(system_end_time.tv_sec - system_start_time.tv_sec),
            (int)(system_end_time.tv_usec - system_start_time.tv_usec));
  }
  else {
    printf("System elapse time : %d.%06d seconds.\n",
            (int)(system_end_time.tv_sec - system_start_time.tv_sec-1),
            (int)(system_end_time.tv_usec - system_start_time.tv_usec+1000000));
  }
*/
  /* calculate cpu time */
/*
  getrusage(RUSAGE_SELF, &usage);

  f = ((float) usage.ru_utime.tv_sec)
    + (((float) usage.ru_utime.tv_usec) / 1000000);
  printf("The CPU running time is %f seconds.\n", f);
  printline();
*/

  print_max_mem();
/*  execl("/bin/ps", "ps", "-flcu eric", (char *)0); */
  /*    execl("/homes/rfsml/eric/bin/mem", "mem", (char *)0); */
  return (0); 
}
/* END of main.c */
#endif


void process_files(int argc, char *argv[]) {
  char *title = 
  "--- State Graph Manipulator (SGM) Log File ---\n\n";

  if (argc==1) {
      exit(1);
  }
  infilename = argv[argc-1];
  logfilename = (char *)malloc(strlen(infilename) + 5);

  if (logfilename == NULL) {
    fprintf(stderr, 
            "Error: Log filename %s creation failed (out of memory?)!\n", 
            logfilename); 
    exit(1);
  }

  strcpy(logfilename, infilename);
  strtok(logfilename,".");
  strcpy(infilename, logfilename);
  strcat(infilename, ".s");
  strcat(logfilename,".log");

  if((argc >= 2) && (argv[argc-1][0] != '-')) {
    yyin = fopen( infilename, "r" );   /* filename was in argv[argc-1] */
    if(yyin == NULL) {
      fprintf(stderr, 
             "*** Error: Specification input file `%s' open failed ***\n",
                                               infilename);
      fflush(stderr);
      yyin=stdin;
      exit(1);
    }
    else 
      printf( "-- Reading specification from input file %s.\n", argv[argc-1]);
  }
  else {
    if (argv[argc-1][1] != 'h' && argv[argc-1][1] != 'i')
      fprintf(stderr, "*** Error: no specification input file given ***\n");
    if (argv[argc-1][1] != 'h') exit(1);
  }
  logfile = fopen( logfilename, "w" );
  out = logfile;       /* stderr for 23tree library */
  if(verbose >= 0)
    fwrite(title, sizeof(char), strlen(title), logfile); 
  if(logfile == NULL) {
    fprintf(stderr, 
           "*** Error: Log file `%s' creation failed (check permission)***\n",
                                   logfilename);
    fflush(stderr);
    exit(1);
  }
  printf("-- Writing log to output file %s.\n", logfilename );
  printline();
}


void program_init() {
  if(tmpstring == NULL) 
    tmpstring = (char *)malloc(sizeof(char)*MAX_CHAR);  
  if(string_buffer == NULL) 
    string_buffer = (char *)malloc(sizeof(char)*MAX_CHAR);  
}

void printline() {
  if(SGM_MODE != GUI && SGM_MODE != PROCESS)
  printf("-----------------------------------------------------------------\n");
}

void intro_output() {
  printf("-----------------------------------------------------------------\n");
  printf("\t    %s (%s) (Version %.1f)\n", 
           tool_name, tool_abbreviation, tool_version);
  printf("\t          %s       \n", authors);
  printf("\t         Institute of Information Science      \n"); 
  printf("\t           Academia Sinica, Taiwan, ROC.       \n");
  printf("-----------------------------------------------------------------\n");
}

void process_flags(int agc, char *agv[]) {
  int i;

  if (agc==1) {
    fprintf(stderr, 
    "Usage: %s [-<flag>]* <input filename>[.s] ('%s -h' for help)\n", 
    agv[0], agv[0]);

    if(SGM_MODE == INTERACTIVE) {
      printf("Entering INTERACTIVE MODE ... (pid = %d).\n", (int)getpid());
      interactive_mode();
    }

    if(SGM_MODE == PROCESS) {
/*
      printf("Entering PROCESS MODE ... (pid = %d).\n", (int)getpid());
*/
      process_mode();
    }

    if(SGM_MODE == GUI) {
      printf("Entering GUI MODE ... (pid = %d).\n", (int)getpid());
      gui_mode();
    }
  }
  else if(agc==2 && !strcmp(agv[1], "-h")) {
    fprintf(stderr, 
    "Usage: %s [-<flag>]* <input filename>[.s] ('%s -h' for help)\n", 
    agv[0], agv[0]);
    process_one_flag(agv[1]+1, 1); 
    exit(0);
  }
  else if(agc==2 && !strcmp(agv[1], "-i")) {
    fprintf(stderr, 
    "Usage: %s [-<flag>]* <input filename>[.s] ('%s -h' for help)\n", 
       agv[0], agv[0]);
    SGM_MODE = INTERACTIVE;
    printf("Entering INTERACTIVE MODE ... (pid = %d).\n", (int)getpid());
    interactive_mode(); }
  else if(agc==2 && !strcmp(agv[1], "-p")) {
    SGM_MODE = PROCESS;
/*
    printf("Entering PROCESS MODE ... (pid = %d).\n", (int)getpid());
*/
    process_mode(); }
  else {
    for(i=1; i<agc; i++) {
      if (agv[i][0] == '-') {
        process_one_flag(agv[i]+1, i);
      }
      else {
        if (i != agc-1) {
          fprintf(stderr, "Warning: illegal flag: %s\n", agv[i]);
          fflush(stderr);
        }
      }
    }
  }
}

void process_one_flag(char *flagstr, int i) {
  if(!strcmp(flagstr, "p")) {
    SGM_MODE = PROCESS;
  }
  else if(!strcmp(flagstr, "cs")) {
    new_manipulator(SC);   clock_shielding = i;
  }
  else if(!strcmp(flagstr, "rn")) {
    new_manipulator(NR);   region_normalization = i;
  }
  else if(!strcmp(flagstr, "tb")) {
    new_manipulator(TB);   internal_trans_bypass = TRUE;
  }
  else if(!strcmp(flagstr, "sv")) {
    new_manipulator(SV);   variables_shielding = TRUE;
  }
  else if(!strcmp(flagstr, "lf")) {
    new_manipulator(LA);   look_forward_reduction = TRUE;
  }
  else if(!strcmp(flagstr, "on")) {
    on_the_fly_normalization = i;
  }
  else if(!strcmp(flagstr, "rt")) {
    EOT = i;
  }
  else if(!strcmp(flagstr, "nmr")) {
    sibling_trans_multi_reduction = FALSE;
  }
  else if(!strcmp(flagstr, "ze")) {
    zone_enclosing = TRUE;
  }
  else if(!strcmp(flagstr, "zc")) {
    zone_collapsing = TRUE;
  }
  else if(!strcmp(flagstr, "ss")) {
    strict_semantics = TRUE;
  }
  else if(!strcmp(flagstr, "s")) {
    stats = TRUE;
  }
  else if(!strcmp(flagstr, "nl")) {
    verbose = -2;
  }
  else if(!strcmp(flagstr, "log")) {
    verbose = 0;
  }
  else if(!strcmp(flagstr, "pt")) {
    verbose = -1;
  }
  else if(!strcmp(flagstr, "vv")) {
    verbose = 2;
  }
  else if(!strcmp(flagstr, "vvv")) {
    verbose = 3;
  }
  else if(!strcmp(flagstr, "kg")) {
    keep_graphs = TRUE;
  }
  else if(!strcmp(flagstr, "lm")) {
    lazy_bdd_merging = TRUE;
  }
  else if(!strcmp(flagstr, "rw")) {
    lazy_bdd_merging = FALSE;
  }
/*
  else if(!strcmp(flagstr, "la")) {
    look_ahead_reduction = TRUE;
  }
  else if(!strcmp(flagstr, "zr")) {
    zone_reduction = TRUE;
  }
  else if(!strcmp(flagstr, "po")) {
    partial_order_reduction = TRUE;
  }
  else if(!strcmp(flagstr, "pos")) {
    partial_order_with_symmetry_reduction = TRUE;
  }
  else if(!strcmp(flagstr, "ln")) {
    verbose = -3;
  }
*/
  else {
    switch (flagstr[0]) {
    case 'o':
      stdoutput = TRUE;
      break;
    case 'v':
      verbose = TRUE;
      break;
    case 'h':
      fprintf(stderr, "Flags currently allowed:\n");
/* MODES */
      fprintf(stderr, 
     " (0) -i   : Interactive Mode (default: GUI mode)\n");
/* MANIPULATORS */
      fprintf(stderr, 
     " (1) -cs  : Clock Shielding (default: no clock shielding)\n");
      fprintf(stderr, 
     " (2) -rn  : Region Normalization (default: no region normalization)\n");
/*
      fprintf(stderr, 
     " (3) -on  : On-the-Fly Normalization (default: no on-the-fly normalization)\n");
*/
      fprintf(stderr, 
     " (3) -rw  : On-the-Fly Read-Write Reduction (default: no)\n");
      fprintf(stderr, 
     " (4) -tb  : Bypass Internal Transition (default: no BIT reduction)\n");
      fprintf(stderr, 
     " (5) -nmr : No Sibling Transition Multiplicity Reduction (default: SMTR)\n");
      fprintf(stderr, 
     " (6) -lm  : Lazy-BDD merging (default: literal-set merging)\n");
/* OTHERS */
      fprintf(stderr, 
     " (7) -ss  : Strict Semantics (default: non-strict semantics)\n");
      fprintf(stderr, 
     " (8) -s   : Dynamic statistics output of manipulations (default: no)\n");
      fprintf(stderr, 
     " (9) -pt           : Log graphs ONLY       (level = -1)!\n");
      fprintf(stderr, 
     "(10) -log          : Log all manipulations (level = 0)!\n");
      fprintf(stderr, 
     "(11) -v, -vv, -vvv : Log verbosity         (level = 1, 2, 3)!\n");
      fprintf(stderr, 
     "(12) -kg  : Keep all merged graphs in memory (default: graphs reset)\n");
      fprintf(stderr, 
     "(13) -h   : This help page!\n");
/*
      fprintf(stderr, 
     "-lf  : Look-Forward Reduction (default: no look-forward reduction)\n");
      fprintf(stderr, 
     "-rt  : Representative Transition only (default: leave all)\n");
      fprintf(stderr, 
     "-ze  : Zone Enclosing (default: no zone enclosing)\n");
      fprintf(stderr, 
     "-zc  : Zone Collapsing (default: no zone collapsing)\n");
      fprintf(stderr, 
     "-la  : Look-Ahead Reduction (default: no look-ahead reduction)\n");
      fprintf(stderr, 
     "-zr  : Zone Reduction (default: no zone reduction)\n");
      fprintf(stderr, 
     "-po  : Partial Order Reduction (default: no PO reduction)\n");
      fprintf(stderr, 
     "-pos : Partial Order with Symmetry Reduction (default: no POS reduction)\n");
      fprintf(stderr, 
     "-ln  : Log Normalization (default: verbosity/log level = 0)\n");
*/
      break;
    default:
      fprintf(stderr, "Warning: illegal flag %s ignored\n", flagstr);
      fflush(stderr);
      break;
    }
  }
}



/* Initialize all data structures */

int ds_init() {
  if(SGM_ERROR != E_NONE) exit(1);
  if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH)
    printf("-- Initializing data structures ... ");
  init_varval();

/* printf("Initializing BDD management ..."); */
  init_bdd();
/* printf(" done\n"); */

/* printf("Installing BDDs for initial boolean expressions of automata ..."); */
  install_init_boolexp_bdds(); 
/* printf(" done\n"); */

/* printf("Installing BDDs for boolean expressions of mode invariants ..."); */
  install_boolexp_bdds(); 
/* printf(" done\n"); */

/* printf("Installing BDDs for trigger conditions of mode transitions ..."); */
  install_trigger_bdds(); 
/* printf(" done\n"); */

  print_prop_bdds();

  remove_duplicate_trans();

/* printf("Initializing mode management ..."); */
  init_mode_management();
/* printf(" done\n"); */

/* printf("Initializing thread management ..."); */
  init_thread_management();
/* printf(" done\n"); */

/*
  if(partial_order_reduction || partial_order_with_symmetry_reduction) {
    printf("Initializing partial order ...");
    init_partial_order();
    printf(" done\n");
  }
*/

/* printf("Initializing zone management ..."); */
  init_zone_management();
/* printf(" done\n"); */

/*  if(region_normalization || on_the_fly_normalization) { */
/* printf("Initializing compact reduction by permutation ..."); */

    init_permutation();

/* printf(" done\n"); */
/*  } */

/* printf("Installing ZONEs for clock expressions of mode invariants ..."); */
  install_clkconj_zones(); 
/* printf(" done\n"); */

  init_clk_literals();

  init_action_management();
/*
  print_manipulations();
*/

  if(SGM_ERROR == E_NONE && (SGM_MODE == INTERACTIVE || SGM_MODE == BATCH))
    printf(" done.\n");

  max_const = get_max_const();
  if(max_const > 0) {
    if(SGM_MODE == INTERACTIVE || SGM_MODE == BATCH)
    printf("-- Maximum clock constant in the specification is %d.\n",max_const);
  }
  return 0;
}

int sequential_merge() {
  int ti, nti;

    action_to_manipulator();
    for(ti = 1, nti = BASIC_THREAD_COUNT;
        ti < BASIC_THREAD_COUNT;
        ti++, nti++)
    {
      /* (1) MERGING */
      CURRENT_GRAPH = &thread[ti];

      gettime(START);
      if(ti == 1) compose(&thread[0], &thread[ti], &thread[nti]);
      else compose(&thread[nti-1], &thread[ti], &thread[nti]);
      gettime(END);

      REDUCED_GRAPH = &thread[nti-1];
      COMPOSED_GRAPH = &thread[nti];

      if(verbose >= -1) {
        output("After merging:\n");
        print_thread(&thread[nti]);
      }
/*
      manipulate_graph(&thread[nti], &thread[nti-1]);
*/
      exec_all_manipulators();
      if(keep_graphs == FALSE && nti > BASIC_THREAD_COUNT) {
        reset_thread(&thread[nti-1]);
      }
   }
  return 0;
}

int exec_all_manipulators() {
  struct manipulator_type *m;
  if(manipulator_list) {
    for(m=manipulator_list->next; ; m=m->next) {
      gettime(START);
      if(exec_manipulator(m) == E_NONE);
        gettime(END);
      if(m==manipulator_list) break;
    }
    free(print_time(COMPOSED_GRAPH));
  }
  return 0;
}

int manipulate_graph(struct thread_type *t, struct thread_type *t_1) {

      /* (2) LOOK-FORWARD REDUCTION */

/*
      if(composed_ta > 2 && look_forward_reduction) {
        changed = TRUE;
        while(changed) {
          changed = look_forward_reduce(t_1, t);
        }
        if(verbose >= -1) {
          output("After look forward reduction:\n");
          composed_ta--;
          print_thread(t_1);
          composed_ta++;
        }
        reset_thread(t);
        compose(t_1, CURRENT_GRAPH, t);
        if(verbose >= -1) {
          output("After look-forward + merging:\n");
          print_thread(t);
        }
      }
*/

      /* (3) CLOCK SHIELDING */
      if(clock_shielding) {
        shield_clocks(t);
        if(verbose >= -1) {
          output("After shielding clocks:\n");
          print_thread(t);
        }
      }

      /* (4) REGION NORMALIZATION */
      if(region_normalization) {
        normalize_regions(t);
        if(verbose >= -1) {
          output("After normalization:\n");
          print_thread(t);
        }
      }

      /* (*) ZONE ENCLOSING */
/*
      if(zone_enclosing) {
        enclose_zones(t);
        if(verbose >= -1 ) {
          output("After enclosing zones:\n");
          print_thread(t);
        }
      }
*/

      /* (*) ZONE COLLAPSING */ 
/*
      if(zone_collapsing) {
        collapse_zones(t);
        if(verbose >= -1 ) {
          output("After collapsing zones:\n");
          print_thread(t);
        }
      }
*/

      /* (*) STATE-SPACE REDUCTION (if not global automata) */

/*
      if(look_ahead_reduction && !lastcomp) {
        gen_btgraph(t);
        gen_quotient_structure(t);
        generic_merge(t->index, CURRENT_GRAPH->index + 1);
        if(verbose >= -1) {
          output("After look-ahead reduction:\n");
          print_thread(t);
        }
      }
*/
  return 0;
}

/*Old method of converging to no change */
/*
      if(look_forward_reduction) {
       old_trans_num = 0;
       new_trans_num = thread[nti-1].trans_count;
       while(old_trans_num != new_trans_num) {
        old_trans_num = new_trans_num;
        look_forward_reduce(&thread[nti-1], &thread[nti]);
        if(verbose >= -1) {
          output("After look forward reduction:\n");
          composed_ta--;
          print_thread(&thread[nti-1]);
          composed_ta++;
        }
        reset_thread(&thread[nti]);
        compose(&thread[nti-1], &thread[ti], &thread[nti]);
        if(verbose >= -1) {
          output("After look-forward + merging:\n");
          print_thread(&thread[nti]);
        }
        if(clock_shielding) {
          shield_clocks(&thread[nti]);
          if(verbose >= -1) {
            output("After shielding clocks:\n");
            print_thread(&thread[nti]);
          }
        }
        new_trans_num = thread[nti-1].trans_count;
       }
      }
*/

                  /* PAIR_WISE merge */
int pairwise_merge() {
  int ti, nti;

  action_to_manipulator();
  for(ti = 0, nti = BASIC_THREAD_COUNT;
      ti < TOTAL_THREAD_COUNT - 2;
      ti += 2, nti++)
  {
    if(nti == TOTAL_THREAD_COUNT-1) lastcomp = TRUE;
/*
    composed_ta++;
*/
    CURRENT_GRAPH = &thread[ti];

    compose(&thread[ti], &thread[ti + 1], &thread[nti]);

    REDUCED_GRAPH = &thread[nti-1];
    COMPOSED_GRAPH = &thread[nti];

    if(verbose >= -1) {
      output("After merging:\n");
      print_thread(&thread[nti]);
    }
/*
    manipulate_graph(&thread[nti], &thread[nti-1]);
*/
    exec_all_manipulators();
/*
    if(keep_graphs == FALSE && nti > BASIC_THREAD_COUNT) {
      reset_thread(&thread[nti - 1]);
    }
*/
  }
  return 0;
}

struct manipulator_type *alloc_manipulator() {
  struct manipulator_type *m;
  m = (struct manipulator_type *)malloc(sizeof(struct manipulator_type));
  m->type = 0;
  m->next = (struct manipulator_type *)NULL;
  return m;
}

int new_manipulator(int type) {
  struct manipulator_type *m;
  m = alloc_manipulator();
  m->type = type;
  if(manipulator_list) {
    m->next = manipulator_list->next;
    manipulator_list->next = m;
  }
  else m->next = m;
  manipulator_list = m;
  return 0;
}

ERROR exec_manipulator(struct manipulator_type *m) {
  switch(m->type) {
    case SC: shield_clocks(COMPOSED_GRAPH);
             if(verbose >= -1) {
               output("After shielding clocks:\n");
               print_thread(COMPOSED_GRAPH);
             }
             break;
    case NR: normalize_regions(COMPOSED_GRAPH);
             if(verbose >= -1) {
               output("After normalization:\n");
               print_thread(COMPOSED_GRAPH);
             }
             break;
    case RW: readwrite_reduce(COMPOSED_GRAPH);
             if(verbose >= -1) {
               output("After read-write reduction:\n");
               print_thread(COMPOSED_GRAPH);
             }
             break;
/*
    case LA: if(composed_ta > 2) {
               changed = TRUE;
               while(changed) {
                 changed = look_forward_reduce(REDUCED_GRAPH, COMPOSED_GRAPH);
               }
               if(verbose >= -1) {
                 output("After look forward reduction:\n");
                 composed_ta--;
                 print_thread(REDUCED_GRAPH);
                 composed_ta++;
               }
               reset_thread(COMPOSED_GRAPH);
               compose(REDUCED_GRAPH, CURRENT_GRAPH, 
                       COMPOSED_GRAPH);
               if(verbose >= -1) {
                 output("After look-forward + merging:\n");
                 print_thread(COMPOSED_GRAPH);
               }
             }
             break;
*/
    case TB: if(internal_transition_bypass(COMPOSED_GRAPH) == E_NONE) {
               if(verbose >= -1) {
                 output("After internal transition bypass reduction:\n");
                 print_thread(COMPOSED_GRAPH);
               }
             }
             else return E_NOTIME;
             break;
    case SV: shield_variables(COMPOSED_GRAPH);
             if(verbose >= -1) {
               output("After variables shielding:\n");
               print_thread(COMPOSED_GRAPH);
             }
             break;
    case PG: print_thread(COMPOSED_GRAPH);
             break;
    default:
             break;
  }
  return E_NONE;
}

int action_to_manipulator() {
  struct action_type *a;

  for(a=action_list; ; a=a->next) {
    switch(a->type) {
      case SM: 
      case PM: break;
      case SC: if(a->g0 == NULL && a->c0 != -2) {
                 new_manipulator(SC);
                 clock_shielding = TRUE;
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      case NR: if(a->g0 == NULL && a->c0 != -2) {
                 new_manipulator(NR);   
                 region_normalization = TRUE;
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      case RW: if(a->g0 == NULL && a->c0 != -2) {
                 new_manipulator(RW);   
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      case TB: if(a->g0 == NULL && a->c0 != -2) {
                 new_manipulator(TB);   
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      case SV: if(a->g0 == NULL && a->c0 != -2) {
                 new_manipulator(SV);   
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      case MC: if(a->g0 == NULL && a->c0 != -2) {
                 model_checking = TRUE;
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      case LA: if(a->g0 == NULL && a->c0 != -2) {
                 new_manipulator(LA); 
                 look_forward_reduction = TRUE;
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      case PG: if(a->g0 == NULL && a->c0 != -2) {
                 new_manipulator(PG);
               }
               else {
                 fprintf(stderr, "Warning: Ignoring action ... ");
                 tmp_int = stdoutput;
                 stdoutput = TRUE;
                 print_action(a);
                 stdoutput = tmp_int;
               }
               break;
      default: fprintf(stderr, "Warning: Ignoring action ... ");
               tmp_int = stdoutput;
               stdoutput = TRUE;
               print_action(a);
               stdoutput = tmp_int;
               break;
    }
    if(a->next == action_list) break;
  }
  return 0;
}
