#include "global.h"
#include "modeman.h"
#include "mcheck.h"
#include "print.h"
#include "lf.h"

int model_check(struct thread_type *, struct spec_type *);
int list_check(struct thread_type *, struct spec_type *, int);
int atom_check(struct thread_type *, struct spec_type *);
char **mode_vector(struct mode_type *);
int mode_spec_check(struct mode_type *m, struct mode_spec_type *ms);
int recursive_mode_spec_check(struct mode_type *m, struct spec_type *s);


int model_check(struct thread_type *t, struct spec_type *s) {
  if(verbose >= 0) {
    sprintf(tmpstring, "Model checking thread %s[%d] against spec:\n", 
            t->name, t->automaton_index);
    output(tmpstring);
    print_tctl_spec(s);
  }
  if(list_check(t, s, NO_PATH)) {
    if(verbose >= 0)
      output("Specification satisfied!\n");
    fprintf(stderr, "Specification satisfied!\n");
    return TRUE;
  }
  else {
    if(verbose >= 0)
      output("Specification not satisfied!\n");
    fprintf(stderr, "Specification not satisfied!\n");
    return FALSE;
  }
  return TRUE;
}

int list_check(struct thread_type *t, struct spec_type *s, int parent_path) {
  int ans = 0, tmp, op = 0;
  struct spec_type *ss;

  if(verbose >= 0) {
    sprintf(tmpstring, "List checking thread %s[%d] against spec:\n", 
            t->name, t->automaton_index);
    output(tmpstring);
    print_tctl_spec(s);
  }

  for(ss=s; ; ss=ss->next) {
    if(ss->sign == NEGATIVE)
      tmp = !(list_check(t, ss->s1, parent_path));
    else if(ss->sign == POSITIVE)
      tmp = list_check(t, ss->s1, parent_path);
    else tmp = atom_check(t, ss);
    if(ss == s) ans = tmp;
    else switch(op) {
      case AND: ans &= tmp;
                break;
      case OR:  ans |= tmp;
                break;
    }
    op = ss->boolcon;
    if(ss->next == s) break;
  }
  if(verbose >= 0) {
    if(ans) sprintf(tmpstring, "List Result = TRUE.\n");
    else sprintf(tmpstring, "List Result = FALSE.\n");
    output(tmpstring);
  }
  return ans;
}

int atom_check(struct thread_type *t, struct spec_type *s) {
  reset_status(t);
  switch(s->path) {
    case NO_PATH:     switch(s->stype) {
                        case HENCEFORTH: list_check(t, s->s1, NO_PATH);
                                         break;
                        case EVENTUALLY: 
                                         break;
                        case UNTIL: 
                                    break;
                      }
                      if(s->ms) 
                        return mode_spec_check(t->starting_mode, s->ms);
                      break;
    case EXISTS_PATH: if(s->ms) 
                        return recursive_mode_spec_check(t->starting_mode, s);
                      break;
    case ALL_PATHS:   if(s->ms) 
                        return recursive_mode_spec_check(t->starting_mode, s);
                      break;
  }
  return 0;
}

char **mode_vector(struct mode_type *m) {
  int i=0;
  char **n = (char **)malloc(sizeof(char *) * system_ta_num);
  char *tc = (char *)malloc(sizeof(char) * (strlen(m->name) + 1));
  char *tmp_name = (char *)malloc(sizeof(char) * (strlen(m->name) + 1));

  for(i=0; i<system_ta_num; i++) n[i] = (char *)NULL;
  tmp_name = strcpy(tmp_name, m->name);
  tc = strtok(tmp_name, ".");
  i = 0;
  n[i++] = str_preserve(tc);
  while((tc = strtok((char *)NULL, "."))) {
    n[i++] = str_preserve(tc);
  }
  free(tmp_name);
  free(tc);
  return n;
}

int mode_spec_check(struct mode_type *m, struct mode_spec_type *ms) {
  int i;
  struct automaton *a;
  char **n = mode_vector(m);

  if(verbose >= 0) {
    sprintf(tmpstring, 
            "    Mode spec checking at mode %s(%d) against mode_spec:\n",
            m->name, m->index);
    output(tmpstring);
    sprintf(tmpstring, "    mode(%s) %s %s\n",
            ms->auto_name, relop_string(ms->relop), ms->mode_name);
    output(tmpstring);
  }
  for(i=0, a=automaton_tail->next; i<system_ta_num; i++, a=a->next) {
    if(!strcmp(automaton_fullname(a), ms->auto_name)) {
      if(!strcmp(n[i], ms->mode_name)) {
        if(verbose >= 0) 
          output("    MODE_SPEC Result = TRUE\n");
        return TRUE;
      }
      else {
        if(verbose >= 0)
          output("    MODE_SPEC Result = FALSE\n");
        return FALSE;
      }
    }
  }
  return 0;
}

int recursive_mode_spec_check(struct mode_type *m, struct spec_type *s) {
  struct transition_type *tr;

  if(m->status == VISITED) return FALSE;
  m->status = VISITED;
  if(s->path == EXISTS_PATH) {
    if(m->out_trans_cycle) { 
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
        if(recursive_mode_spec_check(tr->mode, s)) return TRUE;
        if(tr==m->out_trans_cycle) break;
      }
    }
    else return mode_spec_check(m, s->ms);
    return FALSE;
  }
  else if(s->path == ALL_PATHS) {
    if(m->out_trans_cycle) { 
      for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
        if(!recursive_mode_spec_check(tr->mode, s)) return FALSE;
        if(tr==m->out_trans_cycle) break;
      }
    }
    else return mode_spec_check(m, s->ms);
    return TRUE;
  }
  else return mode_spec_check(m, s->ms);
}

