#ifndef __MODEMAN_H
#define __MODEMAN_H

#define UNVISITED	0
#define VISITED		1
#define REACHABLE	4
#define DELTA		-2

struct mode_perm_type {
  struct mode_type	*mode;
  int *perm;
  struct mode_perm_type	*next;
} ;

extern int comp_stat;
extern struct mode_type *same_mode;
extern struct mode_type *last_mode;


extern int init_mode_management(void);
extern int init_thread_management(void);
extern int reset_thread(struct thread_type *);
extern struct mode_type *alloc_mode(void);
extern int free_mode(struct mode_type *);
extern int free_mode_list(struct mode_type *);
extern int mode_count_inc(struct mode_type  *);
extern int mode_parent_set(struct mode_type *, struct a23tree_type *);
extern struct transition_type  *alloc_transition(void);
extern int free_transition(struct transition_type *);
extern int free_transition_cycle(struct transition_type *, int);
extern int transition_candidate_comp(struct transition_type *, 
                                     struct mode_type *); 
extern struct mode_type *initial_mode(struct thread_type *, 
                                      struct thread_type *);
extern int mode_comp(struct mode_type *, struct mode_type *);
extern struct transition_type *add_transition(struct transition_type *, 
                   struct mode_type *, struct mode_type *);
extern struct transition_type *add_transition_fifo(struct transition_type *, 
                   struct mode_type *, struct mode_type *);
extern struct boolconj_type *compute_dvl(struct mode_type *m,
                                         struct transition_type *tr);
extern struct boolconj_type *extract_dvl(struct boolconj_type *, 
                                         struct variable_type *);
extern int neg_lit_trigger(struct transition_type *tr, struct variable_type *);
extern int pos_lit_trigger(struct transition_type *tr, struct variable_type *);
extern struct boolconj_type *new_dvl_neg(struct transition_type *,
                                         struct mode_type *, 
                                         struct variable_type *);
extern struct boolconj_type *new_dvl_pos(struct transition_type *, 
                                         struct variable_type *, 
                                         struct mode_type *);
extern int not_written_by_others(struct boolconj_type *b);
extern int no_dvar_assign(struct transition_type *tr, struct variable_type *);
extern struct boolconj_type *delete_boolconj(int, int, struct boolconj_type *);
extern bdd dvl_bdd(struct mode_type *m);
extern bdd get_bdd_from_dvl(struct boolconj_type *dvl);
extern int print_dvl_bdd(struct boolconj_type *);

extern int transition_m1_first(struct mode_type *);
extern int transition_m2_first(struct mode_type *);
extern int compose(struct thread_type *, struct thread_type *, 
                   struct thread_type *);

extern struct zone_type *
post_xconj(struct mode_type *,        	/* mode 	   */
           struct clkconj_type *, 	/* xconj_cycle of new submode1 */
           struct clkconj_type *, 	/* xconj_cycle of new submode2 */
           struct clkconj_type *, 	/* clk_trigger_cycle   */
           struct assign_type *, 	/* assign_cycle    */
           struct permutation_type *,	/* permutation on transition */
           struct clkconj_type **     	/* new xconj_cycle */
          );

extern struct clkconj_type *process_max(struct clkconj_type **);
extern struct clkconj_type *advance(struct clkconj_type *, int);
extern struct clkconj_type *conjoin(struct clkconj_type *, int);
extern int transitive_op(int, int);
extern struct clkconj_type *assign(struct clkconj_type *, struct assign_type *);
extern bdd bdd_assign(bdd, struct assign_type *);
extern char *get_name(char ***);
extern char *assign_mode_name(struct mode_type *);
extern char *assign_mode_name1(struct mode_type *, int *, int);

extern int init_clk_literals(void);
extern int append_clkcycle(struct clkconj_type *, struct clkconj_type **);
extern int clkconj_not_in_cycle(struct clkconj_type *, struct clkconj_type **);
extern int append_clkconj(struct clkconj_type *, struct clkconj_type **);

extern struct mode_type *normalize_mode(struct mode_type *m);
extern int *get_mode_perm(struct mode_type *m);

extern int mode_shield_clocks(struct mode_type *m);
extern struct var_link_type *var_link_copy(struct var_link_type *sc);
extern struct var_link_type *clocks_union(struct var_link_type *sc1,
                                   struct var_link_type *sc2);
extern struct var_link_type *insert_var_link(struct var_link_type *c,
                                      struct var_link_type *sc);
extern int print_var_links(struct var_link_type *sc);
extern int shield_clocks_compare(struct var_link_type *, struct var_link_type *);

extern int permutation_comp(struct permutation_type *p1, 
                            struct permutation_type *p2);
extern int perm_count_inc(struct permutation_type *p);
extern int perm_parent_set(struct permutation_type *p, 
                           struct a23tree_type *parent);
extern int free_perm(struct permutation_type *p);
extern struct permutation_type *alloc_permutation(void);
extern int print_permutation(struct permutation_type *);
extern struct var_link_type *sort_var_links(struct var_link_type *sc);

extern struct boolconj_type *initial_dvl1(struct thread_type *, 
                                         struct thread_type *,
                                         int);

extern struct boolconj_type *append_boolconj(struct boolconj_type *,
                                      struct boolconj_type *);
extern struct boolconj_type *boolconj_union(struct boolconj_type *,
                                     struct boolconj_type *);
extern struct boolconj_type *check_boolconj(struct boolconj_type *,
                                     struct boolconj_type *);

extern int print_dvl(struct boolconj_type *);
extern int install_dvl_bdd(struct boolconj_type *);
extern int free_var_links(struct var_link_type *);
extern int merge_iden_red_multi(struct thread_type *);
extern int clear_line();
extern int clean_line();
#endif
