#include "include.h"
#include "global.h"
#include "bddman.h"
#include "modeman.h"
#include "zoneman.h"
#include "symmetry.h"


int output(char *);			   	/* output some string */
int print_auto_dec(void);			/* print automata declarations*/
int print_vars(void);				/* print variables */
int print_init_cond(struct automaton *);	/* print initial condition */
int print_boolconj(struct boolconj_type *);	/* print boolconj */
int print_clkconj(struct clkconj_type *);	/* print clkconj */
int print_literals(struct clkconj_type *);	/* print literals */
int print_boolexp(struct boolexp_type *);	/* print boolexp */
int print_inv(struct mode_type *);		/* print invariant */
int print_trigger(struct transition_type *);	/* print triggering condition */
int print_tr_assignment(struct transition_type *);/* print assignment (trans)*/
int print_assignment(struct assign_type *);   /* print assignment (assign)*/
int print_assignment1(struct assign_type *);  /* print 1 assignment (assign)*/
int print_trans(struct transition_type *);	/* print transition */
int print_mode(struct mode_type *);		/* print mode */
int print_automaton(struct automaton *);	/* print automaton */
int print_automata(void);

char *naming_fn(bdd_manager, bdd, pointer);	/* naming function for bdds */
int print_bdd(bdd);				/* print a bdd */
int print_prop_bdds(void);				/* print proposition BDDs */
int print_sop(bdd);

int print_thread(struct thread_type *);
int outputline(void);

int print_tctl_spec(struct spec_type *);
int print_spec_list(struct spec_type *);
int print_spec(struct spec_type *);
int print_until(struct spec_type *);
int print_graph_list();
int print_action_list(struct action_type *);
int print_action(struct action_type *);

char *nmt_string(int);
char *get_graph_name(struct thread_type *);
int view_graph(struct thread_type *);

int print_sync(struct automaton *);
int print_tr_sync(struct transition_type *);

/********************************************************************
 * Function Definitions 
 ********************************************************************/

int output(char *somestring){
  if(logfile) {
    fprintf(logfile, "%s", somestring);
    fflush(logfile);
    if(stdoutput == TRUE) {
      printf("%s", somestring);
      fflush(stdout);
    }
  }
  else {
    printf("%s", somestring);
    fflush(stdout);
  }
  return 0;
}

int print_auto_dec() {
  struct automaton *a;
  output("\nautomata ");
  if(automaton_tail) {
    for(a=automaton_tail->next; ; a=a->next) {
      if(a->index >= 0) sprintf(tmpstring, "%s[%d]", a->name, a->index);
      else if(a->index == NON_PARAM)
        sprintf(tmpstring, "%s", a->name);
      output(tmpstring);
      if(a == automaton_tail) break;
      else output(", ");
    }
  }
  output(";\n");
  return 0;
}

int print_auto_info() {
  int i;

  printf("Automata Information: ");
  if(automata_num) {
    for(i=1; i<=automata_num[0]; i++) {
      printf("%s : %d", automata_name[i], automata_num[i]);
      if(i < automata_num[0]) printf(", ");
    }
  }
  else printf("NONE.");
  printf(";\n");
  return 0;
}

int print_vars() {
  struct variable_type *x;
  int found = FALSE;
  int first = TRUE;

  if(var_tail) {
    for(x=var_tail->next; ; x=x->next) {
      if(x->type == V_CLOCK) {
        found = TRUE;
        break;
      }
      if(x == var_tail) break;
    }
  }

  if(found) {
    output("clock ");
    for(x = var_tail->next; x; x=x->next) {
      if(x->type == V_CLOCK) {
        if(!first) output(", "); 
        if(x->id>=0) 
          sprintf(tmpstring, "%s[%d]", x->name, x->id);
        else
          sprintf(tmpstring, "%s", x->name);
        output(tmpstring);
        if(first) first = FALSE;
      }
      if(x == var_tail) break;
    }
    output(";\n");
  }
  found = FALSE;
  first = TRUE;

  
  if(var_tail) {
    for(x = var_tail->next; ; x=x->next) {
      if(x->type == V_REGISTER) {
        found = TRUE; 
        break;
      }
      if(x == var_tail) break;
    }
  }

  if(found) {
    output("register ");
    for(x = var_tail->next; ; x=x->next) {
      if(x->type == V_REGISTER) {
        if(!first) output(", "); 
        if(x->id>=0) 
          sprintf(tmpstring, "%s[%d]", x->name, x->id);
        else
          sprintf(tmpstring, "%s", x->name);
        output(tmpstring);
        if(first) first = FALSE;
      }
      if(x == var_tail) break;
    }
    output(";\n");
  }
  output("\n");
  return 0;
}

int print_init_cond(struct automaton *a) {
  if(a->init_clkconj)
    print_clkconj(a->init_clkconj);
  if(a->init_boolexp) {
    if(a->init_clkconj) output(" and ");
    print_boolexp(a->init_boolexp);
  }
  return 0;
}

int print_literals(struct clkconj_type *clkconj_cycle) {
  struct clkconj_type *x;
  int i;

  if(clkconj_cycle) { 
    if(clkconj_cycle->op == TRUE) {
      output("True");
      return 0;
    }
    if(clkconj_cycle->op == FALSE) {
      output("False");
      return 0;
    }
    output("{");
    for(x = clkconj_cycle->next, i = 1; x; x=x->next, i++) {
      if(x->clk1 == DELTA) {
        sprintf(tmpstring, "%s ", var_name(x->clk1)); 
        output(tmpstring); 
      }
      else if(x->clk1) {
        sprintf(tmpstring, "%s[%d] ", var_name(x->clk1), var_id(x->clk1)); 
        output(tmpstring); 
      }
      if(x->clk2 == DELTA) {
        sprintf(tmpstring, "- %s ", var_name(x->clk2)); 
        output(tmpstring); 
      }
      else if(x->clk2) { 
         sprintf(tmpstring, "- %s[%d] ", var_name(x->clk2), var_id(x->clk2));
         output(tmpstring);
      }
      if(x->diff == INF) {
        sprintf(tmpstring, "%s INF", relop_string(x->op));
        output(tmpstring);
      }
      else if(x->diff == -INF) {
        sprintf(tmpstring, "%s -INF", relop_string(x->op));
        output(tmpstring);
      }
      else {
        sprintf(tmpstring, "%s %d", relop_string(x->op), x->diff);
        output(tmpstring);
      }
      if(x == clkconj_cycle) break;
      else output(", ");
      if(i%4 == 0) output("\n\t");
    }
    output("}\n");
  }
  return 0;
}

int print_clkconj(struct clkconj_type *clkconj_cycle) {
  struct clkconj_type *x;
  int i;

  if(clkconj_cycle) { 
    if(clkconj_cycle->op == TRUE) {
      output("True");
      return 0;
    }
    if(clkconj_cycle->op == FALSE) {
      output("False");
      return 0;
    }
    for(x = clkconj_cycle->next, i = 1; x; x=x->next, i++) {
      if(x->clk1 == DELTA || (x->clk1 > 0 && var_id(x->clk1) == NON_PARAM)) {
        sprintf(tmpstring, "%s ", var_name(x->clk1)); 
        output(tmpstring); 
      }
      else if(x->clk1) {
        sprintf(tmpstring, "%s[%d] ", var_name(x->clk1), var_id(x->clk1)); 
        output(tmpstring); 
      }
      if(x->clk2 == DELTA || (x->clk2 > 0 && var_id(x->clk2) == NON_PARAM)) {
        sprintf(tmpstring, "- %s ", var_name(x->clk2)); 
        output(tmpstring); 
      }
      else if(x->clk2) { 
         sprintf(tmpstring, "- %s[%d] ", var_name(x->clk2), var_id(x->clk2));
         output(tmpstring);
      }
      if(x->diff == INF) {
        sprintf(tmpstring, "%s INF", relop_string(x->op));
        output(tmpstring);
      }
      else if(x->diff == -INF) {
        sprintf(tmpstring, "%s -INF", relop_string(x->op));
        output(tmpstring);
      }
      else {
        sprintf(tmpstring, "%s %d", relop_string(x->op), x->diff);
        output(tmpstring);
      }
      if(x == clkconj_cycle) break;
      else output(" & ");
      if(i%4 == 0) output("\n\t");
    }
  }
  return 0;
}

int print_one_clkconj(struct clkconj_type *x) {
  if(x->clk1) {
    sprintf(tmpstring, "%s[%d] ", var_name(x->clk1), var_id(x->clk1)); 
    output(tmpstring); 
  }
  if(x->clk2) { 
    sprintf(tmpstring, "- %s[%d] ", var_name(x->clk2), var_id(x->clk2));
    output(tmpstring);
  }
  if(x->diff == INF) {
    sprintf(tmpstring, "%s INF \n", relop_string(x->op));
    output(tmpstring);
  }
  else {
    sprintf(tmpstring, "%s %d \n", relop_string(x->op), x->diff);
    output(tmpstring);
  }
  return 0;
}
  
int print_boolexp(struct boolexp_type *b) {
  int v;
  struct boolexp_type *y;
  struct boolconj_type *z;

  if(b) {
    for(y=b->next; ; y=y->next) {
     if(y->boolconj)
      for(z = y->boolconj->next; ; z=z->next) {
        v = var_id(z->var1);
        if(v >= 0)
          sprintf(tmpstring, "%s[%d] ", 
                var_name(z->var1), v);
        else 
          sprintf(tmpstring, "%s ", 
                var_name(z->var1));
        output(tmpstring);
        if(z->var2) {
          v = var_id(z->var2);
          if(v >= 0)
            sprintf(tmpstring, "- %s[%d] ", 
                  var_name(z->var2), v);
          else 
            sprintf(tmpstring, "- %s ", var_name(z->var2));
          output(tmpstring);
        }
        sprintf(tmpstring, "%s %d", relop_string(z->op), z->constant);
        output(tmpstring);
        if(z == y->boolconj) break;
        else output(" & ");
      }
      if(y == b) break;
      else output(" | ");
    }
  } 
  return 0;
}

int print_inv(struct mode_type *m) {
  if(m->xconj_cycle) 
    print_clkconj(m->xconj_cycle);
  if(m->boolexp) {
    if(m->xconj_cycle)
      output(" and ");
    print_boolexp(m->boolexp);
  }
  if(!m->xconj_cycle && !m->boolexp)
    output("True");
  output(";\n");
  return 0;
}

int print_boolconj(struct boolconj_type *y) {
  int vid;
  struct boolconj_type *z;
  if(y) {
    if(y->op == TRUE) {
      output("True "); 
      return 0;
    }
    for(z = y->next; ; z = z->next) {
      if(var_id(z->var1) == NON_PARAM)
        sprintf(tmpstring, "%s ", var_name(z->var1));
      else
        sprintf(tmpstring, "%s[%d] ", var_name(z->var1), var_id(z->var1));
      output(tmpstring);
      if(z->var2) {
        if(var_id(z->var2) == NON_PARAM)
          sprintf(tmpstring, "- %s ", var_name(z->var2));
        else
          sprintf(tmpstring, "- %s[%d] ", var_name(z->var2), var_id(z->var2));
        output(tmpstring);
      }
      sprintf(tmpstring, "%s %d", relop_string(z->op), z->constant);
      output(tmpstring);
      if(z->mod_num) {
        sprintf(tmpstring, " (mod %d)", z->mod_num);
        output(tmpstring);
      }
      else if(z->mod_var) {
        sprintf(tmpstring, " (mod %s", var_name(z->mod_var));
        output(tmpstring);
        vid = var_id(z->mod_var);
        if(vid >= 0) 
          sprintf(tmpstring, "[%d])", vid);
        else if(vid == NON_PARAM)
          sprintf(tmpstring, ")");
        output(tmpstring);
      }
      if(z == y) break;
      else output(" & ");
    }
  }
  return 0;
}

int print_trigger(struct transition_type *trans_cycle){
  output("when ");
  if( !(trans_cycle->trn->clk_trigger_cycle) && 
      !(trans_cycle->trn->prp_trigger_cycle) ) {
    output("True");
    return 0;
  }
  if( trans_cycle->trn->clk_trigger_cycle && 
      !(trans_cycle->trn->prp_trigger_cycle) ) {
    if(trans_cycle->trn->clk_trigger_cycle->op == TRUE) {
      output("True"); 
      return 0;
    }
  }
  if( !(trans_cycle->trn->clk_trigger_cycle) && 
      trans_cycle->trn->prp_trigger_cycle ) {
    if(trans_cycle->trn->prp_trigger_cycle->op == TRUE) {
      output("True"); 
      return 0;
    }
  }
  if(trans_cycle->trn->clk_trigger_cycle && 
     trans_cycle->trn->prp_trigger_cycle) {
    if( (trans_cycle->trn->clk_trigger_cycle->op == TRUE) &&
        (trans_cycle->trn->prp_trigger_cycle->op == TRUE) ) {
      output("True"); 
      return 0;
    }
  }
  if(trans_cycle->trn->clk_trigger_cycle) {
    if(trans_cycle->trn->clk_trigger_cycle->op != TRUE)
      print_clkconj(trans_cycle->trn->clk_trigger_cycle);
  }
  if(trans_cycle->trn->prp_trigger_cycle) {
    if(trans_cycle->trn->prp_trigger_cycle->op == TRUE) return 0;
    if(trans_cycle->trn->clk_trigger_cycle)
      if(trans_cycle->trn->clk_trigger_cycle->op != TRUE) 
        output(" and ");
    print_boolconj(trans_cycle->trn->prp_trigger_cycle);
  }
  return 0;
}

int print_tr_assignment(struct transition_type *tr){
  if(tr->trn->assign_cycle)
    print_assignment(tr->trn->assign_cycle);
  return 0;
}

int print_assignment(struct assign_type *assign_cycle){
  struct assign_type *x;

  if(assign_cycle) {
    for(x=assign_cycle->next; ; x=x->next) {
      print_assignment1(x);
/*
      if(var_id(x->var) == NON_PARAM)
        sprintf(tmpstring, "%s := ", var_name(x->var));
      else
        sprintf(tmpstring, "%s[%d] := ",
                        var_name(x->var), var_id(x->var));
      output(tmpstring);
      if(x->use_var) {
        if(var_id(x->use_var) == NON_PARAM)
          sprintf(tmpstring, "%s", var_name(x->use_var));
        else
          sprintf(tmpstring, "%s[%d]",
                        var_name(x->use_var), var_id(x->use_var));
        output(tmpstring);
        if(x->value != 0) {
          sprintf(tmpstring, " + %d", x->value);
          output(tmpstring);
        }
      }
      else {
        sprintf(tmpstring, "%d", x->value);
        output(tmpstring);
      }
      if(x->mod_num) {
        sprintf(tmpstring, " (mod %d); ", x->mod_num);
        output(tmpstring);
      }
      else if(x->mod_var) {
        sprintf(tmpstring, " (mod %s", var_name(x->mod_var));
        output(tmpstring);
        vid = var_id(x->mod_var);
        if(vid >= 0)
          sprintf(tmpstring, "[%d]); ", vid);
        else if(vid == NON_PARAM)
          sprintf(tmpstring, "); ");
        output(tmpstring);
      }
      else output("; ");
*/
      if(x == assign_cycle) break;
      else output(" ");
    }
  }
  return 0;
}

int print_assignment1(struct assign_type *x){
  int vid;

  if(!x) return 0;

  if(var_id(x->var) == NON_PARAM)        /* non-parametric variable */
    sprintf(tmpstring, "%s := ", var_name(x->var));
  else
    sprintf(tmpstring, "%s[%d] := ",
                    var_name(x->var), var_id(x->var));
  output(tmpstring);
  if(x->use_var) {
    if(var_id(x->use_var) == NON_PARAM)
      sprintf(tmpstring, "%s", var_name(x->use_var));
    else
      sprintf(tmpstring, "%s[%d]",
                    var_name(x->use_var), var_id(x->use_var));
    output(tmpstring);
    if(x->value != 0) {
      sprintf(tmpstring, " %s %d", (x->op == ADD) ? "+" : "-", x->value);
      output(tmpstring);
    }
  }
  else {
    sprintf(tmpstring, "%d", x->value);
    output(tmpstring);
  }
  if(x->mod_num) {
    sprintf(tmpstring, " (mod %d); ", x->mod_num);
    output(tmpstring);
  }
  else if(x->mod_var) {
    sprintf(tmpstring, " (mod %s", var_name(x->mod_var));
    output(tmpstring);
    vid = var_id(x->mod_var);
    if(vid >= 0)
      sprintf(tmpstring, "[%d]); ", vid);
    else if(vid == NON_PARAM)
      sprintf(tmpstring, "); ");
    output(tmpstring);
  }
  else output("; ");

  return 0;
}

int print_trans(struct transition_type *tr) {
  if(thread == NULL) {
    sprintf(tmpstring, "(#=%d, A=%d, t=%d) ", 
          tr->index, (tr->automaton==NULL)?-1:tr->automaton->u_index, 
          tr->trn->index);
    output(tmpstring);
  }
  print_trigger(tr);
  print_tr_sync(tr);
  if(tr->mode) {
    sprintf(tmpstring, " may goto %s; ", tr->mode->name);
    output(tmpstring);
  }
  else {
    output("Error: in print_trans() no goto mode name.\n");
  }
  print_tr_assignment(tr);
  output("\n");
  return 0;
}  

int print_mode(struct mode_type *m) {
  struct transition_type *x;

  if(m) {
    sprintf(tmpstring, "  mode %s {\n    invariant ", m->name);
    output(tmpstring);
    print_inv(m);
    if(m->out_trans_cycle) {
      for(x=m->out_trans_cycle->next; ; x=x->next) {
        output("    ");
        print_trans(x);
        if(x == m->out_trans_cycle) break;
      }
    }
    output("  }\n");
  }
  return 0;
}

int print_automaton(struct automaton *a) {
  struct mode_type *x;

  x = a->mode_cycle;
  if(a->index >= 0) 
    sprintf(tmpstring, "automaton %s[%d] {\n", a->name, a->index);
  else 
    sprintf(tmpstring, "automaton %s {\n", a->name);
  output(tmpstring);
  print_sync(a);
  if(a->init_mode) {
    sprintf(tmpstring, "initially %s", a->init_mode);
    output(tmpstring);
  }
  if(a->init_boolexp || a->init_clkconj) {
    output(" and ");
    print_init_cond(a);
  }
  if(a->init_mode || a->init_boolexp || a->init_clkconj) 
    output(";\n");
  if(a->mode_cycle) {
    for(x=a->mode_cycle->next; ; x=x->next) {
      print_mode(x);
      if(x == a->mode_cycle) break;
    }
  }
  output("}\n\n");
  return 0;
}

int print_automata() {
  struct automaton *a;

  sprintf(tmpstring, "--- Parsed the following automata descriptions ---\n");
  output(tmpstring);
  print_auto_dec();
  print_vars();
  if(automaton_tail) {
    for(a=automaton_tail->next; ; a=a->next) {
      print_automaton(a);
      if(a==automaton_tail) break;
    }
  }
  sprintf(tmpstring, "--- End of parsed automata descriptions ---\n");
  output(tmpstring);
  return 0;
}

char *naming_fn(bdd_manager bddm, bdd f, pointer env) {
  char *t = (char *)malloc(sizeof(char)*MAX_CHAR);
  struct proposition_type *p;

  if(prop_tail) { 
    for(p = prop_tail->next; ; p=p->next) {
      if((p->prop_bdd) == f) {
        if(var_id(p->var1) >= 0)
          sprintf(tmpstring, "(%s[%d] ", var_name(p->var1), var_id(p->var1));
        else if(var_id(p->var1) == NON_PARAM)
          sprintf(tmpstring, "(%s ", var_name(p->var1));
        if(p->var2) {
          if(var_id(p->var2) >= 0)
            sprintf(t, "- %s[%d] ", var_name(p->var2), var_id(p->var2));
          else if(var_id(p->var2) == NON_PARAM)
            sprintf(t, "- %s ", var_name(p->var2));
          else {
            fprintf(stderr, "Fatal Error in naming_fn\n");
            exit(1);
          }
          strcat(tmpstring, t);
        }
        sprintf(t, "%s %d)", relop_string(p->op), p->value);
        strcat(tmpstring, t);
        free(t);
        return tmpstring;
      }
      if(p == prop_tail) break;
    }
  }
  fprintf(stderr, "Error in naming_fn() for print_bdd()\n");
  free(t);
  return (char *)NULL;
}

/*
int print_bdd(bdd f) {
*/
/*  
  bdd_print_bdd(bddm, f, bdd_naming_fn_none, bdd_terminal_id_fn_none, 
               (pointer)0, stdout);
  bdd_print_bdd(bddm, f, bdd_naming_fn_none, bdd_terminal_id_fn_none, 
               (pointer)0, logfile);
*/
/*
  if(stdoutput == 1) {
    bdd_print_bdd(bddm, f, naming_fn, bdd_terminal_id_fn_none, 
                  (pointer)0, stdout);
    fflush(stdout);
  }
  bdd_print_bdd(bddm, f, naming_fn, bdd_terminal_id_fn_none, 
                (pointer)0, logfile);
  fflush(logfile);
}
*/

int print_bdd(bdd f) {
  if(f)
    print_sop(f);
  else
   output("NULL BDD\n");
  return 0;
}

int print_prop_bdds() {
  struct proposition_type *p;
/*
  struct proposition_type *q;
  bdd f, g, h;
  int cmp;
*/
  if(verbose > 0) {
    sprintf(tmpstring, "Totally %d BDDs created\n", (int)bdd_vars(bddm));
    output(tmpstring);
    if(prop_tail) {
      for(p = prop_tail->next; ; p = p->next) {
        print_bdd(p->prop_bdd);
        if(p == prop_tail) break;
      }
    }
  }

/**** testing bdd_compare() *****/
/*
  stdoutput = 1;
  f = automaton_tail->next->next->mode_cycle->next->next->boolexp->boolexp_bdd;
  g = automaton_tail->mode_cycle->boolexp->boolexp_bdd; 

  printf("f is ...\n"); print_bdd(f);
  printf("g is ...\n"); print_bdd(g);

  cmp = bdd_compare(bddm, f, g, prop_tail->prop_bdd);
  if(cmp>0) printf("f > g\n");
  else if(cmp < 0) printf("f < g\n");
  else if(cmp == 0) printf("f = g\n");
*/
/*
  f = bdd_and(bddm, p->prop_bdd, q->prop_bdd);
  printf("(%s[%d] - %s[%d] %s %d) AND (%s[%d] - %s[%d] %s %d) is ... \n", 
          var_name(p->var1), var_id(p->var1), 
          var_name(p->var2), var_id(p->var2), 
          relop_string(p->op), p->value, 
          var_name(q->var1), var_id(q->var1), 
          var_name(q->var2), var_id(q->var2), 
	  relop_string(q->op), q->value);
  print_bdd(f);

  printf("\n\n");

  p = q->next;
  q = p->next;
  g = bdd_or(bddm, p->prop_bdd, q->prop_bdd);
  printf("(%s[%d] - %s[%d] %s %d) OR (%s[%d] - %s[%d] %s %d) is ... \n", 
          var_name(p->var1), var_id(p->var1), 
          var_name(p->var2), var_id(p->var2), 
          relop_string(p->op), p->value, 
          var_name(q->var1), var_id(q->var1), 
          var_name(q->var2), var_id(q->var2), 
          relop_string(q->op), q->value);
  print_bdd(g);

  h = bdd_and(bddm, f, g);
  printf("AND of above two bdds is ... \n");
  print_bdd(h);
*/
  return 0;
}

/* used by print_sop */

int print_cube(bdd cube) {
  bdd v;
  bdd next;
  bdd temp;
  int i=0;

  cube=bdd_identity(bddm, cube);
  while (1)
    {
      i++;
      v=bdd_if(bddm, cube);
      temp=bdd_and(bddm, v, cube);
      if (temp == bdd_zero(bddm))
	output("!");
      else
	bdd_free(bddm, temp);
/*
      fputs(naming_fn(bddm, v, (pointer)0), stdout);
*/
      sprintf(tmpstring, "%s", naming_fn(bddm, v, (pointer)0));
      output(tmpstring);
      next=bdd_then(bddm, cube);
      if (next == bdd_zero(bddm))
	next=bdd_else(bddm, cube);
/*
      bdd_free(bddm, cube);
*/
      cube=next;
      if(fmod(i, 4) == 0) output("\n\t");
      if (cube != bdd_one(bddm))
        output(" & ");
      else
	break;
    }
  return 0;
}


/* Some attempt at printing reasonably small SOP expressions... */

int print_sop(bdd f) {
  bdd outside;
  bdd cube;
  bdd temp1, temp2;

  if (f == bdd_one(bddm))
    output("1\n");
  else if (f == bdd_zero(bddm))
    output("0\n");
  else
    {
      f=bdd_identity(bddm, f);
      outside=bdd_not(bddm, f);
      while (1)
	{
	  /* Pick a cube. */
	  cube=bdd_satisfy(bddm, f);
	  temp1=bdd_or(bddm, outside, cube);
	  temp2=cube;
	  /* Expand the cube, but don't include anything outside of f. */
	  cube=bdd_reduce(bddm, temp2, temp1);
	  bdd_free(bddm, temp1);
	  bdd_free(bddm, temp2);
          output("{");
	  print_cube(cube);
          output("}");
	  temp1=bdd_not(bddm, cube);
	  bdd_free(bddm, cube);
	  temp2=f;
	  /* Remove cube from f. */
	  f=bdd_and(bddm, temp1, temp2);
	  bdd_free(bddm, temp1);
	  bdd_free(bddm, temp2);
	  if (f != bdd_zero(bddm))
	    output(" |");
	  else
	    {
	      output("\n");
	      break;
	    }
	}
      bdd_free(bddm, outside);
    }
  return 0;
}

int print_thread(struct thread_type *t) {
struct mode_type *m;
struct transition_type *tr;
  char *c = get_graph_name(t);

  outputline();
  sprintf(tmpstring, 
          "Printing graph %s: %d modes, %d transitions\n", 
          c, t->mode_count, t->trans_count);
  output(tmpstring);
  if(t->starting_mode) {
    for(m = t->starting_mode; m; m=m->cnext) {
      sprintf(tmpstring, 
              "Mode %s (%d) with %d in-trans, %d out-trans.\n", 
              m->name, m->index, m->in_trans_count, m->out_trans_count); 
      output(tmpstring);
      if(clock_shielding) {
        output("    Shielded Clocks: ");
        print_var_links(m->shielded_clocks);
      }
      /*if(!lazy_bdd_merging) { */
        output("    Literals Set: ");
        print_dvl(m->dvl);
      /*} */
/*
      output("    CVL SET: ");
      print_literals(m->cvl);
      output("    Inv BDD: ");
      print_bdd(m->boolexp->boolexp_bdd);
*/
     /* else { */
        output("    Dval BDD: ");
        print_bdd(m->dval_bdd);
     /* } */
      if(total_clocks > 0) {
        output("    Zone: ");
        print_zone(m->zone);
        output("\n");
      }
      if(m->in_trans_count) {
        output("    IN: {"); 
        for(tr=m->in_trans_cycle->next; ; tr=tr->next) {
          sprintf(tmpstring, "[%d] %s(%d)", 
                  tr->index, tr->mode->name, tr->mode->index);
          output(tmpstring);
          if(tr == m->in_trans_cycle) break;
          else output("; ");
        }
        output("}.\n");
      }
      if(m->out_trans_count) {
        output("    OUT: {");
        for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
          sprintf(tmpstring, "[%d] %s(%d)", 
                  tr->index, tr->mode->name, tr->mode->index);
          output(tmpstring);
          if(tr == m->out_trans_cycle) break;
          else output(";  ");
        }
        output("}.\n");
        for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
          sprintf(tmpstring, 
            "\tTrans #%d corr to trans #%d of TA #%d, dest mode #%d", 
            tr->index, tr->trn->index, 
            (tr->automaton==NULL)?-1:tr->automaton->u_index, tr->mode->index);
          output(tmpstring);
          if(tr->bypassed_internal_vars) {
            output(", Shield Vars: ");
            print_var_links(tr->bypassed_internal_vars);
          }
          else output(".\n");
          output("\t  ");
          print_trans(tr);
          if(region_normalization) {
            output("\t  ");
            print_trans_permutation(tr);
          }
          if(tr == m->out_trans_cycle) break;
        }
      }
    }
  }
  sprintf(tmpstring, "------ END of Graph %s ------\n", c);
  output(tmpstring);
  free(c);
  return 0;
}

int print_counts(struct thread_type *t) {
/*
  int t_count = 0, it_count = 0, m_count = 0;
  struct mode_type *m;
  struct transition_type *t;

  for(m = t->starting_mode; m; m=m->cnext) {
    m_count++;
    if(m->out_trans_cycle) {
      for(t = m->out_trans_cycle->next; ; t = t->next) {
        t_count++;
        if(BDD_AND(m->dval_bdd, t->trigger_bdd) == BDD_FALSE) 
          it_count++;
        if(t == m->out_trans_cycle) break;
      }
    }
  }
  if(t->automaton_index >= 0)
    printf(
    "Calculated: %d modes, %d transitions (Graph %s[%d])\n",
        m_count, t_count, t->name, t->automaton_index);
  else
    printf(
    "Calculated: %d modes, %d transitions (Graph %s)\n",
        m_count, t_count, t->name);
*/
  if(t->automaton_index >= 0)
    printf("%d modes, %d transitions (Graph %s[%d])\n",
           t->mode_count, t->trans_count, t->name, t->automaton_index); 
  else
    printf("%d modes, %d transitions (Graph %s)\n",
           t->mode_count, t->trans_count, t->name); 
  fflush(stdout);
  return 0;
}

int outputline() {
  output("-----------------------------------------------------------------\n");
  return 0;
}


/******************************************************************
 **   For specification
 ******************************************************************/

int print_tctl_spec(struct spec_type *s) {
  output("Specification: \n");
  print_spec_list(s);
  output(";\n");
  return 0;
}

int print_spec_list(struct spec_type *s) {
  struct spec_type *t;
  
  for(t=s; ; t=t->next) {
    if(t->sign == NEGATIVE) {
      output("\nnot {");
      print_spec_list(t->s1);
      output("}");
    }
    else if(t->sign == POSITIVE) {
      output("\n{");
      print_spec_list(t->s1);
      output("}");
    }
    else print_spec(t);
    if(t->boolcon == AND || t->boolcon == OR) {
      output(bool_string(t->boolcon));
    }
    if(t->next == s) break;
  }
  return 0;
}

int print_spec(struct spec_type *s) {
  if(s->path) {
    switch(s->path) {
      case EXISTS_PATH: output("\nexists_path ");
                        break;
      case ALL_PATHS: output("\nall_paths ");
                      break;
    }
  }
  output("(");
  if(s->stype == HENCEFORTH) {
    sprintf(tmpstring, "henceforth %s %d ", 
            relop_string(s->relop), s->constant);
    output(tmpstring);
    print_spec_list(s->s1);
  }
  else if(s->stype == EVENTUALLY) {
    sprintf(tmpstring, "eventually %s %d ", 
            relop_string(s->relop), s->constant);
    output(tmpstring);
    print_spec_list(s->s2);
  }
  else if(s->stype == UNTIL) {
    print_spec_list(s->s1);
    sprintf(tmpstring, " until %s %d ", relop_string(s->relop), s->constant);
    output(tmpstring);
    print_spec_list(s->s2);
  }
  else if(s->ms) {
    sprintf(tmpstring, "mode(%s) %s %s", 
            s->ms->auto_name, relop_string(s->ms->relop), s->ms->mode_name);
    output(tmpstring);
  }
  else {
    if(s->b1 || s->c1) {
      if(s->b1) print_boolexp(s->b1);
      if(s->b1 && s->c1) output(" and ");
      if(s->c1) print_clkconj(s->c1);
    }
    else if(s->s1) {
      print_spec_list(s->s1);
    }
    if(s->b2 || s->c2) {
      print_until(s);
      if(s->b2) print_boolexp(s->b2);
      if(s->b2 && s->c2) output(" and ");
      if(s->c2) print_clkconj(s->c2);
    }
    else if(s->s2) {
      print_until(s);
      print_spec(s->s2);
    }
  }
  output(")");
  return 0;
}

int print_until(struct spec_type *s) {
  sprintf(tmpstring, " UNTIL %s %d ", relop_string(s->relop), s->constant);
  output(tmpstring);
  return 0;
}

int print_manipulations() {
  output("\nmanipulation\n");
  print_graph_list();
  print_action_list(action_list);
  return 0;
}

int print_graph_list() {
  struct thread_type *t;
  if(graph_list) {
    output("graph ");
    for(t = graph_list->next; ; t=t->next) {
      if(t != graph_list->next)  output(", ");
      if(t->automaton_index == -1) 
        sprintf(tmpstring, "%s[i]", t->name);
      else if(t->automaton_index == 0)
        sprintf(tmpstring, "%s", t->name);
      else if(t->automaton_index > 0)
        sprintf(tmpstring, "%s[%d]", t->name, t->automaton_index);
      output(tmpstring);
      if(t == graph_list) break;
    }
    output(";\n");
  }
  return 0;
}

int print_action_list(struct action_type *alist) {
  struct action_type *a;
  if(alist) {
    for(a = alist; ; a=a->next) {
      print_action(a);
      if(a->next == alist) break;
    }
  }
  return 0;
}

int print_action(struct action_type *a) {
  char *t0 = NULL, *t1 = NULL, *t2 = NULL;
  if(a->g0) {
    switch(a->g0->automaton_index) {
      case 0:  sprintf(tmpstring, "%s", a->g0->name);
               break;
      default: sprintf(tmpstring, "%s[%d]", 
                                  a->g0->name, a->g0->automaton_index);
               break;
    }
    t0 = str_preserve(tmpstring);
  }
  else if(a->c0 == -1) {
    if(a->p0 > 0) sprintf(tmpstring, "%s[i + %d]", a->n0, a->p0);
    else if(a->p0 < 0) sprintf(tmpstring, "%s[i - %d]", a->n0, -(a->p0));
    else sprintf(tmpstring, "%s[i]", a->n0);
    t0 = str_preserve(tmpstring);
  }
  if(a->g1) {
    switch(a->g1->automaton_index) {
      case 0:  sprintf(tmpstring, "%s", a->g1->name);
               break;
      default: sprintf(tmpstring,
                       "%s[%d]", a->g1->name, a->g1->automaton_index);
               break;
    }
    t1 = str_preserve(tmpstring);
  }
  else if(a->c1 == -1) {
    if(a->p1 > 0) sprintf(tmpstring, "%s[i + %d]", a->n1, a->p1);
    else if(a->p1 < 0) sprintf(tmpstring, "%s[i - %d]", a->n1, -(a->p1));
    else sprintf(tmpstring, "%s[i]", a->n1);
    t1 = str_preserve(tmpstring);
  }
  if(a->g2) {
    switch(a->g2->automaton_index) {
      case 0:  sprintf(tmpstring, "%s", a->g2->name);
               break;
      default: sprintf(tmpstring, 
                       "%s[%d]", a->g2->name, a->g2->automaton_index);
               break;
    }
    t2 = str_preserve(tmpstring);
  }
  else if(a->c2 == -1) {
    if(a->p2 > 0) sprintf(tmpstring, "%s[i + %d]", a->n2, a->p2);
    else if(a->p2 < 0) sprintf(tmpstring, "%s[i - %d]", a->n2, -(a->p2));
    else sprintf(tmpstring, "%s[i]", a->n2);
    t2 = str_preserve(tmpstring);
  }
  switch(a->type) {
    case MG: 
      sprintf(tmpstring, "%s := merge_graph(%s, %s);\n", t0, t1, t2);
      output(tmpstring);
      break;
    case SC: sprintf(tmpstring, "shield_clock(%s);\n", t1?t1:"");
             output(tmpstring);
             break;
    case NR: sprintf(tmpstring, "normalize_region(%s);\n", t1?t1:"");
             output(tmpstring);
             break;
    case LA: if(t1 && t2)
               sprintf(tmpstring, "look_ahead(%s, %s);\n", t1, t2);
             else 
               sprintf(tmpstring, "look_ahead();\n");
             output(tmpstring);
             break;
    case MC: sprintf(tmpstring, "model_check(%s);\n", t1?t1:"");
             output(tmpstring);
             break;
    case PG: sprintf(tmpstring, "print_graph(%s);\n", t1?t1:"");
             output(tmpstring);
             break;
    case SM: sprintf(tmpstring, "sequential_merge();\n");
             output(tmpstring);
             break;
    case PM: sprintf(tmpstring, "pairwise_merge();\n");
             output(tmpstring);
             break;
    case IE: if(a->g1 && a->g2) {
               sprintf(tmpstring, "if(%s(%s) %s %s(%s)) {\n",
                                   nmt_string(a->if_num), t1, 
                                   relop_string(a->if_relop),
                                   nmt_string(a->if_num), t2);
               output(tmpstring);
             }
             else { sprintf(tmpstring, "if(i %s %d) {\n", 
                                       relop_string(a->if_relop), a->if_num);
                    output(tmpstring);
             }
             print_action_list(a->if_list);
             output("}\n");
             if(a->else_list) {
               output("else {\n");
               print_action_list(a->else_list);
               output("}\n");
             }
             break;
    case FL: sprintf(tmpstring, "for(i := %d; i %s %d; i %s= %d) {\n", 
                        a->for_init_num,
                        relop_string(a->for_cond_relop),
                        a->for_cond_num,
                        a->for_step > 0 ? " +":" -",
                        a->for_step
                     );
             output(tmpstring);
             print_action_list(a->loop_list);
             output("}\n");
             break;
    default: fprintf(stderr, 
                     "Error: %d is an undefined action type index.\n", a->type);
             break;
  }
  return 0;
}

char *nmt_string(int nmt) {
  if(nmt == NUM_MODES) return "num_modes";
  else if(nmt == NUM_TRANS) return "num_trans";
  else return NULL;
}

int print_prompt() {
  printf("%s", prompt);
  fflush(stdout);
  return 0;
}

char *get_graph_name(struct thread_type *t) {
  char *n;
  if(t) {
    if(t->automaton_index >= 0) {
      sprintf(string_buffer, "%s[%d]", t->name, t->automaton_index);
      n = str_preserve(string_buffer);
    }
    else n = str_preserve(t->name);
    return n;
  }
  else return (char *)NULL;
}

int view_graph(struct thread_type *t) {
  FILE *f, *l;
  char *n, *c;

  if(t == NULL) return 0;
  c = get_graph_name(t);
  n = (char *)malloc(sizeof(char)*(strlen(c) + 10));
  sprintf(n, "/tmp/%s.g__", c);
  f = fopen(n, "w");
  l = logfile;
  logfile = f;
  print_thread(t);
/*
  sprintf(tmpstring, "textedit -read_only %s&", n);
  system(tmpstring);
*/

  fclose(f);
  logfile = l;
  free(c);
/*
  sprintf(tmpstring, "/bin/rm %s", n);
  system(tmpstring);
*/
  free(n);
  return 0;
}

int print_sync(struct automaton *a) {
  struct var_link_type *vl;
  output("synclabs ");
  if(a->sync_cycle)
  for(vl=a->sync_cycle->next; ; vl=vl->next) {
    sprintf(tmpstring, "%s", vl->var->name);
    output(tmpstring);
    if(vl == a->sync_cycle) break;
    else output(", ");
  }
  output(";\n");
  return 0;
}

int print_tr_sync(struct transition_type *tr) {
  struct var_link_type *vl;
  if(tr->trn->sync_cycle)
    for(vl=tr->trn->sync_cycle->next; ; vl=vl->next) {
      sprintf(tmpstring, " sync %s;", vl->var->name);
      output(tmpstring);
      if(vl == tr->trn->sync_cycle) break;
    }
  return 0;
}
