#include "include.h"
#include "global.h"
#include "bddman.h"
#include "modeman.h"
#include "lf.h"
#include "print.h"
#include "transman.h"

int shield_variables(struct thread_type *);
struct boolconj_type *hide_a_var(int, struct boolconj_type *);
int var_not_read(struct mode_type *, int);
struct boolconj_type *hide_internal(struct boolconj_type *);

struct boolconj_type *hide_vars(struct var_link_type *, 
                                struct boolconj_type *);

int shield_variables(struct thread_type *t) {
  struct mode_type *m;
  struct variable_type *dv;
  char *c;

  NEW_THREAD = t;
  t->mnpl |= SV;
  c = get_graph_name(t);
  printf("    Shielding discrete variables in graph %s.\n", c);
  fflush(stdout);
  if(verbose >= 0) {
    sprintf(tmpstring, "\nShielding discrete variables in graph %s.\n", c);
    output(tmpstring);
  }

  if(t==NULL) return 1;
  if(t->starting_mode) {
    for(m = t->starting_mode; m; m = m->cnext) {
      for(dv = var_tail->next; ; dv = dv->next) {
        reset_status(t);
        if(dv->type == V_REGISTER) { 
          if(is_internal_var(dv, NEW_THREAD) && var_not_read(m, dv->index)) {
            if(verbose >= 0) {
              sprintf(tmpstring, "Var %s can be hidden in mode %s(%d)\n",
				 dv->name, m->name, m->index);
              output(tmpstring);
            }
	    m->dvl = hide_a_var(dv->index, m->dvl);
	  }
        }
        if(dv == var_tail) break;
      }
    }
    printf("    ");
    print_counts(t);
    if(verbose >= 0) {
      sprintf(tmpstring, "\nEND of shielding variables in graph %s.\n", c);
      output(tmpstring);
    }
    free(c);
    merge_iden_red_multi(t);
  }
  return 0;
}

struct boolconj_type *hide_a_var(int var, struct boolconj_type *dvl) {
  struct boolconj_type *bp, *b;

  if(dvl == NULL) return dvl;
  if(dvl->next == dvl) {
    if(dvl->var1 == var) {
      free_boolconj(dvl);
      return NULL;
    }
    else return dvl;
  }
  for(bp=dvl, b=bp->next; ; bp=b, b=b->next) {
    if(var == b->var1) {
      if(dvl->next == dvl) {
        free_boolconj(dvl);
        return NULL;
      }
      bp->next = b->next;
      if(b == dvl) dvl = bp;
      b->next = b;
      free_boolconj(b);
      break;
    }
    if(b==dvl) break;
  }
  return dvl;
}

int var_not_read(struct mode_type *m, int v) {
  struct transition_type *tr;
  struct boolconj_type *b;
  struct assign_type *a;

  if(m->status == VISITED) return TRUE;
  else m->status = VISITED;

  if(m->out_trans_cycle) {
    for(tr = m->out_trans_cycle->next; ; tr=tr->next) {
      if(tr->trn->prp_trigger_cycle) {
        for(b = tr->trn->prp_trigger_cycle->next; ; b = b->next) {
          if(b->var1 == v) return FALSE;
          if(b == tr->trn->prp_trigger_cycle) break;
        }
      }
      if(tr->trn->assign_cycle) {
        for(a = tr->trn->assign_cycle->next; ; a = a->next) {
          if(a->var == v) break;
          else {
            if(var_not_read(tr->mode, v) == FALSE) return FALSE;
          }
          if(a == tr->trn->assign_cycle) break;
        }
      }
      if(tr == m->out_trans_cycle) break;
    }
  }
  return TRUE;
}

struct boolconj_type *hide_internal(struct boolconj_type *dvl) {
  struct boolconj_type *b;
  struct variable_type *dv;

  if(dvl == NULL) return NULL;
  b = boolconj_copy(dvl);
  for(dv = var_tail->next; ; dv=dv->next) {
    if(dv->type == V_REGISTER && is_internal_var(dv, NEW_THREAD)) {
      b = hide_a_var(dv->index, b);
    }
    if(dv == var_tail) break;
  }
  return b;
}

struct boolconj_type *hide_vars(struct var_link_type *vl, 
                                struct boolconj_type *dvl) {
  struct boolconj_type *b, *copy;
  struct var_link_type *vlp;

  if(dvl == NULL) return NULL;
  copy = boolconj_copy(dvl);
  if(vl == NULL) return copy;
  for(b = dvl->next; ; b = b->next) {
    for(vlp = vl->next; ; vlp=vlp->next) {
      if(vlp->var->index == b->var1 || vlp->var->index == b->var2) {
        copy = hide_a_var(vlp->var->index, copy);
        break;
      }
      if(vlp == vl) break;
    }
    if(b == dvl) break;
  }
  return copy;
}
