#include "global.h"
#include "print.h"
#include "modeman.h"
#include "bddman.h"
#include "zoneman.h"
#include "shieldclk.h"
#include "readwrite.h"
#include "symmetry.h"
#include "transman.h"

int *ordered_clocks;
int *sort_array;
int *old_order;
struct mode_type *CURR_MODE;

extern void qsort();
/*
extern void qsort(void *base, size_t nel, size_t width,
                  int (*compar) (const void *, const void *));
*/

int init_permutation(void);
int eliminate_transition(struct transition_type *, struct mode_type *);
int remove_irreachable_mode(struct thread_type *);
int init_order(void);
int init_sort_array(void);
int create_order(void);
int normal_index_compare(const int *, const int *);
int sort_clocks(struct zone_type *);
int different_order(int *, int *);
int move_before_sorted(int, int, struct zone_type *);
int move_after(int, int);
int perm_index(int, int *);
int install_permutation(struct transition_type *, struct mode_type *,
                        struct mode_type *, struct mode_perm_type *,
                        struct thread_type *);
int *inverse_permutation(int *);
int *compose_permutation(int *, int *);
struct zone_type *permute_zone(struct zone_type *, int *, struct thread_type *);
int print_ordered_clocks(void);
struct zone_type *perm_kclosure_zone(struct zone_type *);
struct mode_type *create_new_normal_mode(struct mode_type *, 
                                         struct thread_type *, 
                                         struct mode_type **);
struct mode_perm_type *create_normal_mode_list_using_sort(struct thread_type *);
struct mode_type *find_in_list(struct mode_type *, struct mode_perm_type *);
bdd permute_bdd(bdd, int *);
struct boolconj_type *permute_boolconj(struct boolconj_type *, int *);
struct boolconj_type *permute_boolconj1(struct boolconj_type *, int *);

int check_normal_mode_list(struct mode_perm_type *list, struct thread_type *);
int change_normal_mode(struct mode_type *nm, 
                       struct mode_perm_type *list, struct thread_type *);
struct mode_type *normal_mode(struct mode_type *, struct mode_perm_type *, 
                              struct thread_type *);
int *mode_perm(struct mode_type *, struct mode_perm_type *, 
               struct thread_type *);
int normalize_regions(struct thread_type *);
int free_normal_mode_list(struct mode_perm_type *);
int free_mode_perm(struct mode_perm_type *);
struct mode_type *normalize_thread(struct thread_type *, 
                                   struct mode_perm_type *);
int compact_reduce(struct mode_type *, struct transition_type *, 
                   struct mode_perm_type *, struct thread_type *);
int install_in_permutation(struct transition_type *, struct mode_type *);
int is_normal_zone(void);
int is_normal_bdd(bdd);
int is_normal_mode(struct mode_type *m);
int *get_perm_zone(struct zone_type *z);
int *get_perm_bdd(bdd);
int is_normal_perm(int *);
struct mode_type *find_normal_mode(struct mode_type *, struct thread_type *);
int remove_non_normal_mode(struct thread_type *, struct mode_perm_type *);
int reduce_multigraph(struct thread_type *);
int corr_trans(struct transition_type *, struct transition_type *);
int dvl_compare(struct boolconj_type *, struct boolconj_type *);
int permute_variable(int, int *);
int permute_assignment(struct assign_type *, int *);
struct clkconj_type *permute_clkconj(struct clkconj_type *, int *);
int permute_prp_trigger(struct boolconj_type *, int *);
struct transition_type *permute_transition(struct transition_type *, int *);
int print_trans_permutation(struct transition_type *);
int print_perm(int *);
int is_normal_trans(struct transition_type *);
int copy_clock_order(int *);
char *permute_mode_name(struct mode_type *, int *, char ***);
int *copy_permutation(int *, int *);
int remove_dormant_cycles(struct thread_type *th);
int dfs_visit_mode(struct mode_type *m);

struct thread_type *get_thread(struct automaton *);
struct thread_type *get_thread1(char *, int);
struct automaton *get_automaton(struct thread_type *);

/****** FUNCTIONS **********/

int init_permutation() {
  init_23tree(&perm_tree);
  return 0;
}

int eliminate_transition(struct transition_type *e, struct mode_type *m) {
struct transition_type *t, *pt;

  m->out_trans_count--;
  if(m->out_trans_cycle == m->out_trans_cycle->next) {
    m->out_trans_cycle = NULL;
  }
  else {
    pt = m->out_trans_cycle;
    t = pt->next;
    for(; ; pt=t, t=t->next) {
      if(t == e) {
        pt->next = e->next;
        if(m->out_trans_cycle == e)
          m->out_trans_cycle = pt;
        break;
      }
    }
  }

/*
  printf("Mode %d (%s) now has %d out and %d in trans.\n", 
         m->index, m->name, m->out_trans_count, m->in_trans_count);

  printf("transition %d goes to mode %d (%s) with %d out and %d in trans.\n", 
         e->index, e->mode->index, e->mode->name, e->mode->out_trans_count,
         e->mode->in_trans_count);
*/

  (e->mode->in_trans_count)--;

  if(e->mode->in_trans_cycle == e->mode->in_trans_cycle->next) {
    t = e->mode->in_trans_cycle;
    e->mode->in_trans_cycle = NULL;
  }
  else {
    pt = e->mode->in_trans_cycle;
    t = pt->next;
    for(; ; pt=t, t=t->next) {
      if(t->index == e->index) {
        pt->next = t->next;
        if(e->mode->in_trans_cycle == t)
          e->mode->in_trans_cycle = pt;
        break;
      }
    }
  }
/*
  if(e->index != t->index) {
    printf("Fatal Error in eliminate transition.\n"); getchar(); }
  printf(" ... freeing transition %d.\n", e->index);
*/
  free_transition(e);
/*
  printf(" ... freeing transition %d.\n", t->index);
*/
  free_transition(t);
  return 0;
}

int remove_irreachable_mode(struct thread_type *t) {
  struct mode_type *mp;
  struct mode_type *m;
  struct mode_type *ms;
  int again = FALSE;
  int tmp_count;
  char *c = get_graph_name(t);

/*
  if(verbose >= 0) {
    outputline();
    sprintf(tmpstring, 
   "Removing irreachable modes in graph %s (%d modes, %d transitions).\n", 
     c, t->mode_count, t->trans_count);
    output(tmpstring);
  }
*/

  if(t->starting_mode) 
    mp = t->starting_mode;
  else {
    printf("Graph %s has no mode (NULL starting mode).\n", c);
    printf("\tNot removing irreachable mode!\n");
    if(verbose >= 0) {
      sprintf(tmpstring, "Graph %s has no mode (NULL starting mode).\n", c);
      output(tmpstring);
      output("Not removing irreachable mode!\n");
    }
    free(c);
    return 0;
  }

  if(t->starting_mode->cnext) 
    m = t->starting_mode->cnext;
  else {
    printf("Graph %s has only 1 mode (NULL mode after starting mode).\n", c);
    printf("\tNot removing irreachable mode!\n");
    if(verbose >= 0) {
      sprintf(tmpstring, 
            "Graph %s has only 1 mode (NULL mode after starting mode).\n", c);
      output(tmpstring);
      output("Not removing irreachable mode!\n");
    }
    free(c);
    return 0;
  }

  if(t->starting_mode->cnext->cnext)
    ms = t->starting_mode->cnext->cnext;
  else {
    printf("Graph %s has only 2 modes.\n", c);
    printf("\tNot removing irreachable mode!\n");
    if(verbose >= 0) {
      sprintf(tmpstring, "Graph %s has only 2 modes.\n", c);
      output(tmpstring);
      output("Not removing irreachable mode!\n");
    }
    free(c);
    return 0;
  }

  for(; ; m=ms, ms=ms->cnext) {
    if(m->in_trans_count == 0) {
      if(verbose >= 0) {
        sprintf(tmpstring, "Freeing mode %s(%d) with %d out transitions.\n",
                m->name, m->index, m->out_trans_count);
        output(tmpstring);
      }
      tmp_count = m->out_trans_count;
/*
      i = 0;
      if(m->out_trans_cycle) {
        for(tr = m->out_trans_cycle->next; ; tr = tr->next) {
          if(tr->valid == FALSE) i++;
          if(tr == m->out_trans_cycle) break;
        }
      }
      t->invalid_trans_count -= i;
*/
      free_mode(m);
      t->mode_count--;
      t->trans_count -= tmp_count;
      mp->cnext = ms;
      if(again == FALSE) again = TRUE;
    }
    else mp = m;
    if(ms == NULL) break;
  }
  if(again) remove_irreachable_mode(t);
  else if(verbose >= 0) {
    sprintf(tmpstring, "DONE removing irreachable modes in graph %s.\n", c);
    output(tmpstring);
  }
  free(c);
  return 0;
}

int init_order() {
  int i;

  ordered_clocks = (int *)malloc(sizeof(int) * (system_ta_num+1));
  for(i=0; i<=system_ta_num; i++) {
    ordered_clocks[i] = i;
  }
  return 0;
}

int init_sort_array() {
  int i, j;
  if(sort_array)
    sort_array = (int *)realloc(sort_array, sizeof(int) * composed_ta);
  else
    sort_array = (int *)malloc(sizeof(int) * composed_ta);
  for(i=0, j=0; i<composed_ta; i++) {
    while(!component_graph(j, NEW_THREAD)) j++;
    sort_array[i] = j+1; 
    j++;
  }
  return 0;
}


/* Generate the full permutation: 0, 1, ..., system_ta_num
 * if some process index (i) exists in sort_array[],
 * then ordered_clocks[i] = sort_array[j];
 * else do nothing (already ordered_clocks[i] = i in init_order())  
 */

int create_order() {
  int i, j, k, found = FALSE;

  for(i=1, j=0; i <= system_ta_num && j < composed_ta; i++) {
    for(k=0; k<composed_ta; k++) {
      if(sort_array[k] == i) {
        found = TRUE;
        break;
      }
    }
    if(found) {
      found = FALSE;
      ordered_clocks[i] = sort_array[j++];
    }
  }
  return 0;
}

/* Used for qsort() in create_normal_mode_list_using_sort() ...
 * Purpose: To compare and order two process ids, *i and *j
 */

int normal_index_compare(const int *i, const int *j) {
  int k, ii, jj;
  struct variable_type *v;
  struct proposition_type *p;
  bdd f = BDD_TRUE;

/*
    sprintf(tmpstring, "normal_index_compare(%d, %d) = .\n", *i, *j);
    output(tmpstring);
*/

  if((*i) < 0 && (*j) >= 0) return -1;
  if((*j) < 0 && (*i) >= 0) return 1;
  if((*i) < 0 && (*j) < 0) {
    if( ((*i) - (*j)) < 0 ) return -1;
    else if( ((*i) - (*j)) > 0 ) return 1;
    else return EQUAL;
  }
  if(!component_graph((*i)-1, NEW_THREAD) || 
     !component_graph((*j)-1, NEW_THREAD)) {
    if(perm_index(*i, ordered_clocks) < perm_index(*j, ordered_clocks)) 
      return -1;
    else if(perm_index(*i, ordered_clocks) > perm_index(*j, ordered_clocks)) 
      return 1;
    else return EQUAL;
  }

  /* (0) check mode vector */
  k = strcmp(CURR_MODE->mode_vec[(*i)-1], CURR_MODE->mode_vec[(*j)-1]);
  if(k!=EQUAL) return k;
/*
  tmp_name = str_preserve(CURR_MODE->name);
  tc = strtok(tmp_name, ".");
  if((*i) == 1) tci = tc;
  else if((*j) == 1) tcj = tc;
  k = 1;
  while(tc = strtok((char *)NULL, ".")) {
    k++;
    if((*i) == k) tci = tc;
    if((*j) == k) tcj = tc;
  }
  ii = atoi(&tci[1]);
  jj = atoi(&tcj[1]);
  free(tmp_name);

  if(ii < jj) {
    return -1;
  }
  if(ii > jj) {
    return 1;
  }
*/
  /* (1) check all global discrete variables */
/*
  c = alloc_boolconj();
  for(v=var_tail->next; ; v=v->next) {
    if(v->type == V_REGISTER) {
      c->var1 = v->index;
      c->op = EQ;
      c->constant = (*i);
      install_bdd(c);
      if(BDD_AND(CURR_MODE->dval_bdd, c->prop_bdd) != BDD_FALSE) {
        return (-1);
      }
      c->constant = (*j);
      install_bdd(c);
      if(BDD_AND(CURR_MODE->dval_bdd, c->prop_bdd) != BDD_FALSE) {
        return (1);
      }
    }
    if(v == var_tail) break;
  } 
*/
/*
  free_boolconj(c);
*/

  /* (1) check the discrete variable literal sets */
  for(v=var_tail->next; ; v=v->next) {
    if(v->type == V_REGISTER && v->par_val == TRUE) {
      for(p=prop_tail->next; ; p = p->next) {
        if(p->var1 == v->index && p->value == (*i))
          f = bdd_and(bddm, f, p->prop_bdd);
        if(p->var1 == v->index && p->value == (*j))
          f = bdd_and(bddm, f, bdd_not(bddm, p->prop_bdd));
        if(p == prop_tail) break;
      }
      if(bdd_and(bddm, dvl_bdd(CURR_MODE), f) != BDD_FALSE) {
        return (-1);
      }

      f = BDD_TRUE;
      for(p=prop_tail->next; ; p = p->next) {
        if(p->var1 == v->index && p->value == (*j))
          f = bdd_and(bddm, f, p->prop_bdd);
        if(p->var1 == v->index && p->value == (*i))
          f = bdd_and(bddm, f, bdd_not(bddm, p->prop_bdd));
        if(p == prop_tail) break;
      }
      if(bdd_and(bddm, dvl_bdd(CURR_MODE), f) != BDD_FALSE) {
        return (1);
      }
    }
    if(v == var_tail) break;
  }

  /* local discrete variables are not yet implemented */
  /* (2) check all local clocks */
  ii = new_index(*i, CURR_MODE->zone);
  jj = new_index(*j, CURR_MODE->zone);

  if(CURR_MODE->zone->relop[0][0] != TRUE) {
    if( (CURR_MODE->zone->relop[ii][jj] == LEQ ||
         CURR_MODE->zone->relop[ii][jj] == LESS) &&
        (CURR_MODE->zone->diff[ii][jj] <= 0)
      ) {
      return (-1);
    }
    if( (CURR_MODE->zone->relop[jj][ii] == LEQ || 
         CURR_MODE->zone->relop[jj][ii] == LESS) &&
        (CURR_MODE->zone->diff[jj][ii] <= 0)
      ) {
      return (1);
    }
    if(CURR_MODE->zone->relop[0][ii] != -1) {
      if(CURR_MODE->zone->relop[0][jj] != -1) {
        if(CURR_MODE->zone->diff[0][ii] > CURR_MODE->zone->diff[0][jj])
          return -1;
        if(CURR_MODE->zone->diff[0][ii] < CURR_MODE->zone->diff[0][jj])
          return 1;
        if(CURR_MODE->zone->relop[0][ii] > CURR_MODE->zone->relop[0][jj])
          return -1;
        if(CURR_MODE->zone->relop[0][ii] < CURR_MODE->zone->relop[0][jj])
          return 1;
      }
    }

/* UNFINISHED
    for(k=1; k<=composed_ta; k++) {
      if( (CURR_MODE->zone->relop[*i][k] != -1) ||
          (CURR_MODE->zone->relop[k][*i] != -1) ||
          (CURR_MODE->zone->relop[*j][k] != -1) ||
          (CURR_MODE->zone->relop[k][*j] != -1)
        ) {
        found = TRUE;
        break;
      }
    }
    if(found == FALSE) {
      ovlp = overlap(
    }
*/

  }
  if((*i) < (*j)) return -1;
  else if((*i) > (*j)) return 1;
  else return 0;
}

int sort_clocks(struct zone_type *z) {
  int i, j;

  for(i=1; i<=system_ta_num; i++) {
    old_order[i] = ordered_clocks[i];
  }

/* to move the first equality to the end of the array */
/*
  for(i=1; i<=system_ta_num; i++) {
    for(j=1; j<=system_ta_num; j++) {
      if(i==j) continue;
      if( (z->relop[i][j] == EQ) && (z->diff[i][j] == 0) ) {
        if(i < j) move_after(j, ordered_clocks[system_ta_num]);
        else move_after(i, ordered_clocks[system_ta_num]);
        i = -1;
        break;
      }
    }
    if(i == -1) break;
  }
*/

  for(i=1; i<=system_ta_num; i++) {
    for(j=1; j<=system_ta_num; j++) {
      if(i==j) continue;
      if( (z->relop[i][j] == EQ) && (z->diff[i][j] == 0) ) {
        if(i < j) 
          move_before_sorted(i, j, z);
        else move_before_sorted(j, i, z);
      }
    }
  }

/*
  if(verbose >= 0) {
    output("After processing EQ: "); 
    print_ordered_clocks();
  }
*/

  for(i=1; i<=system_ta_num; i++) {
    for(j=1; j<=system_ta_num; j++) {
      if(i==j) continue;
      if( (z->relop[i][j] == LESS && z->diff[i][j] <= 0) ||
          (z->relop[i][j] == LEQ && z->diff[i][j] < 0) 
        ) {
        move_before_sorted(i, j, z);
      }
    }
  }
  if(different_order(old_order, ordered_clocks)) {
    sort_clocks(z);
  }

/*
  printf("After sort_clocks.\n");
  print_ordered_clocks();
*/

  return 0;
}

int sort_clocks_2(struct zone_type *z) {
  int i, j;

  for(i=1; i<=system_ta_num; i++) {
    old_order[i] = ordered_clocks[i];
  }

  for(i=1; i<=system_ta_num; i++) {
    for(j=1; j<=system_ta_num; j++) {
      if( (i == j)                 ||
          (i == ordered_clocks[1]) || 
          (j == ordered_clocks[1]) ) 
        continue;
      if( (z->relop[i][j] == EQ) && (z->diff[i][j] == 0) ) {
        if(i < j) 
          move_before_sorted(i, j, z);
        else move_before_sorted(j, i, z);
      }
    }
  }
  for(i=1; i<=system_ta_num; i++) {
    for(j=1; j<=system_ta_num; j++) {
      if( (i == j)                 ||
          (i == ordered_clocks[1]) || 
          (j == ordered_clocks[1]) ) 
        continue;
      if( (z->relop[i][j] == LESS && z->diff[i][j] <= 0) ||
          (z->relop[i][j] == LEQ && z->diff[i][j] < 0) 
        ) {
        move_before_sorted(i, j, z);
      }
    }
  }
  if(different_order(old_order, ordered_clocks)) {
    sort_clocks(z);
  }

/*
  printf("After sort_clocks.\n");
  print_ordered_clocks();
*/

  return 0;
}



/* return TRUE if the two orders are different */

int different_order(int *order1, int *order2) {
  int i;

  if(order1 && order2) {
    for(i=1; i<=system_ta_num; i++) {
      if(order1[i] != order2[i]) return TRUE;
    }
  }
  return FALSE;
}


/* assumption: p < q for op == EQ
 * EQ: move p into the partial-array of equal clocks sorted by index
 * LESS: move p into the partial-array of clocks sorted by variable order
 */

int move_before_sorted(int p, int q, struct zone_type *z) {
  int i, op, p_pos = 0, q_pos = 0;

  op = z->relop[p][q];

  for(i=1; i<=system_ta_num; i++) {
    if(ordered_clocks[i] == p) p_pos = i;
    if(ordered_clocks[i] == q) q_pos = i;
  }
  if(q_pos == p_pos) return 0;
  if(q_pos == 1) {
    for(i=p_pos; i>1; i--) {
      ordered_clocks[i] = ordered_clocks[i-1];
    }
    ordered_clocks[1] = p;
  }
  else
  for(i=q_pos-1; i>0; i--) {
    if(op == EQ) {
      if(z->relop[ordered_clocks[i]][q] == EQ && 
         z->diff[ordered_clocks[i]][q] == 0) {
        if(p < ordered_clocks[i]) continue;
        else if(p > ordered_clocks[i]) {
          move_after(p, ordered_clocks[i]);
          break;
        }
        else break;     /* p == ordered_clocks[i] */
      }
      else {
        move_after(p, ordered_clocks[i]);
        break;
      }
    }
    if(op == LESS || LEQ) {
      if(z->relop[ordered_clocks[i]][q] == EQ && 
         z->diff[ordered_clocks[i]][q] == 0) {
        continue;
      }
      if( (z->relop[ordered_clocks[i]][q] == LESS && 
           z->diff[ordered_clocks[i]][q] <= 0) ||
          (z->relop[ordered_clocks[i]][q] == LEQ && 
           z->diff[ordered_clocks[i]][q] < 0) )
        {
        if( (z->relop[p][ordered_clocks[i]] == LESS && 
             z->diff[p][ordered_clocks[i]] <= 0) ||
            (z->relop[p][ordered_clocks[i]] == LEQ && 
             z->diff[p][ordered_clocks[i]] < 0) )
          continue;
        else {
          move_after(p, ordered_clocks[i]);
          break;
        }
      }
      else {
        move_after(p, ordered_clocks[i]);
        break;
      }
    }
  }
  if(verbose >= 0) {
    sprintf(tmpstring, "After move_before_sorted(%d, %d, %d): ", p, q, op);
    output(tmpstring);
    print_ordered_clocks();
  }
  return 0;
}


/* delete j and insert it after k */

int move_after(int j, int k) {
  int i, j_pos = 0, k_pos = 0;
  for(i=1; i<=system_ta_num; i++) {
    if(ordered_clocks[i] == j) j_pos = i;
    if(ordered_clocks[i] == k) k_pos = i;
  }
  if( (j_pos == k_pos) || (j_pos == (k_pos + 1)) ) return 0;
  if(j_pos < k_pos) {
    for(i = j_pos; i < k_pos; i++) {
      ordered_clocks[i] = ordered_clocks[i+1];
    }
    ordered_clocks[k_pos] = j;
  }
  else if(j_pos > k_pos) {
    for(i = j_pos; i > k_pos+1; i--) {
      ordered_clocks[i] = ordered_clocks[i-1];
    }
    ordered_clocks[k_pos+1] = j;
  }
  if(verbose >= 0) {
    sprintf(tmpstring, "After moving %d after %d ", j, k);
    output(tmpstring);
    print_ordered_clocks();
  }
  return 0;
}

int perm_index(int m, int *perm) {
  int i;
/*
  printf("getting perm_index of %d.\n", m);
*/
  if(perm == NULL) {
    return m;
  }
  for(i=0; i<=system_ta_num; i++) {
    if(perm[i] == m) break;
  }
/*
  printf("got perm_index of %d as %d.\n", m, i);
*/
  return i;
}

int install_permutation(struct transition_type *t, struct mode_type *m,
                        struct mode_type *sm,
                        struct mode_perm_type *list, struct thread_type *th){
  int *tmp1 = NULL, *tmp2 = NULL;
  struct permutation_type *p;

  tmp1 = inverse_permutation(mode_perm(m, list, th));

/*
  sprintf(tmpstring, "Mode %s(%d) has inverse ", m->name, m->index);
  output(tmpstring);
  print_perm(tmp1);
*/

/*
  sprintf(tmpstring, "Trans (%d) has ", t->index);
  output(tmpstring);
  print_permutation(t->permutation);
*/

  if(t->permutation)
    tmp2 = compose_permutation(tmp1, t->permutation->perm);
  else tmp2 = copy_permutation(tmp2, tmp1);
  free(tmp1);

/*
  output("Tr+Mode has ");
  print_perm(tmp2);
*/

  tmp1 = compose_permutation(tmp2, mode_perm(sm, list, th));
  free(tmp2);

/*
  output("Final ");
  print_perm(tmp1);
*/

  t->permutation = alloc_permutation();
  t->permutation->perm = tmp1;
  p = (struct permutation_type *)
                   add_23tree(t->permutation, perm_tree, permutation_comp,
                   free_perm, perm_count_inc, perm_parent_set, 
                   print_permutation);

  t->permutation = p;

  if(verbose >= 0) {
    sprintf(tmpstring, 
            "\tInstalled permutation on transition (%d): ", t->index);
    output(tmpstring);
    print_trans_permutation(t);
  }
  return 0;
}

/* Return the inverse permutation of perm */

int *inverse_permutation(int *perm) {
  int i, *new_perm;

  if(is_normal_perm(perm)) 
    return (int *)NULL;
  new_perm = (int *)malloc(sizeof(int) * (system_ta_num + 1));
  for(i=0; i<=system_ta_num; i++) {
    new_perm[perm[i]] = i;
  }
/*
  output("Inverse Permutation of "); print_perm(perm);
  output("is "); print_perm(new_perm);
*/
  return new_perm;
}

/* Return perm2(perm1()) */

int *compose_permutation(int *perm1, int *perm2) {
  int i, *new_perm;

  if(is_normal_perm(perm1) && is_normal_perm(perm2)) 
    return (int *)NULL;

  new_perm = (int *)malloc(sizeof(int) * (system_ta_num + 1));
  if(is_normal_perm(perm1)) {
    for(i=0; i<=system_ta_num; i++) {
      new_perm[i] = perm2[i];
    }
    return new_perm;
  }
  if(is_normal_perm(perm2)) {
    for(i=0; i<=system_ta_num; i++) {
      new_perm[i] = perm1[i];
    }
    return new_perm;
  }
  new_perm[0] = 0;
  for(i=1; i<=system_ta_num; i++) {
    new_perm[perm_index(perm_index(i, perm1), perm2)] = i;
  }
  return new_perm;
}

struct zone_type *permute_zone(struct zone_type *z, int *perm, 
                               struct thread_type *t) {
  int i, j;
  struct zone_type *pz = alloc_zone(t);

/*
  printf("Before permute_zone.\n");
  print_clock_indexes();
*/

  if(z->clk_index == NULL) {
    pz->relop[0][0] = 1;
    pz->diff[0][0] = 0;
    return pz;
  }

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(z->relop[i][j] != -1) {
/*
        pz->relop[new_index(perm_index(old_index(i, z), perm), z)]
                 [new_index(perm_index(old_index(j, z), perm), z)] = 
              z->relop[i][j];
        pz->diff[new_index(perm_index(old_index(i, z), perm), z)]
                [new_index(perm_index(old_index(j, z), perm), z)] = 
              z->diff[i][j];
*/
        pz->relop[i][j] = z->relop[i][j];
        pz->diff[i][j] = z->diff[i][j];
      }
    }
  }
/*
  printf("After permute_zone.\n");
  print_clock_indexes();
*/
  pz->clk_index[0] = z->clk_index[0];
  for(i=1; i<=z->clk_index[0]; i++) {
    pz->clk_index[i] = perm_index(z->clk_index[i], perm);
  }
  return pz;
}

int print_ordered_clocks() {
  int i;
  output("permutation: ( ");
  for(i=1; i<=system_ta_num; i++) {
    sprintf(tmpstring, "%d ", ordered_clocks[i]);
    output(tmpstring);
  }
  output(")\n");
  return 0;
}

struct zone_type *perm_kclosure_zone(struct zone_type *z) {
  int i, j, k;

  for(i=1; i<=z->clk_index[0]; i++) {
    for(j=1; j<=z->clk_index[0]; j++) {
      for(k=0; k<=z->clk_index[0]; k++) {
        if(z->relop[i][j]!=-1 && z->relop[i][k]!=-1 && z->relop[k][j]!=-1) {
          if(z->diff[i][j] == 0 && z->diff[i][k] == 0 && z->diff[k][j] == 0) {
            z->relop[i][j] = transitive_op(z->relop[i][k], z->relop[k][j]);
          }
        }
      }
    }
  }
  return z;
}

struct mode_type *create_new_normal_mode(struct mode_type *m, 
                                         struct thread_type *t, 
                                         struct mode_type **lastmode) {
  struct mode_type *nm;
  nm = alloc_mode();
  nm->name = permute_mode_name(m, ordered_clocks, &(nm->mode_vec));
  nm->dvl = permute_boolconj(m->dvl, ordered_clocks);
  nm->cvl = permute_clkconj(m->cvl, ordered_clocks);
  nm->zone = permute_zone(m->zone, ordered_clocks, t);
  nm->dval_bdd = permute_bdd(m->dval_bdd, ordered_clocks);
  nm->boolexp = alloc_boolexp(); 
  nm->boolexp->boolexp_bdd = 
                           permute_bdd(m->boolexp->boolexp_bdd, ordered_clocks);
  nm->xconj_cycle = zone2clkconj(nm->zone);
  nm->shielded_clocks = 
       sort_var_links(permuted_clocks(m->shielded_clocks, ordered_clocks));

  (*lastmode)->cnext = nm;
  nm->cnext = NULL;
  (*lastmode) = nm;
  t->mode_count++;
  if(verbose >= 0) {
     sprintf(tmpstring, 
             "Mode %s (%d): normal = %s (%d) [CREATED], ", 
             m->name, m->index, nm->name, nm->index);
     output(tmpstring);
  }
  if(verbose >= 3) {
    output("  inv bdd: "); 
    print_bdd(nm->boolexp->boolexp_bdd); 
    output("\n");
    output("  dval bdd: "); 
    print_bdd(nm->dval_bdd); 
    output("\n");
    output("  zone: "); 
    print_zone(nm->zone); 
    output("\n");
  }
  return nm;
}

struct mode_perm_type *
create_normal_mode_list_using_sort(struct thread_type *t) {
  struct mode_perm_type *ml, *list = NULL, *list_tail = NULL;
  struct mode_type *m, *nm, *lastmode, *orig_last = NULL;
  int new_mode = FALSE;

  if(verbose >= 0) {
    outputline();
    sprintf(tmpstring, "Creating normal mode list for thread %s[%d].\n", 
                       t->name, t->index);
    output(tmpstring);
  }

  for(m=t->starting_mode; m; m=m->cnext) {
    if(m->cnext == NULL)
      lastmode = m;
      orig_last = m;
  }

  for(m=t->starting_mode; m; m=m->cnext) {
    CURR_MODE = m;
    init_order();
    init_sort_array();
/*
    qsort(&ordered_clocks[1], system_ta_num, sizeof(int), normal_index_compare);
*/
    qsort(sort_array, composed_ta, sizeof(int), normal_index_compare);

    create_order();

    if(verbose >= 0) {
      sprintf(tmpstring, "For mode %s(%d), ", m->name, m->index);
      output(tmpstring);
      print_ordered_clocks();
    }
    ml = (struct mode_perm_type *)malloc(sizeof(struct mode_perm_type));
    if(new_mode || is_normal_perm(ordered_clocks)) {
      ml->perm = ordered_clocks;
      if(verbose >= 0) {
        sprintf(tmpstring, 
                "Mode %s (%d): normal = %s (%d) [SELF], ", 
                 m->name, m->index, m->name, m->index);
        output(tmpstring);
      }
      ml->mode = m;
      ml->next = NULL;
      if(list_tail) {
        list_tail->next = ml;
        list_tail = ml;
      }
      else {
        list = list_tail = ml;
      }
      ml = NULL;
    }
    else {
      ml->perm = ordered_clocks;
      nm = find_normal_mode(m, t);
      if(nm == NULL) {
        nm = create_new_normal_mode(m, t, &lastmode);
      }
      else {
        if(verbose >= 0) {
          sprintf(tmpstring, "Mode %s (%d): normal = %s (%d) [FOUND], ",
                  m->name, m->index, nm->name, nm->index);
          output(tmpstring);
        }
      }
      ml->mode = nm;
      ml->next = NULL;
      if(list_tail) {
        list_tail->next = ml;
        list_tail = ml;
      }
      else {
        list = list_tail = ml;
      }
      ml = NULL;
    }
    if(verbose >= 0) {
      print_ordered_clocks();
    }
    ordered_clocks = NULL;
    if(m == orig_last) new_mode = TRUE;
  }
  if(verbose >= 0) {
    sprintf(tmpstring, 
            "Finished creating normal mode list for graph %s[%d].\n", 
            t->name, t->index);
    output(tmpstring); 
    outputline();
  }

  if(check_normal_mode_list(list, t) == FALSE) return NULL;

  return list;
}

struct mode_type *find_in_list(struct mode_type *m, 
                               struct mode_perm_type *list) {
  struct mode_perm_type *mp;
  if(list) {
    for(mp = list; mp; mp = mp->next) {
      if(mp->mode->index == m->index) return mp->mode;
    }
  }
  return (struct mode_type *)NULL;
}

bdd permute_bdd(bdd b, int *perm) {
  struct variable_type *v = NULL;
  bdd f = b;
  struct boolconj_type *c = NULL;
  struct assign_type *a = NULL;
  int i;

  if(b == BDD_TRUE) return b;

  if(var_tail) {
    for(v = var_tail->next; ; v=v->next) {
      if(v->type == V_REGISTER) {
          for(i=1; i<=system_ta_num; i++) {
            c = alloc_boolconj();
            c->var1 = v->index;
            c->op = EQ;
            c->constant = i;
            install_bdd(c);
            if(bdd_and(bddm, b, c->prop_bdd) != BDD_FALSE) {
              a = alloc_assign();
              if(v->par_val) {
                a->var = v->index;
                a->value = perm_index(c->constant, perm);
              }
              else if(v->id >= 0) {
                a->var = permute_variable(v->index, perm);
                a->value = c->constant;
              }
              else if(v->id == NON_PARAM) {
                a->var = v->index;
                a->value = c->constant;
              }
              a->next = a;
              f = bdd_assign(f, a);
            }
  /*
            bdd_free(bddm, c->prop_bdd);
  */
            free(c); c = NULL;
            free(a); a = NULL;
          }
      }
      if(v == var_tail) break;
    }
    return f;
  }
  return b;
}

struct boolconj_type *permute_boolconj(struct boolconj_type *b, int *perm) {
  struct boolconj_type *nb = NULL, *bb;
  struct variable_type *v;

/*
  output("Permuting DVL: ");
  print_dvl(b);
*/
  if(b) {
    nb = boolconj_copy(b);
    for(bb = nb->next; ; bb=bb->next) {
      v = var_ptr_from_index(bb->var1);
      if(v->par_val)
        bb->constant = perm_index(bb->constant, perm);
      else if(v->id >= 1) {
        bb->var1 = var_index(var_name(bb->var1), perm_index(v->id, perm));
      }
      if(bb == nb) break;
    }
  }
  install_dvl_bdd(nb);
/*
  output("Permuted DVL: ");
  print_dvl(b);
*/
  return nb;
}

struct boolconj_type *permute_boolconj1(struct boolconj_type *b, int *perm) {
  struct boolconj_type *bb=NULL;
  struct variable_type *v;

/*
  output("Permuting 1 DVL: ");
  print_dvl(b);
*/
  if(b) {
    bb = boolconj_copy1(b);
    v = var_ptr_from_index(bb->var1);
    if(v->par_val)
      bb->constant = perm_index(bb->constant, perm);
    else if(v->id >= 1)
      bb->var1 = var_index(var_name(bb->var1), perm_index(v->id, perm));
    install_dvl_bdd(bb);
  }
/*
  output("Permuted 1 DVL: ");
  print_dvl(b);
*/
  return bb;
}

/* Check if there exists inconsistencies in the normal mode list created */

int check_normal_mode_list(struct mode_perm_type *list, struct thread_type *t) {
  struct mode_type *m, *nm, *nnm;

  for(m = t->starting_mode; m; m=m->cnext) {
    nm = normal_mode(m, list, t);
    nnm = normal_mode(nm, list, t);
    if(m->index != nm->index && 
       nnm->index != nm->index && 
       nnm->index == m->index) {
      fprintf(stderr, 
              "Error in the created normal mode list, not normalizing!\n");
      return FALSE;
/*
      change_normal_mode(nm, list, t);
*/
    }
  }
  return TRUE;
}

int change_normal_mode(struct mode_type *nm, 
                       struct mode_perm_type *list, struct thread_type *t) {
  struct mode_type *mptr;
  struct mode_perm_type *lptr;

  for(lptr = list, mptr = t->starting_mode; lptr; 
      lptr = lptr->next, mptr = mptr->cnext) {
    if(mptr == nm) {
      lptr->mode = nm;
      free(lptr->perm);
      lptr->perm = NULL;
      break;
    }
  }
  return 0;
}


struct mode_type *normal_mode(struct mode_type *m, 
                              struct mode_perm_type *list,
                              struct thread_type *t){
  struct mode_type *mptr;
  struct mode_perm_type *lptr;

  for(lptr = list, mptr = t->starting_mode; lptr; 
      lptr = lptr->next, mptr = mptr->cnext) {
    if(mptr == m) return lptr->mode;
  }
  return NULL;
}

int *mode_perm(struct mode_type *m, struct mode_perm_type *list, 
               struct thread_type *t){
  struct mode_type *mptr;
  struct mode_perm_type *lptr;

  for(lptr = list, mptr = t->starting_mode; lptr; 
      lptr = lptr->next, mptr = mptr->cnext) {
    if(mptr->index == m->index) return lptr->perm;
  }
  return 0;
}

int normalize_regions(struct thread_type *t) {
  int elim, i;
  struct mode_type *m, *ptr;
  struct transition_type *tr, *str;
  struct mode_perm_type *normal_mode_list = (struct mode_perm_type *)NULL;
  char *c;

  if(t->mode_count == 0) {
    fprintf(stderr, "Thread %s[%d] contains no mode.\n",
                     t->name, t->automaton_index);
    return FALSE;
  }

  NEW_THREAD = t;
  t->mnpl |= NR;
  c = get_graph_name(t);
  printf("    (Symmetry-Reducing) Normalizing regions in graph %s.\n", c);
  fflush(stdout);
  if(verbose >= 0) {
    sprintf(tmpstring, 
              "\n(Symmetry-Reducing) Normalizing regions in graph %s.\n", c);
    output(tmpstring);
  }

/*
  printf("\t  Creating normal mode list for graph %s.\n", c);
  fflush(stdout);
*/

  normal_mode_list = create_normal_mode_list_using_sort(t);
  if(normal_mode_list == NULL) return 0;

/*
  printf("\t  Normal mode list created for graph %s.\n", c);
  printf("\t  Normalizing transitions for graph %s.\n", c);
  fflush(stdout);
*/

  if(verbose >= 0) {
    sprintf(tmpstring, "\nNormalizing transitions for graph %s.\n", c);
    output(tmpstring);
  }

  for(m=t->starting_mode,i=1; m; m=m->cnext,i++) {
    if(verbose >= 0) {
      sprintf(tmpstring, 
            "Considering parent mode %s (%d) with %d out transitions.\n",
             m->name, m->index, m->out_trans_count);
      output(tmpstring);
    }
    if(m->out_trans_cycle) {
      for(tr = m->out_trans_cycle->next; ;) {
        elim = compact_reduce(m, tr, normal_mode_list, t);
/*
        printf("(elim = %d) Here, mode %s (%d) has %d out trans.\n",
                elim, m->name, m->index, m->out_trans_count);
*/
        if(tr == m->out_trans_cycle) {
          if(elim) {
            if(verbose >= 0) {
              sprintf(tmpstring,
             "\tEliminating transition (%d) from %s (%d) to %s (%d).\n",
               tr->index, m->name, m->index, tr->mode->name, tr->mode->index);
              output(tmpstring);
            }
            eliminate_transition(tr, m);
          }
          break;
        }
        else {
          if(elim) {
            str = tr->next;
            if(verbose >= 0) {
              sprintf(tmpstring, 
               "\tEliminating transition (%d) from %s (%d) to %s (%d).\n",
               tr->index, m->name, m->index, tr->mode->name, tr->mode->index);
              output(tmpstring);
            }
            eliminate_transition(tr, m);
            tr = str;
          }
          else tr = tr->next;
        }
      }
/*
      printf("After compacting mode %s (%d), it now has %d out trans.\n",
              m->name, m->index, m->out_trans_count);
*/
    }
    if(stats == TRUE && SGM_MODE != GUI && SGM_MODE != PROCESS) {
      clear_line();
      fprintf(stderr, "    Normalized %d modes, Remaining %d modes (%.1f%%)\r",
                      i, t->mode_count-i,
              ((float)(t->mode_count-i) / (float)t->mode_count)*100.0);
    }
  }
  clean_line();

/*
  printf("\t  Transitions normalized for graph %s.\n", c);
  fflush(stdout);
*/

/*
  t->starting_mode = normalize_thread(t, normal_mode_list);
*/
  /* In case the starting_mode itself is not normal */
  m = normal_mode(t->starting_mode, normal_mode_list, t);
  if(t->starting_mode->index != m->index) {
    for(ptr = t->starting_mode; ptr; ptr=ptr->cnext) {
      if(ptr->cnext->index == m->index) break;
    }
    ptr->cnext = m->cnext;        /* Remove m from cnext list */
    m->cnext = t->starting_mode;  /* Prepend m to head of cnext list */
    t->starting_mode = m;         /* Reassign head of cnext list */
  }

  printf("    ");
  print_counts(t);

  printf("\tRemoving irreachable modes in graph %s.\n\t", c);
  fflush(stdout);
  if(verbose >= 0) {
    outputline();
    sprintf(tmpstring,
    "Removing irreachable modes in graph %s (%d modes, %d transitions).\n",
       c, t->mode_count, t->trans_count);
    output(tmpstring);
  }
  remove_irreachable_mode(t);
/*
  if(verbose >= 0) {
    sprintf(tmpstring, "DONE removing irreachable modes in graph %s.\n",c);
    output(tmpstring);
  }
*/
  remove_dormant_cycles(t);
  if(verbose >= 0)
    print_thread(t);
/*
  printf("\tRemoving non-normal modes in graph %s.\n", c);
  sprintf(tmpstring, "Removing non-normal modes in graph %s.\n", c);
  output(tmpstring);
  remove_non_normal_mode(t, normal_mode_list);
  sprintf(tmpstring, "DONE removing non-normal modes in graph %s.\n", c);
  output(tmpstring);
  print_counts(t);
*/

  if(sibling_trans_multi_reduction) {
    printf("\tReducing transitions in (multi)graph %s.\n", c);
    fflush(stdout);
    reduce_multigraph(t);
    if(verbose >= 0) {
      sprintf(tmpstring, "DONE reducing transitions in (multi)graph %s.\n", c);
      output(tmpstring);
    }
    if(SGM_MODE != GUI && SGM_MODE != PROCESS) {
      fprintf(stderr, "\t");
      print_counts(t);
    }
  }
  free_normal_mode_list(normal_mode_list);
  free(c);
  return 0;
}

int free_normal_mode_list(struct mode_perm_type *nml) {
  struct mode_perm_type *mp;
  if(nml) {
    mp = nml;
    while(mp) {
      nml = nml->next;
      free_mode_perm(mp);
      mp = nml;
    }
  }
  return 0;
}

int free_mode_perm(struct mode_perm_type *mp) {
  free(mp->perm);
  free(mp);
  return 0;
}

int compact_reduce(struct mode_type *m, struct transition_type *tr, 
                   struct mode_perm_type *list, struct thread_type *t) {
  struct transition_type *ntr;
  struct mode_type *npm = NULL, *nsm = NULL;

  if(verbose >= 0) {
    sprintf(tmpstring, 
          "    Considering transition (%d) to mode %s (%d):\n", 
          tr->index, tr->mode->name, tr->mode->index);
    output(tmpstring);
  }
/*
  print_zone(tr->mode->zone);
  output("\n");
*/

  npm = normal_mode(m, list, t);
  if(verbose >= 0) {
    sprintf(tmpstring, "\tNormal mode of %s (%d) is %s (%d).\n", 
                        m->name, m->index, npm->name, npm->index);
    output(tmpstring);
  }
  nsm = normal_mode(tr->mode, list, t);
  if(verbose >= 0) {
    sprintf(tmpstring, "\tNormal mode of %s (%d) is %s (%d).\n", 
                        tr->mode->name, tr->mode->index, nsm->name, nsm->index);
    output(tmpstring);
  }

  if(npm == m && nsm == tr->mode) {
    if(verbose >= 0) {
      output("\tBoth source and target modes are normal.\n");
    }
    free(ordered_clocks);
    return FALSE;
  }
  else {
    ntr = add_transition_fifo(tr, npm, nsm);
/*
    c = permute_mode_name(tr->mode, mode_perm(tr->mode, list, t), NULL);
    ntr->goto_mode = c;
*/
    if(verbose >= 0) {
      sprintf(tmpstring, 
          "\tAdded transition (%d) between %s (%d) and %s (%d).\n",
             ntr->index, npm->name, npm->index, nsm->name, nsm->index);
      output(tmpstring);
      output("\t");
      print_trans(ntr);
    }
/*
    print_perm(mode_perm(m, list, t));
*/
    if(is_normal_perm(mode_perm(m, list, t)) == FALSE) {
      if(ntr->automaton == NULL) {
        fprintf(stderr, 
        "Error: Synchronization with Symmetry Reduction not implemented. Exiting!\n");
        exit(1);
      }
      permute_transition(ntr, mode_perm(m, list, t));
    }
    install_permutation(ntr, m, tr->mode, list, t);
    install_in_permutation(ntr, npm);
  }
  return TRUE;
}

/* Point the permutation link of the in-coming trans corresponding to
   the out-going trans "ntr" from mode "npm" to mode "ntr->mode" */

int install_in_permutation(struct transition_type *ntr, struct mode_type *npm) {
  struct transition_type *e;
  for(e=ntr->mode->in_trans_cycle->next; ; e=e->next) {
    if(e->index == ntr->index) {
      e->permutation = ntr->permutation;
      break;
    }
    if(e == ntr->mode->in_trans_cycle) break;
  }
  return 0;
}

int is_normal_zone() {
  int i;
  for(i=1; i<=system_ta_num; i++) {
    if(ordered_clocks[i] != i) return FALSE;
  }
  return TRUE;
}

int is_normal_bdd(bdd f) {
  struct proposition_type *p;
  for(p = prop_tail->next; ; p = p->next) {
    if(BDD_AND(p->prop_bdd, f) != BDD_FALSE) {
      if(p->value == automaton_tail->next->index) return TRUE;
      else if(p->value == 0) return -1;
      else return FALSE;
    }
    if(p == prop_tail) break;
  }
  fprintf(stderr, "Error in reduce.c:is_normal_bdd().\n");
  exit(0);
}

int is_normal_mode(struct mode_type *m) {
  int comp = is_normal_bdd(m->dval_bdd);

  if(comp == FALSE) {
    ordered_clocks = get_perm_bdd(m->dval_bdd);
    ordered_clocks = get_perm_zone(m->zone);
    return FALSE;
  }
  else if(comp == TRUE) {
    init_order();
    ordered_clocks = get_perm_zone(m->zone);
    if(is_normal_perm(ordered_clocks)) return TRUE;
    else return FALSE;
  }
  else if(comp == -1) {
    init_order();
    sort_clocks(m->zone);
    return (is_normal_zone());
  }
  return FALSE;
}

int *get_perm_zone(struct zone_type *z){
/*
  output("Orig Zone: "); print_zone(z); output("\n");
  pz = permute_zone(z, ordered_clocks);
  output("Perm Zone: "); print_zone(pz); output("\n");
*/

  output("Before sort_clocks_2(): "); print_ordered_clocks();
  sort_clocks_2(z);
  output("After sort_clocks_2(): "); print_ordered_clocks();
/*
  for(i=0; i<=composed_ta; i++) sum[i] = 0;
  for(i = 1; i <= composed_ta; i++) {
    for(j = 1; j <= composed_ta; j++) { 
      if(z->relop[i][j] != -1 && z->diff[i][j] == 0) {
        sum[i]++;
      }
    }
  }
  output("Sum: [");
  for(i=1; i<=composed_ta; i++) {
    sprintf(tmpstring, "%d", sum[i]);
    output(tmpstring);
    if(i<composed_ta) output(", ");
  }
  output("]\n");
*/
  return ordered_clocks;
}

int *get_perm_bdd(bdd f) {
  int i, *perm = (int *)malloc(sizeof(int) * (system_ta_num+1));
  struct proposition_type *p;

  for(i=0; i<=system_ta_num; i++) {
    perm[i] = i;
  }
  for(p = prop_tail->next; ; p = p->next) {
    if(BDD_AND(p->prop_bdd, f) != BDD_FALSE) {
      perm[automaton_tail->next->index] = p->value;
      perm[p->value] = automaton_tail->next->index;
      break;
    }
    if(p == prop_tail) break;
  }
  return perm;
}

int is_normal_perm(int *p) {
  int i;
  if(p) {
    for(i=1; i<=system_ta_num; i++) {
      if(p[i] != i) return FALSE;
    }
  }
  return TRUE;
}

struct mode_type *find_normal_mode(struct mode_type *m, struct thread_type *t) {
  struct mode_type *nm;
  struct zone_type *pz;
  struct boolconj_type *nb;
  char *new_name = permute_mode_name(m, ordered_clocks, NULL);
  int comp;
  bdd pb2;

  if(verbose >= 3) {
    sprintf(tmpstring, 
            "\tFinding normal mode of mode %s(%d) in thread %s[%d].\n", 
            m->name, m->index, t->name, t->automaton_index);
    output(tmpstring);
  }
  if(is_normal_perm(ordered_clocks)) {
    output("Permutation is normal, normal mode is itself.\n");
    free(new_name);
    return m;
  }

/*
  output("original zone: "); print_zone(m->zone); output("\n");
*/
  pz = permute_zone(m->zone, ordered_clocks, t);

/*
  for(i=0; i<=system_ta_num; i++) {
    for(j=0; j<=system_ta_num; j++) {
      if(m->zone->relop[i][j] != -1) {
        pz->relop[perm_index(i, ordered_clocks)]
                 [perm_index(j, ordered_clocks)] = m->zone->relop[i][j];
        pz->diff[perm_index(i, ordered_clocks)]
                [perm_index(j, ordered_clocks)] = m->zone->diff[i][j];
      }
    }
  }
*/
/*
  output("permuted zone: "); print_zone(pz); output("\n");
*/

/*
  pb = permute_bdd(m->dval_bdd, ordered_clocks);
*/
  nb = permute_boolconj(m->dvl, ordered_clocks);
  pb2 = get_bdd_from_dvl(nb);

  for(nm=t->starting_mode; nm; nm=nm->cnext) {
    if(nm->index == m->index) continue;
    if(verbose >= 3) {
      sprintf(tmpstring, "\tComparing mode names: %s and %s (%d): ", 
              new_name, nm->name, nm->index);
      output(tmpstring);
    }
    if(strcmp(new_name, nm->name) == 0) {
      if(verbose >= 3) {
        output("same name.\n");
      }
    }
    else {
      if(verbose >= 3) {
        output("different names.\n");
      }
      continue;
    }

/*
    if(verbose >= 3) {
      output("\tComparing bdds: \n");
      output("\t"); print_bdd(pb); 
      output("\t"); print_bdd(nm->dval_bdd);
    }
    if(bdd_comp(pb, nm->dval_bdd) == EQUAL) {
      if(verbose >= 3) {
        output("\tequal bdd.\n");
      }
    }
    else {
      if(verbose >= 3) {
        output("\tdifferent bdds.\n");
      }
      continue;
    }
*/

    if(verbose >= 3) {
      output("\tComparing bdds: \n");
      output("\tpermuted bdd: "); print_bdd(pb2); 
      output("\ttest-mode bdd: "); print_bdd(dvl_bdd(nm));
    }
    if(bdd_comp(pb2, dvl_bdd(nm)) == EQUAL) {
      if(verbose >= 3) {
        output("\tequal dvl bdd.\n");
      }
    }
    else {
      if(verbose >= 3) {
        output("\tdifferent dvl bdds.\n");
      }
      continue;
    }

    if(verbose >= 3) {
      output("\tComparing zones: \n");
      print_zone(pz); 
      output("\nAND\n"); 
      print_zone(nm->zone); 
      output("\n");
    }
    comp = zone_comp(pz, nm->zone);
    if(boob == TRUE) boob = FALSE;
    if(comp == EQUAL) {   /* || comp == CONTAINED || comp == CONTAINS) { */
      if(verbose >= 3) {
        output("\tequal zone.\n");
      }
      if(verbose >= 3) {
        sprintf(tmpstring, 
                "\tFound normal mode %s(%d) of mode %s(%d).\n", 
                nm->name, nm->index, m->name, m->index);
        output(tmpstring);
      }
      free_zone(pz);
      free_boolconj(nb);
      free(new_name);
      return nm;
    }
  }
  /* normal mode not found */
  free_zone(pz);
  free_boolconj(nb);
  free(new_name);
  if(verbose >= 3) {
    sprintf(tmpstring, 
            "Normal mode of mode %s(%d) not found in thread %s[%d].\n",
                       m->name, m->index, t->name, t->automaton_index);
    output(tmpstring);
  }
  return (struct mode_type *)NULL;
}

int remove_non_normal_mode(struct thread_type *t, struct mode_perm_type *list) {
struct mode_type *mp = t->starting_mode;
struct mode_type *m = t->starting_mode->cnext;
struct mode_type *ms = m->cnext;
int again = FALSE;

/*
  printf("Current counts: %d modes, %d transitions, %d invalid trans.\n", 
             t->mode_count, t->trans_count, t->invalid_trans_count);
*/
  for(; ; m=ms, ms=ms->cnext) {
    init_order();
    if(m->zone == NULL) 
      printf("Zone of mode %s (%d) is NULL.\n", m->name, m->index);
    sort_clocks(m->zone);
    if( (is_normal_zone() == FALSE) /*|| 
        (is_normal_bdd(m->dval_bdd) == FALSE)*/ ) {
      if(verbose >= 0) {
        sprintf(tmpstring, "Freeing non-normal mode %d(%s) ...\n ", 
                m->index, m->name);
        output(tmpstring);
      }
/*
      i = 0;
      if(m->out_trans_count) {
        for(tr = m->out_trans_cycle->next; ;tr=tr->next) {
          if(tr->valid == FALSE) i++;
          if(tr == m->out_trans_cycle) break;
        }
      }
*/
      t->mode_count--;
      t->trans_count -= m->out_trans_count;
      t->trans_count -= m->in_trans_count;
/*
      t->invalid_trans_count -= i;
*/

      free_mode(m);
/*
      printf("Current counts: %d modes, %d transitions, %d invalid trans.\n", 
             t->mode_count, t->trans_count, t->invalid_trans_count);
*/
      mp->cnext = ms;
      output(" ... done.\n");
      if(again == FALSE) again = TRUE;
    }
    else mp = m;
    free(ordered_clocks);
    if(ms == NULL) break;
  }
  if(again) remove_non_normal_mode(t, list);
  return 0;
}

/*
int iso_on_perm(struct transition_type *tr, struct transition_type *trr) {
  struct transition_type *t = NULL;
  int ans;

  if(verbose >= 1) {
    sprintf(tmpstring, 
           "\tComparing transitions %d and %d in iso_on_perm:\n", 
            tr->index, trr->index);
    output(tmpstring);
  }
  if(verbose >= 3) {
    output("\t");
    print_trans(tr);
    output("\t");
    print_trans(trr);
  }

  if(trr->permutation) {
    t = copy_transition(trr);
    if(t->permutation)
      permute_transition(t, t->permutation->perm);
    if(tr->trn->index == t->trn->index) ans = TRUE;
    else ans = FALSE;
  }
  else if(tr->permutation) {
    t = copy_transition(tr);
    if(t->permutation)
      permute_transition(t, t->permutation->perm);
    if(trr->trn->index == t->trn->index) ans = TRUE;
    else ans = FALSE;
  }
  else {
    if(tr->trn->index == trr->trn->index) ans = TRUE;
    else ans = FALSE;
  }
  free_transition(t);
  return ans;
}
*/

int assignment_compare(struct assign_type *a1, struct assign_type *a2) {
  int found = FALSE;
  struct assign_type *aa1, *aa2;
  struct assign_type *as1, *as2;

  if(a1 && !a2) return -1;
  if(!a1 && a2) return 1;

  if(a1 && a2) {
    for(aa1 = a1->next; ; aa1 = aa1->next) {
      for(aa2 = a2->next; ; aa2 = aa2->next) {
        if(aa1->var == aa2->var && aa1->value == aa2->value) {
          found = TRUE;
          break;
        }
        if(aa2 == a2) break;
      }
      if(found == FALSE) {
        for(as1 = a1->next, as2 = a2->next; ; as1 = as1->next, as2=as2->next) {
          if(as1->var < as2->var) return -1;
          else if(as1->var > as2->var) return 1;
          else {
            if(as1->value < as2->value) return -1;
            else if(as1->value > as2->value) return 1;
          }
          if(as1 == a1) return -1;
          else if(as2 == a2) return 1;
        }
      }
      else found = FALSE;
      if(aa1 == a1) break;
    }
    for(aa2 = a2->next; ; aa2 = aa2->next) {
      for(aa1 = a1->next; ; aa1 = aa1->next) {
        if(aa1->var == aa2->var && aa1->value == aa2->value) {
          found = TRUE;
          break;
        }
        if(aa1 == a1) break;
      }
      if(found == FALSE) {
        for(as1 = a1->next, as2 = a2->next; ; as1 = as1->next, as2=as2->next) {
          if(as1->var < as2->var) return -1;
          else if(as1->var > as2->var) return 1;
          else {
            if(as1->value < as2->value) return -1;
            else if(as1->value > as2->value) return 1;
          }
          if(as1 == a1) return -1;
          else if(as2 == a2) return 1;
        }
      }
      else found = FALSE;
      if(aa2 == a2) break;
    }
    return EQUAL;
  }
  return EQUAL;
}

int prp_trigger_compare(struct boolconj_type *p1, struct boolconj_type *p2) {
  int found = FALSE;
  struct boolconj_type *pp1, *pp2;
  struct boolconj_type *ps1, *ps2;

  if(!p1 && !p2) return EQUAL;
  if(p1 && !p2) {
    if(p1->op == TRUE) return EQUAL;
    else return -1;
  }
  if(!p1 && p2) {
    if(p2->op == TRUE) return EQUAL;
    return 1;
  }
 
  if(p1 && p2) {
    if(p1->op == TRUE && p2->op == TRUE) return EQUAL;
    if(p1->op == FALSE && p2->op == FALSE) return EQUAL;
    if(p1->op == TRUE && p2->op == FALSE) return -1;
    if(p1->op == FALSE && p2->op == TRUE) return 1;

    for(pp1 = p1->next; ; pp1 = pp1->next) {
      for(pp2 = p2->next; ; pp2 = pp2->next) {
        if(pp1->op == pp2->op && 
           pp1->var1 == pp2->var1 && 
           pp1->constant == pp2->constant) {
           found = TRUE;
           break;
        }
        if(pp2 == p2) break;
      }
      if(found == FALSE) {
        for(ps1 = p1->next, ps2 = p2->next; ; ps1 = ps1->next, ps2=ps2->next) {
          if(ps1->var1 < ps2->var1) return -1;
          if(ps1->var1 > ps2->var1) return 1;
          if(ps1->op < ps2->op) return -1;
          if(ps1->op > ps2->op) return 1;
          if(ps1->constant < ps2->constant) return -1;
          if(ps1->constant > ps2->constant) return 1;
          if(ps1 == p1) return -1;
          else if(ps2 == p2) return 1;
        }
      }
      else found = FALSE;
      if(pp1 == p1) break;
    }
    for(pp2 = p2->next; ; pp2 = pp2->next) {
      for(pp1 = p1->next; ; pp1 = pp1->next) {
        if(pp1->op == pp2->op && 
           pp1->var1 == pp2->var1 && 
           pp1->constant == pp2->constant) {
           found = TRUE;
           break;
        }
        if(pp1 == p1) break;
      }
      if(found == FALSE) {
        for(ps1 = p1->next, ps2 = p2->next; ; ps1 = ps1->next, ps2=ps2->next) {
          if(ps1->var1 < ps2->var1) return -1;
          if(ps1->var1 > ps2->var1) return 1;
          if(ps1->op < ps2->op) return -1;
          if(ps1->op > ps2->op) return 1;
          if(ps1->constant < ps2->constant) return -1;
          if(ps1->constant > ps2->constant) return 1;
          if(ps1 == p1) return -1;
          else if(ps2 == p2) return 1;
        }
      }
      else found = FALSE;
      if(pp2 == p2) break;
    }
  }
  return EQUAL;
}

int clk_trigger_compare(struct clkconj_type *c1, struct clkconj_type *c2) {
  int found = FALSE;
  struct clkconj_type *cc1, *cc2;
  struct clkconj_type *cs1, *cs2;

/*
  output("\tComparing clock triggers:\n");
  print_clkconj(c1); output("\n\t"); print_clkconj(c2); output("\n");
*/

  if(!c1 && !c2) return EQUAL;
  if(c1 && !c2) return -1;
  if(!c1 && c2) return 1;

  if(c1 && c2) {
    if(c1->op == TRUE && c2->op == TRUE) return EQUAL;
    if(c1->op == FALSE && c2->op == FALSE) return EQUAL;
    if(c1->op == TRUE && c2->op == FALSE) return 1;
    if(c1->op == FALSE && c2->op == TRUE) return -1;

    for(cc1 = c1->next; ; cc1 = cc1->next) {
      for(cc2 = c2->next; ; cc2 = cc2->next) {
        if(cc1->clk1 == cc2->clk1 && cc1->clk2 == cc2->clk2 &&
           cc1->op == cc2->op && cc1->diff == cc2->diff) {
          found = TRUE;
          break;
        }
        if(cc2 == c2) break;
      }
      if(found == FALSE) {
        for(cs1 = c1->next, cs2 = c2->next; ; cs1 = cs1->next, cs2=cs2->next) {
          if(cs1->clk1 < cs2->clk1) return -1;
          if(cs1->clk1 > cs2->clk1) return 1;
          if(cs1->clk2 < cs2->clk2) return -1;
          if(cs1->clk2 > cs2->clk2) return 1;
          if(cs1->op < cs2->op) return -1;
          if(cs1->op > cs2->op) return 1;
          if(cs1->diff < cs2->diff) return -1;
          if(cs1->diff > cs2->diff) return 1;
          if(cs1 == c1) return -1;
          else if(cs2 == c2) return 1;
        }
      }
      else found = FALSE;
      if(cc1 == c1) break;
    }
    for(cc2 = c2->next; ; cc2 = cc2->next) {
      for(cc1 = c1->next; ; cc1 = cc1->next) {
        if(cc1->clk1 == cc2->clk1 && cc1->clk2 == cc2->clk2 &&
           cc1->op == cc2->op && cc1->diff == cc2->diff) {
          found = TRUE;
          break;
        }
        if(cc1 == c1) break;
      }
      if(found == FALSE) {
        for(cs1 = c1->next, cs2 = c2->next; ; cs1 = cs1->next, cs2=cs2->next) {
          if(cs1->clk1 < cs2->clk1) return -1;
          if(cs1->clk1 > cs2->clk1) return 1;
          if(cs1->clk2 < cs2->clk2) return -1;
          if(cs1->clk2 > cs2->clk2) return 1;
          if(cs1->op < cs2->op) return -1;
          if(cs1->op > cs2->op) return 1;
          if(cs1->diff < cs2->diff) return -1;
          if(cs1->diff > cs2->diff) return 1;
          if(cs1 == c1) return -1;
          else if(cs2 == c2) return 1;
        }
      }
      else found = FALSE;
      if(cc2 == c2) break;
    }
    return EQUAL;
  }
  return EQUAL;
}

/* Generate a permutation of the automaton_index of t1 and t2 */

int *gen_perm_multi(struct transition_type *t1, struct transition_type *t2) {
  int i;
  int *perm = (int *)malloc(sizeof(int) * (current_ta_num + 1));
  for(i=0; i<=current_ta_num; i++) {
    perm[i] = i;
  }
  perm[t1->automaton->index] = t2->automaton->index;
  perm[t2->automaton->index] = t1->automaton->index;
  return perm;
}

int same_permutation(struct transition_type *t1, struct transition_type *t2) {
  int i;
  if(is_normal_trans(t1)) {
    if(is_normal_trans(t2)) return TRUE;
    else return FALSE;
  }
  if(is_normal_trans(t2)) {
    if(is_normal_trans(t1)) return TRUE;
    else return FALSE;
  }
  for(i=0; i<=system_ta_num; i++) {
    if(t1->permutation->perm[i] != t2->permutation->perm[i]) return FALSE;
  }
  return TRUE;
}

int reduce_multigraph(struct thread_type *t) {
  struct mode_type *m;
  struct transition_type *tr, *trr, *trrn;
  struct zone_type *zz;
  int inv, i, j, k, *elim, count, comp, *tmp_perm;
  char *perm_name;
  struct var_link_type *perm_sc;
  struct boolconj_type *perm_dvl;

  if(verbose >= 0) {
    outputline();
    sprintf(tmpstring, "Reducing multigraph of thread %s[%d].\n", 
                       t->name, t->automaton_index);
    output(tmpstring);
  }

  for(m = t->starting_mode,k=1; m; m=m->cnext,k++) {
    if(verbose >= 0) {
      sprintf(tmpstring, "Considering mode %s(%d) with %d out transitions.\n", 
              m->name, m->index, m->out_trans_count);
      output(tmpstring);
    }

    if(m->out_trans_cycle) {
      elim = (int *)malloc(sizeof(int) * (m->out_trans_count));
      for(i=0; i < m->out_trans_count; i++) {
        elim[i] = FALSE;
      }
/*
      printf("elim = ( ");
      for(i=0; i < m->out_trans_count; i++) {
        printf("%d ", elim[i]);
      }
      printf(")\n");
*/

      for(j = 0, tr = m->out_trans_cycle->next; ; j++, tr = tr->next) {
        if(verbose >= 1) {
          sprintf(tmpstring, 
                "    Considering transition (%d) to %s (%d), ", 
                tr->index, tr->mode->name, tr->mode->index);
          output(tmpstring);
          print_trans_permutation(tr);
        }
        if(elim[j] == TRUE) {
          if(tr == m->out_trans_cycle) break;
          else continue;
        }
        for(i = 0, trr = m->out_trans_cycle->next; ; i++, trr = trr->next) {
/*
            printf("(i = %d) checking transition (%d).\n", i, trr->index);
*/
            if(elim[i] == TRUE) {
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
            if(tr == trr) {
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
            if(trr->mode->index != tr->mode->index) {
              if(verbose >= 1) {
                sprintf(tmpstring, 
        "Tr %d and tr %d have different destination modes %s(%d) and %s(%d).\n",
                  tr->index, trr->index, tr->mode->name, 
                  tr->mode->index, trr->mode->name, trr->mode->index);
                output(tmpstring);
              }
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
            if(tr->automaton && trr->automaton) {
              if(tr->automaton->index == trr->automaton->index) {
                if(tr->trn->index == trr->trn->index) {
                  if(verbose >= 0) {
                    sprintf(tmpstring, "Trans %d and %d are the same: ", 
                            tr->index, trr->index);
                    output(tmpstring);
                  }
                  if(j < i) {
                    elim[i] = TRUE;
                    if(verbose >= 0) {
                      sprintf(tmpstring, "eliminating %d.\n", trr->index);
                      output(tmpstring);
                    }
                    if(trr == m->out_trans_cycle) break;
                    else continue;
                  }
                  else if(j > i) {
                    elim[j] = TRUE;
                    if(verbose >= 0) {
                      sprintf(tmpstring, "eliminating %d.\n", tr->index);
                      output(tmpstring);
                    }
                    break;
                  }
                }
              } 
            }
            if(tr->automaton == NULL && trr->automaton == NULL) {
              if(spec_trans_comp(tr->trn, trr->trn) == EQUAL) {
                if(verbose >= 0) {
                  sprintf(tmpstring, "Sync trans %d and %d are the same: ", 
                          tr->index, trr->index);
                  output(tmpstring);
                }
                if(j < i) {
                  elim[i] = TRUE;
                  if(verbose >= 0) {
                    sprintf(tmpstring, "eliminating %d.\n", trr->index);
                    output(tmpstring);
                  }
                  if(trr == m->out_trans_cycle) break;
                  else continue;
                }
                else if(j > i) {
                  elim[j] = TRUE;
                  if(verbose >= 0) {
                    sprintf(tmpstring, "eliminating %d.\n", tr->index);
                    output(tmpstring);
                  }
                  break;
                }
              }
            }
/*
            if(iso_on_perm(tr, trr) == FALSE) {
              if(verbose >= 1)
                output("\t  Not isomorphic after transition permutation.\n");
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
*/
            if(corr_trans(tr, trr)) {
              if(verbose >= 0) {
                if(tr->automaton->index >= 0) {
                  sprintf(tmpstring, "Trans %d of %s[%d] ",
                    tr->index, tr->automaton->name, tr->automaton->index);
                  output(tmpstring);
                }
                else {
                  sprintf(tmpstring, "Trans %d of %s ",
                    tr->index, tr->automaton->name);
                  output(tmpstring);
                }
                if(trr->automaton->index >= 0) {
                  sprintf(tmpstring, "and %d of %s[%d] are corresponding.\n", 
                    trr->index, trr->automaton->name, trr->automaton->index);
                  output(tmpstring);
                }
                else {
                  sprintf(tmpstring, "and %d of %s are corresponding.\n",
                    trr->index, trr->automaton->name);
                  output(tmpstring);
                }
              }

              if(EOT) {    /* Leave only a representative of a set of 
                            * orthogonal (or corresponding) transitions */
                if(is_normal_trans(tr)) {
                  elim[i] = TRUE;
                  if(verbose >= 0) {
                    sprintf(tmpstring, "Eliminating trans %d.\n", trr->index);
                    output(tmpstring);
                  }
                  if(trr == m->out_trans_cycle) break;
                  else continue;
                }
                else if(is_normal_trans(trr)) { 
                  elim[j] = TRUE;
                  if(verbose >= 0) {
                    sprintf(tmpstring, "Eliminating trans %d.\n", tr->index);
                    output(tmpstring);
                  }
                  break;
                }
              }

              tmp_perm = gen_perm_multi(tr, trr);

              if(verbose >= 3) {
                sprintf(tmpstring, "Permuting process indices: ");
                output(tmpstring);
                print_perm(tmp_perm);
                output("\n");
              }
              /* (a) check mode vectors */
              perm_name = permute_mode_name(m, tmp_perm, NULL);
              if(strcmp(perm_name, m->name)) {
                free(perm_name); perm_name = NULL;
                free(tmp_perm); tmp_perm = NULL;
                if(trr == m->out_trans_cycle) break;
                else continue;
              }
              free(perm_name); perm_name = NULL;
              if(verbose >= 3) 
                output(
                "\tEqual mode vectors after permuting with process indices.\n");

              /* (b) check zones */
              zz = permute_zone(m->zone, tmp_perm, t);
              if(verbose >= 3) {
                output("Parent zone: ");
                print_zone(m->zone);
                output("\n");
                output("Permuted zone: ");
                print_zone(zz);
                output("\n");
              }
              if(zone_comp(m->zone, zz) != EQUAL) {
                if(verbose >= 1) {
                  output(
                   "\t  Unequal zones after permuting with process indices.\n");
                }
                if(boob == TRUE) boob = FALSE;
                                   /*boob is set in zone_comp(), reset it */
                free_zone(zz); zz = NULL;
                free(tmp_perm); tmp_perm = NULL;
                if(trr == m->out_trans_cycle) break;
                else continue;
              }
              if(boob == TRUE) boob = FALSE;
              free_zone(zz); zz = NULL;
              if(verbose >= 3) 
                output("\tEqual zones after permuting with process indices.\n");

              /* (c) check literal sets */
              perm_dvl = permute_boolconj(m->dvl, tmp_perm);
              if(dvl_compare(m->dvl, perm_dvl) != EQUAL) {
                free_boolconj(perm_dvl);
                perm_dvl = NULL;
                free(tmp_perm); tmp_perm = NULL;
                if(trr == m->out_trans_cycle) break;
                else continue;
              }
              free_boolconj(perm_dvl);
              perm_dvl = NULL;
              if(verbose >= 3) 
                output("\tEqual dvl after permuting with process indices.\n");

              /* (d) check shielded clocks */

              if(verbose >= 3) {
                output("Orig clocks: "); 
                print_var_links(m->shielded_clocks);
              }

              perm_sc = permuted_clocks(m->shielded_clocks, tmp_perm);

              if(verbose >= 3) {
                output("Perm clocks: "); 
                print_var_links(perm_sc);
              }

              perm_sc = sort_var_links(perm_sc);

              if(verbose >= 3) {
                output("Sort clocks: "); 
                print_var_links(perm_sc);
              }

              if(shield_clocks_compare(m->shielded_clocks, perm_sc) != EQUAL) {
                free_var_links(perm_sc); perm_sc = NULL;
                free(tmp_perm); tmp_perm = NULL;
                if(trr == m->out_trans_cycle) break;
                else continue;
              }
              free_var_links(perm_sc); perm_sc = NULL;
              free(tmp_perm); tmp_perm = NULL;
              if(verbose >= 3)
                output("\tEqual shielded clocks on permutation.\n");

/*
              if(verbose >= 1) {
                output("\t  Isomorphic and equal zones after permutations.\n");
              }
*/
              if((comp = trans_comp(tr, trr)) < 0)  {
                elim[i] = TRUE;
                if(verbose >= 0) {
                  sprintf(tmpstring, "Eliminating trans %d.\n", trr->index);
                  output(tmpstring);
                }
              }
              else if(comp > 0) { 
                elim[j] = TRUE;
                if(verbose >= 0) {
                  sprintf(tmpstring, "Eliminating trans %d.\n", tr->index);
                  output(tmpstring);
                }
                break;
              }
              else if(comp == 0) {
                if(j < i) {
                  elim[i] = TRUE;
                  if(verbose >= 0) {
                    sprintf(tmpstring, "Eliminating trans %d.\n", trr->index);
                    output(tmpstring);
                  }
                }
                else if(j > i) {
                  elim[j] = TRUE;
                  if(verbose >= 0) {
                    sprintf(tmpstring, "Eliminating trans %d.\n", tr->index);
                    output(tmpstring);
                  }
                  break;
                }
              }
/*
              printf("... this transition (%d) can be eliminated.\n", 
                     trr->index);
              printf("elim = ( ");
              for(i=0; i < m->out_trans_count; i++) {
                printf("%d ", elim[i]);
              }
              printf(")\n");
*/
            }
            if(trr == m->out_trans_cycle) break;
        }
        if(tr == m->out_trans_cycle) break;
      }
/*
      printf("elim = ( ");
      for(i=0; i < m->out_trans_count; i++) {
        printf("%d ", elim[i]);
      }
      printf(")\n");
*/
/*
      for(inv=0, i=0, tr = m->out_trans_cycle->next; ; i++, tr=tr->next) {
        if(elim[i] == TRUE && 
           BDD_AND(tr->trigger_bdd, m->dval_bdd) == BDD_FALSE) 
          inv++;
        if(tr == m->out_trans_cycle) break;
      }
      t->invalid_trans_count -= inv;
*/
      for(inv=0, i=0; i<m->out_trans_count; i++) {
        if(elim[i] == TRUE) inv++;
      }
      t->trans_count -= inv;

      count = m->out_trans_count;
      for(i=0, trr = m->out_trans_cycle->next; i < count; i++) {
        if(elim[i] == TRUE) {
          if(trr == m->out_trans_cycle) {
            if(verbose >= 0) {
              sprintf(tmpstring, 
                   "    Eliminating transition (%d) from %s (%d) to %s (%d).\n",
              trr->index, m->name, m->index, trr->mode->name, trr->mode->index);
              output(tmpstring);
            }
            eliminate_transition(trr, m);
            break;
          }
          else {
            trrn = trr->next;
            if(verbose >= 0) {
              sprintf(tmpstring, 
                   "    Eliminating transition (%d) from %s (%d) to %s (%d).\n",
              trr->index, m->name, m->index, trr->mode->name, trr->mode->index);
              output(tmpstring);
            }
            eliminate_transition(trr, m);
            trr = trrn;
          }
        }
        else trr = trr->next;
      }
      free(elim); elim = NULL;
    } 
    if(stats == TRUE && SGM_MODE != GUI && SGM_MODE != PROCESS) {
      clear_line();
      fprintf(stderr, "\tChecked %d modes, Remaining %d modes (%.1f%%)\r",
                      k, t->mode_count-k,
              ((float)(t->mode_count-k) / (float)t->mode_count)*100.0);
    }
  }
  clean_line();
  return 0;
}

int dvl_compare(struct boolconj_type *dvl1, struct boolconj_type *dvl2) {
  struct boolconj_type *b;
  bdd f1 = BDD_TRUE, f2 = BDD_TRUE;

  if(dvl1) {
    for(b=dvl1->next; ; b=b->next) {
      f1 = BDD_AND(f1, b->prop_bdd);
      if(b==dvl1) break;
    }
  }
  if(dvl2) {
    for(b=dvl2->next; ; b=b->next) {
      f2 = BDD_AND(f2, b->prop_bdd);
      if(b==dvl2) break;
    }
  }
  return bdd_comp(f1, f2);
}

int corr_trans(struct transition_type *tr1, struct transition_type *tr2) {
  struct mode_type *m1, *m2;
  struct transition_type *t1, *t2;

  if(tr1->automaton == NULL || tr2->automaton == NULL) return FALSE;

  if( (tr1->automaton->index == NON_PARAM) ||
      (tr2->automaton->index == NON_PARAM) )
    return FALSE;

  if( (tr1->automaton->index == tr2->automaton->index) &&
      (tr1->trn->index == tr2->trn->index) )
    return TRUE;

  for(m1=get_thread(tr1->automaton)->starting_mode, 
      m2=get_thread(tr2->automaton)->starting_mode; 
      m1 && m2; m1=m1->cnext, m2=m2->cnext) {
    if(m1->out_trans_cycle && m2->out_trans_cycle) {
      for(t1 = m1->out_trans_cycle->next, 
          t2 = m2->out_trans_cycle->next; ;
          t1 = t1->next, t2 = t2->next) {
        if(t1->trn->index == tr1->trn->index) {
          if(t2->trn->index == tr2->trn->index) {
            return TRUE; 
          }
          else return FALSE;
        }
        else if(t2->trn->index == tr2->trn->index) 
          return FALSE;
        if(t1 == m1->out_trans_cycle && t2 == m2->out_trans_cycle) 
          break;
      }
    }
  }
  return FALSE;
}

int same_trans(struct transition_type *tr1, struct transition_type *tr2) {
  if(tr1->automaton == NULL || tr2->automaton == NULL) return FALSE;
  if( (tr1->automaton->u_index == tr2->automaton->u_index) &&
      (tr1->trn->index == tr2->trn->index) ) 
    return TRUE;
  else return FALSE;
}

int reduce_multigraph1(struct thread_type *t) {
  struct mode_type *m;
  struct transition_type *tr, *trr, *trrn;
  int inv, i, j, *elim, count, found = FALSE;

  sprintf(tmpstring, "Reducing multigraph of thread %s[%d].\n",
                     t->name, t->automaton_index);
  output(tmpstring);

  for(m = t->starting_mode; m; m=m->cnext) {
    sprintf(tmpstring, "Considering mode %s (%d) with %d out transitions.\n", 
            m->name, m->index, m->out_trans_count);
    output(tmpstring);

    if(m->out_trans_cycle) {
      elim = (int *)malloc(sizeof(int) * (m->out_trans_count));
      for(i=0; i < m->out_trans_count; i++) {
        elim[i] = FALSE;
      }
/*
      printf("elim = ( ");
      for(i=0; i < m->out_trans_count; i++) {
        printf("%d ", elim[i]);
      }
      printf(")\n");
*/

      for(j = 0, tr = m->out_trans_cycle->next; ; j++, tr = tr->next) {
        if(verbose >= 0) {
          sprintf(tmpstring, 
                  "\tConsidering transition (%d) to %s (%d), ", 
                  tr->index, tr->mode->name, tr->mode->index);
          output(tmpstring);
          print_trans_permutation(tr);
        }

        if(is_normal_trans(tr)) {
          output("\t  This is a normal transition.\n");
          for(i = 0, trr = m->out_trans_cycle->next; ; i++, trr = trr->next) {
/*
            printf("(i = %d) checking transition (%d).\n", i, trr->index);
*/
            if(tr == trr) {
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
            if(trr->mode != tr->mode) {
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
            if(is_normal_trans(trr))  {
              if(trr->index > tr->index)
                elim[i] = TRUE;
            }
            else elim[i] = TRUE;
/*
            printf("... this transition (%d) can be eliminated.\n", trr->index);
            printf("elim = ( ");
            for(i=0; i < m->out_trans_count; i++) {
              printf("%d ", elim[i]);
            }
            printf(")\n");
*/
            if(trr == m->out_trans_cycle) break;
          }
        }
        else {
          output("\t  This is not a normal transition.\n");
          found = FALSE;
          for(i = 0, trr = m->out_trans_cycle->next; ; i++, trr = trr->next) {
            if(tr == trr) {
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
            if(trr->mode != tr->mode) {
              if(trr == m->out_trans_cycle) break;
              else continue;
            }
            if(is_normal_trans(trr))  {
              elim[j] = TRUE;
              found = TRUE;
              break;
            }
            if(trr == m->out_trans_cycle) break;
          }
          if(!found) {
            sprintf(tmpstring, 
                   "\tNo normal transitions found for transition (%d).\n", 
                  tr->index);
            output(tmpstring);
            if(tr->permutation)
              permute_transition(tr, tr->permutation->perm);
/*
            free(tr->permutation->perm);
            tr->permutation->perm = (int *)NULL;
*/
          }
        }
        if(tr == m->out_trans_cycle) break;
      }
/*
      printf("elim = ( ");
      for(i=0; i < m->out_trans_count; i++) {
        printf("%d ", elim[i]);
      }
      printf(")\n");
      for(inv=0, i=0, tr = m->out_trans_cycle->next; ; i++, tr=tr->next) {
        if(elim[i] == TRUE && tr->valid == FALSE) inv++;
        if(tr == m->out_trans_cycle) break;
      }
      t->invalid_trans_count -= inv;
*/

      for(inv=0, i=0; i<m->out_trans_count; i++) {
        if(elim[i] == TRUE) inv++;
      }
      t->trans_count -= inv;

      count = m->out_trans_count;
      for(i=0, trr = m->out_trans_cycle->next; i < count; i++) {
        if(elim[i] == TRUE) {
          if(trr == m->out_trans_cycle) {
            sprintf(tmpstring, 
                    "\t  Eliminating transition (%d) from %s to %s.\n",
                    trr->index, m->name, trr->mode->name);
            output(tmpstring);
            eliminate_transition(trr, m);
            break;
          }
          else {
            trrn = trr->next;
            sprintf(tmpstring, 
                    "\t  Eliminating transition (%d) from %s to %s.\n",
                    trr->index, m->name, trr->mode->name);
            output(tmpstring);
            eliminate_transition(trr, m);
            trr = trrn;
          }
        }
        else trr = trr->next;
      }
      free(elim);
    } 
  }
  return 0;
}

int permute_variable(int v, int *perm) {
  if(perm == NULL)
    return v;
  else
    return var_index( var_name(v), perm_index(var_id(v), perm) );
}

int permute_assignment(struct assign_type *a, int *perm) {
  struct assign_type *aa;
  
  if(a) {
    for(aa = a->next; ; aa = aa->next) {
      if(get_var_type(aa->var) == V_CLOCK) {  
        aa->var = permute_variable(aa->var, perm);
      }
      else {
        if(aa->value == 0) {
          if(aa == a) break;
          else continue;
        }
        else {
          aa->value = perm_index(aa->value, perm);
        }
      }
      if(aa == a) break;
    }
  }
  return 0;
}

struct clkconj_type *permute_clkconj(struct clkconj_type *ct, int *perm) {
  struct clkconj_type *c, *cct = clkconj_copy(ct);


  if(cct) {
    if(cct->op == TRUE) return cct; 
    for(c = cct->next; ; c = c->next) {
      if(c->clk1) {
/*
        printf("Before permuting var %s[%d].\n", 
                var_name(c->clk1), var_id(c->clk1));
*/
        c->clk1 = permute_variable(c->clk1, perm);
/*
        printf("Done permuting var into %s[%d].\n", 
              var_name(c->clk1), var_id(c->clk1));
*/
      }

      if(c->clk2) {
/*
        printf("Before permuting var %s[%d].\n",
                var_name(c->clk2), var_id(c->clk2));
*/
        c->clk2 = permute_variable(c->clk2, perm);
/*
        printf("Done permuting var into %s[%d].\n",
                var_name(c->clk2), var_id(c->clk2));
*/
      }
      if(c == cct) break;
    }
  }
  return cct;
}

int permute_prp_trigger(struct boolconj_type *pt, int *perm) {
  struct boolconj_type *b;
  if(pt) {
    if(pt->op == TRUE) return 0;
    for(b = pt->next; ; b = b->next) {
      b->constant = perm_index(b->constant, perm);
      if(b == pt) break;
    }
  }
  return 0;
}

struct transition_type *
permute_transition(struct transition_type *t, int *perm) {
  struct transition_type *t1, *t2 = NULL;
  struct mode_type *m1, *m2;
  struct thread_type *th;
  int new_automaton_index, found = FALSE;

  if(verbose >= 2) {
    sprintf(tmpstring, "\tPermuting transition (%d) using ", t->index);
    output(tmpstring);
    print_perm(perm);
  
    output("\tOriginal transition:\n\t");
    print_trans(t);
  }

  new_automaton_index = perm_index(t->automaton->index, perm);
  
  th = get_thread1(t->automaton->name, new_automaton_index);
  for(m1=get_thread(t->automaton)->starting_mode, m2=th->starting_mode; 
      m1 && m2; 
      m1=m1->cnext, m2=m2->cnext) {
    for(t1=m1->out_trans_cycle->next,
        t2=m2->out_trans_cycle->next; ; 
        t1=t1->next, t2=t2->next) {
      if(t1->trn->index == t->trn->index) {
        found = TRUE;
        break;
      }
      if(t1 == m1->out_trans_cycle) break;
    }
    if(found) break;
  }
  if(found == FALSE) {
    fprintf(stderr, "Fatal error in permute_transition().\n");
    for(;1;);
  }
  t->trn = t2->trn;
  t->automaton = get_automaton(th);

  if(verbose >= 2) {
    output("\tPermuted transition:\n\t");
    print_trans(t);
  }
  return t;
}

int print_trans_permutation(struct transition_type *t) {
  int i;
  output("permutation: (");
  if(t->permutation) {
    if(t->permutation->perm) {
      for(i=1; i<=system_ta_num; i++) {
        sprintf(tmpstring, "%d", (t->permutation->perm)[i]);
        output(tmpstring);
        if(i < system_ta_num) output(" ");
      }
    }
  }
  else output("NORMAL");
  output(")\n");
  return 0;
}

int print_perm(int *p) {
  int i;
  output("permutation: (");
  if(p) {
    for(i=1; i<=system_ta_num; i++) {
      sprintf(tmpstring, "%d", p[i]);
      output(tmpstring);
      if(i<system_ta_num) output(" ");
    }
  }
  else output("NORMAL");
  output(")\n");
  return 0;
}

int is_normal_trans(struct transition_type *t) {
  if(t->permutation) {
    if(is_normal_perm(t->permutation->perm))
      return TRUE;
    else
      return FALSE;
  }
  else return TRUE;
}


/* copy permutation p2 to p1 where p1 has no space yet allocated */

int *copy_permutation(int *p1, int *p2) {
  int i;
  if(p2) {
    p1 = (int *)malloc(sizeof(int) * (system_ta_num + 1));
    for(i=0; i<=system_ta_num; i++) {
      p1[i] = p2[i];
    }
    return p1;
  }
  else return (int *)NULL;
}

int copy_clock_order(int *perm) {
  int i;

  ordered_clocks = (int *)malloc(sizeof(int) * (system_ta_num+1));
  for(i=1; i<=system_ta_num; i++) {
    ordered_clocks[i] = perm[i];
  }
  return 0;
}

/*
 * (1) permute the name of mode "m" using the permutation "perm",
 * (2) if nmv is not NULL, let it point to the new permuted mode vector, and
 * (3) return the new permuted mode name.
 */
char *permute_mode_name(struct mode_type *m, int *perm, char ***nmv) {
  int i;
  char *n;
  char **v = NULL;

  v = (char **)malloc(sizeof(char *) * system_ta_num);

  for(i=0; i<system_ta_num; i++) {
    v[i] = (char *)NULL;
  }
  for(i=0; i<system_ta_num; i++) {
    if(m->mode_vec[i])
      v[perm_index(i+1, perm)-1] = str_preserve(m->mode_vec[i]);
  }
  n = get_name(&v);

  if(nmv) {
    free(*nmv);
    *nmv = v;
  }
  else {
    for(i=0; i<system_ta_num; i++) {
      if(v[i]) { free(v[i]); v[i] = NULL; }
    }
    free(v); v = NULL;
  }
  return n;
}

char *permute_mode_name1(struct mode_type *m, int *perm) {
  char **n = (char **)malloc(sizeof(char *) * (system_ta_num + 1));
  int i, num_token;
  char *tc = (char *)malloc(sizeof(char) * (strlen(m->name) + 1));
  char *tmp_name = (char *)malloc(sizeof(char)*(strlen(m->name)+1));
  char *new_name = (char *)malloc(sizeof(char)*(strlen(m->name)+1));
  
  if(verbose >= 3) {
    sprintf(tmpstring, "\tPermuting name of mode %s (%d), ", m->name, m->index);
    output(tmpstring);
  }

  for(i=0; i<=system_ta_num; i++) {
    n[i] = (char *)NULL;
  }

  tmp_name = strcpy(tmp_name, m->name);
  tc = strtok(tmp_name, ".");
  
  n[perm_index(1, perm)] = 
                     (char *)malloc(sizeof(char) * (strlen(tc) + 1));
  n[perm_index(1, perm)] = 
                     strcpy(n[perm_index(1, perm)], tc);

/*
  printf("perm_index(1, perm) = %d, token = %s.\n", 
          perm_index(1, perm), tc);
*/

  i = 1;
  while( (tc = strtok((char *)NULL, ".")) ) {
    i++;
/*
    printf("perm_index(%d) = %d, token = %s.\n", 
           i, perm_index(i, perm), tc);
*/
    n[perm_index(i, perm)] = 
                       (char *)malloc(sizeof(char) * (strlen(tc) + 1));
    n[perm_index(i, perm)] = 
                       strcpy(n[perm_index(i, perm)], tc);
  }
  num_token = i;

  new_name = strcpy(new_name, n[1]);
  for(i=2; i<=num_token; i++) {
    new_name = strcat(new_name, ".");
    new_name = strcat(new_name, n[i]);
  }
  if(verbose >= 3) {
    sprintf(tmpstring, "new name is %s.\n", new_name);
    output(tmpstring);
  }
  for(i=0; i<=system_ta_num; i++) {
    free(n[i]);
  }
  free(n);
  free(tmp_name);
  free(tc);
  return new_name;
}

int remove_dormant_cycles(struct thread_type *th) {
  struct mode_type *m;
  struct transition_type *ptr, *tr, *str, *ttr;
  char *c = get_graph_name(th);

  if(verbose >= 0) {
   outputline();
   sprintf(tmpstring, "Removing dormant cycles in graph %s.\n", c);
   output(tmpstring);
  }

  for(m=th->starting_mode; m; m=m->cnext) {
    m->status = 0;
  }

  dfs_visit_mode(th->starting_mode);

  for(m=th->starting_mode; m; m=m->cnext) {
    if(m->status == VISITED) continue;
    if(m->in_trans_cycle) {
      for(ptr=m->in_trans_cycle, tr=ptr->next, str=tr->next; ;
          tr=str, str=str->next) {
        for(ttr=tr->mode->out_trans_cycle->next; ; ttr=ttr->next) {
          if(ttr->index == tr->index) {
            break;
          }
          if(ttr==tr->mode->out_trans_cycle) break;
        }
        eliminate_transition(ttr, tr->mode);
        th->trans_count--;
        if(m->in_trans_cycle == NULL) break;
      }
    }
  }

  remove_irreachable_mode(th);

  if(verbose >= 0) {
  sprintf(tmpstring, "DONE removing dormant cycles in graph %s.\n", c);
  output(tmpstring);
  }
  print_counts(th);
  free(c);
  return 0;
}

 
int dfs_visit_mode(struct mode_type *m) {
  struct transition_type *tr;

  if(m->status == VISITED) return 0;

  m->status = VISITED;

  if(m->out_trans_cycle) {
    for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
      dfs_visit_mode(tr->mode);
      if(tr == m->out_trans_cycle) break;
    }
  }
  return 0;
}

struct thread_type *get_thread(struct automaton *a) {
  int i;

  if(a)
  for(i=0; i<system_ta_num; i++) {
    if((!strcmp(a->name, thread[i].name)) && 
       (a->index == thread[i].automaton_index))
      return &thread[i];
  }
  return (struct thread_type *)NULL;
}

struct thread_type *get_thread1(char *n, int k) {
  int i;

  for(i=0; i<system_ta_num; i++) {
    if((!strcmp(n, thread[i].name)) &&
       (k == thread[i].automaton_index))
      return &thread[i];
  }
  return (struct thread_type *)NULL;
}

struct automaton *get_automaton(struct thread_type *th) {
  struct automaton *a;

  if(th && automaton_tail) {
    for(a=automaton_tail->next; ;a=a->next) {
      if( (!strcmp(a->name, th->name)) && (a->index == th->automaton_index) )
        return a;
      if(a == automaton_tail) break;
    }
  }
  return (struct automaton *)NULL;
}
