%p 4096

%{
#include "include.h"
#include "global.h"
#include "y.tab.h"
%}

%start norm comment1 comment

/* regular definitions */
delim      [ \t\n]
ws         {delim}+
digit      [0-9]
letter     [a-zA-Z]
ident      {letter}({letter}|{digit}|[_])*
number     {digit}+

%%

<norm>[ \t\n]+  ;

<norm>clock     { yylval.v_type =  V_CLOCK; return CLOCK; }
<norm>register  { yylval.v_type =  V_REGISTER; return REGISTER; }
<norm>automata  { return AUTOMATA; }

<norm>total     { return TOTAL; }
<norm>initially { return INITIALLY; }
<norm>automaton { return AUTOMATON; }
<norm>synclabs	{ return SYNCLABS; }
<norm>mode      { return MODE; }
<norm>invariant { return INVARIANT; }
<norm>when      { return WHEN; }
<norm>sync	{ return SYNC; }
<norm>may	{ return MAY; }
<norm>goto      { return GOTO; }

<norm>and       { return P_AND; }
<norm>or        { return P_OR; }
<norm>not       { return P_NOT; }

<norm>mod       { return MOD; }

<norm>True      { return P_TRUE; }
<norm>False     { return P_FALSE; }

<norm>verify      { return VERIFY; }
<norm>until       { return T_UNTIL; }
<norm>exists_path { return T_EXISTS_PATH; }
<norm>all_paths   { return T_ALL_PATHS; }
<norm>eventually  { return T_EVENTUALLY; }
<norm>henceforth  { return T_HENCEFORTH; }

<norm>manipulation     { return MANIPULATION; }
<norm>graph	       { return GRAPH; }
<norm>merge_graph      { return MERGE_GRAPH; }
<norm>shield_clock     { return SHIELD_CLOCK; }
<norm>normalize_region { return NORMALIZE_REGION; }
<norm>read_write       { return READ_WRITE; }
<norm>transition_bypass { return TRANSITION_BYPASS; }
<norm>shield_variables { return SHIELD_VARIABLES; }
<norm>look_ahead       { return LOOK_AHEAD; }
<norm>model_check      { return MODEL_CHECK; }
<norm>sequential_merge { return SEQUENTIAL_MERGE; }
<norm>pairwise_merge   { return PAIRWISE_MERGE; }
<norm>print_graph      { return PRINT_GRAPH; }

<norm>for	{ return FOR; }
<norm>if	{ return IF; }
<norm>else	{ return ELSE; }
<norm>num_modes	{ return M_SIZE; }
<norm>num_trans	{ return T_SIZE; }

<norm>forall    { return FORALL_INDEX; }
<norm>exists    { return EXISTS_INDEX; }

<norm>"++"	{ return PLUS_PLUS; }
<norm>"+="	{ return PLUS_EQ; }
<norm>"=="	{ return EQ_EQ; }
<norm>".."      { return INTERVAL; }
<norm>":"	{ return ':'; }
<norm>":="      { return ASSIGNMENT; }
<norm>"~"	{ return '~'; }
<norm>"&"	{ return '&'; }
<norm>"|"	{ return '|'; }
<norm>","       { return ','; }
<norm>";"       { return ';'; }
<norm>"/"       { return '/'; }
<norm>"+"       { return '+'; }
<norm>"-"       { return '-'; }
<norm>"<"       { yylval.integer = LESS; return RELOP; }
<norm>"<="      { yylval.integer = LEQ; return RELOP; }
<norm>"="       { yylval.integer = EQ; return RELOP; }
<norm>">"       { yylval.integer = GREATER; return RELOP; }
<norm>">="      { yylval.integer = GEQ; return RELOP; }
<norm>"!="      { yylval.integer = NEQ; return RELOP; }
<norm>"{"       { return '{'; }
<norm>"}"       { return '}'; }
<norm>"("       { return '('; }
<norm>")"       { return ')'; }
<norm>"["       { return '['; }
<norm>"]"       { return ']'; }

<norm>{ident}   { yylval.string = str_preserve((char *)yytext); 
                  string_buffer = strcpy(string_buffer, yylval.string); 
	          free(yylval.string);
                  return NAME; }
<norm>{number}  { yylval.integer = atoi((char *)yytext); return INT; }

<norm>"--"	{ BEGIN comment1; }
<comment1>(.)*	;
<comment1>[\n]	{ BEGIN norm; }

<norm>"/\*"	{ BEGIN comment; }
<comment>"\*/"	{ BEGIN norm; }
<comment>[^\*/]*	;

%%

void lexinit() { 
    BEGIN norm;
}

int yywrap() {
    if ((yyin != stdin)) {
	close (yyin);
	yyin = stdin;
    };
   return 1;
}
