%{
#include <stdio.h>
#include <string.h>
#include "include.h"
#include "global.h"
#include "print.h"
#include "action.h"
#include "readwrite.h"
#include "modeman.h"
#include "transman.h"
#include "symmetry.h"

extern FILE *yyin;
extern FILE *logfile;

/* to overcome yacc errors on DECs for redefining YYSTYPE */
#define YYSTYPE HWT_YYSTYPE

/* some versions of yacc use this as a bound on the size of the parse stack */
/* Sun versions seem to ignore it */
#define YYMAXDEPTH 2000

%}        

%start main_program

%union	yystacktype {
  char	*string;
  int	integer;
  int	v_type;
}

%token CLOCK
%token REGISTER
%token INTERVAL

%token AUTOMATA
%token INITIALLY
%token AUTOMATON
%token SYNCLABS
%token MODE
%token INVARIANT
%token WHEN
%token SYNC
%token MAY
%token GOTO
%token ASSIGNMENT
%token TOTAL

%token VERIFY
%token T_UNTIL
%token T_EXISTS_PATH
%token T_ALL_PATHS
%token T_EVENTUALLY
%token T_HENCEFORTH

%token MANIPULATION
%token GRAPH 
%token MERGE_GRAPH
%token SEQUENTIAL_MERGE
%token PAIRWISE_MERGE
%token SHIELD_CLOCK
%token NORMALIZE_REGION
%token READ_WRITE
%token TRANSITION_BYPASS
%token SHIELD_VARIABLES
%token LOOK_AHEAD
%token MODEL_CHECK
%token PRINT_GRAPH

%token FORALL_INDEX
%token EXISTS_INDEX

%token IF
%token ELSE
%token M_SIZE
%token T_SIZE

%token FOR
%token MOD

%left PLUS_PLUS
%left PLUS_EQ

%token P_TRUE
%token P_FALSE

%token EQ_EQ

%left RELOP

%left P_AND

%right '~' 
%right '('
%left '|'
%left '&'

%left P_OR
%right P_NOT


%token INT
%token NAME

%%

main_program : automata_declarations
               var_declarations
               program { install_in_trans(); install_trans_mode();  
                         if(verbose >= 0) print_automata();  }
               specification
               actions 
             ;

var_declarations : { tmp_var_count = var_count; } 
                   variable_type { var_type = yylval.v_type; } 
                   variable ';'
                   { tmp_var_count = var_count - tmp_var_count;
                     add_attribute(var_type); }
                   var_declarations
                 |
                 ;

variable_type : CLOCK
              | REGISTER
              ;

variable : NAME { var_count++; insert_var(string_buffer); }
           next_variable
         | NAME 
           '[' { interval_varname = str_preserve(string_buffer); }
           num { interval_start = last_int; }
           INTERVAL 
           end_int
           ']' { var_count += (interval_end - interval_start + 1); 
                 insert_var(interval_varname);
                 free(interval_varname);
                 interval_varname = NULL;
                 interval_start = -1; interval_end = -1;}
           next_variable
         |
         ;

num : INT { last_int = yylval.integer; }
    | '-' INT { last_int = -yylval.integer; }
    ;

end_int : num { interval_end = last_int; }
        | TOTAL '(' automaton_name
                 { interval_end = total_ta(string_buffer); } 
                ')'
          inc_dec
        ;

inc_dec : '+' num { interval_end += last_int; }
        | '-' num { interval_end -= last_int; }
        |
        ;

next_variable : ',' variable
              |
              ;

automata_declarations : AUTOMATA
                        auto_dec
		        ';'
                      ;

auto_dec : automaton_name { interval_varname = str_preserve(string_buffer); }
	   param_dec
           next_auto_dec
         ;

param_dec : ':'
           param
           RELOP { if(yylval.integer != EQ) {
                     fprintf(stderr, 
                  "\nError (line #%d): no other operators allowed besides =.\n",
                              yylineno); 
                     exit(1); } 
                 }
           num { interval_start = last_int; }
           rest_auto { new_automata(); 
		       free(interval_varname);
                       system_ta_num += interval_end - interval_start + 1; 
                       interval_start = -1; interval_end = -1;}
         |           { new_automata();
		       free(interval_varname);
                       system_ta_num += 1; }
         ;

param :  NAME { parameter = str_preserve(string_buffer); }
      ;

rest_auto : INTERVAL num { interval_end = last_int; }
          | { interval_end = interval_start; }
          ;

next_auto_dec : ',' auto_dec
              |
              ;

program : AUTOMATON
          automaton_name { current_ta_num = total_ta(current_ta->name); 
                           /* system_ta_num += current_ta_num; */
                           init_tails();  init_sync(); init_tr_sync(); }
          automaton
          next_program
        ;

next_program : program
             | { check_automata_defn(); }
             ;

automaton_name : NAME { current_ta = get_ta(string_buffer, 1); }
               | NAME { current_ta = get_ta(string_buffer, 1); }
                 '[' 
                 param { check_param(parameter); 
                         free(parameter); parameter = NULL;}
                 ']' 
               ;

automaton: '{'
	      syncdecl
              prolog { reset_exp_tails(); }
              modelist
           '}' { install_mode(); free_tails(); }
         ;

syncdecl: SYNCLABS 
          synclist { install_sync(); } ';'
        |
        ;

synclist: NAME { insert_sync((char *)yytext); } synclist_next
          |
          ;

synclist_next : ',' synclist
               |
               ;

prolog: INITIALLY 
        NAME { install_init_mode(string_buffer); }
        initial_cond ';'
      |
      ;

initial_cond : P_AND { reset_exp_tails(); 
                       total_exp = current_ta_num; }
               expression { install_init_cond(); }
             | 
             ;

expression : conjunct expressionrest
           | '~' 
             '(' { neg_conj = TRUE; }
             conjunct 
             ')' { neg_conj = FALSE; }
             expressionrest
           ;

conjunct : { sign = 1; need_transform = 0; index_quantifier = -1; }
           subexp 
           RELOP { install_relop(yylval.integer); sign = -1; }
           subexp { transform_clkconj(); }
	   mod_exp
         | P_TRUE { }
         | P_FALSE { }
         ;

expressionrest : P_AND { inside_exp = FALSE; index_relop = -1; }
                 expression
               | '&' { new_boolconj_array(); index_relop = -1; } 
                 expression
               | '|' { new_boolexp(); index_relop = -1; }
                 expression
               | { inside_exp = FALSE; index_relop = -1; }
               ;

subexp : term
       | subexp '+' term
       | subexp '-' { sign = -sign; } term { sign = -sign; }
       ;

term : num         	  { if(inside_exp == FALSE)
                              warn_constant();
                            else install_const(); }
     | NAME         	  { if(inside_exp == FALSE) {
                              if(is_param(string_buffer, current_ta))
                                warn_param();
                              else {
                                new_exp(string_buffer, NULL); 
                                inside_exp = TRUE; 
                              }
                            }
                            if(is_param(string_buffer, current_ta)) {
                              check_param(string_buffer);
                              install_param();
                            }
                            else install_term(string_buffer, NULL); }
     | NAME               { interval_varname = str_preserve(string_buffer); }
       '[' index_range 
       param ']'          { if(inside_exp == FALSE) {
                              new_exp(interval_varname, parameter);
                              inside_exp = TRUE; 
                            }
                            check_param(parameter);
                            install_term(interval_varname, parameter); 
                            free(interval_varname); interval_varname=NULL;
                            free(parameter); parameter=NULL;
                          }
     ;

index_range : FORALL_INDEX { index_quantifier = FORALL; }
              RELOP { index_relop = yylval.integer; }
            | EXISTS_INDEX { index_quantifier = EXISTS; }
              RELOP { index_relop = yylval.integer; }
            |       { index_relop = -1; }
            ;

mod_exp : '(' MOD num ')'    { install_mod_num(); }
	| '(' MOD NAME ')'   { install_mod_var(0); }
	| '(' MOD NAME '['   { interval_varname = str_preserve(string_buffer); }
              num ']' ')'    { install_mod_var(1); free(interval_varname); 
                               interval_varname=NULL;}
	| '(' MOD NAME '['   { interval_varname = str_preserve(string_buffer); }
              param ']' ')'  { install_mod_var(2); 
                               free(parameter); parameter=NULL;
                               free(interval_varname); 
                               interval_varname=NULL;}
        |
	;

modelist : mode next_mode
	 ;

mode : MODE NAME { new_mode(string_buffer); }
      '{' 
	 INVARIANT { reset_exp_tails(); 
                     total_exp = current_ta_num; }
	 expression { install_inv(); 
                      reset_exp_tails(); }
	 ';'
	 transitionlist { install_out_trans(); }
      '}' 
     ;

next_mode : modelist
	  | 
	  ;

transitionlist : transition transitionlist
	       | 
	       ;

transition : WHEN               { reset_exp_tails(); 
                                  total_exp = current_ta_num; 
                                  new_transition(); }
	     expression 	{ install_trigger(); 
                          	  reset_exp_tails(); }
	     trans_synclist     { install_trans_sync(); }
	     MAY GOTO NAME 	{ assign_goto_mode((char *)yytext); }
	     ';'
	     assignmentlist  	{ install_assignment(); }
	   ;

trans_synclist : sync_cond trans_synclist
         |
         ;

sync_cond : SYNC NAME { insert_trans_sync((char *)yytext); }
          ;

assignmentlist : { reset_exp_tails(); }
                 assignment assignmentlist
	       |
	       ;

assignment : { new_assign(); sign = 1; }
             assgt_term 
	     ASSIGNMENT    { sign = -1; }
             right_hand_term
	     mod_assign
             ';'
           ;

right_hand_term : num            { install_assign_value(); }
                | assgt_term
                | assgt_term 
                  oper
                  num            { install_assign_value(); }
                ;

oper : '+'  { install_assign_op(ADD); }
     | '-'  { install_assign_op(SUB); }
     ;

assgt_term : NAME           { if(is_param(string_buffer, current_ta))
                                install_assign_param();
                              else install_assign_var(string_buffer, NULL); }
           | NAME           { interval_varname = str_preserve(string_buffer); }
             '[' param ']'  { check_param(parameter);
                              install_assign_var(interval_varname, parameter);
                              free(interval_varname); interval_varname = NULL;
                              free(parameter); parameter = NULL; }
           ;

mod_assign : '(' MOD num ')'  {install_assign_mod_num(); }
           | '(' MOD NAME ')' {install_assign_mod_var(0); }
           | '(' MOD NAME '[' {interval_varname = str_preserve(string_buffer); }
              num ']' ')'    {install_assign_mod_var(1); free(interval_varname);
                              interval_varname=NULL;}
           | '(' MOD NAME '[' {interval_varname = str_preserve(string_buffer); }
              param ']' ')'   {install_assign_mod_var(2);
                               free(parameter); parameter=NULL;
                               free(interval_varname);
                               interval_varname=NULL;}
           |
           ;

specification : { init_specification(); }
                VERIFY
                tctl_spec { /* print_tctl_spec(vspec); */ }
                ';'
              |
              ;

tctl_spec : tctl_spec_one tctl_spec_rest
          ;

tctl_spec_one : P_NOT 
                '{' { new_spec(NEGATIVE); }
                tctl_spec
                '}' { blevel--; }
              | '{' { new_spec(POSITIVE); } 
                tctl_spec
                '}' { blevel--; }
              | { new_spec(ATOM); init_spec_atom(); } spec_atom
              ;

tctl_spec_rest : P_AND { cspec = set_cspec(); cspec->boolcon = AND; }
                 tctl_spec
               | 
                 P_OR { cspec = set_cspec(); cspec->boolcon = OR; }
                 tctl_spec
               |
               ;

spec_atom : path_spec '(' no_path_formulae ')'
          ;

no_path_formulae : { new_mode_spec(); } mode_spec
                 | expression { install_exp(); }
                 | long_formulae { blevel--; }
                 | short_formulae { blevel--; }
                 ;

path_spec : T_EXISTS_PATH { cspec->path = EXISTS_PATH; }
          |
            T_ALL_PATHS { cspec->path = ALL_PATHS; }
          |
          ;

mode_spec : MODE 
            '(' name_spec
            ')' RELOP 
                { cspec->ms->relop = yylval.integer; }
            NAME { cspec->ms->mode_name = str_preserve(string_buffer); }
          ;

name_spec : NAME               { install_auto_name(0); }
          | NAME '[' param ']' { install_auto_name(-1); }
          | NAME '[' num ']'   { install_auto_name(yylval.integer); }
          ;

long_formulae : tctl_spec
                T_UNTIL { blevel--; 
                          cspec = set_cspec(); 
                          cspec->stype = UNTIL;
                          cspec->relop = GEQ; }
                tctl_spec
              | tctl_spec
                T_UNTIL { blevel--; 
                          cspec = set_cspec(); 
                          cspec->stype = UNTIL; }
                RELOP { cspec->relop = yylval.integer; } 
                num {cspec->constant = yylval.integer; }
                tctl_spec
              ;

short_formulae : T_EVENTUALLY { cspec->stype = EVENTUALLY; cspec->relop = GEQ; }
                 tctl_spec
               | T_EVENTUALLY { cspec->stype = EVENTUALLY; }
                 RELOP { cspec->relop = yylval.integer; } 
                 num {cspec->constant = yylval.integer; }
                 tctl_spec
               | T_HENCEFORTH { cspec->stype = HENCEFORTH; cspec->relop = GEQ; }
                 tctl_spec
               | T_HENCEFORTH { cspec->stype = HENCEFORTH; }
                 RELOP { cspec->relop = yylval.integer; } 
                 num {cspec->constant = yylval.integer; }
                 tctl_spec
               ;

actions : MANIPULATION { init_action_list(); }
          graph_decl
          action_list
        |
        ;

graph_decl : GRAPH { init_graph_list(); }
             gnames ';'
           |
           ;

gnames : NAME { interval_varname = str_preserve(string_buffer); 
                new_graph(); } 
         next_gname
       | NAME
           '[' { interval_varname = str_preserve(string_buffer); }
           num { interval_start = last_int; }
           INTERVAL
           end_int
           ']' { new_graph();
                 interval_start = -1; interval_end = -1;}
         next_gname
       ;

next_gname : ',' gnames
           |
           ;

action_list : { new_action(); } one_action action_list_rest
            ;

action_list_rest : { new_action(); } one_action action_list_rest
                 |
                 ;

one_action : graph_name          { curr_action->type = MG; install_graph(2); }
/* RELOP               { check_relop(yylval.integer, EQ); } */
             ASSIGNMENT
             MERGE_GRAPH 
             '(' graph_name      { install_graph(0); }
             ',' graph_name ')'  { install_graph(1); } ';'
           | SHIELD_CLOCK      { curr_action->type = SC; 
                                 clock_shielding = TRUE; } 
             '(' one_graph ')' ';'
           | NORMALIZE_REGION  { curr_action->type = NR; 
                                 region_normalization = TRUE; }
             '(' one_graph ')' ';'
           | READ_WRITE        { curr_action->type = RW; 
                                 /*lazy_bdd_merging = FALSE;*/ }
             '(' one_graph ')' ';'
           | TRANSITION_BYPASS { curr_action->type = TB; }
             '(' one_graph ')' ';'
           | SHIELD_VARIABLES  { curr_action->type = SV; }
             '(' one_graph ')' ';'
           | LOOK_AHEAD        { curr_action->type = LA; 
                                 look_forward_reduction = TRUE; } 
             '(' la_graphs ')' ';'
           | MODEL_CHECK       { curr_action->type = MC; } 
             '(' one_graph ')' ';'
           | PRINT_GRAPH       { curr_action->type = PG; } 
             '(' one_graph ')' ';'
           | SEQUENTIAL_MERGE  { curr_action->type = SM; merge_type = SM; } 
             '(' ')' ';'
           | PAIRWISE_MERGE    { curr_action->type = PM; merge_type = PM; } 
             '(' ')' ';'
           | if_action
           | loop_action
           ;

one_graph : graph_name { install_graph(0); }
         |
         ;

la_graphs : graph_name { install_graph(0); }
            ',' graph_name { install_graph(1); }
          |
          ;

graph_name : NAME                               { last_int = -1; }
           | NAME '[' param par_inc_dec ']'     { last_int = -2; }
           | NAME '[' end_int ']'		{ last_int = interval_end;
                                                  interval_end = -1; }
           ;

par_inc_dec : '+' num { p_num = last_int; }
            | '-' num { p_num = -last_int; }
            |
            ;

if_action : IF                { curr_action->type = IE; }
            '(' cond_conj ')' '{'
              { tmp_action = curr_action; curr_action = NULL;}
              action_list
              { curr_action = curr_action->next->parent; }
            '}'
            else_part
          ;

else_part : ELSE if_action
          | ELSE '{'
              { tmp_action = curr_action; curr_action = NULL;}
              action_list
              { curr_action = curr_action->next->parent; }
            '}'
          |
          ;

cond_conj : param EQ_EQ { curr_action->if_relop = EQ; }
            num { curr_action->if_num = yylval.integer; }
          | param RELOP { curr_action->if_relop= yylval.integer; }
            num { curr_action->if_num = yylval.integer; }
          | graph_size { install_graph(0); }
            RELOP { curr_action->if_relop = yylval.integer; }
            graph_size { install_graph(1); }
          |
          ;

graph_size : M_SIZE { curr_action->if_num = NUM_MODES; }
             '(' graph_name ')'
           | T_SIZE { curr_action->if_num = NUM_TRANS; }
             '(' graph_name ')'
           ;

loop_action : FOR { curr_action->type = FL; }
              '(' init_list ';' cond_list ';' do_list ')' '{'
              { tmp_action = curr_action; curr_action = NULL;}
              action_list
              { curr_action = curr_action->next->parent; }
              '}'
            ;

init_list : param 
            ASSIGNMENT
            num { curr_action->for_init_num = yylval.integer; }
          |
          ;

cond_list : param 
            RELOP { curr_action->for_cond_relop = yylval.integer; }
            end_int { curr_action->for_cond_num = interval_end;
                      interval_end = 0; }
          |
          ;

do_list : param PLUS_PLUS 
           { curr_action->for_step = 1; }
        | param PLUS_EQ num 
           { curr_action->for_step = yylval.integer; }
        | param ASSIGNMENT param '+' num 
           { curr_action->for_step = yylval.integer; }
        | param ASSIGNMENT param '-' num 
           { curr_action->for_step = -yylval.integer; }
        |
        ;

%%

void
yyerror(const char *yerr_str) {
  fprintf(stderr,"\nError message: %s\n", yerr_str);
  fprintf(stderr,"Error in input on line %d while scanning input: %s\n",
	  yylineno, (char *)yytext);
  fflush(stderr);
  if(SGM_MODE ==  BATCH) exit(1);
  SGM_ERROR = E_PARSE_FAIL;
}
