#include "include.h"
#include "global.h"
#include "zoneman.h"
#include "print.h"
#include "readwrite.h"
#include "modeman.h"
#include "malloc.h"

int zone_count;
int total_clocks;      /* total number of clocks in the system */
int num_clocks;        /* number of clocks accessed by the current graph */
int boob = FALSE;      /* both out of bounds */
int *clock_indexes = (int *)NULL;
struct a23tree_type *zone_tree;

int reset_clock_indexes(void);
int get_clk_num(struct thread_type *);
int assign_new_index(struct thread_type *);
int access_clock(struct variable_type *, struct thread_type *);
int init_zone_management(void);
struct zone_type *alloc_zone(struct thread_type *t);
struct zone_type *init_zone(struct zone_type *, struct clkconj_type *, 
                            struct thread_type *);
struct zone_type *process_zone_max_const(struct zone_type *);
struct zone_type *zone(struct clkconj_type *, struct thread_type *);

int new_index(int, struct zone_type *);
int old_index(int someclk, struct zone_type *);
int clk_not_seen(int *, int);
void print_zone(struct zone_type *z);
int print_clock_indexes(struct zone_type *);

void free_zone(struct zone_type *);
struct zone_type *zone_copy(struct zone_type *);
/*
struct zone_type *zone_and(struct zone_type *, struct zone_type *);
struct zone_type *non_empty_zone_and( struct zone_type *, 
                                      struct zone_type *, 
                                      struct zone_type *);
*/

int zone_comp(struct zone_type *, struct zone_type *);
int install_clkconj_zones(void);
int out_of_bound(struct zone_type *);
struct zone_type *normalize_zone(struct zone_type *);
struct zone_type *init_reduce_zone(struct zone_type *);
struct zone_type *substitute_zone(struct zone_type *);
struct zone_type *reduce_zone(struct zone_type *);
/*
int all_visited(int *, struct zone_type *);
*/
struct zone_type *instantiate_zone(struct zone_type *);
struct clkconj_type *zone2clkconj(struct zone_type *);

struct zone_type *zone_proc_eq(struct zone_type *);
struct zone_type *zone_proc_leq(struct zone_type *);
struct zone_type *kclosure_zone(struct zone_type *);

int both_out_of_bounds(struct zone_type *, struct zone_type *);
int clk_index_compare(struct zone_type *, struct zone_type *);
int clk_index_diff(struct zone_type *, struct zone_type *);


int reset_clock_indexes(void) {
  int i;
  for(i=0; i<=total_clocks; i++) {
    clock_indexes[i] = -1;
  }
  return 0;
}

/* Return the number of clocks accessed by graph t
 * If t == NULL then return all the clocks in the system 
 */

int get_clk_num(struct thread_type *t) {
  int i = -1;
  struct variable_type *v;
  if(var_tail) {
    for(v=var_tail->next, i=0; ; v=v->next) {
      if( (v->type == V_CLOCK) &&
          (t == NULL || access_clock(v, t)) 
        ) i++;
      if(v == var_tail) break;
    }
  }
  return i;
}


/* (1) Assign the index of clocks accessed by graph t 
 *     in the int array clock_indexes[1...] and
 * (2) clock_indexes[0] is the size of the array 
 */

int assign_new_index(struct thread_type *t) {
  int i;
  struct variable_type *v = var_tail;
/*
  if(clock_indexes) free(clock_indexes);
*/
  clock_indexes = (int *)NULL;
  if(v) {
    for(v=v->next, i=0; ; v=v->next) {
      if( (v->type == V_CLOCK) && 
          (t == NULL || access_clock(v, t)) 
        ) {
        if(clk_not_seen(clock_indexes, v->index)) {
	  if(clock_indexes)
            clock_indexes = (int *)realloc(clock_indexes, sizeof(int) * (++i+1));
	  else
            clock_indexes = (int *)malloc(sizeof(int) * (++i+1));
          clock_indexes[0] = i;
          clock_indexes[i] = v->index;
        }
      }
      if(v == var_tail) break;
    }
  }
  return 0;
}


/* TRUE if variable v is accessed by graph t
 * FALSE otherwise
 */

int access_clock(struct variable_type *v, struct thread_type *t) {
  struct transition_type *tr;
  struct mode_type *m;
  struct clkconj_type *c;
  struct assign_type *a;
  int i;

  for(i=0; i<system_ta_num; i++) {
    if(component_graph(i, t)) {
      for(m=thread[i].starting_mode; m ; m=m->cnext) {
        if(m->xconj_cycle) {
          for(c = m->xconj_cycle->next; ; c=c->next) {
            if(c->clk1 == v->index) return TRUE;
            if(c->clk2 == v->index) return TRUE;
            if(c == m->xconj_cycle) break;
          }
        }
        if(m->out_trans_cycle) {
          for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
            if(tr->trn->clk_trigger_cycle) {
              for(c = tr->trn->clk_trigger_cycle->next; ; c=c->next) {
                if(c->clk1 == v->index) return TRUE;
                if(c->clk2 == v->index) return TRUE;
                if(c == tr->trn->clk_trigger_cycle) break;
              }
            }
            if(tr->trn->assign_cycle) {
              for(a=tr->trn->assign_cycle->next; ; a=a->next) {
                if(a->var == v->index) return TRUE;
                if(a == tr->trn->assign_cycle) break;
              }
            }
            if(tr == m->out_trans_cycle) break;
          }
        }
      }
    }
  }

/* Following does not work with subthreads reset */
/*
  if(t->subthread1 == NULL && t->subthread2 == NULL) {
    for(at=automaton_tail->next; ;at=at->next) {
      if(!strcmp(at->name, t->name) && at->index == t->automaton_index)
        break;
    }
    for(m=at->mode_cycle->next; ; m=m->next) {
      if(m->xconj_cycle) {
        for(c = m->xconj_cycle->next; ; c=c->next) {
          if(c->clk1 == v->index) return TRUE;
          if(c->clk2 == v->index) return TRUE;
          if(c == m->xconj_cycle) break;
        }
      }
      if(m->out_trans_cycle) {
        for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
          if(tr->trn->clk_trigger_cycle) {
            for(c = tr->trn->clk_trigger_cycle->next; ; c=c->next) {
              if(c->clk1 == v->index) return TRUE;
              if(c->clk2 == v->index) return TRUE;
              if(c == tr->trn->clk_trigger_cycle) break;
            }
          }
          if(tr->trn->assign_cycle) {
            for(a=tr->trn->assign_cycle->next; ; a=a->next) {
              if(a->var == v->index) return TRUE;
              if(a == tr->trn->assign_cycle) break;
            }
          }
          if(tr == m->out_trans_cycle) break;
        }
      }
      if(m == at->mode_cycle) break;
    }
  }
  else {
    if(access_clock(v, t->subthread1)) return TRUE;
    if(access_clock(v, t->subthread2)) return TRUE;
  }
*/
  return FALSE;
}

int init_zone_management(void) {
  struct variable_type *v = var_tail;
  total_clocks = 0;

  if(v) {
    for(v=var_tail->next; ; v=v->next) {
      if(v->type == V_CLOCK) {
        total_clocks++;
      }
      if(v == var_tail) break;
    }
  }

/*  
  clock_indexes = (int *)malloc((total_clocks+1) * sizeof(int));
  reset_clock_indexes();
  assign_new_index(NULL);
  if(verbose > 0) 
    print_clock_indexes();
*/
  return 0;
}

int clk_not_seen(int clocks[], int c) {
  int i;
  if(c == 0) return TRUE;
  if(clocks) {
    for(i=1; i<=clocks[0]; i++) {
      if(clocks[i] == c) return FALSE;
    }
  }
  return TRUE;
}

struct zone_type *alloc_zone(struct thread_type *t) {
  int i, j;
  struct zone_type *z = (struct zone_type *)malloc(sizeof(struct zone_type));

  num_clocks = get_clk_num(t);
  z->index = ++zone_count;
  z->relop = (int **)malloc((1 + num_clocks) * sizeof(int *));
  z->diff = (int **)malloc((1 + num_clocks) * sizeof(int *));
  for(i=0; i<=num_clocks; i++) {
    z->relop[i] = (int *)malloc((1 + num_clocks) * sizeof(int));
    z->diff[i] = (int *)malloc((1 + num_clocks) * sizeof(int));
  }

  for(i=0; i<=num_clocks; i++) {
    for(j=0; j<=num_clocks; j++) {
      z->relop[i][j] = -1;
      z->diff[i][j] = 0;
    }
  }
/*
  z->clk_index = (int *)NULL;
*/
  assign_new_index(t);
  z->clk_index = clock_indexes;
  return z;
}

void print_zone(struct zone_type *z) {
  int i, j;
  int count = 0;

  if(z->clk_index == NULL) {
    output("ZONE_TRUE");
    return;
  }
  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(i==0 && j==0) {
        if(z->relop[i][j] == TRUE) {
          output("ZONE_TRUE");
          return;
        }
        if(z->relop[i][j] == FALSE) {
          output("ZONE_FALSE");
          return;
        }
      }
      if(z->relop[i][j] >= LESS) {
        if(count > 0) output(" & ");
        else output("\t");
        if(count > 0 && fmod(count, 3) == 0)
          output("\n\t");
        output("(");
        if(i) {
          if(var_id(z->clk_index[i]) == NON_PARAM)
            sprintf(tmpstring, "%s", var_name(z->clk_index[i]));
          else
            sprintf(tmpstring, "%s[%d]", 
                         var_name(z->clk_index[i]), var_id(z->clk_index[i]));
          output(tmpstring);
        }
        if(i && j) output(" - ");
        if(!i && j) output("-");
        if(j) {
          if(var_id(z->clk_index[j]) == NON_PARAM)
            sprintf(tmpstring, "%s", var_name(z->clk_index[j]));
          else 
            sprintf(tmpstring, "%s[%d]", 
                         var_name(z->clk_index[j]), var_id(z->clk_index[j]));
          output(tmpstring);
        }
        if(z->diff[i][j] == INF) {
          sprintf(tmpstring, " %s INF", relop_string(z->relop[i][j]));
          output(tmpstring);
        }
        else if(z->diff[i][j] == -INF) {
          sprintf(tmpstring, " %s -INF", relop_string(z->relop[i][j]));
          output(tmpstring);
        }
        else {
          sprintf(tmpstring, " %s %d", 
                             relop_string(z->relop[i][j]), z->diff[i][j]);
          output(tmpstring);
        }
        output(")");
        count++;
      }
    }
  }
}

int print_clock_indexes(struct zone_type *z) {
  int i;

  sprintf(tmpstring, "There are totally %d clocks.\n", total_clocks);
  output(tmpstring);
  for(i=1; i<=z->clk_index[0]; i++) {
    sprintf(tmpstring, "%s[%d] (old: %d, new: %d)", var_name(z->clk_index[i]), 
                                          var_id(z->clk_index[i]),
                                          z->clk_index[i], i);
    output(tmpstring);
    if(i<total_clocks) output(", ");
  }
  output("\n");
  return 0;
}

struct zone_type *init_zone(struct zone_type *z, struct clkconj_type *c, 
                            struct thread_type *t) {
  struct clkconj_type *x;
  
/*
  assign_new_index(t);
  z->clk_index = clock_indexes;
*/
  if(c) {
    for(x = c->next; ; x = x->next) {
      if(x->op == TRUE) {
        z->relop[0][0] = TRUE;
        return z;
      }
      if(x->op == FALSE) {
        z->relop[0][0] = FALSE;
        return z;
      }
      if(x->clk1 != 0 && x->clk2 == 0) {
        if(x->op == LESS || x->op == LEQ || x->op == EQ) {
          z->relop[new_index(x->clk1, z)][0] = x->op;
          z->diff[new_index(x->clk1, z)][0] = x->diff;
        }
        else if(x->op == GREATER || x->op == GEQ) {
          z->relop[0][new_index(x->clk1, z)] = reverse_op(x->op);
          z->diff[0][new_index(x->clk1, z)] = -(x->diff);
        }
      }
      else if(x->clk1 == 0 && x->clk2 != 0) {
        if(x->op == LESS || x->op == LEQ || x->op == EQ) {
          z->relop[0][new_index(x->clk2, z)] = x->op;
          z->diff[0][new_index(x->clk2, z)] = x->diff;
        }
        else if(x->op == GREATER || x->op == GEQ) {
          z->relop[new_index(x->clk2, z)][0] = reverse_op(x->op);
          z->diff[new_index(x->clk2, z)][0] = -(x->diff);
        }
      }
      else if(x->clk1 != 0 && x->clk2 != 0) {
        if(x->op == LESS || x->op == LEQ || x->op == EQ) {
          z->relop[new_index(x->clk1, z)][new_index(x->clk2, z)] = x->op;
          z->diff[new_index(x->clk1, z)][new_index(x->clk2, z)] = x->diff;
        }
        else if(x->op == GREATER || x->op == GEQ) {
          z->relop[new_index(x->clk2, z)][new_index(x->clk1, z)] = reverse_op(x->op);
          z->diff[new_index(x->clk2, z)][new_index(x->clk1, z)] = -(x->diff);
        }
      }
      if(x == c) break;
    }
  }
  return z;
}

int new_index(int someclk, struct zone_type *z) {
  int i;

  if(someclk == 0) return 0;

  for(i=1; i<=z->clk_index[0]; i++) {
    if(z->clk_index[i] == someclk) {
      return i;
    }
  }
  fprintf(stderr, "Fatal Error: NOT found new index of %d.\n", someclk);
  return -1;         /* not found */
}

struct zone_type *zone(struct clkconj_type *c, struct thread_type *t) {
  int i, j, k;
  struct zone_type *z;

/*
  printf("Taking zone of clkconj:\n");
  print_clkconj(c);
  output("\n");
*/
  z = alloc_zone(t);
  z = init_zone(z, c, t);

  if(c->op == TRUE || c->op == FALSE) {
    return z;
  }

  z = zone_proc_leq(z);
/*
  output("After zone_proc_leq():\n");
  print_zone(z);
  output("\n");
*/

  z = zone_proc_eq(z);
/*
  output("After zone_proc_eq():\n");
  print_zone(z);
  output("\n");
*/

  z = kclosure_zone(z);
/*
  output("After kclosure_zone():\n");
  print_zone(z);
  output("\n");
*/

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(i==0 && j==0) continue;
      for(k=0; k<=z->clk_index[0]; k++) { 
        if(z->relop[i][j]!=-1 && z->relop[i][k]!=-1 && z->relop[k][j]!=-1) {
          if(z->diff[i][j] > z->diff[i][k] + z->diff[k][j]) {
            z->diff[i][j] = z->diff[i][k] + z->diff[k][j];
            z->relop[i][j] = transitive_op(z->relop[i][k], z->relop[k][j]);
          }
        }
      }
    }
  }
/*
  for(i=1; i<=z->clk_index[0]; i++) {
    for(j=1; j<=z->clk_index[0]; j++) {
      if(z->relop[i][0]!=-1 && z->relop[i][j]!=-1 && z->relop[j][0]!=-1) {
        if(z->diff[i][0] > z->diff[i][j] + z->diff[j][0]) {
          z->diff[i][0] = z->diff[i][j] + z->diff[j][0];
          z->relop[i][0] = transitive_op(z->relop[i][j], z->relop[j][0]);
        }
      }
    }
  }
  for(i=1; i<=z->clk_index[0]; i++) {
    for(j=1; j<=z->clk_index[0]; j++) {
      if(z->relop[0][i]!=-1 && z->relop[j][i]!=-1 && z->relop[0][j]!=-1) {
        if(z->diff[0][i] > z->diff[j][i] + z->diff[0][j]) {
          z->diff[0][i] = z->diff[j][i] + z->diff[0][j];
          z->relop[0][i] = transitive_op(z->relop[j][i], z->relop[0][j]);
        }
      }
    }
  }
*/

/*
  if(zone_reduction)
    z = normalize_zone(z);
*/

/*
  z = process_zone_max_const(z);
*/
  return z;
}

void free_zone(struct zone_type *z) {
int i;

/*
  printf("Inside free_zone( "); print_zone(z); printf(" )\n");
*/

  if(z) {
    if(z->clk_index == NULL) {
      free(z->relop[0]);
      free(z->diff[0]);
      free(z->relop);
      free(z->diff);
      z->relop = NULL;
      z->diff = NULL;
      free(z);
      return;
    }
    if(z->relop) {
      for(i=0; i<=z->clk_index[0]; i++) {
        free(z->relop[i]);
        z->relop[i] = NULL;
      }
      free(z->relop);
      z->relop = NULL;
    }
    if(z->diff) {
      for(i=0; i<=z->clk_index[0]; i++) {
        free(z->diff[i]);
        z->diff[i] = NULL;
      }
      free(z->diff);
      z->diff = NULL;
    }
    free(z->clk_index);
    free(z);
  }
/*
  printf("Freed zone.\n");
*/
}

/*
struct zone_type *zone_and(struct zone_type *z1, struct zone_type *z2) {
  struct zone_type *z;

  switch(z1->relop[0][0]) {
    case TRUE:
      switch(z2->relop[0][0]) {
        case TRUE: 	z = alloc_zone();
                        z->relop[0][0] = TRUE;
                   	break;
        case FALSE: 	z = alloc_zone();
                        z->relop[0][0] = FALSE;
                    	break;
        case -1: 	z = zone_copy(z2);
                 	break;
      }
    case FALSE:         z = alloc_zone();
                        z->relop[0][0] = FALSE;
      			break;
    case -1:
      switch(z2->relop[0][0]) {
        case TRUE: 	z = zone_copy(z1);
                   	break;
        case FALSE: 	z = alloc_zone();
                        z->relop[0][0] = FALSE;
                    	break;
        case -1: 	z = alloc_zone();
                        z = non_empty_zone_and(z, z1, z2);
                 	break;
      }
  }
  return z;
}
*/
    

struct zone_type *zone_copy(struct zone_type *z) {
int i, j;
struct zone_type *zc = alloc_zone(NULL);

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      zc->relop[i][j] = z->relop[i][j];
      zc->diff[i][j] = z->diff[i][j];
    }
  }
  return zc;
}


/*
struct zone_type *
non_empty_zone_and( struct zone_type *z, 
                    struct zone_type *z1, 
                    struct zone_type *z2) {
int i, j;

  for(i=1; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(z1->relop[i][j] == -1 && z2->relop[i][j] == -1) continue;
      if(z1->relop[i][j] != -1 && z2->relop[i][j] == -1) {
        z->relop[i][j] = z1->relop[i][j];
        z->diff[i][j] = z1->diff[i][j];
      }
      if(z1->relop[i][j] == -1 && z2->relop[i][j] != -1) {
        z->relop[i][j] = z2->relop[i][j];
        z->diff[i][j] = z2->diff[i][j];
      }
      if(z1->relop[i][j] != -1 && z2->relop[i][j] != -1) {
        z->relop[i][j] = transitive_op(z1->relop[i][j], z2->relop[i][j]);
        z->diff[i][j] =   (z1->diff[i][j] <= z2->diff[i][j]) 
                        ? (z1->diff[i][j]) : (z2->diff[i][j]);
      }
    }
  }
  return z;
}
*/

int clk_index_compare(struct zone_type *z1, struct zone_type *z2) {
  int i, j, found = FALSE;

  if(z1 == NULL && z2 == NULL) return EQUAL;
  if(z1 == NULL && z2 != NULL) return CONTAINS;
  if(z1 != NULL && z2 == NULL) return CONTAINED;

  if(z1->clk_index == NULL && z2->clk_index == NULL) return EQUAL;
  if(z1->clk_index == NULL && z2->clk_index != NULL) return CONTAINS;
  if(z1->clk_index != NULL && z2->clk_index == NULL) return CONTAINED;

  if(z1->clk_index[0] < z2->clk_index[0]) return CONTAINS;
  if(z1->clk_index[0] > z2->clk_index[0]) return CONTAINED;
  for(i=1; i <= z1->clk_index[0]; i++) {
    for(j=1; j <= z2->clk_index[0]; j++) {
      if(z2->clk_index[j] == z1->clk_index[i]) {
        found = TRUE;
        break;
      }
    } 
    if(found == FALSE) return clk_index_diff(z1, z2);
    else found = FALSE;
  }
  return EQUAL;
}

int clk_index_diff(struct zone_type *z1, struct zone_type *z2) {
  int i;
  for(i=1; i <= z1->clk_index[0]; i++) {
    if(z1->clk_index[i] < z2->clk_index[i]) return CONTAINS;
    if(z1->clk_index[i] > z2->clk_index[i]) return CONTAINED;
  }
  return EQUAL;
}

int zone_comp1(struct zone_type *z1, struct zone_type *z2) {
  int i, j, i2, j2;

  if(!z1 && !z2) return EQUAL;
  if(z1 && !z2) return CONTAINED;
  if(!z1 && z2) return CONTAINS;

  i = clk_index_compare(z1, z2);
  if(i != EQUAL) return i;
  
  if(z1->relop[0][0] == TRUE && z2->relop[0][0] == TRUE) return EQUAL;
  if(z1->relop[0][0] == FALSE && z2->relop[0][0] == FALSE) return EQUAL;
  if(z1->relop[0][0] == TRUE && z2->relop[0][0] == FALSE) return CONTAINS;
  if(z1->relop[0][0] == FALSE && z2->relop[0][0] == TRUE) return CONTAINED;

  if(both_out_of_bounds(z1, z2)) {
    boob = TRUE; 
    return EQUAL;
  }
  for(i=0; i<=z1->clk_index[0]; i++) {
    for(j=0; j<=z1->clk_index[0]; j++) {
      if(i==0 && j==0) continue;
      i2 = new_index(old_index(i, z1), z2);
      j2 = new_index(old_index(j, z1), z2);
/*
      if(i == 0 && z1->relop[i][j] == LEQ && z1->diff[i][j] == 0 
                && z2->relop[i2][j2] == -1) continue;
      if(i == 0 && z2->relop[i2][j2] == LEQ && z2->diff[i2][j2] == 0 
                && z1->relop[i][j] == -1) continue;
*/
      if(z1->relop[i][j] == -1 && z2->relop[i2][j2] == -1) continue;
      if(z1->relop[i][j] != -1 && z2->relop[i2][j2] == -1) {
        return CONTAINED;
      }
      if(z1->relop[i][j] == -1 && z2->relop[i2][j2] != -1) {
        return CONTAINS;
      }
      if(z1->relop[i][j] != -1 && z2->relop[i2][j2] != -1) {
        if(z1->relop[i][j] < z2->relop[i2][j2]) return LATER;
        else if(z1->relop[i][j] > z2->relop[i2][j2]) return EARLIER;
        else {      		/* z1->relop[i][j] == z2->relop[i2][j2] */
          if(z1->diff[i][j] < z2->diff[i2][j2]) return LATER;
          else if(z1->diff[i][j] > z2->diff[i2][j2]) return EARLIER;
          else continue; 	/* z1->diff[i][j] == z2->diff[i2][j2] */
        }
      }
    }
  }
  return EQUAL;
}


int zone_comp(struct zone_type *z1, struct zone_type *z2) {
  int i, j, i2, j2;
  int ans = EQUAL;

  if(!z1 && !z2) return EQUAL;
  if(z1 && !z2) return CONTAINS;
  if(!z1 && z2) return CONTAINED;

  i = clk_index_compare(z1, z2);
  if(i != EQUAL) return i;

  if(z1->relop[0][0] == TRUE && z2->relop[0][0] == TRUE) return EQUAL;
  if(z1->relop[0][0] == FALSE && z2->relop[0][0] == FALSE) return EQUAL;
  if(z1->relop[0][0] == TRUE && z2->relop[0][0] == FALSE) return CONTAINS;
  if(z1->relop[0][0] == FALSE && z2->relop[0][0] == TRUE) return CONTAINED;

  if(both_out_of_bounds(z1, z2)) {
    boob = TRUE;
/*
    sprintf(tmpstring,
            "(max_const = %d) Both zones out of bounds: ", max_const); 
    output(tmpstring);
    print_zone(z1);
    output(" AND "); print_zone(z2); output("\n");
*/
    return EQUAL;
  }
  for(i=0; i<=z1->clk_index[0]; i++) {
    for(j=0; j<=z1->clk_index[0]; j++) {
      if(i==0 && j==0) continue;
      i2 = new_index(old_index(i, z1), z2);
      j2 = new_index(old_index(j, z1), z2);
      if(i == 0 && z1->relop[i][j] == LEQ && z1->diff[i][j] == 0 
                && z2->relop[i2][j2] == -1) continue;
      if(i == 0 && z2->relop[i2][j2] == LEQ && z2->diff[i2][j2] == 0 
                && z1->relop[i][j] == -1) continue;
      if(z1->relop[i][j] == -1 && z2->relop[i2][j2] == -1) continue;
      if(z1->relop[i][j] != -1 && z2->relop[i2][j2] == -1) {
        if(z1->relop[i][j] == EQ && z1->diff[i][j] == 0) {
          if(z2->relop[j2][i2] == LESS && z2->diff[j2][i2] > 0) {
            if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
            else ans = CONTAINED;
          }
          else if(z2->relop[j2][i2] == LEQ && z2->diff[j2][i2] >= 0) {
            if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
            else ans = CONTAINED;
          }
          else if(z2->relop[j2][i2] == EQ && z2->diff[j2][i2] == 0) 
            continue;
          else
            ans = INCOMPARABLE;
        }
        else {
          if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
          else ans = CONTAINED;
        }
      }
      if(z1->relop[i][j] == -1 && z2->relop[i2][j2] != -1) {
        if(z2->relop[i2][j2] == EQ && z2->diff[i2][j2] == 0) {
          if(z1->relop[j][i] == LESS && z1->diff[j][i] > 0) {
            if(ans == CONTAINED || ans == INCOMPARABLE) ans = INCOMPARABLE;
            else ans = CONTAINS;
          }
          else if(z1->relop[j][i] == LEQ && z1->diff[j][i] >= 0) {
            if(ans == CONTAINED || ans == INCOMPARABLE) ans = INCOMPARABLE;
            else ans = CONTAINS;
          }
          else if(z1->relop[j][i] == EQ && z1->diff[j][i] == 0) 
            continue;
          else
            ans = INCOMPARABLE;
        }
        else {
          if(ans == CONTAINED || ans == INCOMPARABLE) ans = INCOMPARABLE;
          else ans = CONTAINS;
        }
      }
      if(z1->relop[i][j] != -1 && z2->relop[i2][j2] != -1) {
/*
        sprintf(tmpstring, "ans = %d, comparing: i=%d and j=%d, ", ans, i, j);
        output(tmpstring);
*/
        switch(z1->relop[i][j]) {
          case LESS:
            switch(z2->relop[i2][j2]) {
              case LESS:
                if(z1->diff[i][j]<z2->diff[i2][j2]) {
                  if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i2][j2]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = INCOMPARABLE;
                }
                break;
              case LEQ:
                if(z1->diff[i][j]<=z2->diff[i2][j2]) {
                  if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i2][j2]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = INCOMPARABLE;
                }
                break;
              case EQ:
                if(z1->diff[i][j]<=z2->diff[i2][j2]) {
                  ans = INCOMPARABLE;
                }
                else if(z1->diff[i][j]>z2->diff[i2][j2]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = INCOMPARABLE;
                }
                break;
            }
            break;
          case LEQ:
            switch(z2->relop[i2][j2]) {
              case LESS:
                if(z1->diff[i][j]<z2->diff[i2][j2]) {
                  if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>=z2->diff[i2][j2]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = INCOMPARABLE;
                }
                break;
              case LEQ:
                if(z1->diff[i][j]<z2->diff[i2][j2]) {
                  if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i2][j2]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = INCOMPARABLE;
                }
                break;
              case EQ:
                if(z1->diff[i][j]<z2->diff[i2][j2]) {
                  ans = INCOMPARABLE;
                }
                else if(z1->diff[i][j]>=z2->diff[i2][j2]) {
                  if(ans == EQUAL || ans == CONTAINS) ans = CONTAINS;
                  else ans = INCOMPARABLE;
                }
                break;
            }
            break;
          case EQ:
            switch(z2->relop[i2][j2]) {
              case LESS:
                if(z1->diff[i][j]<z2->diff[i2][j2]) {
                  if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>=z2->diff[i2][j2]) {
                  ans = INCOMPARABLE;
                }
                break;
              case LEQ:
                if(z1->diff[i][j]<=z2->diff[i2][j2]) {
                  if(ans == CONTAINS || ans == INCOMPARABLE) ans = INCOMPARABLE;
                  else ans = CONTAINED;
                }
                else if(z1->diff[i][j]>z2->diff[i2][j2]) {
                  ans = INCOMPARABLE;
                }
                break;
              case EQ:
                if(z1->diff[i][j] != z2->diff[i2][j2]) {
                  ans = INCOMPARABLE;
                }
                break;
            }
            break;
        } /* end switch */
/*
        sprintf(tmpstring, "ans = %d\n", ans);
        output(tmpstring);
*/
      } 
    }
  }

  if(ans == INCOMPARABLE) {
    for(i=0; i<=z1->clk_index[0]; i++) {
      for(j=0; j<=z1->clk_index[0]; j++) {
        if(i==0 && j==0) continue;
        i2 = new_index(old_index(i, z1), z2);
        j2 = new_index(old_index(j, z1), z2);
        if(z1->relop[i][j] != -1 && z2->relop[i2][j2] == -1) 
          return LATER;
        if(z1->relop[i][j] == -1 && z2->relop[i2][j2] != -1) 
          return EARLIER;
        if(z1->relop[i][j] != -1 && z2->relop[i2][j2] != -1) {
          if(z1->relop[i][j] < z2->relop[i2][j2]) return LATER;
          else if(z1->relop[i][j] > z2->relop[i2][j2]) return EARLIER;
          else {      		/* z1->relop[i][j] == z2->relop[i2][j2] */
            if(z1->diff[i][j] < z2->diff[i2][j2]) return LATER;
            else if(z1->diff[i][j] > z2->diff[i2][j2]) return EARLIER;
            else continue; 	/* z1->diff[i][j] == z2->diff[i2][j2] */
          }
        }
      }
    }
  }
/*
  output("ZONE comparison: ");
  print_zone(z1); output(" AND "); print_zone(z2);
  sprintf(tmpstring, " IS %d.\n", ans); 
  output(tmpstring);
*/
  return ans;
}

int install_clkconj_zones() {
  struct thread_type *t;
  struct mode_type *m;
  int i;
/*
  int tmp_zone_reduction = zone_reduction;

  zone_reduction = FALSE;
*/
  for(i=0, t = &thread[i]; i<system_ta_num; t = &thread[++i]) {
    for(m = t->starting_mode; m; m = m->cnext) {
      if(m->xconj_cycle) {
        m->zone = zone(m->xconj_cycle, t);
        if(verbose >= 2) {
          sprintf(tmpstring,
                  "Installed zone for clock invariant of mode %s: ", m->name);
          output(tmpstring);
          print_zone(m->zone);
          output("\n");
        }
      }
      else {
       m->xconj_cycle = alloc_clkconj();
       m->xconj_cycle->op = TRUE;
       m->zone = zone(m->xconj_cycle, t);
      }
    }
  }
/*
  zone_reduction = tmp_zone_reduction;
*/
  return 0;
}

int out_of_bound(struct zone_type *z) {
  int i, j;

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(z->diff[i][j] > max_const) 
        return TRUE;
      if(z->diff[i][j] < -max_const) 
        return TRUE;
      if(z->diff[i][j] == -max_const && z->relop[i][j] == LESS) 
        return TRUE;
    }
  }
  return FALSE;
}

struct zone_type *process_zone_max_const(struct zone_type *z1) {
  int i, j;

/*
  for(i=0; i<=z1->clk_index[0]; i++) {
    for(j=0; j<=z1->clk_index[0]; j++) {
      if(z->relop[i][j] != -1) {
        if(z->diff[i][j] > max_const) z->diff[i][j] = INF;
        if(z->diff[i][j] < -max_const) z->diff[i][j] = -INF;
      }
    }
  }
*/
  for(i=0; i<=z1->clk_index[0]; i++) 
    if (   (z1->diff[0][i] < -max_const && z1->relop[0][i] == LEQ)
	|| (z1->diff[0][i] <= -max_const && z1->relop[0][i] == LESS)
	) {
      z1->relop[0][i] = LESS; 
      z1->diff[0][i] = -max_const; 
      z1->relop[i][0] = -1; 
      z1->diff[i][0] = 0; 
      for (j = 1; j <= z1->clk_index[0]; j++) 
	if (i != j) {
	  z1->relop[i][j] = -1; 
	  z1->relop[j][i] = -1; 
	  z1->diff[i][j] = 0; 
	  z1->diff[j][i] = 0; 
	}
    }
  z1 = zone_proc_leq(z1);
  z1 = zone_proc_eq(z1);
  return z1;
}

int both_out_of_bounds(struct zone_type *z1, struct zone_type *z2) {
  int i, j, ans = FALSE;

/*
  for(i=0; i<=z1->clk_index[0]; i++) {
    for(j=0; j<=z1->clk_index[0]; j++) {
      if(i==j) continue;
      if(z1->relop[i][j] != -1 && z1->diff[i][j] > max_const) 
        z1->diff[i][j] = max_const;
      if(z1->relop[i][j] != -1 && z1->diff[i][j] < -max_const) 
        z1->diff[i][j] = -max_const;

      if(z2->relop[i][j] != -1 && z2->diff[i][j] > max_const) 
        z2->diff[i][j] = max_const;
      if(z2->relop[i][j] != -1 && z2->diff[i][j] < -max_const) 
        z2->diff[i][j] = -max_const;
    }
  }
*/
  for(i=0; i<=z1->clk_index[0]; i++) {
    for(j=0; j<=z1->clk_index[0]; j++) {
      if(i==j) continue;
      if((z1->relop[i][j] != -1) && (z2->relop[i][j] != -1)) {
        if(z1->diff[i][j] > max_const && z2->diff[i][j] > max_const) {
          ans |= TRUE;
          continue;
        }
        if(z1->diff[i][j] < -max_const && z2->diff[i][j] < -max_const) {
          ans |= TRUE;
          continue;
        }
        if(z1->diff[i][j] <= -max_const && z1->relop[i][j] == LESS &&
           z2->diff[i][j] <= -max_const && z2->relop[i][j] == LESS) {
          ans |= TRUE;
          continue;
        }
/*
        if(z1->diff[i][j] < -max_const && z1->relop[i][j] == LEQ &&
           z2->diff[i][j] < -max_const && z2->relop[i][j] == LEQ) {
          ans |= TRUE;
          continue;
        }
*/
        if(z1->relop[i][j] == z2->relop[i][j] &&
           z1->diff[i][j] == z2->diff[i][j])
          continue;
        return FALSE;       /* a single term different => different zones */
      }
    }
  }
  return ans;
}

struct zone_type *normalize_zone(struct zone_type *z) {

  if(z->relop[0][0] == TRUE || z->relop[0][0] == FALSE) return z;

  z = instantiate_zone(z);
  if(z->relop[0][0] == FALSE) return z;

  z = init_reduce_zone(z);
  if(z->relop[0][0] == FALSE) return z;

  z = substitute_zone(z);
  if(z->relop[0][0] == FALSE) return z;

  z = init_reduce_zone(z);
  if(z->relop[0][0] == FALSE) return z;

/* newly added */
/*
  z = substitute_zone(z);                 
  if(z->relop[0][0] == FALSE) return z;
*/

  z = instantiate_zone(z);
  if(z->relop[0][0] == FALSE) return z;

  z = reduce_zone(z);

  return z;
}

/* reduce x[i] - x[j] <= diff    AND    x[j] - x[i] <= -(diff)
     INTO               x[i] - x[j] = diff
*/
struct zone_type *init_reduce_zone(struct zone_type *z) {
int i, j;

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(z->relop[i][j] == LEQ && z->relop[j][i] == LEQ) {
        if(z->diff[i][j] == -(z->diff[j][i])) {
          if( (z->diff[i][j] > 0) || (z->diff[i][j] == 0 && i < j) ) {
            z->relop[i][j] = EQ;
            z->relop[j][i] = -1;
            z->diff[j][i] = 0;
          }
          else if( (z->diff[i][j] < 0) || (z->diff[i][j] == 0 && i > j) ) {
            z->relop[j][i] = EQ;
            z->relop[i][j] = -1;
            z->diff[i][j] = 0;
          }
        }
      }
    }
  }
/*
  output("After init_reduce_zone: ");
  print_zone(z); output("\n");
*/
  return z;
}

struct zone_type *substitute_zone(struct zone_type *z) {
  int i, j, k, m, n;

  /* if x[i] - x[j] = 0  and  i < j then substitute all x[j] for x[i] 
  */
  for(i=1; i<=z->clk_index[0]; i++) {
    for(j=1; j<=z->clk_index[0]; j++) {

      if(z->relop[i][j] == EQ && z->diff[i][j] == 0) {
        if(i > j) {
          if(z->relop[j][i] == -1) {
            z->relop[i][j] = -1;
            z->relop[j][i] = EQ;
            z->diff[j][i] = 0;
          }
          else {
            switch(z->relop[j][i]) {
              case LESS:
                if(z->diff[j][i] <= 0) {
                  z->relop[0][0] = FALSE;
                  return z;
                }
                break;
              case LEQ:
                if(z->diff[j][i] < 0) {
                  z->relop[0][0] = FALSE;
                  return z;
                }
                break;
              case EQ:
                if(z->diff[j][i] != 0) {
                  z->relop[0][0] = FALSE;
                  return z;
                }
                break;
            } /* end switch */
            z->relop[i][j] = -1;
            z->relop[j][i] = EQ;
            z->diff[j][i] = 0;
          } /* end else */
        } /* end if(i > j) */
        m = (i > j) ? i : j; /* m = max(i, j); */
        n = (i < j) ? i : j; /* n = min(i, j); */
        for(k=0; k<=z->clk_index[0]; k++) {
          if(k == m || k == n) continue;
                                /* (x[m] - x[k] ~ d) ---> (x[n] - x[k] ~ d) */
          if(z->relop[m][k] != -1) {
            if(z->relop[n][k] == -1) {
              z->relop[n][k] = z->relop[m][k];
              z->diff[n][k] = z->diff[m][k];
              z->relop[m][k] = -1;
              z->diff[m][k] = 0;
            }
            else {
              switch(z->relop[m][k]) {
                case LESS:
                  switch(z->relop[n][k]) {
                    case LESS:
                      if(z->diff[m][k] < z->diff[n][k]) {
                        z->diff[n][k] = z->diff[m][k];
                      }
                      break;
                    case LEQ:
                      if(z->diff[m][k] <= z->diff[n][k]) {
                        z->relop[n][k] = LESS;
                        z->diff[n][k] = z->diff[m][k];
                      }
                      break;
                    case EQ:
                      if(z->diff[m][k] <= z->diff[n][k]) {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                  }
                  break;
                case LEQ:
                  switch(z->relop[n][k]) {
                    case LESS:
                      if(z->diff[m][k] < z->diff[n][k]) {
                        z->relop[n][k] = LEQ;
                        z->diff[n][k] = z->diff[m][k];
                      }
                      break;
                    case LEQ:
                      if(z->diff[m][k] < z->diff[n][k]) {
                        z->diff[n][k] = z->diff[m][k];
                      }
                      break;
                    case EQ:
                      if(z->diff[m][k] < z->diff[n][k]) {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                  }
                  break;
                case EQ:
                  switch(z->relop[n][k]) {
                    case LESS:
                      if(z->diff[m][k] < z->diff[n][k]) {
                        z->relop[n][k] = EQ;
                        z->diff[n][k] = z->diff[m][k];
                      }
                      else {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                    case LEQ:
                      if(z->diff[m][k] <= z->diff[n][k]) {
                        z->relop[n][k] = EQ;
                        z->diff[n][k] = z->diff[m][k];
                      }
                      else {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                    case EQ:
                      if(z->diff[m][k] != z->diff[n][k]) {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                  }
                  break;
              } /* end switch */
              z->relop[m][k] = -1;
              z->diff[m][k] = 0;
            } /* end else */
          }
                                /* (x[k] - x[m] ~ d) ---> (x[k] - x[n] ~ d) */
          if(z->relop[k][m] != -1) {
            if(z->relop[k][n] == -1) {
              z->relop[k][n] = z->relop[k][m];
              z->diff[k][n] = z->diff[k][m];
              z->relop[k][m] = -1;
              z->diff[k][m] = 0;
            }
            else {
              switch(z->relop[k][m]) {
                case LESS:
                  switch(z->relop[k][n]) {
                    case LESS:
                      if(z->diff[k][m] < z->diff[k][n]) {
                        z->diff[k][n] = z->diff[k][m];
                      }
                      break;
                    case LEQ:
                      if(z->diff[k][m] <= z->diff[k][n]) {
                        z->relop[k][n] = LESS;
                        z->diff[k][n] = z->diff[k][m];
                      }
                      break;
                    case EQ:
                      if(z->diff[k][m] <= z->diff[k][n]) {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                  }
                  break;
                case LEQ:
                  switch(z->relop[k][n]) {
                    case LESS:
                      if(z->diff[k][m] < z->diff[k][n]) {
                        z->relop[k][n] = LEQ;
                        z->diff[k][n] = z->diff[k][m];
                      }
                      break;
                    case LEQ:
                      if(z->diff[k][m] < z->diff[k][n]) {
                        z->diff[k][n] = z->diff[k][m];
                      }
                      break;
                    case EQ:
                      if(z->diff[k][m] < z->diff[k][n]) {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                  }
                  break;
                case EQ:
                  switch(z->relop[k][n]) {
                    case LESS:
                      if(z->diff[k][m] < z->diff[k][n]) {
                        z->relop[k][n] = EQ;
                        z->diff[k][n] = z->diff[k][m];
                      }
                      else {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                    case LEQ:
                      if(z->diff[k][m] <= z->diff[k][n]) {
                        z->relop[k][n] = EQ;
                        z->diff[k][n] = z->diff[k][m];
                      }
                      else {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                    case EQ:
                      if(z->diff[k][m] != z->diff[k][n]) {
                        z->relop[0][0] = FALSE;
                        return z;
                      }
                      break;
                  }
                  break;
              } /* end switch */
              z->relop[k][m] = -1;
              z->diff[k][m] = 0;
            } /* end else */
          }
        }
      }

    }
  }

/*
  output("After substitute_zone: ");
  print_zone(z); output("\n");
*/
  return z;
}

/* minimize clock constraints in zone: 3 reduce_zone functions 
 */

/*   first reduce_zone, using recursive: problem is deadlock occurs */
/*
struct zone_type *reduce_zone(struct zone_type *z) {
int i, j;  
int min, *adj;
adj = (int *)malloc(sizeof(int));
*adj = TRUE;

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(z->relop[i][j] != -1) {
        printf("Considering x[%d] - x[%d]:\n", i,j);
        min = get_min(z, i, j, 0, adj);
        printf("For x[%d]-x[%d], min = %d\n", i,j,min);
        if( (min < z->diff[i][j]) ||
            (min == z->diff[i][j] && *adj == FALSE) ) {
          z->relop[i][j] = -1;
          z->diff[i][j] = 0;
        }
      } 
    }
  }
  output("After reduce_zone: ");
  print_zone(z); output("\n");

  return z;
}

int get_min(struct zone_type *z, int i, int j, int sum, int *adj) {
int k, tmp, noway = TRUE, min = 1000;
  
  printf("getting min of x[%d] --> x[%d]:\n", i,j);
  for(k=0; k<=z->clk_index[0]; k++) {
    if(z->relop[i][k] != -1) {
      if(noway) noway = FALSE;
      if(k == j) {
        if(min > sum + z->diff[i][k]) 
          min = (sum + z->diff[i][k]);
      }
      else {
        tmp = get_min(z, k, j, (sum + (z->diff[i][k])), adj);
        if(min > tmp) min = tmp;
      }
    }
  }
  if(noway) return 2000;
  printf("min[ x[%d] --> x[%d] = %d\n",i,j,min);
  return min;
}
*/

/* second reduce_zone: problem is Dijkstra's algorithm is only for +ve edge
   weights */
/*
struct zone_type *reduce_zone(struct zone_type *z) {
int i, j, k, l;
int min, min_id, first_iter = TRUE, adj = FALSE;
int visited[z->clk_index[0] + 1];
int sum[z->clk_index[0] + 1];
int all_visited = FALSE;


  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      for(k=0; k<=z->clk_index[0]; k++) {
        visited[k] = FALSE;
        sum[k] = 1000;
      }
      if(z->relop[i][j] != -1) {
        printf("Considering: x[%d] - x[%d] %s %d\n", 
             i, j, relop_string(z->relop[i][j]), z->diff[i][j]); 
        fflush(stdout);
        visited[i] = TRUE;
        sum[i] = 0;
        for(k=0; k<=z->clk_index[0]; k++) {
          if(z->relop[i][k] != -1) {
            sum[k] = z->diff[i][k];
            printf("sum[%d] = %d, ", k, sum[k]); 
          }
        }
        printf("\n"); fflush(stdout);
        while(!all_visited) {
          all_visited = TRUE;
          min = 1000;
          for(k=0; k<=z->clk_index[0]; k++) {
            if(visited[k]) {
              for(l=0; l<=z->clk_index[0]; l++) {
                if( (z->relop[k][l] != -1) && (visited[l] == FALSE) ) {
                  if(min > sum[l]) {
                    min = sum[l];
                    min_id = l;
                  }
                }
              }
            }
          }
          printf("min = %d, min_id = %d\n", min, min_id); fflush(stdout);
          visited[min_id] = TRUE;
          if(first_iter && min_id == j) adj = TRUE;
          else if(first_iter) first_iter = FALSE;
          printf("adj = %d\n", adj);
          for(l=0; l<=z->clk_index[0]; l++) {
            if( (visited[l] == FALSE) && (z->relop[min_id][l] != -1) ) { 
              all_visited = FALSE;
              printf("reevaluating: sum[%d]\n", l); fflush(stdout);
              if(min + z->diff[min_id][l] < sum[l])
                sum[l] = min + z->diff[min_id][l];
            }
          }
        } * end while *
      } * end if(z->relop[i][j] != -1) *
      all_visited = FALSE;
    }
    if( (adj == FALSE && z->diff[i][j] == sum[j]) ||
        (z->diff[i][j] > sum[j]) ) {
      z->relop[i][j] = -1;
      z->diff[i][j] = 0;
    }
  }
  return z;
}
*/

/* third reduce_zone: brute force */

struct zone_type *reduce_zone(struct zone_type *z) {
int i, j, k, l, m, n, o, p, q, r, sum;

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(z->relop[i][j] != -1) {
        for(k=0; k<=z->clk_index[0]; k++) {
          if(z->relop[i][k]!=-1) {
            if(z->relop[k][j]!=-1) {
              if( (z->diff[i][j] > z->diff[i][k] + z->diff[k][j]) 
                ||
                  ( z->diff[i][j] == (z->diff[i][k] + z->diff[k][j])
                                 &&
                    (z->relop[i][j] >= transitive_op(z->relop[i][k], 
                                                   z->relop[k][j])
                    )
                  )
                ) {
                z->relop[i][j] = -1;
                z->diff[i][j] = 0;
                break;
              } 
            }
          }
        } /* end for(k=0; k<=z->clk_index[0]; k++) */
      } /* end if(z->relop[i][j] != -1) */
    }
  }

  if(composed_ta > 2) {
    for(i=0; i<=z->clk_index[0]; i++) {
      for(j=0; j<=z->clk_index[0]; j++) {
        if(z->relop[i][j] != -1) {
          for(k=0; k<=z->clk_index[0]; k++) {
            if(k == j) continue;
            if(z->relop[i][k]!=-1) {
              for(l=0; l<=z->clk_index[0]; l++) {
                if(l == j) continue;
                if(z->relop[k][l]!=-1) {
                  if(z->relop[l][j]!=-1) {
                    sum = z->diff[i][k] + z->diff[k][l] + z->diff[l][j];
                    if(   (z->diff[i][j] > sum ) ||
                         ((z->diff[i][j] == sum )
                       && (z->relop[i][j] >= 
                           transitive_op( transitive_op(z->relop[i][k], 
                                                        z->relop[k][l]),
                                          z->relop[l][j])))
                      ) {
                      z->relop[i][j] = -1;
                      z->diff[i][j] = 0;
                      break;
                    }
                  } 
                }
              }
              if(z->relop[i][j] == -1) break;
            }
          }
        }
      }
    }
  }
  
  if(composed_ta > 3) {
    for(i=0; i<=z->clk_index[0]; i++) {
      for(j=0; j<=z->clk_index[0]; j++) {
        if(z->relop[i][j] != -1) {
          for(k=0; k<=z->clk_index[0]; k++) {
            if(k == j) continue;
            if(z->relop[i][k]!=-1) {
              for(l=0; l<=z->clk_index[0]; l++) {
                if(l == j) continue;
                if(z->relop[k][l]!=-1) {
                  for(m=0; m<=z->clk_index[0]; m++) {
                    if(m == j) continue;
                    if(z->relop[l][m]!=-1) {
                      if(z->relop[m][j]!=-1) {
                        sum = z->diff[i][k] + z->diff[k][l] + z->diff[l][m] +
                              z->diff[m][j];
                        if( (z->diff[i][j] > sum ) ||
                          ( (z->diff[i][j] == sum )
                            && (z->relop[i][j] >=
                                transitive_op( transitive_op(z->relop[i][k],
                                                             z->relop[k][l]),
                                               transitive_op(z->relop[l][m],
                                                             z->relop[m][j])))))
                        {
                          z->relop[i][j] = -1;
                          z->diff[i][j] = 0;
                          break;
                        }
                      }
                    }
                  }
                  if(z->relop[i][j] == -1) break;
                }
              }
              if(z->relop[i][j] == -1) break;
            }
          }
        }
      }
    }
  }

  if(composed_ta > 4) {
    for(i=0; i<=z->clk_index[0]; i++) {
      for(j=0; j<=z->clk_index[0]; j++) {
        if(z->relop[i][j] != -1) {
          for(k=0; k<=z->clk_index[0]; k++) {
            if(z->relop[i][k]!=-1) {
              for(l=0; l<=z->clk_index[0]; l++) {
                if(z->relop[k][l]!=-1) {
                  for(m=0; m<=z->clk_index[0]; m++) {
                    if(z->relop[l][m]!=-1) {
                      for(n=0; n<=z->clk_index[0]; n++) {
                        if(z->relop[m][n]!=-1) {
                          if(z->relop[n][j]!=-1) {
                            sum =  z->diff[i][k] + z->diff[k][l] + z->diff[l][m]
                                 + z->diff[m][n] + z->diff[n][j];
                            if( (z->diff[i][j] > sum) ||
                              ( (z->diff[i][j] == sum )
                                && (z->relop[i][j] >=
                                    transitive_op( transitive_op( 
                                                   transitive_op(z->relop[i][k],
                                                                 z->relop[k][l]),
                                                   transitive_op(z->relop[l][m],
                                                                 z->relop[m][n])
                                                   ), z->relop[n][j])))) {
                              z->relop[i][j] = -1;
                              z->diff[i][j] = 0;
                              break;
                            }
                          }
                        }
                      }
                      if(z->relop[i][j] == -1) break;
                    }
                  }
                  if(z->relop[i][j] == -1) break;
                }
              }
              if(z->relop[i][j] == -1) break;
            }
          }
        } /* end if(z->relop[i][j] != -1) */
      } 
    }
  }

  if(composed_ta > 5) {
   for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
     if(z->relop[i][j] != -1) {
      for(k=0; k<=z->clk_index[0]; k++) {
       if(z->relop[i][k]!=-1) {
        for(l=0; l<=z->clk_index[0]; l++) {
         if(z->relop[k][l]!=-1) {
          for(m=0; m<=z->clk_index[0]; m++) {
           if(z->relop[l][m]!=-1) {
            for(n=0; n<=z->clk_index[0]; n++) {
             if(z->relop[m][n]!=-1) {
              for(o=0; o<=z->clk_index[0]; o++) {
               if(z->relop[n][o]!=-1) {
                if(z->relop[o][j]!=-1) {
                  sum = z->diff[i][k] + z->diff[k][l] + z->diff[l][m]
                      + z->diff[m][n] + z->diff[n][o] + z->diff[o][j];
                  if( (z->diff[i][j] > sum) ||
                     (   (z->diff[i][j] == sum )
                      && (z->relop[i][j] >=
                          transitive_op( transitive_op( 
                                          transitive_op(z->relop[i][k],
                                                        z->relop[k][l]),
                                          transitive_op(z->relop[l][m],
                                                        z->relop[m][n])
                                         ), 
                                         transitive_op(z->relop[n][o],
                                                       z->relop[o][j])
                                        )
                         )
                     )
                    ) {
                      z->relop[i][j] = -1;
                      z->diff[i][j] = 0;
                      break;
                  }
                }
               }
              }
              if(z->relop[i][j] == -1) break;
             }
            }
            if(z->relop[i][j] == -1) break;
           }
          }
          if(z->relop[i][j] == -1) break;
         }
        }
        if(z->relop[i][j] == -1) break;
       }
      }
     } /* end if(z->relop[i][j] != -1) */
    } 
   }
  }

  if(composed_ta > 6) {
   for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
     if(z->relop[i][j] != -1) {
      for(k=0; k<=z->clk_index[0]; k++) {
       if(z->relop[i][k]!=-1) {
        for(l=0; l<=z->clk_index[0]; l++) {
         if(z->relop[k][l]!=-1) {
          for(m=0; m<=z->clk_index[0]; m++) {
           if(z->relop[l][m]!=-1) {
            for(n=0; n<=z->clk_index[0]; n++) {
             if(z->relop[m][n]!=-1) {
              for(o=0; o<=z->clk_index[0]; o++) {
               if(z->relop[n][o]!=-1) {
                for(p=0; p<=z->clk_index[0]; p++) {
                 if(z->relop[o][p]!=-1) {
                  if(z->relop[p][j]!=-1) {
                   sum = z->diff[i][k] + z->diff[k][l] + z->diff[l][m]
                       + z->diff[m][n] + z->diff[n][o] + z->diff[o][p]
                       + z->diff[p][j];
                   if( (z->diff[i][j] > sum) ||
                      (   (z->diff[i][j] == sum )
                       && (z->relop[i][j] >=
                           transitive_op( 
                             transitive_op( 
                               transitive_op(z->relop[i][k], z->relop[k][l]),
                               transitive_op(z->relop[l][m], z->relop[m][n])
                             ), 
                             transitive_op( 
                               transitive_op(z->relop[n][o], z->relop[o][p]),
                               z->relop[p][j]
                             )
                           )
                          )
                      )
                     ) {
                       z->relop[i][j] = -1;
                       z->diff[i][j] = 0;
                       break;
                   }
                  }
                 }
                }
                if(z->relop[i][j] == -1) break;
               }
              }
              if(z->relop[i][j] == -1) break;
             }
            }
            if(z->relop[i][j] == -1) break;
           }
          }
          if(z->relop[i][j] == -1) break;
         }
        }
        if(z->relop[i][j] == -1) break;
       }
      }
     } /* end if(z->relop[i][j] != -1) */
    } 
   }
  }

  if(composed_ta > 7) {
   for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
     if(z->relop[i][j] != -1) {
      for(k=0; k<=z->clk_index[0]; k++) {
       if(z->relop[i][k]!=-1) {
        for(l=0; l<=z->clk_index[0]; l++) {
         if(z->relop[k][l]!=-1) {
          for(m=0; m<=z->clk_index[0]; m++) {
           if(z->relop[l][m]!=-1) {
            for(n=0; n<=z->clk_index[0]; n++) {
             if(z->relop[m][n]!=-1) {
              for(o=0; o<=z->clk_index[0]; o++) {
               if(z->relop[n][o]!=-1) {
                for(p=0; p<=z->clk_index[0]; p++) {
                 if(z->relop[o][p]!=-1) {
                  for(q=0; q<=z->clk_index[0]; q++) {
                   if(z->relop[p][q]!=-1) {
                    if(z->relop[q][j]!=-1) {
                     sum = z->diff[i][k] + z->diff[k][l] + z->diff[l][m]
                         + z->diff[m][n] + z->diff[n][o] + z->diff[o][p]
                         + z->diff[p][q] + z->diff[q][j];
                     if( (z->diff[i][j] > sum) ||
                        (   (z->diff[i][j] == sum )
                         && (z->relop[i][j] >=
                             transitive_op( 
                               transitive_op( 
                                 transitive_op(z->relop[i][k], z->relop[k][l]),
                                 transitive_op(z->relop[l][m], z->relop[m][n])
                               ), 
                               transitive_op( 
                                 transitive_op(z->relop[n][o], z->relop[o][p]),
                                 transitive_op(z->relop[p][q], z->relop[q][j])
                               )
                             )
                            )
                        )
                       ) {
                         z->relop[i][j] = -1;
                         z->diff[i][j] = 0;
                         break;
                     }
                    }
                   }
                  }
                  if(z->relop[i][j] == -1) break;
                 }
                }
                if(z->relop[i][j] == -1) break;
               }
              }
              if(z->relop[i][j] == -1) break;
             }
            }
            if(z->relop[i][j] == -1) break;
           }
          }
          if(z->relop[i][j] == -1) break;
         }
        }
        if(z->relop[i][j] == -1) break;
       }
      }
     } /* end if(z->relop[i][j] != -1) */
    } 
   }
  }

  if(composed_ta > 8) {
   for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
     if(z->relop[i][j] != -1) {
      for(k=0; k<=z->clk_index[0]; k++) {
       if(z->relop[i][k]!=-1) {
        for(l=0; l<=z->clk_index[0]; l++) {
         if(z->relop[k][l]!=-1) {
          for(m=0; m<=z->clk_index[0]; m++) {
           if(z->relop[l][m]!=-1) {
            for(n=0; n<=z->clk_index[0]; n++) {
             if(z->relop[m][n]!=-1) {
              for(o=0; o<=z->clk_index[0]; o++) {
               if(z->relop[n][o]!=-1) {
                for(p=0; p<=z->clk_index[0]; p++) {
                 if(z->relop[o][p]!=-1) {
                  for(q=0; q<=z->clk_index[0]; q++) {
                   if(z->relop[p][q]!=-1) {
                    for(r=0; r<=z->clk_index[0]; r++) {
                     if(z->relop[q][r]!=-1) {
                      if(z->relop[r][j]!=-1) {
                       sum = z->diff[i][k] + z->diff[k][l] + z->diff[l][m]
                           + z->diff[m][n] + z->diff[n][o] + z->diff[o][p]
                           + z->diff[p][q] + z->diff[q][r] + z->diff[r][j];
                       if( (z->diff[i][j] > sum) ||
                          (   (z->diff[i][j] == sum )
                           && (z->relop[i][j] >=
                               transitive_op(
                                transitive_op( 
                                 transitive_op( 
                                  transitive_op(z->relop[i][k], z->relop[k][l]),
                                  transitive_op(z->relop[l][m], z->relop[m][n])
                                 ), 
                                 transitive_op( 
                                  transitive_op(z->relop[n][o], z->relop[o][p]),
                                  transitive_op(z->relop[p][q], z->relop[q][r])
                                 )
                                ),
                                z->relop[r][j]
                               )
                              )
                          )
                         ) {
                           z->relop[i][j] = -1;
                           z->diff[i][j] = 0;
                           break;
                       }
                      }
                     }
                    }
                    if(z->relop[i][j] == -1) break;
                   }
                  }
                  if(z->relop[i][j] == -1) break;
                 }
                }
                if(z->relop[i][j] == -1) break;
               }
              }
              if(z->relop[i][j] == -1) break;
             }
            }
            if(z->relop[i][j] == -1) break;
           }
          }
          if(z->relop[i][j] == -1) break;
         }
        }
        if(z->relop[i][j] == -1) break;
       }
      }
     } /* end if(z->relop[i][j] != -1) */
    } 
   }
  }
/*
  output("After reduce_zone: ");
  print_zone(z); output("\n");
*/
  return z;
}

/*
int all_visited(int *v, struct zone_type *z) {
  int i, j, k;
  int tmp_v[z->clk_index[0] + 1];

  for(i=0; i<=z->clk_index[0]; i++) {
    tmp_v[i] = 0;
  }
  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      printf("Reached 2 in all_visited(), i = %d, j = %d\n", i, j); 
      fflush(stdout);
      if(z->relop[i][j] != -1) {
        if(!tmp_v[i]) tmp_v[i] = 1;
        if(!tmp_v[j]) tmp_v[j] = 1;
      }
    }
  }

  printf("v = [");
  for(i=0; i<=z->clk_index[0]; i++)
    printf("%d ",v[i]);
  printf("]\n"); 

  printf("tmp_v = [");
  for(i=0; i<=z->clk_index[0]; i++)
    printf("%d ",tmp_v[i]);
  printf("]\n"); 

  for(i=0; i<=z->clk_index[0]; i++) {
    if( (v[i] == 0) && (tmp_v[i] == 1) ) return FALSE;
  }
  return TRUE;
}
*/

/* instantiate_zone()
 *
 * PURPOSE: to instantiate all clock conjuncts: "x[i] = c" and "-x[i] = c" 
 */

struct zone_type *instantiate_zone(struct zone_type *z) {
  int i, j;
  int tmpdiff;
  int new_eq = FALSE;

  /* transform all "-x[i] = c" into "x[i] = c" if possible  */

  for(i=1; i<=z->clk_index[0]; i++) {
    if(z->relop[0][i] == EQ) {         /* -x[i] = some_const */
      if(z->relop[i][0] == -1) {
        z->relop[i][0] = EQ;
        z->diff[i][0] = -(z->diff[0][i]);
      }
      else {                          /* there is already an inequality in  */
        switch(z->relop[i][0]) {      /* "x[i] <op> some_const" */
          case LESS:
            if(z->diff[i][0] > -(z->diff[0][i])) {
              z->relop[i][0] = EQ;
              z->diff[i][0] = -(z->diff[0][i]);
            }
            else {
              z->relop[0][0] = FALSE;
              return z;
            }
            break;
          case LEQ:
            if(z->diff[i][0] >= -(z->diff[0][i])) {
              z->relop[i][0] = EQ;
              z->diff[i][0] = -(z->diff[0][i]);
            }
            else {
              z->relop[0][0] = FALSE;
              return z;
            }
            break;
          case EQ:
            if(z->diff[i][0] != -(z->diff[0][i])) {
              z->relop[0][0] = FALSE;
              return z;
            }
            break;
        }
      }
      z->relop[0][i] = -1;
      z->diff[0][i] = 0;
    }
  }


  /* for each "x[i] = c", substitute all x[i] occurring in:
   * CASE (0): -x[i] ~ d ----> remove if not contradictory, o/w return FALSE;
   * CASE (1): x[i] - x[j] ~ d  ----> -x[j] ~ d-c
   *      (2): x[j] - x[i] ~ d  ---->  x[j] ~ d+c
   */

  for(i=1; i<=z->clk_index[0]; i++) {
    if(z->relop[i][0] == EQ) {             /* x[i] = c, c = some constant */
      if(z->relop[0][i] == LESS) {                   /* CASE (0) */
        if(-(z->diff[0][i]) >= z->diff[i][0]) {
          z->relop[0][0] = FALSE;
          return z;
        }
        else {
          z->relop[0][i] = -1;
          z->diff[0][i] = 0;
        }
      }
      else if(z->relop[0][i] == LEQ){
        if(-(z->diff[0][i]) > z->diff[i][0]) {
          z->relop[0][0] = FALSE;
          return z;
        }
        else {
          z->relop[0][i] = -1;
          z->diff[0][i] = 0;
        }
      }
      for(j=1; j<=z->clk_index[0]; j++) {
        if(z->relop[i][j] != -1) {                   /* CASE (1) */
          tmpdiff = z->diff[i][j] - z->diff[i][0];
          if(z->relop[0][j] == -1) {
            z->relop[0][j] = z->relop[i][j];
            z->diff[0][j] = tmpdiff;
            if(z->relop[0][j] == EQ) new_eq = TRUE;
          }
          else {              /* there is an inequality in [0][j] already */
            switch(z->relop[0][j]) {
              case LESS:
                switch(z->relop[i][j]) {
                  case LESS:
                    if(tmpdiff < z->diff[0][j]) {
                      z->diff[0][j] = tmpdiff;
                    }
                    break;
                  case LEQ:
                    if(tmpdiff < z->diff[0][j]) {
                      z->diff[0][j] = tmpdiff;
                      z->relop[0][j] = LESS;
                    }
                    break;
                  case EQ:
                    if(tmpdiff >= z->diff[0][j]) {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                }
                break;
              case LEQ:
                switch(z->relop[i][j]) {
                  case LESS:
                    if(tmpdiff <= z->diff[0][j]) {
                      z->relop[0][j] = LESS;
                      z->diff[0][j] = tmpdiff;
                    }
                    break;
                  case LEQ:
                    if(tmpdiff < z->diff[0][j]) {
                      z->diff[0][j] = tmpdiff;
                    }
                    break;
                  case EQ:
                    if(tmpdiff <= z->diff[0][j]) {
                      z->relop[0][j] = EQ;
                      z->diff[0][j] = tmpdiff;
                      new_eq = TRUE;
                    }
                    else {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                }
                break;
              case EQ:
                switch(z->relop[i][j]) {
                  case LESS:
                    if(tmpdiff <= z->diff[0][j]) {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                  case LEQ:
                    if(tmpdiff < z->diff[0][j]) {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                  case EQ:
                    if(tmpdiff != z->diff[0][j]) {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                }
                break;
            } /* end switch */
          } /* end else */
          z->relop[i][j] = -1;
          z->diff[i][j] = 0;
        }

        if(z->relop[j][i] != -1) {	                 /* CASE (2) */
          tmpdiff = z->diff[j][i] + z->diff[i][0];
          if(z->relop[j][0] == -1) {
            z->relop[j][0] = z->relop[j][i];
            z->diff[j][0] = tmpdiff;
            if(z->relop[j][0] == EQ) new_eq = TRUE;
          }
          else {       /* there is an inequality in [j][0] already */
            switch(z->relop[j][0]) {
              case LESS:
                switch(z->relop[j][i]) {
                  case LESS:
                    if(tmpdiff < z->diff[j][0]) {
                      z->diff[j][0] = tmpdiff;
                    }
                    break;
                  case LEQ:
                    if(tmpdiff < z->diff[j][0]) {
                      z->relop[j][0] = LEQ;
                      z->diff[j][0] = tmpdiff;
                    }
                    break;
                  case EQ:
                    if(tmpdiff < z->diff[j][0]) {
                      z->relop[j][0] = EQ;
                      z->diff[j][0] = tmpdiff;
                      new_eq = TRUE;
                    }
                    else {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                }
                break;
              case LEQ:
                switch(z->relop[j][i]) {
                  case LESS:
                    if(tmpdiff <= z->relop[j][0]) {
                      z->relop[j][0] = LESS;
                      z->diff[j][0] = tmpdiff;
                    }
                    break;
                  case LEQ:
                    if(tmpdiff < z->relop[j][0]) {
                      z->diff[j][0] = tmpdiff;
                    }
                    break;
                  case EQ:
                    if(tmpdiff <= z->relop[j][0]) {
                      z->relop[j][0] = EQ;
                      z->diff[j][0] = tmpdiff;
                      new_eq = TRUE;
                    }
                    else {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                }
                break;
              case EQ:
                switch(z->relop[j][i]) {
                  case LESS:
                    if(tmpdiff <= z->diff[j][0]) {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                  case LEQ:
                    if(tmpdiff < z->diff[j][0]) {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                  case EQ:
                    if(tmpdiff != z->diff[j][0]) {
                      z->relop[0][0] = FALSE;
                      return z;
                    }
                    break;
                }
                break;
            } /* end switch */
          } /* end else */
          z->relop[j][i] = -1;
          z->diff[j][i] = 0;
        }
      }
    }
  }
  if(new_eq) {
    z = instantiate_zone(z);
  }
/*
  output("After instantiate_zone: ");
  print_zone(z); output("\n");
*/
  return z;
}
/* instantiate_zone */

struct clkconj_type *zone2clkconj(struct zone_type *z) {
  int i, j;
  struct clkconj_type *x = (struct clkconj_type *)NULL;
  struct clkconj_type *y = (struct clkconj_type *)NULL;

  if(z) {
    if(z->relop[0][0] == TRUE) {
      x = alloc_clkconj();
      x->op = TRUE;
      return x;
    }
    if(z->relop[0][0] == FALSE) {
      x = alloc_clkconj();
      x->op = FALSE;
      return x;
    }
    for(i=0; i<=z->clk_index[0]; i++) {
      for(j=0; j<=z->clk_index[0]; j++) {
        if(i==0 && j==0) continue;
        if(z->relop[i][j] != -1) {
          y = alloc_clkconj();
          y->clk1 = old_index(i, z);
          y->clk2 = old_index(j, z);
          y->op = z->relop[i][j];
          y->diff = z->diff[i][j];
          if(x) {
            y->next = x->next;
            x->next = y;
          }
          x = y;
        }
      }
    }
  }
  else return (struct clkconj_type *) NULL;
/*
  printf("Zone: "); print_zone(z); printf("\n");
  printf("Clkconj: "); print_clkconj(x); printf("\n");
*/
  return x;
}

int old_index(int someclk, struct zone_type *z) {
  if(someclk == 0) return 0;
  else return z->clk_index[someclk];
}

/* The following are used with non-reduced zones */


/* if "x[i] = x[j]" then ensure "x[j] = x[i]" */

struct zone_type *zone_proc_eq(struct zone_type *z) {
  int i, j;

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(i == j) continue;

      if(z->relop[i][j] == EQ && z->diff[i][j] == 0) {
        if(z->relop[j][i] == -1) {
          z->relop[j][i] = EQ;
          z->diff[j][i] = 0;
        }
        else {
          switch(z->relop[j][i]) {
            case LESS:
              if(z->diff[j][i] <= 0) {
                z->relop[0][0] = FALSE;
                return z;
              }
              break;
            case LEQ:
              if(z->diff[j][i] < 0) {
                z->relop[0][0] = FALSE;
                return z;
              }
              break;
            case EQ:
              if(z->diff[j][i] != 0) {
                z->relop[0][0] = FALSE;
                return z;
              }
              break;
          } /* end switch */
          z->relop[j][i] = EQ;
          z->diff[j][i] = 0;
        } /* end else */
      }
    }
  }
  return z;
}

/* reduce x[i] - x[j] <= diff    AND    x[j] - x[i] <= -(diff)
     INTO               x[i] - x[j] = diff
*/
struct zone_type *zone_proc_leq(struct zone_type *z) {
int i, j;

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(z->relop[i][j] == LEQ && z->relop[j][i] == LEQ) {
        if(z->diff[i][j] == -(z->diff[j][i])) {
          z->relop[i][j] = EQ;
          z->relop[j][i] = EQ;
        }
      }
    }
  }
  return z;
}

/* The Kleene's closure of a zone */

struct zone_type *kclosure_zone(struct zone_type *z) {
  int i, j, k;

  for(i=0; i<=z->clk_index[0]; i++) {
    for(j=0; j<=z->clk_index[0]; j++) {
      if(i==0 && j==0) continue;
      if(i == j) continue;

      for(k=0; k<=z->clk_index[0]; k++) {
        if(z->relop[i][k] != -1 && z->relop[k][j] != -1) {
          if(z->relop[i][j] == -1) {
            z->diff[i][j] = z->diff[i][k] + z->diff[k][j];
            z->relop[i][j] = transitive_op(z->relop[i][k], z->relop[k][j]);
          }
          else if(z->diff[i][j] > z->diff[i][k] + z->diff[k][j]) {
            z->diff[i][j] = z->diff[i][k] + z->diff[k][j];
            z->relop[i][j] = transitive_op(z->relop[i][k], z->relop[k][j]);
          }
        }
      }
    }
  }
  return z;
}

